﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Core;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Common base-class for IEntityRelationEngine implementations
    /// </summary>
    abstract public class BaseEngine : IEntityRelationEngine
    {

        private EngineId id;
        private string description;
        private bool trainingCompleted = false;
        private double minRatingValue = 1;
        private double maxRatingValue = 5;
        private IEntityRelationDataProvider backend;
        private RelationType[] registerRelationTypes;
        
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="id">Engine ID</param>
        /// <param name="description">Engine description</param>
        /// <param name="registerRelationTypes">List of relation types to register update handlers for</param>
        public BaseEngine(EngineId id, string description, RelationType[] registerRelationTypes)
        {
            this.id = id;
            this.description = description;
            this.registerRelationTypes = registerRelationTypes;
        }

        /// <summary>
        /// Gets/sets the IEntityRelationDataProvider to use
        /// </summary>
        public IEntityRelationDataProvider EntityRelationDataProvider
        {
            get
            {
                return backend;
            }
            set
            {
                if (backend != null)
                {
                    throw new InvalidOperationException("The EntityRelationDataProvider can only be set once.");
                }
                backend = value;
                foreach (RelationType relationType in registerRelationTypes)
                {
                    backend.RegisterForRelationUpdates(relationType, this);
                }
                backend.RegisterForEntityUpdates(EntityType.User, this);
                backend.RegisterForEntityUpdates(EntityType.CatalogItem, this);
            }
        }

        /// <summary>
        /// Checks if this engine is configured properly
        /// </summary>
        protected virtual void verifyConfiguration()
        {
            if (EntityRelationDataProvider == null) throw new Exception("No IEntityRelationDataProvider reference configured");
        }

        /// <summary>
        /// Gets the ID of this engine
        /// </summary>
        public EngineId Id
        {
            get
            {
                return id;
            }
        }

        /// <summary>
        /// Gets a description of this engine
        /// </summary>
        public string Description
        {
            get
            {
                return description;
            }
        }

        /// <summary>
        /// Trains this engine
        /// </summary>
        public virtual void Train()
        {
            trainingCompleted = false;
            verifyConfiguration();
            doTrain();
            trainingCompleted = true;
            // just notify that we're done
            if (TrainingCompleted != null) TrainingCompleted.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Should be overridden to do the actual training
        /// </summary>
        protected virtual void doTrain() {
        }

        /// <summary>
        /// Gets the training completed state
        /// </summary>
        /// <returns>True if training is completed, false otherwise</returns>
        public virtual bool IsTrainingCompleted()
        {
            return trainingCompleted;
        }

        /// <summary>
        /// Training completed event handler
        /// </summary>
        public EventHandler<EventArgs> TrainingCompleted
        {
            get;
            set;
        }

        /// <summary>
        /// Predicts the value for the specified relation
        /// </summary>
        /// <param name="relationType">Relation type</param>
        /// <param name="key">Relation key</param>
        /// <returns>Predicted value for the specified relation type and value</returns>
        public abstract double PredictRelation(RelationType relationType, int[] key);

        /// <summary>
        /// Optinal method for the algorithm to learn from the discrepancy between actual and predicted value.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="predictedValue">The predicted value.</param>
        /// <param name="actualValue">The actual value.</param>
        public virtual void SubmitRelationFeedback(RelationType relationType, int[] key, double predictedValue, double actualValue)
        {
        }

        /// <summary>
        /// Predicted rating
        /// </summary>
        protected class PredictedRating {
            /// <summary>
            /// Gets/sets the predicted rating
            /// </summary>
            public double Rating { get; set; }
            /// <summary>
            /// Gets/sets the item ID
            /// </summary>
            public int ID { get; set; }

            /// <summary>
            /// Constructor
            /// </summary>
            /// <param name="id">Item id</param>
            /// <param name="rating">Predicted rating for that item</param>
            public PredictedRating(int id, double rating)
            {
                ID = id;
                Rating = rating;
            }
        }

        /// <summary>
        /// Generates a top N of entities for a particular key
        /// </summary>
        /// <param name="entityType">Entity type to create a top-N for</param>
        /// <param name="keys">Optional set of keys</param>
        /// <param name="n">Size of the top-N</param>
        /// <returns>A top N of entities</returns>
        public virtual int[] PredictEntity(EntityType entityType, int[] keys, int n)
        {
            switch (entityType)
            {
                case EntityType.CatalogItem:
                    if ((keys == null) || (keys.Length != 1)) throw new Exception("Exactly one key (user ID) needed");

                    List<PredictedRating> ratings = new List<PredictedRating>(); // using list because #items might be < n
                    double minRating = 0;
                    IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
                    for (dr.Open(); dr.Read(); )
                    {
                        int itemID = dr.GetInt32(0);
                        double rating = PredictRelation(RelationType.Rated, new int[] { keys[0], itemID });
                        if ((ratings.Count < n) || (minRating < rating))
                        {
                            int index = 0;
                            while ((index < ratings.Count) && (ratings[index].Rating > rating)) index++;
                            ratings.Insert(index, new PredictedRating(itemID, rating));
                            if (ratings.Count > n) ratings.RemoveAt(n);
                            minRating = ratings[ratings.Count - 1].Rating;
                        }
                    }
                    int[] result = new int[ratings.Count];
                    for (int i = 0; i < result.Length; i++)
                    {
                        result[i] = ratings[i].ID;
                    }
                    return result;

                default:
                    throw new Exception("Can only predict items");
            }
        }

        /// <summary>
        /// Relation has been added
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="values">The values.</param>
        public virtual void AddRelation(RelationType relationType, int[] key, object[] values) { }

        /// <summary>
        /// Relation has been updated
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="values">The values.</param>
        public virtual void UpdateRelation(RelationType relationType, int[] key, object[] values) { }

        /// <summary>
        /// Entity has been added
        /// </summary>
        /// <param name="entityType">The entityType.</param>
        /// <param name="entityId">The entityId.</param>
        /// <param name="values">The values.</param>
        public virtual void AddEntity(EntityType entityType, int entityId, object[] values) { }

        /// <summary>
        /// Entity has been updated
        /// </summary>
        /// <param name="entityType">The entityType.</param>
        /// <param name="entityId">The entityId.</param>
        /// <param name="values">The values.</param>
        public virtual void UpdateEntity(EntityType entityType, int entityId, object[] values) { }

        /// <summary>
        /// Entity has been removed
        /// </summary>
        /// <param name="entityType">The entityType.</param>
        /// <param name="entityId">The entityId.</param>
        public virtual void RemoveEntity(EntityType entityType, int entityId) { }

        /// <summary>
        /// Relation has been removed
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        public virtual void RemoveRelation(RelationType relationType, int[] key) { }

        /// <summary>
        /// Gets/sets the maximum allowed rating value (default = 5)
        /// </summary>
        public double MaxRatingValue 
        {
            get
            {
                return maxRatingValue;
            }
            set
            {
                maxRatingValue = value;
            }
        }

        /// <summary>
        /// Gets/sets the minimum allowed rating value (default = 1)
        /// </summary>
        public double MinRatingValue
        {
            get
            {
                return minRatingValue;
            }
            set
            {
                minRatingValue = value;
            }
        }
    }
}
