﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Algorithms.Utils;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Predictor that uses Case-Based Reasoning for its predictions. This predictor
    /// bases its predictions on similar items that the user has already rated in the
    /// past.
    /// <p>
    /// There is a tuning attribute available:
    /// <dl>
    /// <dt><code>similarityThreshold</code></dt>
    /// <dd>The minimum similarity that is required between the content item for
    /// which a prediction is calculated and other content items. Only content items
    /// that are at least <i>that</i> similar will be taken into account while
    /// calculating a prediction. The value of this tuning attribute is a double
    /// between 0 and 1. The default value is 0.5.</dd>
    /// </dl></p>
    /// </summary>
    public class CBR : BaseEngine
    {
        /// <summary>
        /// Default rating model to use, stores all ratings and optimizes for user retrieval
        /// </summary>
        class CBRRatingModel : IRatingModel
        {
            private float min;
            private float max;

            public CBRRatingModel(float min, float max)
            {
                this.min = min;
                this.max = max;
            }

            public IEntityRelationDataProvider EntityRelationDataProvider { get; set; }

            private Dictionary<int, UserRatings> userRatingsCache = new Dictionary<int, UserRatings>();
            private RatingList ratings = new RatingList();

            public void initialize()
            {
                IEntityRelationDataReader dr = EntityRelationDataProvider.GetRelation(RelationType.Rated);
                for (dr.Open(); dr.Read(); )
                {
                    addRating(dr.GetInt32(0), dr.GetInt32(1), (float)dr.GetDouble(2));
                }
            }

            public void addRating(int userId, int itemId, float rating)
            {

                Rating r = new Rating(userId, itemId, rating);
                ratings.Add(r);

                UserRatings userRatings;
                if (userRatingsCache.ContainsKey(userId))
                {
                    userRatings = userRatingsCache[userId];
                }
                else
                {
                    userRatings = new UserRatings(userId);
                    userRatingsCache[userId] = userRatings;
                }
                userRatings.Add(r);
            }

            public void removeRating(int userId, int itemId)
            {
                if (userRatingsCache.ContainsKey(userId))
                {
                    Rating r = userRatingsCache[userId].Remove(itemId);
                    if (r != null) ratings.Remove(r);
                }
            }

            public void removeUser(int userId)
            {
                if (userRatingsCache.ContainsKey(userId))
                {
                    UserRatings ratings = userRatingsCache[userId];
                    userRatingsCache.Remove(userId);
                    foreach (Rating r in ratings.getRatings())
                    {
                        removeRating(userId, r.ItemId);
                    }
                }
            }

            public void removeItem(int itemId)
            {
            }


            #region IReadOnlyRatingModel Members

            public ItemRatings getItemRatings(int itemId)
            {
                // not implemented
                return new ItemRatings(itemId);
            }

            public UserRatings getUserRatings(int userId)
            {
                if (userRatingsCache.ContainsKey(userId))
                {
                    return userRatingsCache[userId];
                }

                return new UserRatings(userId);
            }

            public List<int> getItemsRatedByBoth(int userIdA, int userIdB)
            {
                return new List<int>();
            }

            public Correlation getUserCorrelation(int userId, int otherUserId)
            {
                UserRatings userRatings = getUserRatings(userId);
                UserRatings otherUserRatings = getUserRatings(otherUserId);

                return getCorrelation(userRatings, otherUserRatings);
            }

            public Correlation getCorrelation(UserRatings ratingsA, UserRatings ratingsB)
            {
                Correlation c = null;
                List<double> aResult = new List<double>();
                List<double> bResult = new List<double>();

                foreach (Rating rating in ratingsA.getRatings())
                {
                    Rating otherRating = ratingsB.getRating(rating.ItemId);
                    if (otherRating != null)
                    {
                        aResult.Add(rating.Value);
                        bResult.Add(otherRating.Value);
                    }
                }
                int n = aResult.Count;
                if (n > 0)
                {
                    c = new Correlation(Statistics.pearson(aResult.ToArray(), bResult.ToArray()), n);
                }
                return c;
            }

            public IEnumerator<int> getUserEnumerator()
            {
                IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.User);
                for (dr.Open(); dr.Read(); )
                {
                    yield return dr.GetInt32(0);
                }
            }

            public IEnumerator<int> getItemEnumerator()
            {
                IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
                for (dr.Open(); dr.Read(); )
                {
                    yield return dr.GetInt32(0);
                }
            }

            public int getUserCount()
            {
                // lazy impl
                IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.User);
                int count = 0;
                for (dr.Open(); dr.Read(); ) count++;

                return count;
            }

            public int getItemCount()
            {
                // lazy impl
                IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
                int count = 0;
                for (dr.Open(); dr.Read(); ) count++;

                return count;
            }

            public IEnumerator<Rating> getRatingEnumerator()
            {
                return ratings.getRatings().GetEnumerator();
            }

            public RatingList getAllRatings()
            {
                return ratings;
            }


            public Rating getRating(int userId, int itemId)
            {
                UserRatings ratings = getUserRatings(userId);
                if (ratings == null) return null;
                return ratings.getRating(itemId);
            }

            public float MaxValue
            {
                get
                {
                    return max;
                }
            }

            public float MinValue
            {
                get
                {
                    return min;
                }
            }
            #endregion
        }

        /// <summary>
        /// Constructor
        /// </summary>
        public CBR() : base(new MyMediaProject.RecommenderSystem.Core.EngineId(new Guid(13, 13, 13, new byte[] { 13, 13, 13, 13, 13, 13, 13, 21 })), "CBR", new RelationType[] { RelationType.Rated, RelationType.ItemKeyword})
        {
            SimilarityThreshold = 0.5;
        }

        /// <summary>
        /// Gets/sets the rating model to be used
        /// </summary>
        public IRatingModel RatingModel { get; set; }

        /// <summary>
        /// Gets/sets the item similarity model to be used
        /// </summary>
        public IItemSimilarityModel ItemSimilarityModel { get; set; }

        /// <inheritdoc />
        protected override void doTrain()
        {
            if (RatingModel == null)
            {
                CBRRatingModel crm = new CBRRatingModel((float)MinRatingValue, (float)MaxRatingValue);
                crm.EntityRelationDataProvider = EntityRelationDataProvider;
                crm.initialize();
                RatingModel = crm;
            }
            if (ItemSimilarityModel == null)
            {
                KeywordSimilarityModel ksm = new KeywordSimilarityModel();
                ksm.EntityRelationDataProvider = EntityRelationDataProvider;
                ksm.CacheKeywords = true;
                ksm.initialize();
                ItemSimilarityModel = ksm;
            }
        }

        /// <inheritdoc />
        public override double PredictRelation(RelationType relationType, int[] key)
        {
            if (relationType == RelationType.Rated)
            {
                int userID = key[0];
                int itemID = key[1];

                double predValue = 0;
                double totalSim = 0;
                Dictionary<int, ItemSimilarity> similarities = ItemSimilarityModel.getItemSimilarities(itemID, RatingModel.getUserRatings(userID).getItemIds().GetEnumerator());
                IEnumerator<ItemSimilarity> simIter = similarities.Values.GetEnumerator();
                while (simIter.MoveNext())
                {
                    ItemSimilarity similarity = simIter.Current;
                    double sim = similarity.Similarity;
                    if (sim > SimilarityThreshold)
                    {
                        sim *= sim;
                        int otherID = similarity.OtherItemId;
                        double otherrating = RatingModel.getRating(userID, otherID).Value;
                        predValue += otherrating * sim;
                        totalSim += sim;
                    }
                }

                if (totalSim > 0)
                {
                    predValue = predValue / totalSim;
                }
                if (predValue < RatingModel.MinValue) predValue = RatingModel.MinValue;
                if (predValue > RatingModel.MaxValue) predValue = RatingModel.MaxValue;

                return predValue;
            }
            throw new Exception("Unsupported relation type");
        }

        /// <summary>
        /// Gets/sets the threshold value for similarity.
        /// </summary>
        public double SimilarityThreshold { get; set; }

        /// <inheritdoc />
        public override void SubmitRelationFeedback(RelationType relationId, int[] key, double predictedValue, double actualValue)
        {
            if (relationId == RelationType.Rated)
            {
                RatingModel.addRating(key[0], key[1], (float)actualValue);
            }
        }

        /// <inheritdoc />
        public override void AddRelation(MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm.RelationType relation_id, int[] key, object[] values)
        {
            if (relation_id == RelationType.ItemKeyword)
            {
                ItemSimilarityModel.removeItem(key[0]);
                ItemSimilarityModel.getItemSimilarities(key[0], RatingModel.getItemEnumerator());
            }
            else if (relation_id == RelationType.Rated)
            {
                RatingModel.addRating(key[0], key[1], Convert.ToSingle(values[0]));
            }
        }

        /// <inheritdoc />
        public override void UpdateRelation(MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm.RelationType relation_id, int[] key, object[] values)
        {
            if (relation_id == RelationType.ItemKeyword)
            {
                ItemSimilarityModel.removeItem(key[0]);
                ItemSimilarityModel.getItemSimilarities(key[0], RatingModel.getItemEnumerator());
            }
            else if (relation_id == RelationType.Rated)
            {
                RatingModel.addRating(key[0], key[1], Convert.ToSingle(values[0]));
            }
        }

        /// <inheritdoc />
        public override void RemoveEntity(EntityType entity_type, int entity_id)
        {
            if (entity_type == EntityType.User)
            {
                RatingModel.removeUser(entity_id);
            }
            else if (entity_type == EntityType.CatalogItem)
            {
                RatingModel.removeItem(entity_id);
                ItemSimilarityModel.removeItem(entity_id);
            }
        }

        /// <inheritdoc />
        public override void RemoveRelation(RelationType relation_id, int[] key)
        {
            if (relation_id == RelationType.Rated)
            {
                RatingModel.removeRating(key[0], key[1]);
            }
            else if (relation_id == RelationType.ItemKeyword)
            {
                ItemSimilarityModel.removeItem(key[0]);
                ItemSimilarityModel.getItemSimilarities(key[0], RatingModel.getItemEnumerator());
            }
        }
    }
}
