﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Algorithms.Utils;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Collaborative filter, Duine implementation
    /// </summary>
    class CollaborativeFilter : BaseEngine
    {
        /// <summary>
        /// Constructor
        /// </summary>
        public CollaborativeFilter()
            : base(new MyMediaProject.RecommenderSystem.Core.EngineId(new Guid(14, 14, 14, new byte[] { 14, 14, 14, 14, 14, 14, 14, 17 })), "CollaborativeFilter", new RelationType[] { RelationType.Rated })
        {
        }

        /// <summary>
        /// The minimum number of rated items that are both rated by this user and the other, for the other to be 
        /// considered as user that is similar to this user.
        /// </summary>
	    private const int minNoItemsInCommon = 50;

        /// <summary>
        /// Gets/sets the rating model used. If not set then the default (Duine) will be used
        /// </summary>
        public IRatingModel RatingModel { get; set; }

        /// <summary>
        /// Gets/sets the user similarity model used. If not set then the default (Duine) will be used
        /// </summary>
        public IUserSimilarityModel UserSimilarityModel { get; set; }

        protected override void doTrain()
        {
            if (RatingModel == null)
            {
                DuineRatingModel dm = new DuineRatingModel((float)MinRatingValue, (float)MaxRatingValue);
                dm.EntityRelationDataProvider = EntityRelationDataProvider;
                dm.initialize();
                RatingModel = dm;
            }
            if (UserSimilarityModel == null)
            {
                DuineUserSimilarityModel dusm = new DuineUserSimilarityModel();
                dusm.EntityRelationDataProvider = EntityRelationDataProvider;
                dusm.RatingModel = RatingModel;
                dusm.initialize();
                UserSimilarityModel = dusm;
            }
        }

        public override double PredictRelation(RelationType relationType, int[] key)
        {
            if (relationType == RelationType.Rated)
            {
		        double prediction = 0.0;
		        double confidence = 0.0;

                ItemRatings rc = RatingModel.getItemRatings(key[0]);
		        UserRatings rc_me = RatingModel.getUserRatings(key[1]);
		        List<Rating> v = rc.getRatings();
		        double tR = 0.0;
		        double tW = 0.0;
		        int nSim = 0;

		        List<int> userIDs = new List<int>(rc.getUserIds());
                Dictionary<int, UserSimilarity> userSims = UserSimilarityModel.getUserSimilarities(key[0], userIDs.GetEnumerator());

		        foreach (Rating r in v) {
			        UserRatings rc_other = RatingModel.getUserRatings(r.UserId);
                    UserSimilarity userSim = null;
                    userSims.TryGetValue(r.UserId, out userSim);
			        double cor = (userSim != null ? userSim.Similarity : 0);
			        int sampleSize = (userSim != null ? userSim.SampleSize : 0);
			        if (sampleSize < minNoItemsInCommon) {
				        // The other user is not that simular to me,
				        // so reduce the correlation in a linear way.
				        cor *= ((double) sampleSize) / minNoItemsInCommon;
			        }
			        double c = ((r.Value - rc_other.getAverage()) / rc_other.getStdev()) * cor;
			        if (!Double.IsNaN(c)) {
				        tR += c;
				        tW += Math.Abs(cor);
				        nSim++;
			        }
		        }
		        prediction = rc_me.getAverage();
		        double dev = rc_me.getStdev() * (tR / tW);
		        if (!Double.IsNaN(dev)) {
			        prediction += dev;
		        }
		        if (nSim > 0) {
			        confidence = tW / nSim;
		        }
		        if (Double.IsNaN(prediction)) {
			        throw new Exception("NaN is not a valid prediction value");
		        }
                if (prediction > RatingModel.MaxValue) prediction = RatingModel.MaxValue;
                if (prediction < RatingModel.MinValue) prediction = RatingModel.MinValue;
                return prediction;
            }
            throw new Exception("Unsupported relation type");
        }

        public override void SubmitRelationFeedback(RelationType relationId, int[] key, double predictedValue, double actualValue)
        {
            if (relationId == RelationType.Rated)
            {
                RatingModel.addRating(key[0], key[1], (float)actualValue);
                UserSimilarityModel.removeUser(key[0]);
                UserSimilarityModel.getUserSimilarities(key[0], RatingModel.getUserEnumerator());
            }
        }

        public override void AddRelation(RelationType relation_id, int[] key, object[] values)
        {
            if (relation_id == RelationType.Rated)
            {
                RatingModel.addRating(key[0], key[1], Convert.ToSingle(values[0]));
                UserSimilarityModel.removeUser(key[0]);
                UserSimilarityModel.getUserSimilarities(key[0], RatingModel.getUserEnumerator());
            }
        }

        public override void UpdateRelation(RelationType relation_id, int[] key, object[] values)
        {
            if (relation_id == RelationType.Rated)
            {
                RatingModel.addRating(key[0], key[1], Convert.ToSingle(values[0]));
                UserSimilarityModel.removeUser(key[0]);
                UserSimilarityModel.getUserSimilarities(key[0], RatingModel.getUserEnumerator());
            }
        }

        public override void AddEntity(EntityType entity_type, int entity_id, object[] values)
        {
        }

        public override void UpdateEntity(EntityType entity_type, int entity_id, object[] values)
        {
        }

        public override void RemoveEntity(EntityType entity_type, int entity_id)
        {
            if (entity_type == EntityType.User)
            {
                RatingModel.removeUser(entity_id);
                UserSimilarityModel.removeUser(entity_id);
            }
            else if (entity_type == EntityType.CatalogItem)
            {
                RatingModel.removeItem(entity_id);
            }
        }

        public override void RemoveRelation(RelationType relation_id, int[] key)
        {
            if (relation_id == RelationType.Rated)
            {
                RatingModel.removeRating(key[0], key[1]);
                UserSimilarityModel.removeUser(key[0]);
                UserSimilarityModel.getUserSimilarities(key[0], RatingModel.getUserEnumerator());
            }
        }
    }
}
