﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Combining several predictors with a weighted ensemble.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public class EnsembleWeighted : Ensemble
    {
        /// <summary>
        /// List of weights.
        /// </summary>
        public List<double> weights = new List<double>();

        /// <inheritdoc />
        public override double PredictRelation(RelationType relation_id, int[] key)
        {
            double result = 0;
            for (int i = 0; i < engines.Count; i++)
            {
                result += weights[i] * engines[i].PredictRelation(relation_id, key);
            }
            return result;
        }

        /// <inheritdoc />
        public override int[] PredictEntity(EntityType entityType, int[] keys, int n)
        {
            // TODO: works only for rating data.
            return RatingPredictor._PredictEntity(EntityRelationDataProvider, entityType, keys, n, this);
        }
    }
    /// <summary>
    /// Abtract class for combining several prediction methods.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public abstract class Ensemble : IEntityRelationEngine
    {
        /// <summary>
        /// List of engines.
        /// </summary>
        public List<IEntityRelationEngine> engines = new List<IEntityRelationEngine>();

        /// <inheritdoc />
        public abstract int[] PredictEntity(EntityType entityType, int[] keys, int n);
        /// <inheritdoc />
        public abstract double PredictRelation(RelationType relationType, int[] key);


        /// <inheritdoc />
        public void Train()
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.Train();
            }
        }

        /// <inheritdoc />
        public bool IsTrainingCompleted()
        {
            bool result = true;
            foreach (IEntityRelationEngine engine in engines)
            {
                result = result && engine.IsTrainingCompleted();
            }
            return result;
        }

        /// <inheritdoc />
        public void AddRelation(RelationType relation_id, int[] key, object[] values)
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.AddRelation(relation_id, key, values);
            }
        }

        /// <inheritdoc />
        public void UpdateRelation(RelationType relation_id, int[] key, object[] values)
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.UpdateRelation(relation_id, key, values);
            }
        }

        /// <inheritdoc />
        public void RemoveRelation(RelationType relation_id, int[] key)
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.RemoveRelation(relation_id, key);
            }
        }

        /// <inheritdoc />
        public void AddEntity(EntityType entity_type, int entity_id, object[] values)
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.AddEntity(entity_type, entity_id, values);
            }
        }
        /// <inheritdoc />
        public void UpdateEntity(EntityType entity_type, int entity_id, object[] values)
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.UpdateEntity(entity_type, entity_id, values);
            }
        }
        /// <inheritdoc />
        public void RemoveEntity(EntityType entity_type, int entity_id)
        {
            foreach (IEntityRelationEngine engine in engines)
            {
                engine.RemoveEntity(entity_type, entity_id);
            }
        }

        #region IEntityRelationEngine Members

        /// <inheritdoc />
        public void SubmitRelationFeedback(RelationType relationId, int[] key, double predictedValue, double actualValue)
        {
            throw new NotImplementedException();
        }

        /// <inheritdoc />
        public string Description
        {
            get { return "EnsembleEngine"; }
        }

        private MyMediaProject.RecommenderSystem.Core.EngineId engineId;

        /// <inheritdoc />
        public MyMediaProject.RecommenderSystem.Core.EngineId Id
        {
            get
            {
                if (engineId == null)
                {
                    engineId = new MyMediaProject.RecommenderSystem.Core.EngineId(new Guid());
                }
                return engineId;
            }
        }

        /// <inheritdoc />
        public EventHandler<EventArgs> TrainingCompleted
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        #endregion

        #region IEntityRelationEngine Members


        /// <inheritdoc />
        public IEntityRelationDataProvider EntityRelationDataProvider
        {
            get
            {
                //Assuming that each engine is using the same data reader.
                return this.engines.First().EntityRelationDataProvider;
            }
            set
            {
                foreach (IEntityRelationEngine engine in this.engines)
                {
                    engine.EntityRelationDataProvider = value;
                }
            }
        }

        #endregion
    }
}
