using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Core;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Abstract class for item recommenders
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public abstract class ItemRecommender : IEntityRelationEngine
    {
        private bool training_completed = false;
        /// <summary>
        /// The <see cref="IEntityRelationDataProvider"/> used by the instance.
        /// </summary>
        protected IEntityRelationDataProvider backend;
        /// <inheritdoc />
        public abstract double PredictWeight(int user_id, int item_id);
        /// <inheritdoc />
        public virtual void AddFeedback(int user_id, int item_id) { }
        /// <inheritdoc />
        public virtual void RemoveFeedback(int user_id, int item_id) { }
        /// <inheritdoc />
        public virtual void AddUser(int user_id) { }
        /// <inheritdoc />
        public virtual void AddItem(int item_id) { }
        /// <inheritdoc />
        public virtual void RemoveUser(int user_id) { }
        /// <inheritdoc />
        public virtual void RemoveItem(int item_id) { }
        /// <inheritdoc />
        protected abstract void _Train();
        
        #region IEngine Members

        /// <inheritdoc />
        protected RelationType target_relation;

        /// <inheritdoc />
        public ItemRecommender(RelationType target_relation)
        {
            switch (target_relation) 
            {
                case RelationType.Purchased: break;
                case RelationType.Viewed: break;
                default: throw new NotImplementedException();
            }
            this.target_relation = target_relation;
        }

        /// <inheritdoc />
        public bool IsTrainingCompleted()
        {
            return training_completed;
        }

        /// <inheritdoc />
        public void Train()
        {
            training_completed = false;
            _Train();
            training_completed = true;
            if (TrainingCompleted != null)
            {
                TrainingCompleted(this, null);
            }
        }

        /// <inheritdoc />
        public void SubmitRelationFeedback(RelationType relationId, int[] key, double predictedValue, double actualValue)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Weighted items class
        /// </summary>
        protected class WeightedItem : IComparable
        {
            /// <summary>
            /// Item id
            /// </summary>
            public int item_id;
            /// <summary>
            /// Weight
            /// </summary>
            public double weight;

            /// <summary>
            /// Compares to.
            /// </summary>
            /// <param name="o">The o.</param>
            /// <returns></returns>
            public int CompareTo(Object o)
            {
                if (((WeightedItem)o).weight > weight)
                {
                    return 1;
                }
                else if (((WeightedItem)o).weight < weight)
                {
                    return -1;
                }
                else
                {
                    return 0;
                }
            }

        }

        /// <inheritdoc />
        public int[] PredictEntity(EntityType entityType, int[] keys, int N)
        {
            if (entityType == EntityType.CatalogItem)
            {
                IEntityRelationDataReader items = backend.GetEntity(EntityType.CatalogItem);

                List<WeightedItem> result = new List<WeightedItem>();
                while (items.Read())
                {
                    WeightedItem i = new WeightedItem();
                    i.item_id = items.GetInt32(0);
                    i.weight = PredictWeight(keys[0], i.item_id);
                    result.Add(i);
                }
                result.Sort();
                if (N > result.Count) { N = result.Count; };
                int[] return_array = new int[N];
                for (int i = 0; i < N; i++)
                {
                    return_array[i] = result[i].item_id;
                }
                return return_array;
            }
            else
            {
                throw new NotImplementedException();
            }
        }
        #endregion

        /// <inheritdoc />
        public IEntityRelationDataProvider EntityRelationDataProvider
        {
            get
            {
                return this.backend;
            }
            set
            {
                if (this.backend != null)
                {
                    throw new InvalidOperationException("The EntityRelationDataProvider can only be set once.");
                }
                this.backend = value;
                backend.RegisterForRelationUpdates(target_relation, this);
                backend.RegisterForEntityUpdates(EntityType.User, this);
                backend.RegisterForEntityUpdates(EntityType.CatalogItem, this);
            }
        }

        /// <inheritdoc />
        public double PredictRelation(RelationType relation_id, int[] key)
        {
            if (relation_id == target_relation)
            {
                return PredictWeight(key[0], key[1]);
            }
            else
            {
                throw new Exception("can only predict the target relation");
            }
        }

        /// <inheritdoc />
        public void AddRelation(RelationType relation_id, int[] key, object[] values)
        {
            if (relation_id == target_relation)
            {
                AddFeedback(key[0], key[1]);
            }
        }

        /// <inheritdoc />
        public void UpdateRelation(RelationType relation_id, int[] key, object[] values)
        {
            // only remove and add should be possible
        }

        /// <inheritdoc />
        public void RemoveRelation(RelationType relation_id, int[] key)
        {
            if (relation_id == target_relation)
            {
                RemoveFeedback(key[0], key[1]);
            }
        }

        /// <inheritdoc />
        public void AddEntity(EntityType entity_type, int entity_id, object[] values)
        {
            switch (entity_type)
            {
                case EntityType.CatalogItem:
                    AddItem(entity_id);
                    break;
                case EntityType.User:
                    AddUser(entity_id);
                    break;
                default: 
                    throw new Exception("unknown entity_type for recommender");
            }
        }

        /// <inheritdoc />
        public void UpdateEntity(EntityType entity_type, int entity_id, object[] values)
        {
            switch (entity_type)
            {
                case EntityType.CatalogItem:
                    // UpdateItem(entity_id);
                    break;
                case EntityType.User:
                    // UpdateUser(entity_id);
                    break;
                default:
                    throw new Exception("unknown entity_type for recommender");
            }
        }

        /// <inheritdoc />
        public void RemoveEntity(EntityType entity_type, int entity_id)
        {
            switch (entity_type)
            {
                case EntityType.CatalogItem:
                    RemoveItem(entity_id);
                    break;
                case EntityType.User:
                    RemoveUser(entity_id);
                    break;
                default:
                    throw new Exception("unknown entity_type for recommender");
            }
        }

        #region IEntityRelationEngine Members


        /// <inheritdoc />
        public string Description
        {
            get { return "ItemRecommender"; }
        }

        private EngineId engineId = new EngineId(new Guid());

        /// <inheritdoc />
        public MyMediaProject.RecommenderSystem.Core.EngineId Id
        {
            get { return engineId; }
        }

        /// <inheritdoc />
        public EventHandler<EventArgs> TrainingCompleted
        {
            get
            {
                // TODO
                return null;
            }
            set
            {
                // TODO
                ;
            }
        }

        #endregion

    }




}
