using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Algorithms.Utils;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Weighted matrix factorization method proposed by Hu et al. and Pan et al.:
    /// Y. Hu, Y. Koren, and C. Volinsky:
    /// Collaborative filtering for implicit feedback datasets.
    /// In IEEE International Conference on Data Mining (ICDM 2008), pages 263--272, 2008.
    /// 
    /// R. Pan, Y.Zhou, B. Cao, N. N. Liu, R. M. Lukose, M. Scholz, and Q. Yang.
    /// One-class collaborative filtering.
    /// In IEEE International Conference on Data Mining (ICDM 2008), pages 502--511, 2008.
    /// 
    /// We use the fast computation method proposed by Hu et al. and we allow a global 
    /// weight to penalize obsorved/unobserved values.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public class ItemRecommenderWRMF : ItemRecommenderMF
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ItemRecommenderWRMF"/> class.
        /// </summary>
        /// <param name="target_relation">The target_relation.</param>
        public ItemRecommenderWRMF(RelationType target_relation) : base(target_relation) { }

        /// <summary>
        /// Regularization
        /// </summary>
        public double regularization = 0.015;
        /// <summary>
        /// Initial f mean
        /// </summary>
        public double init_f_mean = 0;
        /// <summary>
        /// Initial f standard deviation
        /// </summary>
        public double init_f_stdev = 0.1;
        /// <summary>
        /// Number of iterations
        /// </summary>
        public int num_iter = 30;
        /// <summary>
        /// C position
        /// </summary>
        public double c_pos = 1;

        /// <inheritdoc />
        protected override void _Train()
        {
            base._Train();
            MatrixUtils.initNormal(user_feature, init_f_mean, init_f_stdev);
            MatrixUtils.initNormal(item_feature, init_f_mean, init_f_stdev);
            
            for (int iter = 1; iter < num_iter; iter++)
            {
                optimize(data_user, user_feature, item_feature);
                optimize(data_item, item_feature, user_feature);
            }

        }

        /// <summary>
        /// Optimizes the specified data.
        /// </summary>
        /// <param name="data">The data.</param>
        /// <param name="W">The W.</param>
        /// <param name="H">The H.</param>
        protected void optimize(SparseBooleanMatrix data, Matrix<double> W, Matrix<double> H) 
        {
            Matrix<double> HH = new Matrix<double>(num_feature, num_feature);
            Matrix<double> HCIH = new Matrix<double>(num_feature, num_feature);
            double[] HCp = new double[num_feature];

            MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(num_feature, num_feature);
            MathNet.Numerics.LinearAlgebra.Matrix m_inv; // = new MathNet.Numerics.LinearAlgebra.Matrix(num_feature);


            // (1) create HH in O(f^2|I|)
            // HH is symmetric
            for (int f_1 = 0; f_1 < num_feature; f_1++)
            {
                for (int f_2 = 0; f_2 < num_feature; f_2++)
                {
                    double d = 0;
                    for (int i = 0; i < H.dim1; i++)
                    {
                        d += H.get(i, f_1) * H.get(i, f_2);
                    }
                    HH.set(f_1, f_2, d);	
                }

            }
            // (2) optimize all U
            // HCIH is symmetric
            for (int u = 0; u < W.dim1; u++)
            {
                HashSet<int> row = data.getRow(u);
                // create HCIH in O(f^2|S_u|)
                for (int f_1 = 0; f_1 < num_feature; f_1++)
                {
                    for (int f_2 = 0; f_2 < num_feature; f_2++)
                    {
                        double d = 0;
                        foreach (int i in row)
                        {
                            d += H.get(i, f_1) * H.get(i, f_2) * (c_pos - 1);
                        }
                        HCIH.set(f_1, f_2, d);
                    }
                }
                // create HCp in O(f|S_u|)
                for (int f = 0; f < num_feature; f++)
                {
                    double d = 0;
                    foreach (int i in row)
                    {
                        d += H.get(i, f) * c_pos * 1;
                    }
                    HCp[f] = d;
                }
                // create m = HH + HCp + gamma*I
                // m is symmetric
                // the inverse m_inv is symmetric
                for (int f_1 = 0; f_1 < num_feature; f_1++)
                {
                    for (int f_2 = 0; f_2 < num_feature; f_2++)
                    {
                        double d = HH.get(f_1, f_2) + HCIH.get(f_1, f_2);
                        if (f_1 == f_2)
                        {
                            d += regularization;
                        }
                        m[f_1, f_2] = d;
                    }
                }
                m_inv = m.Inverse();
                // write back optimal W
                for (int f = 0; f < num_feature; f++)
                {
                    double d = 0;
                    for (int f_2 = 0; f_2 < num_feature; f_2++)
                    {
                        d += m_inv[f, f_2] * HCp[f_2];
                    }
                    W.set(u, f, d);
                }
            }
        }

    }

}