﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Core;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Abstract class for rating predictors
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public abstract class RatingPredictor : IEntityRelationRatingPredictor
    {
        private bool training_completed = false;
        /// <summary>
        /// The <see cref="IEntityRelationDataProvider"/> used by the instance.
        /// </summary>
        public IEntityRelationDataProvider backend;
        /// <inheritdoc />
        public abstract double PredictRating(int user_id, int item_id);
        /// <inheritdoc />
        public virtual void AddRating(int user_id, int item_id, double rating) { }
        /// <inheritdoc />
        public virtual void UpdateRating(int user_id, int item_id, double rating) { }
        /// <inheritdoc />
        public virtual void RemoveRating(int user_id, int item_id) { }
        /// <inheritdoc />
        public virtual void AddUser(int user_id) { }
        /// <inheritdoc />
        public virtual void AddItem(int item_id) { }
        /// <inheritdoc />
        public virtual void RemoveUser(int user_id) { }
        /// <inheritdoc />
        public virtual void RemoveItem(int item_id) { }
        /// <inheritdoc />
        protected abstract void _Train();
        /// <inheritdoc />
        protected double max_data_value = 5;
        /// <inheritdoc />
        protected double min_data_value = 1;

        #region IEngine Members


        /// <inheritdoc />
        public bool IsTrainingCompleted()
        {
            return training_completed;
        }

        /// <inheritdoc />
        public void Train()
        {
            training_completed = false;
            _Train();
            training_completed = true;
            if (TrainingCompleted != null)
            {
                TrainingCompleted(this, null);
            }
        }

        /// <inheritdoc />
        public void SubmitRelationFeedback(RelationType relationId, int[] key, double predictedValue, double actualValue)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// 
        /// </summary>
        protected class WeightedItem : IComparable 
        {
            /// <summary>
            /// Item id
            /// </summary>
            public int item_id;
            /// <summary>
            /// Weight
            /// </summary>
            public double weight;

            /// <summary>
            /// Compares to.
            /// </summary>
            /// <param name="o">The o.</param>
            /// <returns></returns>
            public int CompareTo(Object o)
            {
                if (((WeightedItem)o).weight > weight)
                {
                    return 1;
                }
                else if (((WeightedItem)o).weight < weight)
                {
                    return -1;
                }
                else
                {
                    return 0;
                }
            }

        }
        /// <summary>
        /// Predicts entity.
        /// </summary>
        /// <param name="backend">The backend.</param>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="keys">The keys.</param>
        /// <param name="N">The N.</param>
        /// <param name="engine">The engine.</param>
        /// <returns></returns>
        public static int[] _PredictEntity(IEntityRelationDataProvider backend, EntityType entityType, int[] keys, int N, IEntityRelationEngine engine)
        {
            if (entityType == EntityType.CatalogItem)
            {
                IEntityRelationDataReader items = backend.GetEntity(EntityType.CatalogItem);

                List<WeightedItem> result = new List<WeightedItem>();
                while (items.Read())
                {
                    WeightedItem i = new WeightedItem();
                    i.item_id = items.GetInt32(0);
                    i.weight = engine.PredictRelation(RelationType.Rated, new int[] {keys[0], i.item_id});
                    result.Add(i);
                }
                result.Sort();
                if (N > result.Count) { N = result.Count; };
                int[] return_array = new int[N];
                for (int i = 0; i < N; i++)
                {
                    return_array[i] = result[i].item_id;
                }
                return return_array;     

            }
            else
            {
                throw new NotImplementedException();
            }
        }

        /// <inheritdoc />
        public int[] PredictEntity(EntityType entityType, int[] keys, int N)
        {
            return _PredictEntity(backend, entityType, keys, N, this);
        }
        #endregion



        /// <inheritdoc />
        public IEntityRelationDataProvider EntityRelationDataProvider
        {
            get
            {
                return this.backend;
            }
            set
            {
                if (this.backend != null)
                {
                    throw new InvalidOperationException("The EntityRelationDataProvider can only be set once.");
                }
                this.backend = value;
                backend.RegisterForRelationUpdates(RelationType.Rated, this);
                backend.RegisterForEntityUpdates(EntityType.User, this);
                backend.RegisterForEntityUpdates(EntityType.CatalogItem, this);
            }
        }

        /// <inheritdoc />
        public double PredictRelation(RelationType relation_id, int[] key)
        {
            switch (relation_id)
            {
                case RelationType.Rated:
                    return PredictRating(key[0], key[1]);
                default:
                    throw new Exception("unknown relation_id for recommender");
            }
        }

        /// <inheritdoc />
        public void AddRelation(RelationType relation_id, int[] key, object[] values)
        {
            switch (relation_id)
            {
                case RelationType.Rated:
                    AddRating(key[0], key[1], (double)values[0]);
                    break;
                default:
                    throw new Exception("unknown relation_id for recommender");
            }
        }

        /// <inheritdoc />
        public void UpdateRelation(RelationType relation_id, int[] key, object[] values)
        {
            switch (relation_id)
            {
                case RelationType.Rated:
                    UpdateRating(key[0], key[1], (double)values[0]);
                    break;
                default:
                    throw new Exception("unknown relation_id for recommender");
            }
        }

        /// <inheritdoc />
        public void RemoveRelation(RelationType relation_id, int[] key)
        {
            switch (relation_id)
            {
                case RelationType.Rated:
                    RemoveRating(key[0], key[1]); 
                    break;
                default:
                    throw new Exception("unknown relation_id for recommender");
            }
        }

        /// <inheritdoc />
        public void AddEntity(EntityType entity_type, int entity_id, object[] values)
        {
            switch (entity_type)
            {
                case EntityType.CatalogItem:
                    AddItem(entity_id);
                    break;
                case EntityType.User:
                    AddUser(entity_id);
                    break;
                default:
                    throw new Exception("unknown entity_type for recommender");
            }
        }
        /// <inheritdoc />
        public void UpdateEntity(EntityType entity_type, int entity_id, object[] values)
        {
            switch (entity_type)
            {
                case EntityType.CatalogItem:
                    // UpdateItem(entity_id);
                    break;
                case EntityType.User:
                    // UpdateUser(entity_id);
                    break;
                default:
                    throw new Exception("unknown entity_type for recommender");
            }
        }
        /// <inheritdoc />
        public void RemoveEntity(EntityType entity_type, int entity_id)
        {
            switch (entity_type)
            {
                case EntityType.CatalogItem:
                    RemoveItem(entity_id);
                    break;
                case EntityType.User:
                    RemoveUser(entity_id);
                    break;
                default:
                    throw new Exception("unknown entity_type for recommender");
            }
        }

        #region IEntityRelationEngine Members

        /// <inheritdoc />
        public string Description
        {
            get { return "RatingPredictor"; }
        }

        private EngineId engineId = new EngineId(new Guid());

        /// <inheritdoc />
        public MyMediaProject.RecommenderSystem.Core.EngineId Id
        {
            get { return engineId; }
        }

        /// <inheritdoc />
        public EventHandler<EventArgs> TrainingCompleted
        {
            get
            {
                // TODO
                return null ;
            }
            set
            {
                // TODO
                ;
            }
        }

        #endregion

        /// <summary>
        /// Gets or sets the max rating value.
        /// </summary>
        /// <value>The max rating value.</value>
        public double MaxRatingValue
        {
            get
            {
                return this.max_data_value;
            }
            set
            {
                this.max_data_value = value;
            }
        }

        /// <summary>
        /// Gets or sets the min rating value.
        /// </summary>
        /// <value>The min rating value.</value>
        public double MinRatingValue
        {
            get
            {
                return this.min_data_value;
            }
            set
            {
                this.min_data_value = value;
            }
        }
    }
    
    

  
}
