﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{
    /// <summary>Evaluation class</summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public class Evaluate
    {
        /// <summary>
        /// Evaluates a rating predictor for RMSE.
        /// </summary>
        /// <param name="engine">Rating prediction engine</param>
        /// <param name="ratings">Test cases</param>
        static public void evaluateRated(IEntityRelationEngine engine, IEntityRelationDataReader ratings)
        {
            double rmse = 0;
            double mae = 0;
            int cnt = 0;
            ratings.Open();
            while (ratings.Read()) 
            {
                double prediction = engine.PredictRelation(RelationType.Rated, new int[2] { ratings.GetInt32(0), ratings.GetInt32(1) });
                double error = (prediction - ratings.GetDouble(2));
                rmse += error*error;
                mae += Math.Abs(error);
                cnt++;
            }
            mae /= cnt;
            rmse = Math.Sqrt(rmse / cnt);
            Console.WriteLine("RMSE/MAE " + rmse + "/" + mae); 
        }

        /// <summary>
        /// Evaluation for rankings of item recommenders. Computes the AUC.
        /// </summary>
        /// <param name="engine">Item recommender engine</param>
        /// <param name="test_reader">Data reader for the test cases</param>
        /// <param name="train_reader">Data reader for the train cases</param>
        /// <param name="items_reader">Data reader to list all items</param>
        static public void evaluateItemRecommender(IEntityRelationEngine engine, IEntityRelationDataReader test_reader, IEntityRelationDataReader train_reader, IEntityRelationDataReader items_reader)
        {
            // (0) read items
            HashSet<int> items = new HashSet<int>();
            items_reader.Open();
            while (items_reader.Read())
            {
                items.Add(items_reader.GetInt32(0));
            }
            
            // (1) read test 
            SparseBooleanMatrix test = new SparseBooleanMatrix();
            test_reader.Open();
            while (test_reader.Read())
            {
                int user_id = test_reader.GetInt32(0);
                int item_id = test_reader.GetInt32(1);
                if (items.Contains(item_id)) {
                    test.addEntry(user_id, item_id);
                }
            }
            // (2) read train
            SparseBooleanMatrix train = new SparseBooleanMatrix();
            train_reader.Open();
            while (train_reader.Read())
            {
                int user_id = train_reader.GetInt32(0);
                int item_id = train_reader.GetInt32(1);
                if (items.Contains(item_id) && test.ContainsKey(user_id)) 
                {
                    if (test[user_id].Contains(item_id))
                    {
                        throw new Exception("Overlap in train and test set for user " + user_id);
                    }
                    train.addEntry(user_id, item_id);
                }
            }

            double sum_auc = 0;
            int num_user = 0;
            foreach (KeyValuePair<int, HashSet<int>> user in test)
            {
                int user_id = user.Key;
                HashSet<int> test_items = user.Value;
                int[] prediction = engine.PredictEntity(EntityType.CatalogItem, new int[] { user_id }, Int32.MaxValue);

                if (prediction.Length != items.Count)
                {
                    throw new Exception("not all items have been ranked.");
                }
                int train_item_size = 0;
                if (train.ContainsKey(user_id)) {
                    train_item_size = train[user_id].Count;
                }
                int num_eval_items = items.Count - train_item_size;
	            int num_eval_pairs = (num_eval_items - test_items.Count) * test_items.Count;
	
                // start with the highest weighting item...
	            int num_correct_pairs = 0;
	            int num_pos_above = 0;
	            foreach (int item_id in prediction) {
		            if (! train[user_id].Contains(item_id)) {
                        if (test_items.Contains(item_id)) {
			                num_pos_above++;
                        } else{
                            num_correct_pairs += num_pos_above;
                        }
                    }
                }

                double user_auc = ((double) num_correct_pairs) / num_eval_pairs;
                sum_auc += user_auc;
                num_user++;
            }
            double auc = sum_auc / num_user;
            Console.WriteLine("AUC " + auc);

        }

    }
}
