﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using System.IO;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{

    /// <summary>
    /// Class that holds the similarity of 2 items
    /// </summary>
    public class ItemSimilarity
    {

        /// <summary>
        /// The similarity value
        /// </summary>
        private double similarity = 0;

        /// <summary>
        /// The id of the item
        /// </summary>
        private int itemId = 0;

        /// <summary>
        /// The id of the other item
        /// </summary>
        private int otherItemId = 0;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="itemId">Item id</param>
        /// <param name="otherItemId">Other item id</param>
        /// <param name="similarity">Similarity between both items</param>
        public ItemSimilarity(int itemId, int otherItemId, double similarity)
        {
            this.itemId = itemId;
            this.otherItemId = otherItemId;
            this.similarity = similarity;
        }

        /// <summary>
        /// Constructor, sets similarity to 0
        /// </summary>
        /// <param name="itemId">Item id</param>
        /// <param name="otherItemId">Other item id</param>
        public ItemSimilarity(int itemId, int otherItemId)
        {
            this.itemId = itemId;
            this.otherItemId = otherItemId;
        }

        /// <summary>
        /// Gets the similarity value
        /// </summary>
        public double Similarity
        {
            get
            {
                return similarity;
            }
        }

        /// <summary>
        /// Gets the item id
        /// </summary>
        public int ItemId
        {
            get
            {
                return itemId;
            }
        }

        /// <summary>
        /// Gets the other item id
        /// </summary>
        public int OtherItemId
        {
            get
            {
                return otherItemId;
            }
        }

        /// <summary>
        /// Gets the hash code
        /// </summary>
        /// <returns>The hash code for this instance</returns>
        public override int GetHashCode()
        {
            int prime = 31;
            int result = 1;
            result = prime * result + itemId;
            result = prime * result + otherItemId;
            return result;
        }

        /// <summary>
        /// Compares this object to another object
        /// </summary>
        /// <param name="obj">Object to compare with</param>
        /// <returns>True if the objects are equal, false otherwise</returns>
        public override bool Equals(object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (!(obj is ItemSimilarity))
                return false;
            ItemSimilarity other = (ItemSimilarity)obj;
            if (!itemId.Equals(other.ItemId))
                return false;
            return otherItemId.Equals(other.OtherItemId);
        }

    }

    /// <summary>
    /// An interface that provides access to the similarity of items.
    /// </summary>
    public interface IItemSimilarityModel
    {
        /// <summary>
        /// Get the similarity between the given content items
        /// </summary>
        /// <param name="itemId">the item id</param>
        /// <param name="otherItemId">the id of the other item</param>
        /// <returns>An object describing the similarities between the items. If an
        /// item does not exist in the model an object is returned with similarity value 0.</returns>
        ItemSimilarity getItemSimilarity(int itemId, int otherItemId);

        /// <summary>
        /// Gets the similarities between the given content item and a list of other items
        /// </summary>
        /// <param name="itemId">the item id</param>
        /// <param name="otherItemIds">enumerator with ids of the other items</param>
        /// <returns>A dictionary describing the similarities between the item and the other items</returns>
        Dictionary<int, ItemSimilarity> getItemSimilarities(int itemId, IEnumerator<int> otherItemIds);

        /// <summary>
        /// Removes the similarities between the specified item and all others
        /// </summary>
        /// <param name="itemID">Item to remove</param>
        void removeItem(int itemID);
    }

    /// <summary>
    /// Base ItemSimilarityModel
    /// </summary>
    public abstract class BaseItemSimilarityModel : IItemSimilarityModel
    {
        /// <summary>
        /// Similarity map
        /// </summary>
        protected Dictionary<int, Dictionary<int, ItemSimilarity>> similarityMap = new Dictionary<int, Dictionary<int, ItemSimilarity>>();

        /// <summary>
        /// Gets/sets the IEntityRelationDataProvider to be used by this model
        /// </summary>
        public IEntityRelationDataProvider EntityRelationDataProvider { get; set; }

        /// <summary>
        /// Gets/sets the name of the cache file where similarity values are cached. If null (the default) no caching is done.
        /// </summary>
        public string CacheFileName { get; set; }

        /// <summary>
        /// Constructor
        /// </summary>
        public BaseItemSimilarityModel()
        {
        }

        #region IItemSimilarityModel Members

        /// <inheritdoc />
        public ItemSimilarity getItemSimilarity(int itemId, int otherItemId)
        {
            if (!similarityMap.ContainsKey(itemId))
            {
                // not in cache -> compute them
                similarityMap[itemId] = new Dictionary<int, ItemSimilarity>();
                computeSimilarities(itemId, 0);
            }
            if (!similarityMap.ContainsKey(otherItemId))
            {
                // not in cache -> compute them
                similarityMap[otherItemId] = new Dictionary<int, ItemSimilarity>();
                computeSimilarities(otherItemId, 0);
            }
            if (similarityMap.ContainsKey(itemId))
            {
                Dictionary<int, ItemSimilarity> level2SimilarityMap = similarityMap[itemId];
                if (level2SimilarityMap.ContainsKey(otherItemId)) return level2SimilarityMap[otherItemId];
            }
            if (similarityMap.ContainsKey(otherItemId))
            {
                Dictionary<int, ItemSimilarity> level2SimilarityMap = similarityMap[otherItemId];
                if (level2SimilarityMap.ContainsKey(itemId)) return level2SimilarityMap[itemId];
            }

            return new ItemSimilarity(itemId, otherItemId, 0);
        }

        /// <inheritdoc />
        public Dictionary<int, ItemSimilarity> getItemSimilarities(int itemId, IEnumerator<int> otherItemIds)
        {
            Dictionary<int, ItemSimilarity> results = new Dictionary<int, ItemSimilarity>();
            while (otherItemIds.MoveNext())
            {
                int otherItemId = otherItemIds.Current;
                results[otherItemId] = getItemSimilarity(itemId, otherItemId);
            }
            return results;
        }

        /// <inheritdoc />
        public void addItemSimilarity(int itemId, int otherItemId, double similarity)
        {
            Dictionary<int, ItemSimilarity> level2SimilarityMap;

            ItemSimilarity newSim = new ItemSimilarity(itemId, otherItemId, similarity);


            // add new
            if (similarityMap.ContainsKey(otherItemId))
            {
                level2SimilarityMap = similarityMap[otherItemId];
                level2SimilarityMap[itemId] = newSim;

                // remove old, if exists in other table
                if (similarityMap.ContainsKey(itemId))
                {
                    level2SimilarityMap = similarityMap[itemId];
                    if (level2SimilarityMap.ContainsKey(otherItemId)) level2SimilarityMap.Remove(otherItemId);
                }
            }
            else
            {
                if (similarityMap.ContainsKey(itemId))
                {
                    level2SimilarityMap = similarityMap[itemId];
                }
                else
                {
                    level2SimilarityMap = new Dictionary<int, ItemSimilarity>();
                    similarityMap[itemId] = level2SimilarityMap;
                }
                level2SimilarityMap[otherItemId] = newSim;
            }
        }

        /// <inheritdoc />
        public void removeItem(int itemId)
        {
            // removes all similarities for the specified item, this will cause a recompute upon the next similarity request
            // for this user
            similarityMap.Remove(itemId);
        }
        #endregion

        /// <summary>
        /// Initializes the model
        /// </summary>
        public virtual void initialize()
        {
            if (EntityRelationDataProvider == null) throw new Exception("No EntityRelationDataProvider reference configured");

            // try to load pre-computed data
            string fileName = CacheFileName;
            if (fileName != null)
            {
                FileInfo file = new FileInfo(fileName);
                if ((file.Exists) && (file.Length > 0))
                {
                    using (BinaryReader reader = new BinaryReader(new FileStream(fileName, FileMode.Open)))
                    {
                        load(reader);
                    }
                    return;
                }
            }
            computeSimilarities();
            if (fileName != null)
            {
                using (BinaryWriter writer = new BinaryWriter(new FileStream(CacheFileName, FileMode.Create)))
                {
                    save(writer);
                }
            }
        }

        /// <summary>
        /// Computes the similarities between all items in the data set
        /// </summary>
        protected virtual void computeSimilarities()
        {
            IEntityRelationDataReader items = EntityRelationDataProvider.GetEntity(EntityType.User);
            items.Open();
            int startIndex = 1;
            while (items.Read()) 
            {
                int itemID = items.GetInt32(0);
                // item similarity is considered symmetrical, so skip the items we already processed
                similarityMap[itemID] = new Dictionary<int, ItemSimilarity>();
                computeSimilarities(itemID, startIndex);
                startIndex++;
            }
        }

        /// <summary>
        /// Loads cached similarities from stream
        /// </summary>
        /// <param name="reader">Stream to load from</param>
        protected virtual void load(BinaryReader reader)
        {
            long itemSize = sizeof(int) + sizeof(int) + sizeof(double);
            long count = reader.BaseStream.Length / itemSize;
            while (count > 0)
            {
                int itemID = reader.ReadInt32();
                int otherItemID = reader.ReadInt32();
                double similarity = reader.ReadDouble();

                addItemSimilarity(itemID, otherItemID, similarity);
            }
        }

        /// <summary>
        /// Saves the similarity values to a stream
        /// </summary>
        /// <param name="writer">Stream to write to</param>
        public virtual void save(BinaryWriter writer)
        {
            IEntityRelationDataReader items = EntityRelationDataProvider.GetEntity(EntityType.User);
            items.Open();
            int startIndex = 1;
            while (items.Read())
            {
                // item similarity is considered symmetrical, so skip the items we already processed
                IEntityRelationDataReader otherItems = EntityRelationDataProvider.GetEntity(EntityType.User);
                otherItems.Open();
                for (int i = 0; i < startIndex; i++)
                {
                    otherItems.Read();
                }

                while (otherItems.Read())
                {
                    int itemID = items.GetInt32(0);
                    int otherItemID = otherItems.GetInt32(0);

                    ItemSimilarity itemSimilarity = getItemSimilarity(itemID, otherItemID);
                    writer.Write(itemID);
                    writer.Write(otherItemID);
                    writer.Write(itemSimilarity.Similarity);
                }
            }
        }

        /// <summary>
        /// Computes the similarities between a item and a (subset of) all other items. Should be overridden
        /// by subclasses that can use their specific similarity metric here
        /// </summary>
        /// <param name="itemId">Item id</param>
        /// <param name="startIndex">Index in the item set at which to start computing</param>
        protected abstract void computeSimilarities(int itemId, int startIndex);
    }
}
