﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{
    /// <summary>
    /// Collection of statistical methods.
    /// </summary>
    public class Statistics
    {
        private Statistics()
        {
        }

        /// <summary>
        /// Calculates the sum of the values in the array <c>x</c>. By definition, the sum of an empty array is 0.
        /// </summary>
        /// <param name="x">The array with values to sum</param>
        /// <returns>The sum of the values in the array. When the array is empty then 0.0 is returned.</returns>
        public static double sum(double[] x)
        {
            double sum = 0.0;
            for (int i = 0; i < x.Length; i++)
            {
                sum += x[i];
            }
            return sum;
        }

        /// <summary>
        /// Calculates the average of the values in the array <c>x</c>. By definition, the average value of an empty array is 0.
        /// </summary>
        /// <param name="x">The array with values to use</param>
        /// <returns>The average of the values in the array. When the array is empty then 0.0 is returned.</returns>
        public static double avg(double[] x)
        {
            return x.Length > 0 ? sum(x) / x.Length : 0.0;
        }

        /// <summary>
        /// Calculates the standard deviation of the values in the array <c>x</c>. By definition, the standard deviation of an empty array is 0.
        /// </summary>
        /// <param name="x">The array with values to use</param>
        /// <returns>The the standard deviation of the values in the array <c>x</c>. When the array is empty then 0.0 is returned.</returns>
        public static double stdvp(double[] x)
        {
            double sum = 0.0;
            double sumsqr = 0.0;
            int n = x.Length;
            for (int i = 0; i < n; i++)
            {
                sum += x[i];
                sumsqr += x[i] * x[i];
            }
            return n == 0 ? 0.0 : Math.Sqrt((n * sumsqr - sum * sum) / (n * n));
        }

        /// <summary>
        /// Calculates the pearson coefficient (amount of correlation) between the arrays <c>x</c> and <c>y</c>.
        /// By definition, the pearson coefficient of two empty arrays is 0.
        /// </summary>
        /// <param name="x">The first array.</param>
        /// <param name="y">y The second array.</param>
        /// <returns>The pearson coefficient the two arrays.</returns>
        public static double pearson(double[] x, double[] y)
        {
            if (x.Length != y.Length)
            {
                throw new ArgumentException("pearson - Arrays must be equal sized");
            }
            double avg_x = avg(x);
            double avg_y = avg(y);
            double sum_xx = 0.0;
            double sum_xy = 0.0;
            double sum_yy = 0.0;
            for (int i = 0; i < x.Length; i++)
            {
                double dev_x = x[i] - avg_x;
                double dev_y = y[i] - avg_y;
                sum_xx += dev_x * dev_x;
                sum_xy += dev_x * dev_y;
                sum_yy += dev_y * dev_y;
            }
            double coeff = sum_xy / Math.Sqrt(sum_xx * sum_yy);
            if (Double.IsNaN(coeff))
            {
                coeff = 0.0;
            }
            return coeff;
        }

        /// <summary>
        /// Computes entropy of a set of probabilities
        /// </summary>
        /// <param name="probabilities">Set of probabilities</param>
        /// <returns>Entropy of the set</returns>
        public static float Entropy(float[] probabilities)
        {
            float E = 0;
            int s = probabilities.Length;
            for (int i = 0; i < s; i++)
            {
                float p = probabilities[i];
                if (p > 0.0)
                {
                    E += p * (float)Math.Log((1.0f / p));
                }
            }
            return E;
        }

        /// <summary>
        /// Computes relative entropy of a set of probabilities
        /// </summary>
        /// <param name="probabilities">Set of probabilities</param>
        /// <param name="baseProbs">Base set of probabilities</param>
        /// <returns>Relative entropy of the set</returns>
        public static float RelativeEntropy(float[] probabilities, float[] baseProbs)
        {
            float E = 0.0f;
            float p1;
            float p2;
            int s = probabilities.Length;
            if (s > baseProbs.Length) throw new ArgumentException("baseProbs must have at least the same number of elements as probabilities");

            for (int i = 0; i < s; i++)
            {
                p1 = probabilities[i];
                p2 = baseProbs[i];
                if (p1 > 0.0 && p2 > 0.0)
                { //dat laatste zou niet mogen gebeuren
                    E += p1 * (float)Math.Log((p1 / p2));
                }
            }

            return E;
        }

        /// <summary>
        /// Computes relative entropy of a set of probabilities
        /// </summary>
        /// <param name="probabilities">Sparse set of probabilities</param>
        /// <param name="baseProbs">Base set of probabilities</param>
        /// <returns>Relative entropy of the set</returns>
        public static float RelativeEntropy(SparseVector<float> probabilities, float[] baseProbs)
        {
            float E = 0.0f;
            float p1;
            float p2;
            int s = baseProbs.Length;

            for (int i = 0; i < s; i++)
            {
                p1 = probabilities[i];
                p2 = baseProbs[i];
                if (p1 > 0.0 && p2 > 0.0)
                { //dat laatste zou niet mogen gebeuren
                    E += p1 * (float)Math.Log((p1 / p2));
                }
            }

            return E;
        }

        /// <summary>
        /// Computes relative entropy of a set of probabilities
        /// </summary>
        /// <param name="probabilities">Sparse set of probabilities</param>
        /// <param name="baseProbs">Sparse base set of probabilities</param>
        /// <returns>Relative entropy of the set</returns>
        public static float RelativeEntropy(SparseVector<float> probabilities, SparseVector<float> baseProbs)
        {
            float E = 0.0f;
            float p1;
            float p2;

            IEnumerator<int> it = probabilities.getKeyIteratorUnion(baseProbs);
            int i;

            while (it.MoveNext())
            {
                i = it.Current;
                p1 = probabilities[i];
                p2 = baseProbs[i];

                if (p1 > 0.0 && p2 > 0.0)
                { //dat laatste zou niet mogen gebeuren
                    E += p1 * (float)Math.Log((p1 / p2));
                }
            }

            return E;
        }

        /// <summary>
        /// Computes the JSD between two sparse vectors
        /// </summary>
        /// <param name="p1">Vector 1</param>
        /// <param name="p2">Vector 2</param>
        /// <returns>The JSD between vector 1 and vector 2</returns>
        public static float JSDivergence(SparseVector<float> p1, SparseVector<float> p2)
        {
            return JSDivergence(p1, p2, 10.0f);
        }

        /*
                        private static SparseVector<float> meanDistr(SparseVector<float> p1, SparseVector<float> p2, float w)
                        {
                            SparseVector<float> p = new SparseVector<float>(p1.getNrOfNonZeroValues());

                            IEnumerator<int> it = p1.getKeyIteratorUnion(p2);
                            int t;
                            while (it.MoveNext())
                            {
                                t = it.Current;
                                p[t] = w * p1[t] + (1 - w) * p2[t];
                            }
                            return p;
                        }
                */
        /*
                private static SparseVector<float> meanDistrHalf(SparseVector<float> p1, SparseVector<float> p2)
                {
                    SparseVector<float> p = new SparseVector<float>(p1.getNrOfNonZeroValues());

                    IEnumerator<int> it = p1.getKeyIteratorUnion(p2);
                    int t;
                    while (it.MoveNext())
                    {
                        t = it.Current;
                        p[t] = 0.5f * (p1[t] + p2[t]);
                    }
                    return p;
                }

                public static float JSDivergence(SparseVector<float> probs1, SparseVector<float> probs2, float maxJSD)
                {
                    float E;
                    SparseVector<float> m = meanDistrHalf(probs1, probs2);
            
                    E = 0.5f * RelativeEntropy(probs1, m);
                    if (E > 0.9 * maxJSD)
                    {
                        E = 2f * E;
                    }
                    else
                    {
                        E += 0.5f * RelativeEntropy(probs2, m);
                    }

                    return E;
                }
         */
        /*        public static float Entropy(SparseVector<float> probabilities)
                {
                    float E = 0;

                    IEnumerator<int> it = probabilities.getKeyIterator();
                    while (it.MoveNext())
                    {
                        int i = it.Current;
                        float p = probabilities[i];
                        if (p > 0.0)
                        {
                            E += p * (float)Math.Log((1f / p));
                        }
                    }
                    return E;
                }

                public static float JSDivergence(SparseVector<float> p1, SparseVector<float> p2, float maxJSD)
                {
                    float E = 0f;

                    IEnumerator<int> it = p1.getKeyIteratorUnion(p2);
                    int t;
                    float p_t;
                    while (it.MoveNext())
                    {
                        t = it.Current;
                        float p1_t = p1[t];
                        float p2_t = p2[t];
                        p_t = 0.5f * (p1_t + p2_t);
                        E += p_t * (float)Math.Log(p_t);
                    }
                    return 2 * E - Entropy(p1) - Entropy(p2);
                }*/

        private readonly static float LOG2TWICE = 2.0f * (float)Math.Log(2);

        /// <summary>
        /// Computes the JSD between two sparse vectors
        /// </summary>
        /// <param name="p1">Vector 1</param>
        /// <param name="p2">Vector 2</param>
        /// <param name="maxJSD">Stop computing if the intermediate result goes above this value</param>
        /// <returns>The JSD between vector 1 and vector 2</returns>
        public static float JSDivergence(SparseVector<float> p1, SparseVector<float> p2, float maxJSD)
        {
            SparseVector<float> indexVector;
            SparseVector<float> otherVector;
            if (p1.getNrOfNonZeroValues() < p2.getNrOfNonZeroValues())
            {
                indexVector = p1;
                otherVector = p2;
            }
            else
            {
                indexVector = p2;
                otherVector = p1;
            }
            IEnumerator<int> indexIter = indexVector.getKeyIterator();
            float jsd = LOG2TWICE;
            while (indexIter.MoveNext())
            {
                float p1f = otherVector[indexIter.Current];
                if (p1f != 0f) {
                    float p2f = indexVector[indexIter.Current];

                    float sum = p1f + p2f;
                    jsd = jsd + (sum * (float)Math.Log(1.0f / sum)) - (p1f * (float)Math.Log(1.0f / p1f)) - (p2f * (float)Math.Log(1.0f / p2f));
                }
            }
            return jsd;
        }
    }
}
