﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{

    /// <summary>
    /// Class that holds the similarity of 2 users
    /// </summary>
    public class UserSimilarity {
        /// <summary>
        /// The similarity value
        /// </summary>
	    private double similarity = 0;
        /// <summary>
        /// Sample size
        /// </summary>
        private int sampleSize = 0;	

        /// <summary>
        /// The id of the user
        /// </summary>
        private int userId = 0;

        /// <summary>
        /// The id of an other user
        /// </summary>
        private int otherUserId = 0;
    	
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="userId">User id</param>
        /// <param name="otherUserId">Other user id</param>
        /// <param name="similarity">Similarity value</param>
        /// <param name="sampleSize">Sample size on which the similarity is based</param>
	    public UserSimilarity(int userId, int otherUserId, double similarity, int sampleSize) {
		    this.userId = userId;
		    this.otherUserId = otherUserId;
		    this.similarity = similarity;
		    this.sampleSize = sampleSize;
	    }

        /// <summary>
        /// Constructor, sets the similarity to 0
        /// </summary>
        /// <param name="userId">User id</param>
        /// <param name="otherUserId">Other user id</param>
	    public UserSimilarity(int userId, int otherUserId) {
		    this.userId = userId;
		    this.otherUserId = otherUserId;
	    }

        /// <summary>
        /// Gets/sets the similarity value
        /// </summary>
	    public double Similarity {
            get {
                return similarity;
            }
            set {
                similarity = value;
            }
	    }

        /// <summary>
        /// Gets/sets the sample size
        /// </summary>
	    public int SampleSize {
            get {
                return sampleSize;
            }
            set {
                sampleSize = value;
            }
        }

        /// <summary>
        /// Gets the user Id
        /// </summary>
	    public int UserId {
            get {
    		    return userId;
            }
	    }

        /// <summary>
        /// Gets the other user Id
        /// </summary>
	    public int OtherUserId {
            get {
		        return otherUserId;
            }
	    }

        /// <summary>
        /// Gets the hash code
        /// </summary>
        /// <returns>The hash code for this instance</returns>
	    public override int GetHashCode() {
		    int prime = 31;
		    int result = 1;
		    result = prime * result + (int)userId;
            result = prime * result + (int)otherUserId;
		    return result;
	    }

        /// <summary>
        /// Compares this object to another object
        /// </summary>
        /// <param name="obj">Object to compare with</param>
        /// <returns>True if the objects are equal, false otherwise</returns>
	    public override bool Equals(Object obj) {
		    if (this == obj)
			    return true;
		    if (obj == null)
			    return false;
		    if (!(obj is UserSimilarity))
			    return false;
		    UserSimilarity other = (UserSimilarity) obj;
		    if (userId != other.UserId)
			    return false;
		    if (otherUserId != other.OtherUserId)
			    return false;
		    return true;
	    }
    	
        /// <summary>
        /// Get a string representation of this instance
        /// </summary>
        /// <returns>A string representation of this instance</returns>
	    public override String ToString() {
            StringBuilder sb = new StringBuilder();
            sb.Append("user similarity");
            sb.Append(" (u1=");
            sb.Append(userId);
            sb.Append(", u2=");
            sb.Append(", sim=");
            sb.Append(similarity);
            sb.Append(sampleSize);
            return sb.ToString();
	    }
    }

    /// <summary>
    /// Interface for similarities of users. The model is assumed to be symmetric. Exchanging the parameter values of userId and otherUserId in 
    /// the method calls has no effect on the result.
    /// </summary>
    public interface IUserSimilarityModel 
    {
        /// <summary>
        /// Retrieves the similarity between the given users. If a specified user is unknown the returned UserSimilarity has similarity and samplesize 0
        /// </summary>
        /// <param name="userId"></param>
        /// <param name="otherUserId"></param>
        /// <returns></returns>
        UserSimilarity getUserSimilarity(int userId, int otherUserId);

        /// <summary>
        /// Retrieves the similarity between the given user and the other users. The size of the returned dictionary always equals the size of otherUserIds. 
        /// For unknown users in otherUserIds an entry is returned with similarity and samplesize 0.
        /// </summary>
        /// <param name="userId"></param>
        /// <param name="otherUserIds"></param>
        /// <returns></returns>
        Dictionary<int, UserSimilarity> getUserSimilarities(int userId, IEnumerator<int> otherUserIds);

        /// <summary>
        /// Adds a user similarity. If a similarity between the users already exists it is updated.
        /// </summary>
        /// <param name="userId"></param>
        /// <param name="otherUserId"></param>
        /// <param name="similarity"></param>
        /// <param name="sampleSize"></param>
        void addUserSimilarity(int userId, int otherUserId, double similarity, int sampleSize);

        /// <summary>
        /// Removes the similarities between the specified user and all others
        /// </summary>
        /// <param name="userId">User to remove</param>
        void removeUser(int userId);
    }

    /// <summary>
    /// Base UserSimilarityModel
    /// </summary>
    public abstract class BaseUserSimilarityModel : IUserSimilarityModel
    {
        /// <summary>
        /// Similarity map
        /// </summary>
        protected Dictionary<int, Dictionary<int, UserSimilarity>> similarityMap = new Dictionary<int, Dictionary<int, UserSimilarity>>();

        /// <summary>
        /// Gets/sets the IEntityRelationDataProvider to be used by this model
        /// </summary>
        public IEntityRelationDataProvider EntityRelationDataProvider { get; set; }

        /// <summary>
        /// Gets/sets the name of the cache file where similarity values are cached. If null (the default) no caching is done.
        /// </summary>
        public string CacheFileName { get; set; }

        /// <summary>
        /// Constructor
        /// </summary>
        public BaseUserSimilarityModel()
        {
        }

        #region IUserSimilarityModel Members

        /// <inheritdoc />
        public UserSimilarity getUserSimilarity(int userId, int otherUserId)
        {
            if (!similarityMap.ContainsKey(userId))
            {
                // not in cache -> compute them
                similarityMap[userId] = new Dictionary<int, UserSimilarity>();
                computeSimilarities(userId, 0);
            }
            if (!similarityMap.ContainsKey(otherUserId))
            {
                // not in cache -> compute them
                similarityMap[otherUserId] = new Dictionary<int, UserSimilarity>();
                computeSimilarities(otherUserId, 0);
            }
            if (similarityMap.ContainsKey(userId))
            {
                Dictionary<int, UserSimilarity> level2SimilarityMap = similarityMap[userId];
                if (level2SimilarityMap.ContainsKey(otherUserId)) return level2SimilarityMap[otherUserId];
            }
            if (similarityMap.ContainsKey(otherUserId))
            {
                Dictionary<int, UserSimilarity> level2SimilarityMap = similarityMap[otherUserId];
                if (level2SimilarityMap.ContainsKey(userId)) return level2SimilarityMap[userId];
            }

            return new UserSimilarity(userId, otherUserId, 0, 0);
        }

        /// <inheritdoc />
        public Dictionary<int, UserSimilarity> getUserSimilarities(int userId, IEnumerator<int> otherUserIds)
        {
            Dictionary<int, UserSimilarity> results = new Dictionary<int, UserSimilarity>();
            while (otherUserIds.MoveNext())
            {
                int otherUserId = otherUserIds.Current;
                results[otherUserId] = getUserSimilarity(userId, otherUserId);
            }
            return results;
        }

        /// <inheritdoc />
        public void addUserSimilarity(int userId, int otherUserId, double similarity, int sampleSize)
        {
            Dictionary<int, UserSimilarity> level2SimilarityMap;

            UserSimilarity newSim = new UserSimilarity(userId, otherUserId, similarity, sampleSize);


            // add new
            if (similarityMap.ContainsKey(otherUserId))
            {
                level2SimilarityMap = similarityMap[otherUserId];
                level2SimilarityMap[userId] = newSim;

                // remove old, if exists in other table
                if (similarityMap.ContainsKey(userId))
                {
                    level2SimilarityMap = similarityMap[userId];
                    if (level2SimilarityMap.ContainsKey(otherUserId)) level2SimilarityMap.Remove(otherUserId);
                }
            }
            else
            {
                if (similarityMap.ContainsKey(userId))
                {
                    level2SimilarityMap = similarityMap[userId];
                }
                else
                {
                    level2SimilarityMap = new Dictionary<int, UserSimilarity>();
                    similarityMap[userId] = level2SimilarityMap;
                }
                level2SimilarityMap[otherUserId] = newSim;
            }
        }

        /// <inheritdoc />
        public void removeUser(int userId)
        {
            // removes all similarities for the specified user, this will cause a recompute upon the next similarity request
            // for this user
            similarityMap.Remove(userId);
        }
        #endregion

        /// <summary>
        /// Initializes the model
        /// </summary>
        public virtual void initialize()
        {
            if (EntityRelationDataProvider == null) throw new Exception("No EntityRelationDataProvider reference configured");

            // try to load pre-computed data
            string fileName = CacheFileName;
            if (fileName != null)
            {
                FileInfo file = new FileInfo(fileName);
                if ((file.Exists) && (file.Length > 0))
                {
                    using (BinaryReader reader = new BinaryReader(new FileStream(fileName, FileMode.Open)))
                    {
                        load(reader);
                    }
                    return;
                }
            }
            computeSimilarities();
            if (fileName != null)
            {
                using (BinaryWriter writer = new BinaryWriter(new FileStream(CacheFileName, FileMode.Create)))
                {
                    save(writer);
                }
            }
        }

        /// <summary>
        /// Computes the similarities between all users in the data set
        /// </summary>
        protected virtual void computeSimilarities() {
            IEntityRelationDataReader users = EntityRelationDataProvider.GetEntity(EntityType.User);
            users.Open();
            int startIndex = 1;
            while (users.Read()) {
                int userID = users.GetInt32(0);
                // user similarity is considered symmetrical, so skip the users we already processed
                similarityMap[userID] = new Dictionary<int, UserSimilarity>();
                computeSimilarities(userID, startIndex);
                startIndex++;
            }
        }

        /// <summary>
        /// Loads cached similarities from stream
        /// </summary>
        /// <param name="reader">Stream to load from</param>
        protected virtual void load(BinaryReader reader)
        {
            long itemSize = sizeof(int) + sizeof(int) + sizeof(double) + sizeof(int);
            long count = reader.BaseStream.Length / itemSize;
            while (count > 0)
            {
                int userID = reader.ReadInt32();
                int otherUserID = reader.ReadInt32();
                double similarity = reader.ReadDouble();
                int sampleSize = reader.ReadInt32();

                addUserSimilarity(userID, otherUserID, similarity, sampleSize);
                count--;
            }
        }

        /// <summary>
        /// Saves the similarity values to a stream
        /// </summary>
        /// <param name="writer">Stream to write to</param>
        public virtual void save(BinaryWriter writer) {

            IEntityRelationDataReader users = EntityRelationDataProvider.GetEntity(EntityType.User);
            users.Open();
            int startIndex = 1;
            while (users.Read()) 
            {
                // user similarity is considered symmetrical, so skip the users we already processed
                IEntityRelationDataReader otherUsers = EntityRelationDataProvider.GetEntity(EntityType.User);
                otherUsers.Open();
                for (int i = 0; i < startIndex; i++)
                {
                    otherUsers.Read();
                }

                while (otherUsers.Read())
                {
                    int userID = users.GetInt32(0);
                    int otherUserID = otherUsers.GetInt32(0);
                    UserSimilarity userSimilarity = getUserSimilarity(userID, otherUserID);
                    if (userSimilarity.SampleSize != 0)
                    {
                        writer.Write(userID);
                        writer.Write(otherUserID);
                        writer.Write(userSimilarity.Similarity);
                        writer.Write(userSimilarity.SampleSize);
                    }
                }
            }
        }

        /// <summary>
        /// Computes the similarities between a user and a (subset of) all other users. Should be overridden
        /// by subclasses that can use their specific similarity metric here
        /// </summary>
        /// <param name="userId">User id</param>
        /// <param name="startIndex">Index in the user set at which to start computing</param>
        protected abstract void computeSimilarities(int userId, int startIndex);
    }
}
