﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Globalization;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{
    /// <summary>
    /// Duine UserSimilarityModel
    /// </summary>
    public class DuineUserSimilarityModel : BaseUserSimilarityModel
    {
        /// <summary>
        /// Constructor
        /// </summary>
        public DuineUserSimilarityModel()
        {
        }

        /// <summary>
        /// Gets/sets the rating model to use
        /// </summary>
        public IRatingModel RatingModel { get; set; }

        /// <summary>
        /// Gets/sets the treshold for computed user similarity
        /// </summary>
        public double Treshold { get; set; }

        /// <summary>
        /// Compute the similarities of the specified user with all others
        /// </summary>
        /// <param name="userId">User ID to compute similarities for</param>
        /// <param name="startIndex">Optional start index, the first startIndex users will be skipped</param>
        protected override void computeSimilarities(int userId, int startIndex)
        {
            IEntityRelationDataReader otherUsers = EntityRelationDataProvider.GetEntity(EntityType.User);
            otherUsers.Open();
            for (int i = 0; i < startIndex; i++)
            {
                otherUsers.Read();
            }

            similarityMap[userId] = new Dictionary<int, UserSimilarity>();
            UserRatings userRatings = RatingModel.getUserRatings(userId);
            while (otherUsers.Read())
            {
                int otherUserID = otherUsers.GetInt32(0);
                if (otherUserID != userId)
                {
                    UserRatings otherUserRatings = RatingModel.getUserRatings(otherUserID);
                    Correlation correlation = RatingModel.getCorrelation(userRatings, otherUserRatings);
                    if ((correlation != null) && (correlation.CorrelationFactor >= Treshold))
                    {
                        addUserSimilarity(userId, otherUserID, correlation.CorrelationFactor, correlation.SampleSize);
                    }
                }
            }
        }
    }

    /// <summary>
    /// JSD UserSimilarityModel, By Christian Wartena
    /// </summary>
    public class JSDUserSimilarityModel : BaseUserSimilarityModel
    {
        /// <summary>
        /// Rating probability struct
        /// </summary>
        public struct RatingProbability
        {
            /// <summary>
            /// Positive rating probability
            /// </summary>
            public float PositiveProbability { get; set; }
            /// <summary>
            /// Negative rating probability
            /// </summary>
            public float NegativeProbability { get; set; }

            /// <summary>
            /// Gets an empty (0, 0) RatingProbability instance
            /// </summary>
            public static RatingProbability Empty
            {
                get
                {
                    JSDUserSimilarityModel.RatingProbability empty = new JSDUserSimilarityModel.RatingProbability();
                    empty.NegativeProbability = 0;
                    empty.PositiveProbability = 0;
                    return empty;
                }
            }
        }

        private const double WARMTECONSTANTE = 20;

        private Dictionary<int, float> positiveTotalsPerItem;
        private Dictionary<int, float> negativeTotalsPerItem;
        private Dictionary<int, float> totalsPerUser;
        private Dictionary<int, SparseVector<float>> cachedCoRatings = new Dictionary<int, SparseVector<float>>();

        /// <summary>
        /// Constructor
        /// </summary>
        public JSDUserSimilarityModel()
        {
            positiveTotalsPerItem = new Dictionary<int, float>();
            negativeTotalsPerItem = new Dictionary<int, float>();
            totalsPerUser = new Dictionary<int, float>();
        }

        /// <summary>
        /// Gets/sets the rating model to use
        /// </summary>
        public IRatingModel RatingModel { get; set; }

        internal static RatingProbability convertRating(float rating, float min, float max)
        {
            RatingProbability rp = new RatingProbability();
            rp.PositiveProbability = (rating - min) / (max - min);
            rp.NegativeProbability = 1f - rp.PositiveProbability;

            return rp;
        }

        /// <summary>
        /// Initializes this model
        /// </summary>
        public override void initialize()
        {
            base.initialize();
        }

        /// <summary>
        /// Adds all ratings to this model
        /// </summary>
        private void addAllUserRatings()
        {
            IEnumerator<int> userSet = RatingModel.getUserEnumerator();
            while (userSet.MoveNext())
            {
                addUserRatings(userSet.Current);
            }
        }

        /// <summary>
        /// Pre-computes all user similarities
        /// </summary>
        protected override void  computeSimilarities()
        {
            addAllUserRatings();
 	        base.computeSimilarities();
        }

        /// <summary>
        /// Loads cached similarities etc. from disk
        /// </summary>
        /// <param name="reader">Stream to read from</param>
        protected override void load(BinaryReader reader)
        {
            addAllUserRatings();
            int userCount = reader.ReadInt32();
            while (userCount > 0)
            {
                int userID = reader.ReadInt32();
                SparseVector<float> currentVector = new SparseVector<float>();
                cachedCoRatings[userID] = currentVector;
                int vectorCount = reader.ReadInt32();
                while (vectorCount > 0)
                {
                    int index = reader.ReadInt32();
                    float value = reader.ReadSingle();
                    currentVector[index] = value;
                    vectorCount--;
                }
                userCount--;
            }
            base.load(reader);
        }

        /// <summary>
        /// Saves all cached similarities etc. to disk
        /// </summary>
        /// <param name="writer">Stream to write to</param>
        public override void save(BinaryWriter writer)
        {
            writer.Write(cachedCoRatings.Count);
            foreach (int userId in cachedCoRatings.Keys)
            {
                writer.Write(userId);
                SparseVector<float> currentVector = cachedCoRatings[userId];
                writer.Write(currentVector.getNrOfNonZeroValues());
                IEnumerator<int> indexes = currentVector.getKeyIterator();
                while (indexes.MoveNext())
                {
                    writer.Write(indexes.Current);
                    writer.Write(currentVector[indexes.Current]);
                }
            }
            base.save(writer);
        }

        /// <summary>
        /// Adds all ratings of a single user to this model
        /// </summary>
        /// <param name="UserId">ID of the user</param>
        private void addUserRatings(int UserId) {
            // add user ratings
            UserRatings userRatings = RatingModel.getUserRatings(UserId);
            foreach (Rating rating in userRatings.getRatings()) {
                // convert rating to (P, N) tuple
                RatingProbability rp = convertRating((float)rating.Value, RatingModel.MinValue, RatingModel.MaxValue);
        		
                // add positive and negative rating
                if (!positiveTotalsPerItem.ContainsKey(rating.ItemId))
                {
                    positiveTotalsPerItem[rating.ItemId] = 0;
                    negativeTotalsPerItem[rating.ItemId] = 0;
                }
                positiveTotalsPerItem[rating.ItemId] += rp.PositiveProbability;
                negativeTotalsPerItem[rating.ItemId] += rp.NegativeProbability;

                if (!totalsPerUser.ContainsKey(UserId))
                {
                    totalsPerUser[UserId] = 0;
                }
                // @MW: totalsPerUser is always +1 (positiveRating + negativeRating), unless rating == 0 (no rating)
                totalsPerUser[UserId] = totalsPerUser[UserId] + rp.PositiveProbability + rp.NegativeProbability;
            }
        }

        // This function is rather expensive to compute and has to be computed rather frequently
        // We need some caching for it.
        // Possible regime: recompute coRatings(z) only if a prediction for z has to be computed
        private SparseVector<float> coRatings(int UserId)
        {
            /*
             * 
             * 
            if (!users.containsKey(UserId))
            {
                return null; //TODO throw exception
            }
            int z = users.get(UserId);
            SparseVector<Float> p = new SparseVector<Float>();

            float Nzd, Ntd;
            SparseVector<Float> distrZ = index.get(z);
            SparseVector<Float> distrT;

            for (int t = 0; t < nrOfUsers; t++)
            {
                distrT = index.get(t);
                float p_t = 0;
                int d;
                Iterator<Integer> i = distrT.getKeyIteratorUnion(distrZ);
                while (i.hasNext())
                {
                    d = i.next();
                    Nzd = distrZ.getFloatAt(d);
                    Ntd = distrT.getFloatAt(d);
                    p_t += ((float)Nzd / (float)totalsPerUser.get(z)) * ((float)Ntd / (float)totalsPerItem.get(d));
                }
                if (p_t > 0)
                {
                    p.setAt(t, p_t);
                }
            }

            return p;
            */
            // use cached if available
            if (cachedCoRatings.ContainsKey(UserId)) return cachedCoRatings[UserId];

            // fallback, if totalsPerUser for this user does not exist then
            // add all the ratings for that user
            if (!totalsPerUser.ContainsKey(UserId))
            {
                // add user ratings
                addUserRatings(UserId);
            }

            SparseVector<float> p = new SparseVector<float>();

            // utility empty value
            RatingProbability empty = RatingProbability.Empty;
            RatingProbability Nzd, Ntd;

            // distriZ == all ratings of UserId
            UserRatings distrZ = RatingModel.getUserRatings(UserId);
            UserRatings distrT;

            // iterate over all users
            IEnumerator<int> otherUserSet = RatingModel.getUserEnumerator();
            while (otherUserSet.MoveNext())
            {
                // distrT == all ratings for the other user
                distrT = RatingModel.getUserRatings(otherUserSet.Current);
                float p_t = 0;
                int d;
                // iterate over all ratings given by this user and the other user
                IEnumerator<int> i = distrT.getKeyIteratorUnion(distrZ);
                while (i.MoveNext())
                {
                    d = i.Current;

                    // get ratings of this user and other user for item d
                    Rating RNzd = distrZ.getRating(d);
                    Rating RNtd = distrT.getRating(d);

                    // convert both ratings to a (N, P) tuple
                    Nzd = RNzd == null ? empty : convertRating((float)RNzd.Value, RatingModel.MinValue, RatingModel.MaxValue);
                    Ntd = RNtd == null ? empty : convertRating((float)RNtd.Value, RatingModel.MinValue, RatingModel.MaxValue);

                    // use converted rating in co-rating computation
                    p_t += ((float)Nzd.PositiveProbability / totalsPerUser[UserId]) * ((float)Ntd.PositiveProbability / positiveTotalsPerItem[d]);
                    p_t += ((float)Nzd.NegativeProbability / totalsPerUser[UserId]) * ((float)Ntd.NegativeProbability / negativeTotalsPerItem[d]);
                }
                // store p_t, if useful
                if (p_t > 0)
                {
                    p[(int)otherUserSet.Current] = p_t;
                }
            }

            // cache computed coRating
            cachedCoRatings[UserId] = p;

            return p;
        }

        /// <summary>
        /// Clears the cache (similarity map and co ratings)
        /// </summary>
        public void clearCache()
        {
            cachedCoRatings.Clear();
            similarityMap.Clear();
        }

        /// <summary>
        /// Compute the similarities of the specified user with all others
        /// </summary>
        /// <param name="userId">User ID to compute similarities for</param>
        /// <param name="startIndex">Optional start index, the first startIndex users will be skipped</param>
        protected override void computeSimilarities(int userId, int startIndex)
        {
            /*
             * 
             * 
                        SparseVector<Float> userModel = coRatings(userNr);

                        for (int u = 0; u < nrOfUsers; u++)
                        {
                            if (u == userNr)
                            {
                                continue;
                            }
                            SparseVector<Float> ratings = index.get(userNr);
            //                float ratingN = ratings.getFloatAt(itemNr);
            //                float ratingP = ratings.getFloatAt(itemNr + 1);
            //                if (ratingN > 0 || ratingP > 0)
            //                {
                                SparseVector<Float> userModel1 = coRatings(u);
                                float div = Statistics.JSD(userModel, userModel1);
                                float weight = (float)Math.pow(2, (-1 * div)); //TODO this is the similarity between two users; This is extremely costly to compute; we should do here some caching.
            //                    weigthSum += weight;
            //                    ratingSumN += weight * ratingN;
            //                    ratingSumP += weight * ratingP;
            //                }
                        }
            //            ratingProbs[0] = ratingSumN / weigthSum;
            //            ratingProbs[1] = ratingSumP / weigthSum;

                        return ratingProbs;
            */
            SparseVector<float> userModel = coRatings(userId);

            IEntityRelationDataReader otherUsers = EntityRelationDataProvider.GetEntity(EntityType.User);
            otherUsers.Open();
            for (int i = 0; i < startIndex; i++)
            {
                otherUsers.Read();
            }

            while (otherUsers.Read())
            {
                int otherUserID = otherUsers.GetInt32(0);
                if (otherUserID != userId)
                {
                    SparseVector<float> userModel1 = coRatings(otherUserID);
                    float div = Statistics.JSDivergence(userModel, userModel1);
                    float weight = (float)Math.Pow(2, (-WARMTECONSTANTE * div)); //TODO this is the similarity between two users; This is extremely costly to compute; we should do here some caching.

                    addUserSimilarity(userId, otherUserID, weight, 1);
                }
            }
        }
    }
}
