﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EMIC.Design;
using System.Globalization;
using MyMediaProject.RecommenderSystem.Core.Config;


namespace MyMediaProject.RecommenderSystem.Core.ActionLogging
{
    /// <summary>
    /// Provides the functionality to log actions/events that might be useful for statistical purposes
    /// </summary>
    public static class ActionLogger
    {
        /// <summary>
        /// String corresponding to the Value of a configurational property, signalizing that the logging category is enabled.
        /// </summary>
        internal const String EnabledConfigPropertyValueFlag = "Enabled";

        /// <summary>
        /// String corresponding to the Value of a configurational property, signalizing that the logging category is disabled.
        /// </summary>
        internal const String DisabledConfigPropertyValueFlag = "Disabled";

        /// <summary>
        /// If the specified category is not registered in the system yet and this parameter is true, the category will be created and enabled for logging right immediately; 
        /// Otherwise, it will be registered and disabled.
        /// </summary>
        private const bool IsEnablingNewCategoryByDefaultOn = true;

        /// <summary>
        /// Description of a configuration property, used when creating a new category
        /// </summary>
        internal const String DescriptionOfLoggingActionConfigProperty = "If the value is set to '" + EnabledConfigPropertyValueFlag  + "', the actions of this category will be logged. (The PrivateValue field holds the categoryID entry stored in the database that is generated and handled programmatically and should not be changed.)";

        /// <summary>
        /// Description of the LogAction configuration module(will be used to create the configuration file if not created yet)
        /// </summary>
        private const String LogActionModuleDescription = "Module specifying which logging categories should be logged.";

        /// <summary>
        /// Logs action related to a local user
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="user">Initialized User object with meaningful UserId property.</param>
        public static void LogAction(String categoryName, User user)
        {
            int userId = 0;
            if (user != null) userId = user.UserId;

            LogAction(categoryName, null, userId, default(int), default(int), default(int), default(int));
        }

        /// <summary>
        /// Logs action related to a local user
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="user">Initialized User object with meaningful UserId property.</param>
        public static void LogAction(String categoryName, String propertyValue, User user)
        {
            int userId = 0;
            if (user != null) userId = user.UserId;

            LogAction(categoryName, propertyValue, userId, default(int), default(int), default(int), default(int));
        }

        /// <summary>
        /// Logs the action related to an external system user
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="externalUser">Initialized ExternalSystemUser object with meaningful ExternalSystemUserId property.</param>
        public static void LogAction(String categoryName, String propertyValue, ExternalSystemUser externalUser)
        {
            int extUserId = 0;

            if (externalUser != null) extUserId = externalUser.ExternalSystemUserId;

            LogAction(categoryName, propertyValue, default(int), default(int), default(int), extUserId, default(int));
        }


        /// <summary>
        /// Logs the specified action
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="user"></param>
        /// <param name="item"></param>
        /// <param name="userAction"></param>
        public static void LogAction(String categoryName, String propertyValue, User user, CatalogItem item, UserAction userAction)
        {
            int userId = 0;
            int itemId = 0;
            int userActionId = 0;

            if (user != null) userId = user.UserId;
            if (item != null) itemId = item.ItemId;
            if (userAction != null) userActionId = userAction.Key;

            LogAction(categoryName, propertyValue, userId, itemId, userActionId, userActionId, default(int));
        }


        /// <summary>
        /// Logs the specified action
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="user"></param>
        /// <param name="item"></param>
        public static void LogAction(String categoryName, String propertyValue, User user, CatalogItem item)
        {
            int userId = 0;
            int itemId = 0;

            if (user != null) userId = user.UserId;
            if (item != null) itemId = item.ItemId;

            LogAction(categoryName, propertyValue, userId, itemId, default(int), default(int), default(int));
        }

        /// <summary>
        /// Logs the action related to a virtual profile
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="virtualProfile"></param>
        public static void LogVirtualProfileAction(String categoryName, String propertyValue, VirtualProfile virtualProfile)
        {
            int virtualProfileId = 0;
            if (virtualProfile != null) virtualProfileId = virtualProfile.VirtualProfileId;

            LogAction(categoryName, propertyValue, default(int), default(int), default(int), default(int), virtualProfileId);
        }

        /// <summary>
        /// Logs the action related to a virtual profile
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="virtualProfile"></param>
        /// <param name="user"></param>
        /// <param name="externalSystemUser"></param>
        public static void LogVirtualProfileAction(String categoryName, String propertyValue, VirtualProfile virtualProfile, User user, ExternalSystemUser externalSystemUser)
        {
            int userId = 0;
            int extUserId = 0;
            int virtualProfileId = 0;

            if (user != null) userId = user.UserId;
            if (externalSystemUser != null) extUserId = externalSystemUser.ExternalSystemUserId;
            if (virtualProfile != null) virtualProfileId = virtualProfile.VirtualProfileId;

            LogAction(categoryName, propertyValue, userId, default(int), default(int), extUserId, virtualProfileId);
        }

        /// <summary>
        /// Logs the action related to a virtual profile
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="virtualProfile"></param>
        /// <param name="user"></param>
        /// <param name="propertyValue"></param>
        public static void LogVirtualProfileAction(String categoryName, String propertyValue, VirtualProfile virtualProfile, User user)
        {
            int userId = 0;
            int virtualProfileId = 0;

            if (user != null) userId = user.UserId;
            if (virtualProfile != null) virtualProfileId = virtualProfile.VirtualProfileId;

            LogAction(categoryName, propertyValue, userId, default(int), default(int), default(int), virtualProfileId);
        }

        /// <summary>
        /// Logs the action related to a virtual profile
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="propertyValue"></param>
        /// <param name="virtualProfile"></param>
        /// <param name="extUser"></param>
        public static void LogVirtualProfileAction(String categoryName, String propertyValue, VirtualProfile virtualProfile, ExternalSystemUser extUser)
        {
            int extUserId = 0;
            int virtualProfileId = 0;

            if (extUser != null) extUserId = extUser.ExternalSystemUserId;
            if (virtualProfile != null) virtualProfileId = virtualProfile.VirtualProfileId;

            LogAction(categoryName, propertyValue, default(int), default(int), default(int), extUserId, virtualProfileId);
        }

        /// <summary>
        /// Logs data about the specified action. Does not log if the logging of the specified category is explicitly disabled through configuration.
        /// </summary>
        /// <param name="categoryName">Textual category name (category identifier)</param>
        /// <param name="propertyValue">textual property to be logged (or null if not relevant)</param>
        /// <param name="userId">MyMedia user id (or 0, if not relevant)</param>
        /// <param name="itemId">Item id (or 0, if not relevant)</param>
        /// <param name="userAction">MyMedia action id (or 0, if not relevant)</param>
        /// <param name="externalSystemUserId">External system user id (or 0, if not relevant)</param>
        /// <param name="virtualProfileId">Virtual profile id (or 0, if not relevant)</param>
        private static void LogAction(String categoryName, String propertyValue, int userId, int itemId, int userAction, int externalSystemUserId, int virtualProfileId)
        {
            TraceLog.Write(LogEventType.Verbose, "Action logging request: LogAction('{0}','{1}','{2}','{3}','{4}','{5}','{6}')", categoryName, propertyValue, userId, itemId, userAction, externalSystemUserId, virtualProfileId);

            int categoryId = GetCategoryIdFromConfiguration(categoryName);
            if (categoryId <= 0)
            {
                // this is a new category, previously unregistered category(unless config files are incorrect)
                // that needs to be registered in the DB & configuration
                categoryId = GetCategoryIdFromDB(categoryName, true); // insert a new category to the DB, if missing
                RegisterLoggingCategoryToConfiguration(categoryName, categoryId, IsEnablingNewCategoryByDefaultOn);
            }

            if (!IsLogCategoryEnabled(categoryName))
            {
                TraceLog.Write(LogEventType.Verbose, "Not logging a logAction for disabled category '{0}'.", categoryName);
                return;
            }

            // use the light weight short-lived non-static data context to prevent concurrency problems
            using (MyMediaDataContext Context = new MyMediaDataContext(ConfigManager.DBConnectionString))
            {
                LogAction logAction = new LogAction();
                logAction.LogCategoryId = categoryId;

                logAction.UserId = userId;
                logAction.ExternalSystemUserId = externalSystemUserId;
                logAction.ActionId = userAction;
                logAction.ItemId = itemId;
                logAction.VirtualProfileId = virtualProfileId;
                logAction.Property1 = propertyValue;
                logAction.Created = DateTime.Now;

                Context.LogActions.InsertOnSubmit(logAction);
                Context.SubmitChanges();
            }
        }

        /// <summary>
        /// Returns log category Id by the specified category name. 
        /// If the category is not registered in the DB, it will be registered and the Id of the newly created category will be returned
        /// </summary>
        /// <param name="categoryName">Name of the logging category</param>
        /// <returns>Id of the category</returns>
        private static int GetCategoryIdFromConfiguration(String categoryName)
        {
            int categoryId = 0;

            if (!ConfigManager.PropertyValueExists(ConfigConstants.LogActionModuleName, categoryName))
            {
                return 0;
            }

            ConfigProperty configProperty = ConfigManager.Modules[ConfigConstants.LogActionModuleName].Properties[categoryName];
            if (configProperty != null)
            {
                // the PrivateValue of the property holds the real logCategoryId (as registered in the the DB)
                categoryId = GetIntOrNullFromString(configProperty.PrivateValue);
                return categoryId;
            }

            return 0;
        }


        /// <summary>
        /// Creates configuration module(in memory only) if ConfigManager does not have it yet 
        /// </summary>
        private static void CreateConfigurationModuleIfMissing()
        {
            if (!ConfigManager.Modules.ContainsKey(ConfigConstants.LogActionModuleName))
            {
                ConfigModule configModule = new ConfigModule(ConfigConstants.LogActionModuleName, ConfigConstants.LogActionModuleName, LogActionModuleDescription);

                ConfigManager.Modules.Add(ConfigConstants.LogActionModuleName, configModule);
            }
        }

        /// <summary>
        /// Utility method to convert string to integer
        /// </summary>
        /// <param name="strValue"></param>
        /// <returns>Integer extracted from a string, or 0 if the convertion fails.</returns>
        private static int GetIntOrNullFromString(String strValue)
        {
            if (strValue == null || strValue.Length == 0)
            {
                return 0;
            }

            try
            {
                int iValue = Convert.ToInt32(strValue, CultureInfo.InvariantCulture);
                return iValue;
            }
            catch (FormatException)
            {
                return 0;
            }
        }

        /// <summary>
        /// Registers the specified logging category in the configuration (so that it could be disabled conveniently)
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="categoryId"></param>
        /// <param name="isEnablingNewCategoryByDefaultOn"></param>
        private static void RegisterLoggingCategoryToConfiguration(String categoryName, int categoryId, bool isEnablingNewCategoryByDefaultOn)
        {
            CreateConfigurationModuleIfMissing();

            String propertyValue = DisabledConfigPropertyValueFlag;
            if (isEnablingNewCategoryByDefaultOn)
            {
                propertyValue = EnabledConfigPropertyValueFlag;
            }

            // are we rewriting the existing property?
            if (ConfigManager.PropertyValueExists(ConfigConstants.LogActionModuleName, categoryName))
            {
                propertyValue = ConfigManager.Modules[ConfigConstants.LogActionModuleName].Properties[categoryName].Value;
                // remove the current configuration property to reinsert it
                ConfigManager.Modules[ConfigConstants.LogActionModuleName].Properties.Remove(categoryName);
            }

            
            ConfigProperty property = new ConfigProperty(categoryName, propertyValue, DescriptionOfLoggingActionConfigProperty);
            property.PrivateValue = categoryId.ToString(CultureInfo.InvariantCulture);

            ConfigManager.Modules[ConfigConstants.LogActionModuleName].Properties.Add(categoryName, property);

            ConfigManager.SaveModuleConfiguration(ConfigConstants.LogActionModuleName);
        }

        /// <summary>
        /// Registers new category in the DB
        /// </summary>
        /// <param name="categoryName"></param>
        /// <returns>Id of the newly registered category</returns>
        private static int InsertNewLogCategoryIntoDB(String categoryName)
        {
            // use the light weight short-lived non-static data context to prevent concurrency problems
            using (MyMediaDataContext Context = new MyMediaDataContext(ConfigManager.DBConnectionString))
            {
                LogCategory logCategory = new LogCategory();
                logCategory.Created = DateTime.Now;
                logCategory.Category = categoryName;

                Context.LogCategories.InsertOnSubmit(logCategory);
                Context.SubmitChanges();
                return logCategory.LogCategoryId;
            }
        }

        /// <summary>
        /// Retrieves categoryId from the DB by the specified category name. Returns -1 if the category is not found.
        /// </summary>
        /// <param name="categoryName"></param>
        /// <param name="bCreateNewIfMissing"></param>
        /// <returns>category id (existing or newly created), or -1 if the category is not registered in the DB</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        private static int GetCategoryIdFromDB(String categoryName, bool bCreateNewIfMissing)
        {
            // use the light weight short-lived non-static data context to prevent concurrency problems
            using (MyMediaDataContext Context = new MyMediaDataContext(ConfigManager.DBConnectionString))
            {
                LogCategory logCategory = Context.LogCategories.FirstOrDefault<LogCategory>(c => c.Category == categoryName);
                if (logCategory != null)
                {
                    int logCategoryId = logCategory.LogCategoryId;
                    return logCategoryId;
                }

                if (bCreateNewIfMissing)
                {
                    return InsertNewLogCategoryIntoDB(categoryName);
                }
            }

            return -1;
        }

        /// <summary>
        /// Verifies if a logging category is enabled through the configuration 
        /// (all logging categories are disabled by default)
        /// </summary>
        /// <param name="categoryName">Name of the category</param>
        /// <returns>true if category is enabled(if logging actions for this category should be logged), false otherwise </returns>
        private static bool IsLogCategoryEnabled(String categoryName)
        {
            if (ConfigManager.PropertyValueExists(ConfigConstants.LogActionModuleName, categoryName))
            {
                ConfigProperty configProperty = ConfigManager.Modules[ConfigConstants.LogActionModuleName].Properties[categoryName];
                if (configProperty != null)
                {
                    if (EnabledConfigPropertyValueFlag.Equals(configProperty.Value))
                    {
                        return true;
                    }
                }
            }

            return false;
        }
    }
}
