﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core.Config
{
    /// <summary>
    /// Central entry point for handling(retrieving, dynamically updating) configuration of the application.
    /// </summary>
    public static partial class ConfigManager
    {
        private static Dictionary<String, ConfigModule> modules = new Dictionary<string,ConfigModule>();
        private static bool IsInitialized; // is set 'false' by the runtime
        private const String ConfigDirectoryName = "config";
        // there could be situation when configuration is loaded twice on concurrent access. Use the lock to avoid it.
        private static Object configurationDataLoadingLockSync = new Object();

        // object providing operations of loading from/saving to the concrete storage
        private static IConfigLoader configLoader;

        /// <summary>
        /// Configuration properties for modules.
        /// </summary>
        public static Dictionary<String, ConfigModule> Modules 
        {
            get
            {
                //TraceLog.Write(LogEventType.Verbose, "MyMediaProject.RecommenderSystem.Core.Config.ConfigManager: Modules object requested!");
                ConfigManager.Initialize();
                return modules;
            }
        }

        /// <summary>
        /// Reference to the default module (provided as a shortcut for convenience)
        /// </summary>
        public static SortedDictionary<String, ConfigProperty> DefaultModuleProperties
        {
            get
            {
                // Initialization will be performed by accessing Modules data member
                return Modules[ConfigConstants.DefaultModuleName].Properties;
            }
        }

        /// <summary>
        /// Loads configuration
        /// </summary>
        /// <param name="isReloadEnforced">if true, the configuration will be reloaded even if it has been loaded already</param>
        public static void Initialize(bool isReloadEnforced)
        {
            if (IsInitialized && !isReloadEnforced)
            {
                return;
            }

            String configDirectoryPath = "";
            try
            {
                configDirectoryPath = TryToGuessConfigDirectoryPath();
            }
            catch (OperationCanceledException)
            {
                TraceLog.Write(LogEventType.Error, "Initialize(): Configuration folder not specified by user. Throwing an exception.");
                throw;
            }
            Initialize(configDirectoryPath, isReloadEnforced);
        }

        /// <summary>
        /// Loads configuration, if not loaded yet
        /// </summary>
        /// <param name="configDirectoryPath">The path to the directory holding configuration files of various modules</param>
        /// <param name="isReloadEnforced"></param>
        public static void Initialize(String configDirectoryPath, bool isReloadEnforced)
        {
            if (IsInitialized && !isReloadEnforced)
            {
                return;
            }

            lock (configurationDataLoadingLockSync)
            {
                // ensure that the configuration hasn't been loaded while waiting for the lock release
                if (IsInitialized && !isReloadEnforced)
                {
                    return;
                }

                TraceLog.Write(LogEventType.Information, "MyMediaProject.RecommenderSystem.Core.Config.ConfigManager: Initializing ConfigManager with configDirectoryPath = '{0}'.", configDirectoryPath);

                configLoader = new XMLConfigLoader(configDirectoryPath);

                modules = configLoader.GetAllModulesDefaultProperties();
                Dictionary<String, ConfigModule> modifiedModules = configLoader.GetAllModulesModifiedProperties();

                // update modules' object data with the modified values, 
                // and set the DefaultValue to the one taken from the default configuration
                foreach (String moduleName in modules.Keys)
                {
                    foreach (String propertyName in modules[moduleName].Properties.Keys)
                    {
                        // set the default value to the one from the default configuration 
                        // (so that the value's modification could be determined easily by comparing to the default one)
                        modules[moduleName].Properties[propertyName].DefaultValue = modules[moduleName].Properties[propertyName].Value;

                        // if modified properties exist, rewrite default ones with the modified values
                        if (modifiedModules != null && modifiedModules.Count > 0)
                        {
                            // if this setting has been locally modified
                            if (modifiedModules.ContainsKey(moduleName)
                                && modifiedModules[moduleName].Properties.ContainsKey(propertyName))
                            {
                                modules[moduleName].Properties[propertyName].Value = modifiedModules[moduleName].Properties[propertyName].Value;
                            }
                        }
                    }

                    // add modified properties that exist in the modified version only (if any)
                    if (modifiedModules != null && modifiedModules.Count > 0 && modifiedModules.ContainsKey(moduleName))
                    {
                        foreach (String propertyName in modifiedModules[moduleName].Properties.Keys)
                        {
                            // if property is not in the resulting 'modules' object yet, add the property
                            if (!modules[moduleName].Properties.ContainsKey(propertyName))
                            {
                                modules[moduleName].Properties.Add(propertyName, modifiedModules[moduleName].Properties[propertyName]);
                                TraceLog.Write(LogEventType.Verbose, "Loading a property that exists in the modified version only: {0}", new object[] { modifiedModules[moduleName].Properties[propertyName] });
                            }
                        }
                    }
                }

                // are there any MODULES that exist in the local version only?
                if (modifiedModules != null){
                    foreach(String moduleName in modifiedModules.Keys){
                        if (modules.ContainsKey(moduleName))
                        {
                            continue;
                        }
                        TraceLog.Write(LogEventType.Verbose, "Adding a module that exists in the modified version only: {0}", new object[] { moduleName });
                        modules.Add(moduleName, modifiedModules[moduleName]);
                    }
                }

                IsInitialized = true;
            }
        }

        /// <summary>
        /// Private initialization method without forcing configuration reload
        /// </summary>
        private static void Initialize()
        {
            Initialize(false);
        }

        /// <summary>
        /// Private initialization method without forcing configuration reload
        /// </summary>
        /// <param name="configDirectoryPath"></param>
        private static void Initialize(String configDirectoryPath)
        {
            Initialize(configDirectoryPath, false);
        }

        /// <summary>
        /// Checks if the property with a given name has been configured for the specified module
        /// </summary>
        /// <param name="moduleName">Unique module's name that unifies set of properties</param>
        /// <param name="propertyName">Property name(unique within a module)</param>
        /// <returns>true - if property exists in the loaded configuration (even if its value is null), false otherwise</returns>
        public static bool PropertyValueExists(String moduleName, String propertyName)
        {
            ConfigManager.Initialize();

            if (!modules.ContainsKey(moduleName) || !modules[moduleName].Properties.ContainsKey(propertyName))
            {
                return false;
            }
            return true;
        }

        /// <summary>
        /// Checks if the property with a given name has been configured for the default module
        /// </summary>
        /// <param name="propertyNameInDefaultModule">Property name(unique within a module) of the default module</param>
        /// <returns>true - if property exists in the loaded configuration (even if its value is null), false otherwise</returns>
        public static bool PropertyValueExists(String propertyNameInDefaultModule)
        {
            ConfigManager.Initialize();

            if (!ConfigManager.modules.ContainsKey(ConfigConstants.DefaultModuleName) || !ConfigManager.modules[ConfigConstants.DefaultModuleName].Properties.ContainsKey(propertyNameInDefaultModule))
            {
                TraceLog.Write(LogEventType.Information, "MyMediaProject.RecommenderSystem.Core.Config.ConfigManager: PropertyValueExists({0}) returns false.", propertyNameInDefaultModule);
                return false;
            }
            return true;
        }

        /// <summary>
        /// Saves the settings of the individual module to the storage
        /// </summary>
        /// <param name="moduleName">The name of the module that unifies set of properties</param>
        public static void SaveModuleConfiguration(String moduleName)
        {
            if (!modules.ContainsKey(moduleName))
            {
                // module with the specified name does not exist
                return;
            }

            configLoader.SaveModuleModifiedConfiguration(modules[moduleName]);
        }

        /// <summary>
        /// Tries to find the location of the configuration folder according to the following sequence:
        ///     - previous path specified by user (via a folder browsing dialog)
        ///     - folder holding the binaries of the currently executed application 
        ///     - closest directory with the 'ConfigDirectoryName' in the folders hierarchical tree (walking up the tree from the binary folder)
        ///     - tries to ask user(via a folder browsing dialog) to specify the location explicitly (would save and use this location for later)
        ///  If folder is not specified by user explicitly in the last step, an OperationCanceledException is thrown.
        /// </summary>
        /// <returns>Path to the configuration folder</returns>
        private static String TryToGuessConfigDirectoryPath()
        {
            String Result;
            TraceLog.Write(LogEventType.Verbose, "Try to guess the configuration directory.");
            
            String previousPath = ApplicationConfigurationUtility.GetApplicationSetting("LastConfigurationFolderPath");
            if (!String.IsNullOrEmpty(previousPath) && Directory.Exists(previousPath))
            {
                TraceLog.Write(LogEventType.Verbose, "Using previously specified configuration directory path ('{0}').", new object[] { previousPath });
                return previousPath;
            }

            bool IsDirectoryFound = true;
            String BinDirPath = AppDomain.CurrentDomain.RelativeSearchPath;
            if(null == BinDirPath)
            {
                BinDirPath = Path.GetDirectoryName(Assembly.GetAssembly(typeof(ConfigManager)).Location);
            }
            TraceLog.Write(LogEventType.Information, "Using {0} as the root path to search for the configuration files.", BinDirPath);
            
            String configParentDirPath = BinDirPath;

            // walk up directories tree to find <ConfigDirectoryName> directory
            while (!Directory.Exists(Path.Combine(configParentDirPath, ConfigDirectoryName)))
            {
                DirectoryInfo configParentDirectory = Directory.GetParent(configParentDirPath);
                if (configParentDirectory == null || configParentDirectory.ToString().Length == 0)
                {
                    IsDirectoryFound = false;
                    break;
                }
                configParentDirPath = configParentDirectory.ToString();
            }

            if (!IsDirectoryFound)
            {
                // the directory has not been found, let the best guess be the assembly folder itself
                FolderSpecificationControl folderControl = new FolderSpecificationControl();

                Result = folderControl.AskConfigurationFolderPath(); // throws OperationCanceledException

                // Keep user's choice to use it next time
                ApplicationConfigurationUtility.SaveApplicationSetting("LastConfigurationFolderPath", Result);

                TraceLog.Write(LogEventType.Verbose, "User specified configuration folder location: '{0}'.", Result);
            }
            else
            {
                Result = Path.Combine(configParentDirPath, ConfigDirectoryName);
            }

            TraceLog.Write(LogEventType.Verbose, "Guessed location of the config folder : '{0}'.", Result);

            return Result;
        }

        /// <summary>
        /// Looks for the locally modified settings (as opposed to configuration files that 
        /// are up-to-date with the repository) and overwrites the already loaded configuration 
        /// with them. Should be called AFTER loading non-locally modified settings.
        /// </summary>
        private static void OverwriteConfigWithLocalSettings()
        {
            configLoader.GetAllModulesDefaultProperties();
        }
    }
}