﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MyMediaProject.RecommenderSystem.Core.Config
{
    /// <summary>
    /// The entity unifying a set of configuration properties by some structural, logical or conceptual criterion
    /// </summary>
    public class ConfigModule : IComparable<ConfigModule>
    {
        /// <summary>
        /// The display order index value that implies the module would be listed last by configuration utilities
        /// </summary>
        private const int LAST_DISPLAY_ORDER_INDEX = -1;

        /// <summary>
        /// The unique identificator of the module that is used to reference the module
        /// </summary>
        public String Name { get; set; }

        /// <summary>
        /// Human readable name of the module that can be used to identify the module to an admin user. Should be short to fit various UIs of the configuration utilities.
        /// </summary>
        public String ShortName { get; set; }

        /// <summary>
        /// Human understandable explanation of the module's concept
        /// </summary>
        public String Description { get; set; }
        /// <summary>
        /// Could be used to display modules in a certain order. The lower the value(starting with '1') the earlier the module would be listed.
        /// Can be empty. The values '0', '', LAST_DISPLAY_ORDER_INDEX  would mean the module should be listed last.
        /// </summary>
        public int DisplayOrderIndex { get; set; }

        /// <summary>
        /// Set of properties that is unified by this module
        /// </summary>
        private SortedDictionary<String, ConfigProperty> properties = new SortedDictionary<String, ConfigProperty>();

        /// <summary>
        /// Set of properties that is unified by this module
        /// </summary>
        public SortedDictionary<String, ConfigProperty> Properties
        {
            get
            {
                return properties;
            }
        }

        /// <summary>
        /// Configurational module's data holder
        /// </summary>
        /// <param name="name">The (textual)ID of the moduleName(could be e.g. full class name). Can contain spaces if needed.</param>
        /// <param name="shortName">The human friendly short name of the moduleName. Can be listed as the name of the configurational section in GUI. Can contain spaces if needed. Should be unique to avoid user confusions.</param>
        /// <param name="description">Relatively short human friendly description of the configurational moduleName.</param>
        /// <param name="displayOrderIndex">An integer, specifying the order in which the module should be presented to the admin user (by utilities with UI)</param>
        public ConfigModule(String name, String shortName, String description, int displayOrderIndex)
        {
            this.Name = name;
            this.ShortName = shortName;
            this.Description = description;
            this.DisplayOrderIndex = displayOrderIndex;
            if (this.DisplayOrderIndex < 1)
            {
                this.DisplayOrderIndex = LAST_DISPLAY_ORDER_INDEX;
            }
        }

        /// <summary>
        /// Configurational module's data holder (when display order is insignificant (will be listed last))
        /// </summary>
        /// <param name="name">The (textual)ID of the moduleName(could be e.g. full class name). Can contain spaces if needed.</param>
        /// <param name="shortName">The human friendly short name of the moduleName. Can be listed as the name of the configurational section in GUI. Can contain spaces if needed. Should be unique to avoid user confusions.</param>
        /// <param name="description">Relatively short human friendly description of the configurational moduleName.</param>
        public ConfigModule(String name, String shortName, String description)
        {
            this.Name = name;
            this.ShortName = shortName;
            this.Description = description;
            this.DisplayOrderIndex = LAST_DISPLAY_ORDER_INDEX; // L
            if (this.DisplayOrderIndex < 1)
            {
                this.DisplayOrderIndex = LAST_DISPLAY_ORDER_INDEX;
            }
        }

        /// <summary>
        /// Configurational module's data holder
        /// </summary>
        /// <param name="name">The (textual)ID of the moduleName(could be e.g. full class name). Can contain spaces if needed.</param>
        /// <param name="shortName">The human friendly short name of the moduleName. Can be listed as the name of the configurational section in GUI. Can contain spaces if needed. Should be unique to avoid user confusions.</param>
        /// <param name="description">Relatively short human friendly description of the configurational moduleName.</param>
        /// <param name="displayOrderIndex">An integer, specifying the order in which the module should be presented to the admin user (by utilities with UI)</param>
        /// <param name="modulesProperties">The properties object to set for the module</param>
        public ConfigModule(String name, String shortName, String description, int displayOrderIndex, SortedDictionary<String, ConfigProperty> modulesProperties)
        {
            this.Name = name;
            this.ShortName = shortName;
            this.Description = description;
            this.DisplayOrderIndex = displayOrderIndex;
            if (this.DisplayOrderIndex < 1)
            {
                this.DisplayOrderIndex = LAST_DISPLAY_ORDER_INDEX;
            }
            this.properties = modulesProperties;
        }

        /// <summary>
        /// Adds property into the module's object in the memory
        /// </summary>
        /// <param name="property">ConfigProperty to be added to the module's object in the memory</param>
        public void AddOrUpdateProperty(ConfigProperty property)
        {
            if (Properties.ContainsKey(property.Name))
            {
                this.Properties[property.Name] = property;
            }
            else
            {
                this.Properties.Add(property.Name, property);
            }
        }

        /// <summary>
        /// Sets property's value to its default one
        /// </summary>
        /// <param name="propertyName"></param>
        public void ResetPropertyToDefault(String propertyName)
        {
            if (!Properties.ContainsKey(propertyName))
            {
                return;
            }
            Properties[propertyName].Value = Properties[propertyName].DefaultValue;
        }


        /// <summary>
        /// Needed to order modules by their displayOrderIndex in ascending order: 1..XX. With one exception:
        /// if displayOrderIndex=LAST_DISPLAY_ORDER_INDEX, the module must be the last one (ordered alphabetically, if several modules must be the last ones).
        /// </summary>
        /// <param name="module1">The module1.</param>
        /// <param name="module2">The module2.</param>
        /// <returns></returns>
        public static int Compare(ConfigModule module1, ConfigModule module2)
        {
            if (module1.Equals(module2))
            {
                return 0;
            }

            // if one of the indecies is not set
            if (module1.DisplayOrderIndex < 1)
            {
                // if both indecies are not set, sort alphabetically
                if (module2.DisplayOrderIndex < 1)
                {
                    return string.Compare(module1.ShortName, module2.ShortName, StringComparison.Ordinal);
                }

                return 1;
            }

            if (module2.DisplayOrderIndex < 1)
            {
                return -1;
            }

            int result = module1.DisplayOrderIndex.CompareTo(module2.DisplayOrderIndex);
            if (result == 0)
            {
                // sort alphabetically
                return string.Compare(module1.ShortName, module2.ShortName, StringComparison.Ordinal);
            }
            return result;
        }

        /// <summary>
        /// Alias for calling Compare(ConfigModule m1, ConfigModule m2) method
        /// </summary>
        /// <param name="other">The module to be compared to this instance of the ConfigModule</param>
        /// <returns></returns>
        public int CompareTo(ConfigModule other)
        {
            return Compare(this, other);
        }
    }
}
