﻿using System;
using System.Globalization;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.IO;
using System.Reflection;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core.Config
{
    /// <summary>
    /// Configuration loader that stores config settings for various modules in XML files.
    /// </summary>
    class XMLConfigLoader : IConfigLoader
    {
        /// <summary>
        /// Pattern of the XML configuration files names that will be loaded automatically, 
        /// once placed in to the specified directory
        /// </summary>
        public const String DefaultConfigurationFilesExtension = "mmconf.xml";
        public const String ModifiedConfigurationFilesExtension = "mmconf.local.xml";

        protected const string XML_ROOT_ELEMENT = "ModuleConfig";
        protected const string XML_ELEMENT_MODULE_INFO = "ModuleInfo";
        protected const string XML_MODULE_ATTRIBUTE_SHORT_NAME = "ShortName";
        protected const string XML_MODULE_ATTRIBUTE_DESCRIPTION = "Description";
        protected const string XML_MODULE_ATTRIBUTE_DISPLAY_ORDER_INDEX = "DisplayOrderIndex";
        protected const string XML_ELEMENT_PROPERTIES_LIST = "PropertiesList";

        protected const string xPathShortName = "/ModuleConfig/ModuleInfo/ShortName";
        protected const string xPathMDescription = "/ModuleConfig/ModuleInfo/Description";
        protected const string xPathMDisplayOrderIndex = "/ModuleConfig/ModuleInfo/DisplayOrderIndex";
        protected const string xPathModulesProperties = "/ModuleConfig/PropertiesList/Property";

        protected const string XML_ELEMENT_PROPERTY = "Property";
        protected const string XML_PROPERTY_ATTRIBUTE_NAME = "Name";
        protected const string XML_PROPERTY_ATTRIBUTE_VALUE = "Value";
        protected const string XML_PROPERTY_ATTRIBUTE_DESCRIPTION = "Description";
        protected const string XML_PROPERTY_ATTRIBUTE_PRIVATE_VALUE = "PrivateValue";

        private const string BREAK_LINE_CHARACTERS = "\r\n";
        private const string TAB_CHARACTER = "\t";

        protected String ConfigDirectoryPath;

        protected Dictionary<String, ConfigModule> Modules;

        /// <summary>
        /// Constructor 
        /// </summary>
        /// <param name="_configDirectoryPath">Path to the directory that holds config files</param>
        public XMLConfigLoader(String _configDirectoryPath)
        {
            ConfigDirectoryPath = _configDirectoryPath;
        }

        /// <summary>
        /// Returns absolute path to the configuration file, specified by module and configuration files' extension
        /// </summary>
        /// <param name="module"></param>
        /// <param name="configurationFilesExtension"></param>
        /// <returns></returns>
        private String GetConfigurationFilePath(ConfigModule module, String configurationFilesExtension)
        {
            return ConfigDirectoryPath + "/" + module.Name + "." + configurationFilesExtension;
        }

        /// <summary>
        /// Saves modified configuration of the specified module
        /// </summary>
        /// <param name="configModule">The configuration module object to be saved</param>
        public void SaveModuleModifiedConfiguration(ConfigModule configModule)
        {
            String xmlModulesConfigPath = GetConfigurationFilePath(configModule, ModifiedConfigurationFilesExtension);
            TraceLog.Write(LogEventType.Information, "MyMediaProject.RecommenderSystem.Core.Config.XMLConfigLoader:SaveModuleModifiedConfiguration(): Saving module '{0}' configuration to the file '{1}'.", configModule.Name, xmlModulesConfigPath);

            if (!IsModifiedPropertyPresent(configModule))
            {
                // delete the file, if exists
                if (File.Exists(xmlModulesConfigPath))
                {
                    File.Delete(xmlModulesConfigPath);
                    TraceLog.Write(LogEventType.Information, "XMLConfigLoader:SaveModuleModifiedConfiguration(): Deleting configuration file '{0}' (no modified properties exist).", xmlModulesConfigPath);
                }
                return;
            }

            // there are modified properties, save them into the configuration file
            XmlTextWriter textWriter = new XmlTextWriter(xmlModulesConfigPath, null);

            // Opens the document
            textWriter.WriteStartDocument();

            // Write first element
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS);
            textWriter.WriteStartElement(XML_ROOT_ELEMENT);
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER);
            // Write module's header
            textWriter.WriteStartElement(XML_ELEMENT_MODULE_INFO);
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER + TAB_CHARACTER);

            textWriter.WriteElementString(XML_MODULE_ATTRIBUTE_SHORT_NAME, configModule.ShortName);
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER + TAB_CHARACTER);
            textWriter.WriteElementString(XML_MODULE_ATTRIBUTE_DESCRIPTION, configModule.Description);
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER + TAB_CHARACTER);
            textWriter.WriteElementString(XML_MODULE_ATTRIBUTE_DISPLAY_ORDER_INDEX, configModule.DisplayOrderIndex.ToString(new NumberFormatInfo()));
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER);
            textWriter.WriteEndElement();
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER);

            // Write properties
            textWriter.WriteStartElement(XML_ELEMENT_PROPERTIES_LIST);
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER);

            foreach (ConfigProperty property in configModule.Properties.Values)
            {
                // save the settings that differ from the default ones only (or settings that are programmatically generated).
                if (!property.IsChanged && property.PrivateValue != null && property.PrivateValue.Length > 0)
                {
                    continue;
                }

                textWriter.WriteWhitespace(TAB_CHARACTER);
                textWriter.WriteStartElement(XML_ELEMENT_PROPERTY);
                textWriter.WriteAttributeString(XML_PROPERTY_ATTRIBUTE_NAME, property.Name);
                textWriter.WriteAttributeString(XML_PROPERTY_ATTRIBUTE_VALUE, property.Value);
                textWriter.WriteAttributeString(XML_PROPERTY_ATTRIBUTE_DESCRIPTION, property.Description);
                textWriter.WriteAttributeString(XML_PROPERTY_ATTRIBUTE_PRIVATE_VALUE, property.PrivateValue);
                
                textWriter.WriteEndElement();
                textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS + TAB_CHARACTER);
            }
            textWriter.WriteEndElement();
            textWriter.WriteWhitespace(BREAK_LINE_CHARACTERS);
            textWriter.WriteEndElement();

            textWriter.WriteEndDocument();

            // close writer
            textWriter.Close();
        }


        /// <summary>
        /// Determines if the gievn module holds any modified properties 
        /// </summary>
        /// <param name="configModule"></param>
        /// <returns></returns>
        public static bool IsModifiedPropertyPresent(ConfigModule configModule)
        {
            foreach (ConfigProperty property in configModule.Properties.Values)
            {
                if (property.IsChanged)
                {
                    return true;
                }
            }

            // no modified properties found
            return false;
        }

        /// <summary>
        /// Reads in, parses and returns modified configuration for all modules stored 
        /// in the XML files(according to the file DefaultConfigurationFilesExtension) of the specified directory
        /// </summary>
        /// <returns></returns>
        public Dictionary<String, ConfigModule> GetAllModulesModifiedProperties()
        {
            return GetAllModulesPropertiesByFileExtension(ModifiedConfigurationFilesExtension);
        }

        /// <summary>
        /// Reads in, parses and returns default configuration for all modules stored 
        /// in the XML files(according to the file DefaultConfigurationFilesExtension) of the specified directory
        /// </summary>
        /// <returns></returns>
        public Dictionary<String, ConfigModule> GetAllModulesDefaultProperties()
        {
            return GetAllModulesPropertiesByFileExtension(DefaultConfigurationFilesExtension);
        }


        /// <summary>
        /// Reads in, parses and returns configuration for all modules stored 
        /// in the XML files(according to the specified configuration files extension pattern ) of the specified directory
        /// </summary>
        /// <param name="configurationFilesExtension">The configuration files extension (e.g. "xml")</param>
        /// <returns></returns>
        private Dictionary<String, ConfigModule> GetAllModulesPropertiesByFileExtension(String configurationFilesExtension)
        {
            Modules = new Dictionary<String, ConfigModule>();
            String ConfigFilesExtensionPattern = "*." + configurationFilesExtension;
            TraceLog.Write(LogEventType.Information, "MyMediaProject.RecommenderSystem.Core.Config.XMLConfigLoader:GetAllModulesPropertiesByFileExtension(): Loading all modules' properties");
            

            DirectoryInfo di = new DirectoryInfo(ConfigDirectoryPath);
            FileInfo[] arrFiles = di.GetFiles(ConfigFilesExtensionPattern);

            foreach (FileInfo fileInfo in arrFiles)
            {
                String moduleName = fileInfo.Name.Substring(0, (int)(fileInfo.Name.Length - ConfigFilesExtensionPattern.Length + 1));

                XmlDocument xDoc = new XmlDocument();
                String filePath = fileInfo.FullName;
                TraceLog.Write(LogEventType.Information, "MyMediaProject.RecommenderSystem.Core.Config.XMLConfigLoader:GetAllModulesPropertiesByFileExtension(): Loading configuration from the file: '{0}'.", filePath);

                try
                {
                    xDoc.Load(filePath);

                    XmlNamespaceManager nsMgr = new XmlNamespaceManager(new NameTable());

                    String moduleShortName = (xDoc.SelectSingleNode(xPathShortName, nsMgr) as XmlElement).InnerText;
                    String moduleDescription = (xDoc.SelectSingleNode(xPathMDescription, nsMgr) as XmlElement).InnerText;
                    String moduleDisplayOrderIndex = (xDoc.SelectSingleNode(xPathMDisplayOrderIndex, nsMgr) as XmlElement).InnerText;
                    int iDisplayOrderIndex = 0;
                    if (moduleDisplayOrderIndex != null && moduleDisplayOrderIndex.Length > 0)
                    {
                        iDisplayOrderIndex = Convert.ToInt32(moduleDisplayOrderIndex, new NumberFormatInfo());
                    }

                    // read in module's list of properties
                    SortedDictionary<String, ConfigProperty> modulesProperties = new SortedDictionary<String, ConfigProperty>();

                    XmlNodeList nodesList = xDoc.SelectNodes(xPathModulesProperties, nsMgr);
                    foreach (XmlNode propertyNode in nodesList)
                    {
                        String PropertyName = propertyNode.Attributes.GetNamedItem(XML_PROPERTY_ATTRIBUTE_NAME).InnerText;
                        String PropertyValue = propertyNode.Attributes.GetNamedItem(XML_PROPERTY_ATTRIBUTE_VALUE).InnerText;
                        String PropertyDescription = propertyNode.Attributes.GetNamedItem(XML_PROPERTY_ATTRIBUTE_DESCRIPTION).InnerText;
                        XmlNode privateValueNode = propertyNode.Attributes.GetNamedItem(XML_PROPERTY_ATTRIBUTE_PRIVATE_VALUE);
                        String PropertyPrivateValue = "";
                        if (privateValueNode != null){
                            PropertyPrivateValue = privateValueNode.InnerText;
                        }
                        
                        if (PropertyName == null || PropertyName.Length == 0)
                        {
                            continue;
                        }
                        ConfigProperty property = new ConfigProperty(PropertyName, PropertyValue, PropertyDescription);
                        property.PrivateValue = PropertyPrivateValue;
                        if (modulesProperties.ContainsKey(PropertyName))
                        {
                            modulesProperties[PropertyName] = property;
                            TraceLog.Write(LogEventType.Error, "MyMediaProject.RecommenderSystem.Core.Config.XMLConfigLoader:GetAllModulesPropertiesByFileExtension(): Error trying to load configuration from the file: '{0}'. The property '{1}' is duplicated.", new object[]{filePath, PropertyName});
                        }
                        else
                        {
                            modulesProperties.Add(PropertyName, property);
                        }
                    }

                    Modules.Add(moduleName, new ConfigModule(moduleName, moduleShortName, moduleDescription, iDisplayOrderIndex, modulesProperties));
                }
                catch (NullReferenceException ex)
                {
                    // if there is a configuration file declared incorrectly, just skip it
                    TraceLog.Write(LogEventType.Error, "MyMediaProject.RecommenderSystem.Core.Config.XMLConfigLoader:GetAllModulesPropertiesByFileExtension(): Error trying to load configuration from the file: '{0}'. Exception caught:{1}.", filePath, ex.Message);
                    continue;
                }
            }

            return Modules;
        }
    }
}
