﻿using System.Collections.ObjectModel;
using System.Collections.Generic;
using System.Linq;
using System;
using EMIC.Design;
using System.Globalization;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// This class is internally used by the recommender class. It
    /// implements a broker object that can collect context status information from multiple 
    /// context sources and forward them to the registered context sinks according to the
    /// context status types they have registered for.
    /// </summary>
    class ContextBroker: IContextBroker
    {
        private Dictionary<RecommenderId, ContextSource> contextSources;
        private Dictionary<RecommenderId, IContextSink> contextSinks;

        internal ContextBroker()
        {
            contextSources = new Dictionary<RecommenderId, ContextSource>();
            contextSinks = new Dictionary<RecommenderId, IContextSink>();
        }

        internal void AddContextSource(ContextSource contextSource)
        {
            lock (contextSources)
            {
                if (contextSources.ContainsKey(contextSource.Id))
                {
                    throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, "Context source already in the list: id {0}", contextSource.Id));
                }

                contextSource.SetContextBroker(this);
                contextSources.Add(contextSource.Id, contextSource);
                TraceLog.Write(LogEventType.Verbose, "Registered new context source: id {0}", contextSource.Id);
            }
        }

        internal void RemoveContextSource(ContextSource contextSource)
        {
            lock (contextSources)
            {
                if (!contextSources.ContainsKey(contextSource.Id))
                {
                    throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, "Context source cannot be removed, because it is currently not in the list: id {0}", contextSource.Id));
                }

                contextSource.SetContextBroker(null);
                contextSources.Remove(contextSource.Id);
                TraceLog.Write(LogEventType.Verbose, "Unregistered context source: id {0}", contextSource.Id);
            }
        }

        internal void AddContextSink(IContextSink contextSink)
        {
            lock (contextSinks)
            {
                if (contextSinks.ContainsKey(contextSink.Id))
                {
                    throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, "Context sink already in the list: id {0}", contextSink.Id));
                }

                contextSinks.Add(contextSink.Id, contextSink);
                TraceLog.Write(LogEventType.Verbose, "Registered new context sink: id {0}", contextSink.Id);
            }
        }

        internal void RemoveContextSink(IContextSink contextSink)
        {
            lock (contextSinks)
            {
                if (!contextSinks.ContainsKey(contextSink.Id))
                {
                    throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, "Context sink cannot be removed, because it is currently not in the list: id {0}", contextSink.Id));
                }

                contextSinks.Remove(contextSink.Id);
                TraceLog.Write(LogEventType.Verbose, "Unregistered context sink: id {0}", contextSink.Id);
            }
        }

        #region IContextBroker Members

        public void ContextStatusChanged(ContextStatus contextStatus)
        {
            lock (contextSinks)
            {
                foreach (IContextSink contextSink in contextSinks.Values)
                {
                    if (contextSink.HandledContextTypes().Contains(contextStatus.GetType()))
                    {
                        contextSink.ReceiveContextStatus(contextStatus);
                    }
                }
            }
        }

        #endregion
    }
}
