﻿namespace MyMediaProject.RecommenderSystem.Core
{
    using System;
    using System.Text;
    using System.Collections.Generic;

    /// <summary>
    /// User registered in an external system (e.g. external social network)
    /// </summary>
    partial class ExternalSystemUser
    {
        /// <summary>
        /// First name of the external user (NB! Is volatile: not stored in the DB because of the external privacy policies)
        /// </summary>
        public String FirstName { get; set; }

        /// <summary>
        /// Last name of the external user (NB! Is volatile: not stored in the DB because of the external privacy policies)
        /// </summary>
        public String LastName { get; set; }

        /// <summary>
        /// Display name to be shown to an end-user
        /// </summary>
        /// <returns></returns>
        public String DisplayName
        {
            get
            {
                // if we know the first name
                if (!String.IsNullOrEmpty(this.FirstName))
                {
                    String Result = this.FirstName;
                    if (!String.IsNullOrEmpty(this.LastName))
                    {
                        Result += " " + this.LastName;
                    }

                    return Result;
                }
                else // if first name not shown, nothing better to do than showing an external ID
                {
                    return this.ExternalSystem.Name + " user '" + this.ExternalSystemUserId + "'";
                }
            }
        }

        /// <summary>
        /// Textual representation of the user for tracing purposes
        /// </summary>
        /// <returns></returns>
        public override String ToString()
        {
            String Result = "";

            // objects might be not initialized entirely, avoid NullPointerException-s
            if (!String.IsNullOrEmpty(this.FirstName))
            {
                Result = this.FirstName + " " + this.LastName + " ";
            }

            Result += "(";
            if (this.ExternalSystem != null)
            {
                Result += this.ExternalSystem.Name + " ";
            }

            if (this.ExternalSystemUserId != default(int)){
                Result += "Id=" + this.ExternalSystemUserId;
            }

            if (!String.IsNullOrEmpty(this.UserExternalIdentifier))
            {
                Result += " UserExternalIdentifier=" + this.UserExternalIdentifier;
            }

            if (this.LocalUserId != null)
            {
                Result += " LocalMyMedia user='" + this.LocalUserId + "'";
            }

            Result += ")";

            return Result;
        }

        /// <summary>
        /// Static method to get serialized IDs(ExternalUserIdentifier) of the list of ExternalSystemUser objects conveniently
        /// </summary>
        /// <returns>serialized list of external IDs separated by given separator</returns>
        public static String SerializeIds(ICollection<ExternalSystemUser> externalUsersList, String separator)
        {
            String result = "";
            foreach (ExternalSystemUser extUser in externalUsersList)
            {
                if (String.IsNullOrEmpty(result))
                {
                    result += separator;
                }
                result += extUser.UserExternalIdentifier;
            }
            return result;
        }

    }


    /// <summary>
    /// Externally registered grouping of users
    /// </summary>
    partial class ExternalGroup
    {
        /// <summary>
        /// /Displays data about the external group in the convenient for tracing and logging format.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            String result = "ExternalGroup (id=" + this.ExternalGroupId + ", ExternalSystem='"
                + this.ExternalSystem.Name + "', ExternalIdentifier = '" + this.GroupExternalIdentifier + "'";
            if (this.Name != null && Name.Length > 0)
            {
                result += ", Name='" + this.Name + "'";
            }
            result += ").";

            return result;
        }
    }


    /// <summary>
    /// Integration Component that provides connectivity to certain external system
    /// </summary>
    public partial class ExternalSystemIntegrator
    {
        /// <summary>
        /// The name of the config property, specifying external system's name of the integration component
        /// </summary>
        public const String ConfigProperty_ExternalSystemName = "ExternalSystemName";

    }

    /// <summary>
    /// External System (like an external social network)
    /// </summary>
    public partial class ExternalSystem
    {
        /// <summary>
        /// The name of the config property, specifying external system's name
        /// </summary>
        public const String ConfigProperty_Name = "Name";


        /// <summary>
        /// The name of the config property, specifying external system's URL to its logo
        /// </summary>
        public const String ConfigProperty_UrlToLogo = "UrlToLogo";

        /// <summary>
        /// The name of the config property, specifying external system's URL to the start page
        /// </summary>
        public const String ConfigProperty_UrlToStartPage = "UrlToStartPage";


        /// <summary>
        /// Updates data of the feature if it differs from the data passed
        /// </summary>
        /// <param name="urlToLogo"></param>
        /// <param name="urlToStartPage"></param>
        /// <returns></returns>
        public bool UpdateObjectIfNeeded(String urlToLogo, String urlToStartPage)
        {
            // if data differs
            if (!urlToLogo.Equals(this.UrlToLogo) || !urlToStartPage.Equals(this.UrlToStartPage))
            {
                this.UrlToLogo = urlToLogo;
                this.UrlToStartPage = urlToStartPage;

                return true;
            }
            return false;
        }
    }

    /// <summary>
    /// General feature that could be provided by multiple integration components
    /// </summary>
    partial class ExternalIntegrationFeature
    {
        /// <summary>
        /// Updates data of the feature if it differs from the data passed
        /// </summary>
        /// <param name="description"></param>
        /// <param name="isVisible"></param>
        /// <returns>true - if data has been updated, false - if passed data didn't differ from what's in the DB.</returns>
        public bool UpdateObjectIfNeeded(String description, bool isVisible)
        {
            // if data differs
            if (!description.Equals(this.Description) || isVisible != this.IsVisible)
            {
                this.Description = description;
                this.IsVisible = isVisible;

                return true;
            }
            return false;
        }
    }
}