﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;

using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// The base for a recommendation algorithm.
    /// </summary>
    /// <remarks>
    /// All recommendation algorithms are derived either from this base class or from <see cref="PluggableEngine&lt;TCatalogItemType, TUserActionType&gt;"/>. Whereas PluggableEngine
    /// allows you to construct an engine from the individual parts (<see cref="IPredictionAlgorithm&lt;TCatalogItemType&gt;"/>, <see cref="INewCatalogItemHandler&lt;TCatalogItemType&gt;"/> and
    /// <see cref="INewUserActionHandler&lt;TCatalogItemType, TUserActionType&gt;"/>), this class requires you to implement their functionality in resp.
    /// <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.Predict"/>, <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewCatalogItems"/> and <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewFeedback"/>.
    /// The framework ensures that engines will always receive all catalog item and feedback changes, even if they have been off-line
    /// for some time. As soon as they are part of any flow, they will receive all (missed) update changes.
    /// Engines must be added to a <see cref="Flow&lt;TCatalogItemType, TUserActionType&gt;"/>. The flow is then used to generate predictions. All engines that are part of a flow will
    /// also be called in case of catalog updates (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewCatalogItems"/>) and if feedback is received (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewFeedback"/>).
    /// </remarks>
    public abstract class Engine : Engine<CatalogItem, UserAction>
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="engineId">The engine id.</param>
        /// <param name="description">The description.</param>
        protected Engine(EngineId engineId, string description)
            : base(engineId, description)
        {
        }
    }
    /// <summary>
    /// The base for a recommendation algorithm.
    /// </summary>
    /// <typeparam name="TCatalogItemType">The CatalogItem type (must be derived from <see cref="CatalogItem"/>.</typeparam>
    /// <typeparam name="TUserActionType">The IUserAction type (must be derived from <see cref="UserAction"/>.</typeparam>
    /// <remarks>
    /// All recommendation algorithms are derived either from this base class or from <see cref="PluggableEngine&lt;TCatalogItemType, TUserActionType&gt;"/>. Whereas PluggableEngine
    /// allows you to construct an engine from the individual parts (<see cref="IPredictionAlgorithm&lt;TCatalogItemType&gt;"/>, <see cref="INewCatalogItemHandler&lt;TCatalogItemType&gt;"/> and
    /// <see cref="INewUserActionHandler&lt;TCatalogItemType, TUserActionType&gt;"/>), this class requires you to implement their functionality in resp.
    /// <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.Predict"/>, <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewCatalogItems"/> and <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewFeedback"/>.
    /// The framework ensures that engines will always receive all catalog item and feedback changes, even if they have been off-line
    /// for some time. As soon as they are part of any flow, they will receive all (missed) update changes.
    /// Engines must be added to a <see cref="Flow&lt;TCatalogItemType, TUserActionType&gt;"/>. The flow is then used to generate predictions. All engines that are part of a flow will
    /// also be called in case of catalog updates (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewCatalogItems"/>) and if feedback is received (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewFeedback"/>).
    /// </remarks>
    public abstract class Engine<TCatalogItemType, TUserActionType>
        where TCatalogItemType : CatalogItem
        where TUserActionType : UserAction
    {
        private EngineId engineId;
        private string description;

        private ICatalogItemProvider<TCatalogItemType> iCatalogItemProvider;
        private IUserActionProvider<TCatalogItemType, TUserActionType> iUserActionProvider;
        private string dbConnectionString;
        private bool registered;

        /// <summary>
        /// Database connection string for direct access to the MyMedia data storage database.
        /// </summary>
        protected string DatabaseConnectionString
        {
            get { return dbConnectionString; }
            set { dbConnectionString = value; }
        }

        /// <summary>
        /// Occurs when engine providers are registered.
        /// </summary>
        protected event EventHandler EngineProvidersRegistered;

        /// <summary>
        /// Occurs when engine providers are unregistered.
        /// </summary>
        protected event EventHandler EngineProvidersUnregistered;

        /// <summary>
        /// Do not use this; for inheritance only.
        /// </summary>
        protected Engine() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="engineId">The engine id.</param>
        /// <param name="description">The description.</param>
        protected Engine(EngineId engineId, string description)
        {
            this.engineId = engineId;
            this.description = description;
        }

        /// <summary>
        /// Gets the id.
        /// </summary>
        /// <value>The id.</value>
        public EngineId Id 
        {
            get
            {
                return this.engineId;
            }
        }

        /// <summary>
        /// Gets the description.
        /// </summary>
        /// <value>The description.</value>
        public string Description
        {
            get
            {
                return this.description;
            }
        }

        /// <summary>
        /// Internal use only. Please ignore.
        /// </summary>
        public RecommenderEventHandler NewUserActionEventHandler
        {
            get
            {
                return this.InternalHandleNewUserAction;
            }
        }

        private void InternalHandleNewUserAction(EventArgs args)
        {
            Contract.Precondition(this.registered, "Engine isn't registered with a provider.");

            StoreChangedEventArgs scargs = args as StoreChangedEventArgs;

            if (scargs != null)
            {
                InternalHandleNewUserAction(scargs.NewChangeNumber);
            }
        }

        private void InternalHandleNewUserAction(int changeNumber)
        {
            List<TUserActionType> feedback = new List<TUserActionType>();

            foreach (IDataStoreChange sc in this.iUserActionProvider.GetFeedbackChanges(changeNumber, changeNumber))
            {
                if (sc.ChangeType == StoreChangeType.Added)
                {
                    TUserActionType feedbackItem;
                    TCatalogItemType catalogItem;

                    this.iUserActionProvider.RetrieveUserAction(sc.ChangedItemId, out feedbackItem, out catalogItem);

                    feedback.Add(feedbackItem);
                }
            }

            this.HandleNewFeedback(new ReadOnlyCollection<TUserActionType>(feedback));
        }

        /// <summary>
        /// Retrieves previous (in time) user specific feedback.
        /// </summary>
        /// <param name="userId">The user id.</param>
        /// <param name="maxHistoryLength">The maximum amount of feedback items to return</param>
        /// <returns>The feedback</returns>
        protected IEnumerable<TUserActionType> RetrieveLastFeedback(User userId, int maxHistoryLength)
        {
            Contract.Precondition(this.registered, "Engine isn't registered with a provider.");

            return this.iUserActionProvider.RetrieveLastFeedback(userId, maxHistoryLength);
        }

        /// <summary>
        /// Resets the local user settings of the engine.
        /// </summary>
        public void ResetLocalUserSettings()
        {
            this.InternalHandleNewCatalogItems(this.iCatalogItemProvider.ChangeNumber);
            this.InternalHandleNewUserAction(this.iUserActionProvider.ChangeNumber);
        }

        /// <summary>
        /// Resets all settings.
        /// </summary>
        public void ResetAllSettings()
        {
            this.InternalHandleNewCatalogItems(this.iCatalogItemProvider.ChangeNumber);
            this.InternalHandleNewUserAction(this.iUserActionProvider.ChangeNumber);
        }

        /// <summary>
        /// Internal use only. Please ignore.
        /// </summary>
        /// <value>The new catalog item event handler.</value>
        public RecommenderEventHandler NewCatalogItemEventHandler
        {
            get
            {
                return this.InternalHandleNewCatalogItems;
            }
        }

        private void InternalHandleNewCatalogItems(EventArgs args)
        {
            Contract.Precondition(this.registered, "Engine isn't registered with a provider.");

            StoreChangedEventArgs scargs = args as StoreChangedEventArgs;

            if (scargs != null)
            {
                InternalHandleNewCatalogItems(scargs.NewChangeNumber);
            }
        }

        private void InternalHandleNewCatalogItems(int newChangeNumber)
        {
            List<TCatalogItemType> items = new List<TCatalogItemType>();

            foreach (IDataStoreChange sc in this.iCatalogItemProvider.GetCatalogChanges(newChangeNumber, newChangeNumber))
            {
                if (sc.ChangeType == StoreChangeType.Added)
                {
                    TCatalogItemType catalogItem;

                    this.iCatalogItemProvider.GetCatalogItem(sc.ChangedItemId, out catalogItem);
                    items.Add(catalogItem);
                }
            }

            this.HandleNewCatalogItems(new ReadOnlyCollection<TCatalogItemType>(items));
        }

        internal void RegisterProviders(
            ICatalogItemProvider<TCatalogItemType> catalogItemProvider,
            IUserActionProvider<TCatalogItemType, TUserActionType> userActionProvider,
            string connectionString)
        {
            Contract.Precondition(this.registered == false, "Engine still registered. Please unregister first.");

            this.iCatalogItemProvider = catalogItemProvider;
            this.iUserActionProvider = userActionProvider;
            this.dbConnectionString = connectionString;

            this.registered = true;

            EventHandler handler = this.EngineProvidersRegistered;
            if (null != handler)
            {
                handler.Invoke(this, new EventArgs());
            }
        }

        internal void UnregisterProviders()
        {
            Contract.Precondition(this.registered, "Engine not registered.");

            EventHandler handler = this.EngineProvidersUnregistered;
            if (null != handler)
            {
                handler.Invoke(this, new EventArgs());
            }

            this.iCatalogItemProvider = null;
            this.iUserActionProvider = null;
            this.dbConnectionString = string.Empty;

            this.registered = false;
        }

        /// <summary>
        /// Handles new catalog items.
        /// </summary>
        /// <remarks>
        /// This method can be used for engines that need to update their settings (e.g. do statistical analysis on the items meta-data)
        /// for new catalog items. 
        /// </remarks>
        /// <param name="catalogItems">The catalog items.</param>
        protected abstract void HandleNewCatalogItems(ReadOnlyCollection<TCatalogItemType> catalogItems);

        /// <summary>
        /// Handles new feedback.
        /// </summary>
        /// This method can be used to process feedback that comes from the user (explicit) or from the system (implicit).
        /// <param name="feedback">The feedback.</param>
        protected abstract void HandleNewFeedback(ReadOnlyCollection<TUserActionType> feedback);

        /// <summary>
        /// Calculates a prediction (recommendation) for each catalog item in the query.
        /// </summary>
        /// <param name="query">The query.</param>
        /// <param name="user">The user.</param>
        /// <returns>a list of predictions.</returns>
        /// <remarks>
        /// This forms the heart of the recommendation algorithm. It has to generate a prediction for each catalog item in the query.
        /// </remarks>
        public abstract IList<Prediction<TCatalogItemType>> Predict(IEnumerable<TCatalogItemType> query, User user);
    }
}
