﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;
using System.Data.SqlClient;
using System.Data;
using System.Configuration;

namespace MyMediaProject.RecommenderSystem.Core.EntityRelationshipEngine
{
    /// <summary>
    /// Base implementation of the <see cref="IEntityRelationDataProvider"/> interface.
    /// </summary>
    /// <remarks>
    /// This base implementation requires that the <see cref="IEntityRelationDataProvider.GetRelation(RelationType, int[])"/>
    /// be implemented.  The Un/RegisterForEntityUpdates methods are implemented as well.  The Un/RegisterForRelationUpdates methods
    /// are partially implemented, requiring 
    /// All other methods are "optional," depending on the requirements of the engine.
    /// </remarks>
    public abstract class EntityRelationDataProvider : IEntityRelationDataProvider
    {
        private Dictionary<RelationType, HashSet<IEntityRelationEngine>> registeredRelationUpdateEngines = new Dictionary<RelationType, HashSet<IEntityRelationEngine>>();
        private Dictionary<EntityType, HashSet<IEntityRelationEngine>> registeredEntityUpdateEngines = new Dictionary<EntityType, HashSet<IEntityRelationEngine>>();

        /// <summary>
        /// Gets the registered relation update engines.
        /// </summary>
        /// <value>The registered relation update engines.</value>
        protected Dictionary<RelationType, HashSet<IEntityRelationEngine>> RegisteredRelationUpdateEngines
        {
            get { return registeredRelationUpdateEngines; }
        }

        /// <summary>
        /// Gets the registered entity update engines.
        /// </summary>
        /// <value>The registered entity update engines.</value>
        protected Dictionary<EntityType, HashSet<IEntityRelationEngine>> RegisteredEntityUpdateEngines
        {
            get { return registeredEntityUpdateEngines; }
        }


        /// <summary>
        /// Initializes a new instance of the <see cref="EntityRelationDataProvider"/> class.
        /// </summary>
        protected EntityRelationDataProvider()
        {
        }

        /// <summary>
        /// same as GetRelation but returns all attribute columns
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <returns></returns>
        virtual public IEntityRelationDataReader GetRelation(RelationType relationType)
        {
            return this.GetRelation(relationType, new int[] { });
        }

        /// <summary>
        /// Gets the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="attributes">The attributes.</param>
        /// <returns></returns>
        abstract public IEntityRelationDataReader GetRelation(RelationType relationType, int[] attributes);

        /// <summary>
        /// same as GetEntity but returns all attribute columns
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <returns></returns>
        virtual public IEntityRelationDataReader GetEntity(EntityType entityType)
        {
            return this.GetEntity(entityType, new int[] { });
        }

        /// <summary>
        /// Gets the entity.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="attributes">The attributes.</param>
        /// <returns></returns>
        virtual public IEntityRelationDataReader GetEntity(EntityType entityType, int[] attributes)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Registers for relation updates.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="engine">The engine.</param>
        virtual public void RegisterForRelationUpdates(RelationType relationType, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engineList = null;
            RegisteredRelationUpdateEngines.TryGetValue(relationType, out engineList);
            if (engineList == null)
            {
                engineList = new HashSet<IEntityRelationEngine>();
                RegisteredRelationUpdateEngines.Add(relationType, engineList);
            }
            engineList.Add(engine);
        }

        /// <summary>
        /// Registers for entity updates.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="engine">The engine.</param>
        virtual public void RegisterForEntityUpdates(EntityType entityType, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engineList = null;
            RegisteredEntityUpdateEngines.TryGetValue(entityType, out engineList);
            if (engineList == null)
            {
                engineList = new HashSet<IEntityRelationEngine>();
                RegisteredEntityUpdateEngines.Add(entityType, engineList);
            }
            engineList.Add(engine);
        }

        /// <summary>
        /// Unregisters for relation updates.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="engine">The engine.</param>
        virtual public void UnregisterForRelationUpdates(RelationType relationType, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engineList = null;
            RegisteredRelationUpdateEngines.TryGetValue(relationType, out engineList);
            if (engineList != null)
            {
                engineList.Remove(engine);
            }
        }

        /// <summary>
        /// Unregisters for entity updates.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="engine">The engine.</param>
        virtual public void UnregisterForEntityUpdates(EntityType entityType, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engineList = null;
            RegisteredEntityUpdateEngines.TryGetValue(entityType, out engineList);
            if (engineList != null)
            {
                engineList.Remove(engine);
            }
        }

        /// <summary>
        /// Adds the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="values">The values.</param>
        virtual protected void AddRelation(RelationType relationType, int[] key, object[] values)
        {
            this.ModifyRelation(Modification.Add, relationType, key, values);
        }
        /// <summary>
        /// Updates the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="values">The values.</param>
        virtual protected void UpdateRelation(RelationType relationType, int[] key, object[] values)
        {
            this.ModifyRelation(Modification.Update, relationType, key, values);
        }
        /// <summary>
        /// Removes the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        virtual protected void RemoveRelation(RelationType relationType, int[] key)
        {
            this.ModifyRelation(Modification.Remove, relationType, key, null);
        }
        /// <summary>
        /// Adds the entity.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="entityId">The entity id.</param>
        /// <param name="values">The values.</param>
        virtual protected void AddEntity(EntityType entityType, int entityId, object[] values)
        {
            this.ModifyEntity(Modification.Add, entityType, entityId, values);
        }
        /// <summary>
        /// Updates the entity.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="entityId">The entity id.</param>
        /// <param name="values">The values.</param>
        virtual protected void UpdateEntity(EntityType entityType, int entityId, object[] values)
        {
            this.ModifyEntity(Modification.Update, entityType, entityId, values);
        }
        /// <summary>
        /// Removes the entity.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="entityId">The entity id.</param>
        virtual protected void RemoveEntity(EntityType entityType, int entityId)
        {
            this.ModifyEntity(Modification.Remove, entityType, entityId, null);
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "modification")]
        private void ModifyRelation(Modification modification, RelationType relationType, int[] key, object[] values)
        {
            HashSet<IEntityRelationEngine> engines = null;
            RegisteredRelationUpdateEngines.TryGetValue(relationType, out engines);
            if (engines != null)
            {
                foreach (IEntityRelationEngine engine in engines)
                {
                    engine.AddRelation(relationType, key, values);
                }
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "modification")]
        private void ModifyEntity(Modification modification, EntityType entityType, int entityId, object[] values)
        {
            HashSet<IEntityRelationEngine> engines = null;
            RegisteredEntityUpdateEngines.TryGetValue(entityType, out engines);
            if (engines != null)
            {
                foreach (IEntityRelationEngine engine in engines)
                {
                    engine.AddEntity(entityType, entityId, values);
                }
            }
        }
        private enum Modification
        {
            Undefined = 0,
            Add,
            Update,
            Remove,
        }
    }
}
