﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.EntityRelationshipEngine;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm
{
    /// <summary>
    /// Base class for all Engines which consume an <see cref="IEntityRelationEngine"/> based recommendation algorithm.
    /// </summary>
    public class EntityRelationEngine : Engine, IEntityRelationPredictor
    {
        private IEntityRelationEngine entityRelationAlgorithm;
        private IEntityRelationDataProvider dataProvider;

        /// <summary>
        /// Gets the entity relation algorithm.
        /// </summary>
        /// <value>The entity relation algorithm.</value>
        protected IEntityRelationEngine EntityRelationAlgorithm
        {
            get
            {
                return this.entityRelationAlgorithm;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="EntityRelationEngine"/> class, with the default <see cref="IEntityRelationDataProvider"/>, 
        /// <see cref="MyMediaEntityRelationDataProvider"/>.
        /// </summary>
        /// <param name="algorithm">The algorithm.</param>
        public EntityRelationEngine(IEntityRelationEngine algorithm)
            : this(algorithm, new MyMediaEntityRelationDataProvider())
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="EntityRelationEngine"/> class.
        /// </summary>
        /// <param name="algorithm">The algorithm.</param>
        /// <param name="customDataProvider">The custom data provider.</param>
        public EntityRelationEngine(IEntityRelationEngine algorithm, IEntityRelationDataProvider customDataProvider)
            : base(algorithm.Id, algorithm.Description)
        {
            this.entityRelationAlgorithm = algorithm;
            this.dataProvider = customDataProvider;
            this.entityRelationAlgorithm.EntityRelationDataProvider = this.dataProvider;

            this.RegisterEngineProviderEvents();
        }

        private void RegisterEngineProviderEvents()
        {
            this.EngineProvidersRegistered += new EventHandler(EntityRelationEngine_EngineProvidersRegistered);
            this.EngineProvidersUnregistered += new EventHandler(EntityRelationEngine_EngineProvidersUnregistered);
        }

        /// <summary>
        /// Trains this algorithm.
        /// </summary>
        public void Train()
        {
            this.EntityRelationAlgorithm.Train();
        }

        /// <summary>
        /// Handles new catalog items.
        /// </summary>
        /// <param name="catalogItems">The catalog items.</param>
        /// <remarks>
        /// This implementation does nothing; updates from the system are handled via the <see cref="IEntityRelationDataProvider"/> interface.
        /// </remarks>
        protected override void HandleNewCatalogItems(System.Collections.ObjectModel.ReadOnlyCollection<CatalogItem> catalogItems)
        {            
            //Do nothing.
        }

        /// <summary>
        /// Handles new feedback.
        /// </summary>
        /// <param name="feedback">The feedback.</param>
        /// <remarks>
        /// This implementation does nothing; updates from the system are handled via the <see cref="IEntityRelationDataProvider"/> interface.
        /// </remarks>
        protected override void HandleNewFeedback(System.Collections.ObjectModel.ReadOnlyCollection<UserAction> feedback)
        {
            //Do nothing.
        }

        /// <summary>
        /// Calculates a prediction (recommendation) for each catalog item in the query.
        /// </summary>
        /// <param name="query">The query.</param>
        /// <param name="user">The user.</param>
        /// <returns>a list of predictions.</returns>
        /// <remarks>
        /// This forms the heart of the recommendation algorithm. It has to generate a prediction for each catalog item in the query.
        /// </remarks>
        public override IList<Prediction<CatalogItem>> Predict(IEnumerable<CatalogItem> query, User user)
        {
            throw new NotImplementedException();
        }

        private void EntityRelationEngine_EngineProvidersRegistered(object sender, EventArgs e)
        {
            if (this.dataProvider is IMyMediaEntityRelationDataProvider)
            {
                IMyMediaEntityRelationDataProvider myMediaDataProvider = this.dataProvider as IMyMediaEntityRelationDataProvider;
                myMediaDataProvider.RegisterDataStoreContext(base.DatabaseConnectionString);
            }
        }

        private void EntityRelationEngine_EngineProvidersUnregistered(object sender, EventArgs e)
        {
            if (this.dataProvider is IMyMediaEntityRelationDataProvider)
            {
                IMyMediaEntityRelationDataProvider myMediaDataProvider = this.dataProvider as IMyMediaEntityRelationDataProvider;
                myMediaDataProvider.UnregisterDataStoreContext(base.DatabaseConnectionString);
            }
        }

        #region IEntityRelationPredictor Members

        /// <summary>
        /// Given an identifier for the relation type (e.g. "rating") and the proper key, the predicted value
        /// for this relation is returned.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="keys">The keys.</param>
        /// <returns></returns>
        public double PredictRelation(RelationType relationType, int[] keys)
        {
            return this.entityRelationAlgorithm.PredictRelation(relationType, keys);
        }

        /// <summary>
        /// This is the TopN method.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="keys">Combination of other relation's keys. E.g. for predicting top-n items, keys[0] contains the user_id for which top-n ratings should be predicted.</param>
        /// <param name="count">The count.</param>
        /// <returns>An ordered array of keys to that entity.</returns>
        public int[] PredictEntity(EntityType entityType, int[] keys, int count)
        {
            return this.entityRelationAlgorithm.PredictEntity(entityType, keys, count);
        }

        #endregion
    }
}
