﻿using MyMediaProject.RecommenderSystem.Core.Databases;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Configuration;
using System.Globalization;
using System.Data;
using System.Data.Common;
using System.Data.Linq.Mapping;
using System.Data.SqlClient;
using System.IO;
using System.Linq;
using System.Text;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Core.EntityRelationshipEngine
{
    /// <summary>
    /// Provides a default <see cref="IEntityRelationDataProvider"/> implementaiton for the MyMediaFramework.  This class
    /// can be extened by overriding the virtual methods, which allows for supporting additional <see cref="EntityType"/> 
    /// and <see cref="RelationType"/>.
    /// </summary>
    public class MyMediaEntityRelationDataProvider : EntityRelationDataProvider, IMyMediaEntityRelationDataProvider
    {
        string dbConnectionString;

        /// <summary>
        /// Initializes a new instance of the <see cref="MyMediaEntityRelationDataProvider"/> class.
        /// </summary>
        public MyMediaEntityRelationDataProvider()
        {
            //Register for changes to all of the data stores.
            new FeedbackStoreChangedEvent().Subscribe(this.NewUserActionEventHandler);
            new CatalogStoreChangedEvent().Subscribe(this.NewCatalogItemEventHandler);
            new UserStoreChangedEvent().Subscribe(this.NewUserEventHandler);
        }

        /// <summary>
        /// Gets the entity.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="attributes">The attributes.</param>
        /// <returns></returns>
        public sealed override IEntityRelationDataReader GetEntity(EntityType entityType, int[] attributes)
        {
            using (MyMediaDataContext context = new MyMediaDataContext(this.dbConnectionString))
            {
                IQueryable query = this.GetEntity(entityType, context);
                return MyMediaEntityRelationDataProvider.ConvertQueryToDataReader(query, context);
            }
        }

        /// <summary>
        /// Extensibility point for derived classes.  This allows derived classes to change the entity information returned.
        /// </summary>
        /// <remarks>
        /// Here are the <see cref="EntityType"/> supported by this implementation:
        /// <see cref="EntityType.Keyword"/> returns {[KeywordId], [Keyword]}, 
        /// where KeywordId is the entity key as an integer, and Keyword is a unicode string representing the key word.
        /// </remarks>
        /// <example>
        /// Here is a simple example of a derived implementation which would extend <see cref="EntityType.User"/>
        /// to include the UserId and UserName in the entity information.
        /// 
        /// IQueryable query = null;
        /// switch (entityType)
        /// {
        ///     case EntityType.User:
        ///         query = myMediaDataContext.Users.Select(user => new { user.UserId, user.UserName });
        ///         break;
        ///     default:
        ///         base.GetEntity(entityType, myMediaDataContext);
        ///     break;
        /// }
        /// return query;
        ///
        /// </example>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="dataStoreContext">My media data context.</param>
        /// <returns></returns>
        protected virtual IQueryable GetEntity(EntityType entityType, MyMediaDataContext dataStoreContext)
        {
            IQueryable query = null;

            switch (entityType)
            {
                case EntityType.Keyword:
                    query = dataStoreContext.Keywords.Select(keyword => new { keyword.KeywordId, keyword.Word });
                    break;
                case EntityType.Undefined:
                    throw new ArgumentException("Must not be undefined.", "entityType");
                //break;
                case EntityType.User:
                case EntityType.CatalogItem:
                case EntityType.Tag:
                case EntityType.VirtualProfile:
                default:
                    throw new NotImplementedException(string.Format(CultureInfo.InvariantCulture,
                        "Getting relations for type {0} are not implemented.", entityType));
                //break;
            }
            return query;
        }

        /// <summary>
        /// Gets the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="attributes">The attributes.</param>
        /// <returns></returns>
        override sealed public IEntityRelationDataReader GetRelation(RelationType relationType, int[] attributes)
        {
            using (MyMediaDataContext context = new MyMediaDataContext(this.dbConnectionString))
            {
                IQueryable query = this.GetRelation(relationType, context);
                return MyMediaEntityRelationDataProvider.ConvertQueryToDataReader(query, context);
            }
        }

        /// <summary>
        /// Extensibility point for derived classes.  This allows derived classes to change the relation information returned.
        /// </summary>
        /// <remarks>
        /// Deriving classes can override this method to extend the relation types supported by this data adapter.
        /// Here are the <see cref="RelationType"/> supported by this implementation:
        /// <see cref="RelationType.Rated"/> returns {[UserId],[ItemId],[RatingValue]}, 
        /// where UserId and ItemId are the entity keys as integers, and the rating value is a double of range [-1,1].
        /// <see cref="RelationType.Tagged"/> returns {[UserId],[ItemId],[TagId]}, 
        /// where UserId, ItemId and TagId are the entity keys as integers.
        /// <see cref="RelationType.Viewed"/> returns {[UserId],[ItemId]}, 
        /// where UserId and ItemId are the entity keys as integers.
        /// </remarks>
        /// <see cref="RelationType.ItemKeyword"/> returns {[ItemId],[KeywordId]}, 
        /// where ItemId and KeywordId are the entity keys as integers.
        /// <see cref="RelationType.VirtualProfileRated"/> returns {[UserId],[ItemId],[VirtualProfileId],[RatingValue]}, 
        /// where UserId, ItemId, and VirtualProfileId (optional; can be null) are the entity keys as integers, and the rating value is a double of range [-1,1].
        /// <see cref="RelationType.VirtualProfileMembers"/> returns {[VirtualProfileId],[UserId]}, 
        /// where VirtualProfileId and UserId are the entity keys as integers.
        /// <example>
        /// Here is a simple example of a derived implementation which would extend <see cref="RelationType.Rated"/>
        /// to include the virtual profile id used while rating an item.
        /// 
        /// IQueryable query = null;
        /// switch (relationType)
        /// {
        ///     case RelationType.Rated:
        ///         query = dataStoreContext.UserActions.Where(rating =&gt; rating is RatingAction).Cast&lt;RatingAction&gt;().
        ///                         Select(rating =&gt; new { rating.UserId, rating.ItemId, rating.RatingValue.Value, rating.VirtualProfileId });
        ///         break;
        ///     default:
        ///         query = base.GetRelation(relationType, myMediaDataContext);
        ///         break;
        /// }
        /// return query;
        /// 
        /// </example>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="dataStoreContext">The data store context.</param>
        /// <returns></returns>
        protected virtual IQueryable GetRelation(RelationType relationType, MyMediaDataContext dataStoreContext)
        {
            IQueryable query = null;

            switch (relationType)
            {
                case RelationType.Rated:
                    query = dataStoreContext.UserActions.Where(action => action.TestAction == false).Where(rating => rating is RatingAction).Cast<RatingAction>().
                                    Select(rating => new { rating.UserId, rating.ItemId, rating.Rating.Value });
                    break;
                case RelationType.Tagged:
                    query = dataStoreContext.UserActions.Where(action => action.TestAction == false).Where(tagAction => tagAction is TaggingAction).Cast<TaggingAction>().
                        Select(tagAction => new { tagAction.UserId, tagAction.ItemId, tagAction.TagId });
                    break;
                case RelationType.Viewed:
                    query = dataStoreContext.UserActions.Where(action => action.TestAction == false).Where(viewing => viewing is ViewedAction).Cast<ViewedAction>().
                        Select(viewing => new { viewing.UserId, viewing.ItemId });
                    break;
                case RelationType.ItemKeyword:
                    query = dataStoreContext.ItemKeywords.Select(itemKeyword => new { itemKeyword.ItemId, itemKeyword.KeywordId });
                    break;
                case RelationType.VirtualProfileRated:
                    query = dataStoreContext.UserActions.Where(rating => rating is RatingAction).Cast<RatingAction>().
                        Select(rating => new { rating.UserId, rating.ItemId, rating.VirtualProfileId, RatingValue = rating.Rating });
                    break;
                case RelationType.VirtualProfileMembers:
                    query = dataStoreContext.VirtualProfileUsers.Select(vp => new {vp.VirtualProfileId, vp.UserId });
                    break;
                case RelationType.Undefined:
                    throw new ArgumentException("Must not be undefined.", "relationType");
                //break;
                case RelationType.Purchased:
                default:
                    throw new NotImplementedException(string.Format(CultureInfo.InvariantCulture,
                        "Getting relations for type {0} are not implemented.", relationType));
                //break;
            }

            return query;
        }

        /// <summary>
        /// Gets the new user action event handler.
        /// </summary>
        /// <value>The new user action event handler.</value>
        public RecommenderEventHandler NewUserActionEventHandler
        {
            get
            {
                return this.InternalHandleFeedbackEvent;
            }
        }

        /// <summary>
        /// Gets the new catalog item event handler.
        /// </summary>
        /// <value>The new catalog item event handler.</value>
        public RecommenderEventHandler NewCatalogItemEventHandler
        {
            get
            {
                return this.InternalHandleCatalogItemEvents;
            }
        }

        /// <summary>
        /// Gets the new user event handler.
        /// </summary>
        /// <value>The new user event handler.</value>
        public RecommenderEventHandler NewUserEventHandler
        {
            get
            {
                return this.InternalHandleUserEvents;
            }
        }

        private static IEntityRelationDataReader ConvertQueryToDataReader(IQueryable query, MyMediaDataContext context)
        {
            DbCommand command = context.GetCommand(query);
            if (command.Connection.State == ConnectionState.Closed)
            {
                command.Connection.Open();
            }
            return new SqlEntityRelationDataReader(command.ExecuteReader(System.Data.CommandBehavior.CloseConnection));
        }

        /// <summary>
        /// Registers the database connection string.
        /// </summary>
        /// <param name="databaseConnectionString">The database connection string.</param>
        public void RegisterDataStoreContext(string databaseConnectionString)
        {
            this.dbConnectionString = databaseConnectionString;
        }

        /// <summary>
        /// Unregisters the database connection string.
        /// </summary>
        /// <param name="databaseConnectionString">The database connection string.</param>
        public void UnregisterDataStoreContext(string databaseConnectionString)
        {
            Contract.Precondition(this.dbConnectionString == databaseConnectionString, "Attempting to unregister a different connection string.");
            this.dbConnectionString = string.Empty;
        }

        /// <summary>
        /// Gets the data context, which must be disposed of after use.
        /// </summary>
        /// <returns></returns>
        protected MyMediaDataContext GetDataContext()
        {
            return new MyMediaDataContext(this.dbConnectionString);
        }

        private void InternalHandleCatalogItemEvents(EventArgs args)
        {
            StoreChangedEventArgs scargs = args as StoreChangedEventArgs;

            if (scargs != null)
            {
                InternalHandleCatalogItemEvents(scargs.NewChangeNumber);
            }
        }

        private void InternalHandleCatalogItemEvents(int newChangeNumber)
        {
            List<CatalogItem> newItems = new List<CatalogItem>();
            List<CatalogItem> updatedItems = new List<CatalogItem>();
            List<int> deletedItems = new List<int>();

            using (CatalogItemStore store = new CatalogItemStore(new CatalogItemStoreDataContext(this.dbConnectionString)))
            {
                store.Initialize();
                foreach (IDataStoreChange sc in store.GetDifferences(newChangeNumber, newChangeNumber))
                {
                    CatalogItem catalogItem;
                    switch (sc.ChangeType)
                    {
                        case StoreChangeType.Added:
                            catalogItem = store.Single(item => item.ItemId == sc.ChangedItemId);
                            newItems.Add(catalogItem);
                            break;
                        case StoreChangeType.Changed:
                            catalogItem = store.Single(item => item.ItemId == sc.ChangedItemId);
                            updatedItems.Add(catalogItem);
                            break;
                        case StoreChangeType.Removed:
                            deletedItems.Add(sc.ChangedItemId);
                            break;
                        default:
                            break;
                    }
                }

                this.HandleNewCatalogItems(new ReadOnlyCollection<CatalogItem>(newItems));
                this.HandleChangedCatalogItems(new ReadOnlyCollection<CatalogItem>(updatedItems));
                this.HandleRemovedCatalogItems(new ReadOnlyCollection<int>(deletedItems));
            }
        }

        /// <summary>
        /// Handles the removed catalog items.
        /// </summary>
        /// <param name="removedItems">The removed items.</param>
        protected virtual void HandleRemovedCatalogItems(ReadOnlyCollection<int> removedItems)
        {
            foreach (int itemId in removedItems)
            {
                this.RemoveEntity(EntityType.CatalogItem, itemId);
            }
        }

        /// <summary>
        /// Handles the changed catalog items.
        /// </summary>
        /// <param name="changedItems">The changed items.</param>
        protected virtual void HandleChangedCatalogItems(ReadOnlyCollection<CatalogItem> changedItems)
        {
            foreach (CatalogItem item in changedItems)
            {
                this.UpdateEntity(EntityType.CatalogItem, item.ItemId, new object[]{});
            }
        }

        /// <summary>
        /// Handles the new catalog items.
        /// </summary>
        /// <param name="newItems">The new items.</param>
        protected virtual void HandleNewCatalogItems(ReadOnlyCollection<CatalogItem> newItems)
        {
            foreach (CatalogItem item in newItems)
            {
                this.AddEntity(EntityType.CatalogItem, item.ItemId, new object[]{});
            }
        }

        private void InternalHandleUserEvents(EventArgs args)
        {
            StoreChangedEventArgs scargs = args as StoreChangedEventArgs;

            if (scargs != null)
            {
                InternalHandleUserEvents(scargs.NewChangeNumber);
            }
        }

        private void InternalHandleUserEvents(int newChangeNumber)
        {
            List<User> newItems = new List<User>();
            List<User> updatedItems = new List<User>();
            List<int> deletedItems = new List<int>();

            using (UserStore store = new UserStore(new MyMediaDataContext(this.dbConnectionString)))
            {
                store.Initialize();
                foreach (IDataStoreChange sc in store.GetDifferences(newChangeNumber, newChangeNumber))
                {
                    User User;
                    switch (sc.ChangeType)
                    {
                        case StoreChangeType.Added:
                            User = store.Single(item => item.UserId == sc.ChangedItemId);
                            newItems.Add(User);
                            break;
                        case StoreChangeType.Changed:
                            User = store.Single(item => item.UserId == sc.ChangedItemId);
                            updatedItems.Add(User);
                            break;
                        case StoreChangeType.Removed:
                            deletedItems.Add(sc.ChangedItemId);
                            break;
                        default:
                            break;
                    }
                }

                this.HandleNewUsers(new ReadOnlyCollection<User>(newItems));
                this.HandleChangedUsers(new ReadOnlyCollection<User>(updatedItems));
                this.HandleRemovedUsers(new ReadOnlyCollection<int>(deletedItems));
            }
        }

        /// <summary>
        /// Handles the removed catalog items.
        /// </summary>
        /// <param name="removedItems">The removed items.</param>
        protected virtual void HandleRemovedUsers(ReadOnlyCollection<int> removedItems)
        {
            foreach (int itemId in removedItems)
            {
                this.RemoveEntity(EntityType.User, itemId);
            }
        }

        /// <summary>
        /// Handles the changed catalog items.
        /// </summary>
        /// <param name="changedItems">The changed items.</param>
        protected virtual void HandleChangedUsers(ReadOnlyCollection<User> changedItems)
        {
            foreach (User item in changedItems)
            {
                this.UpdateEntity(EntityType.User, item.UserId, new object[] { });
            }
        }

        /// <summary>
        /// Handles the new catalog items.
        /// </summary>
        /// <param name="newItems">The new items.</param>
        protected virtual void HandleNewUsers(ReadOnlyCollection<User> newItems)
        {
            foreach (User item in newItems)
            {
                this.AddEntity(EntityType.User, item.UserId, new object[] { });
            }
        }

        private void InternalHandleFeedbackEvent(EventArgs args)
        {
            StoreChangedEventArgs scargs = args as StoreChangedEventArgs;

            if (scargs != null)
            {
                InternalHandleFeedbackEvent(scargs.NewChangeNumber);
            }
        }

        private void InternalHandleFeedbackEvent(int newChangeNumber)
        {
            List<UserAction> newItems = new List<UserAction>();
            List<UserAction> updatedItems = new List<UserAction>();
            List<int> deletedItems = new List<int>();

            using (UserActionStore store = new UserActionStore(new CatalogItemStoreDataContext(this.dbConnectionString)))
            {
                store.Initialize();
                foreach (IDataStoreChange sc in store.GetDifferences(newChangeNumber, newChangeNumber))
                {
                    UserAction userAction;
                    switch (sc.ChangeType)
                    {
                        case StoreChangeType.Added:
                            userAction = store.Single(item => item.Key == sc.ChangedItemId); 
                            newItems.Add(userAction);
                            break;
                        case StoreChangeType.Changed:
                            userAction = store.Single(item => item.Key == sc.ChangedItemId); 
                            updatedItems.Add(userAction);
                            break;
                        case StoreChangeType.Removed:
                            deletedItems.Add(sc.ChangedItemId);
                            break;
                        default:
                            break;
                    }
                }

                this.HandleNewFeedback(new ReadOnlyCollection<UserAction>(newItems));
                this.HandleChangedFeedback(new ReadOnlyCollection<UserAction>(updatedItems));
                this.HandleRemovedFeedback(new ReadOnlyCollection<int>(deletedItems));
            }
        }

        /// <summary>
        /// Handles the removed feedback.
        /// </summary>
        /// <param name="removedItems">The removed items.</param>
        protected virtual void HandleRemovedFeedback(ReadOnlyCollection<int> removedItems)
        {
            //foreach (int item in removedItems)
            //{
            //    RelationType relationType = RelationType.Undefined;
            //    int[] keys;
            //    object [] values;
            //    this.GetRelationType(item, out relationType, out keys, out values);
            //    this.RemoveRelation(relationType, keys);
            //}
        }

        /// <summary>
        /// Handles the changed feedback.
        /// </summary>
        /// <param name="changedItems">The changed items.</param>
        protected virtual void HandleChangedFeedback(ReadOnlyCollection<UserAction> changedItems)
        {
            foreach (UserAction item in changedItems)
            {
                RelationType relationType = RelationType.Undefined;
                int[] keys = new int[] { };
                object[] values = new object[]{};

                this.GetRelationType(item, out relationType, out keys, out values);
                this.UpdateRelation(relationType, keys, values);
            }
        }

        /// <summary>
        /// Handles the new feedback.
        /// </summary>
        /// <param name="newItems">The new items.</param>
        protected virtual void HandleNewFeedback(ReadOnlyCollection<UserAction> newItems)
        {
            foreach (UserAction item in newItems)
            {
                RelationType relationType = RelationType.Undefined;
                int[] keys = new int[] { };
                object[] values = new object[] { };

                this.GetRelationType(item, out relationType, out keys, out values);
                this.AddRelation(relationType, keys, values);
            }
        }

        /// <summary>
        /// Gets the type of the relation.
        /// </summary>
        /// <param name="userAction">The user action.</param>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="keys">The keys.</param>
        /// <param name="values">The values.</param>
        protected virtual void GetRelationType(UserAction userAction, out RelationType relationType, out int [] keys, out object [] values)
        {
            if (userAction is RatingAction)
            {
                relationType = RelationType.Rated;
                keys = new int[] { userAction.UserId, userAction.ItemId };
                values = new object[] { (userAction as RatingAction).Rating };
            }
            else
            {
                throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, "RelationType not defined for UserAction type {0}.", typeof(UserAction)));
            }
        }
    }
}
