﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// The base for a recommendation algorithm.
    /// </summary>
    /// <remarks>
    /// In contrast to a <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>, this class is a container for the individual algorithm 
    /// (<see cref="IPredictionAlgorithm&lt;TCatalogItemType&gt;"/>), the catalog update handler (<see cref="INewCatalogItemHandler&lt;TCatalogItemType&gt;"/>) and
    /// the feedback handler (<see cref="INewUserActionHandler&lt;TCatalogItemType, TUserActionType&gt;"/>).
    /// The framework ensures that engines will always receive all catalog item and feedback changes, even if they have been off-line
    /// for some time. As soon as they are part of any flow, they will receive all (missed) update changes.
    /// Engines must be added to a <see cref="Flow&lt;TCatalogItemType, TUserActionType&gt;"/>. The flow is then used to generate predictions. All engines that are part of a flow will
    /// also be called in case of catalog updates (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewCatalogItems(ReadOnlyCollection&lt;TCatalogItemType&gt;)"/>) and if feedback is received (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewFeedback"/>).
    /// </remarks>
    public class PluggableEngine : PluggableEngine<CatalogItem, UserAction> 
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="PluggableEngine&lt;TCatalogItemType, TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="algorithm">The algorithm.</param>
        /// <param name="newCatalogItemHandler">The new catalog item handler.</param>
        /// <param name="newUserActionHandler">The new feedback item handler.</param>
        /// <param name="engineId">The engine id.</param>
        /// <param name="description">The description.</param>
        public PluggableEngine(
            IPredictionAlgorithm algorithm,
            INewCatalogItemHandler newCatalogItemHandler,
            INewUserActionHandler newUserActionHandler,
            EngineId engineId,
            string description)
            : base(algorithm, newCatalogItemHandler, newUserActionHandler, engineId, description)
        {
        }
    }
    /// <summary>
    /// The base for a recommendation algorithm.
    /// </summary>
    /// <remarks>
    /// In contrast to a <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>, this class is a container for the individual algorithm 
    /// (<see cref="IPredictionAlgorithm&lt;TCatalogItemType&gt;"/>), the catalog update handler (<see cref="INewCatalogItemHandler&lt;TCatalogItemType&gt;"/>) and
    /// the feedback handler (<see cref="INewUserActionHandler&lt;TCatalogItemType, TUserActionType&gt;"/>).
    /// The framework ensures that engines will always receive all catalog item and feedback changes, even if they have been off-line
    /// for some time. As soon as they are part of any flow, they will receive all (missed) update changes.
    /// Engines must be added to a <see cref="Flow&lt;TCatalogItemType, TUserActionType&gt;"/>. The flow is then used to generate predictions. All engines that are part of a flow will
    /// also be called in case of catalog updates (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewCatalogItems(ReadOnlyCollection&lt;TCatalogItemType&gt;)"/>) and if feedback is received (<see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;.HandleNewFeedback"/>).
    /// </remarks>
    /// <typeparam name="TCatalogItemType">The type of the catalog item type.</typeparam>
    /// <typeparam name="TUserActionType">The type of the feedback item type.</typeparam>
    public class PluggableEngine<TCatalogItemType, TUserActionType> : Engine<TCatalogItemType, TUserActionType>
        where TCatalogItemType : CatalogItem
        where TUserActionType : UserAction
    {
        private IPredictionAlgorithm<TCatalogItemType> algorithm;
        private INewCatalogItemHandler<TCatalogItemType> newCatalogItemHandler;
        private INewUserActionHandler<TCatalogItemType, TUserActionType> newUserActionHandler;

        /// <summary>
        /// Initializes a new instance of the <see cref="PluggableEngine&lt;TCatalogItemType, TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="algorithm">The algorithm.</param>
        /// <param name="newCatalogItemHandler">The new catalog item handler.</param>
        /// <param name="newUserActionHandler">The new feedback item handler.</param>
        /// <param name="engineId">The engine id.</param>
        /// <param name="description">The description.</param>
        public PluggableEngine(
            IPredictionAlgorithm<TCatalogItemType> algorithm,
            INewCatalogItemHandler<TCatalogItemType> newCatalogItemHandler,
            INewUserActionHandler<TCatalogItemType, TUserActionType> newUserActionHandler,
            EngineId engineId,
            string description)

            : base(engineId, description)
        {
            this.algorithm = algorithm;
            this.newCatalogItemHandler = newCatalogItemHandler;
            this.newUserActionHandler = newUserActionHandler;
        }

        /// <summary>
        /// Internal use only. Please ignore.
        /// </summary>
        protected override void HandleNewCatalogItems(ReadOnlyCollection<TCatalogItemType> catalogItems)
        {
            if (this.newCatalogItemHandler != null)
            {
                this.newCatalogItemHandler.HandleNewCatalogItems(catalogItems);
            }
        }

        /// <summary>
        /// Internal use only. Please ignore.
        /// </summary>
        protected override void HandleNewFeedback(ReadOnlyCollection<TUserActionType> feedback)
        {
            if (this.newUserActionHandler != null)
            {
                this.newUserActionHandler.HandleNewFeedback(feedback);
            }
        }

        /// <summary>
        /// Internal use only. Please ignore.
        /// </summary>
        /// 
        public override IList<Prediction<TCatalogItemType>> Predict(IEnumerable<TCatalogItemType> query, User user)
        {
            return this.algorithm.Predict(query, user);
        }
    }
}
