﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EMIC.Design;
using System.Collections;
using System.Threading;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Defines a method that handles a <see cref="RecommenderEvent"/>.
    /// </summary>
    /// <param name="args">The recommender event arguments.</param>
    public delegate void RecommenderEventHandler(RecommenderEventArgs args);

    internal class Dispatcher: IDisposable
    {
        List<RecommenderEventHandler> handler = new List<RecommenderEventHandler>();
        List<object> handlerChanges = new List<object>();
        AutoResetEvent syncEvent = new AutoResetEvent(true);

        internal Dispatcher() { }

        internal Dispatcher(RecommenderEventHandler handler)
        {
            TraceLog.EnterMethod();
            TraceLog.Write(LogEventType.Verbose, handler);
            this.handler.Add(handler);
            TraceLog.ExitMethod();
        }

        internal Dispatcher Parent
        {
            get;
            set;
        }

        internal void AddHandler(RecommenderEventHandler handle)
        {
            Contract.Invariant(handler != null);
            Contract.Invariant(handlerChanges != null);
            lock (handlerChanges)
            {
                handlerChanges.Add(handle);
            }
        }

        private void ChangeHandlerList()
        {
            Contract.Invariant(handler != null);
            Contract.Invariant(handlerChanges != null);
            lock (handler)
            {
                lock (handlerChanges)
                {
                    object change;
                    for (int i = 0; i < handlerChanges.Count; i++)
                    {
                        change = handlerChanges[i];
                        Predicate<RecommenderEventHandler> remove = change as Predicate<RecommenderEventHandler>;
                        RecommenderEventHandler add = change as RecommenderEventHandler;

                        if (remove != null)
                        {
                            handler.RemoveAll(remove);
                        }
                        else if (add != null)
                        {
                            handler.Add(add);
                        }
                    }
                    handlerChanges.Clear();
                }
            }
        }

        internal void RemoveHandler(Predicate<RecommenderEventHandler> match)
        {
            Contract.Invariant(handler != null);
            Contract.Invariant(handlerChanges != null);
            lock (handlerChanges)
            {
                handlerChanges.Add(match);
            }
        }

        internal void Dispatch(RecommenderEventArgs args)
        {
            TraceLog.EnterMethod();
            Contract.Invariant(handler != null);
            syncEvent.WaitOne();
            try
            {
                ChangeHandlerList();
                foreach (RecommenderEventHandler d in handler)
                {
                    TraceLog.Write(LogEventType.Verbose, d);
                    d(args);
                }
            }
            finally
            {
                syncEvent.Set();
            }
            TraceLog.ExitMethod();
        }

        #region IDisposable Members

        ~Dispatcher()
        {
            Dispose(false);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (syncEvent != null)
                {
                    syncEvent.Close();
                }
            }
        }

        #endregion
    }

    internal static class EventManager
    {
        // Used for thread synchronization.
        private static object monitor = new object();

        // Maps the event types to dispatchers that contain the corresponding subscribers.
        private static Dictionary<string, Dispatcher> typeSubscriberMapping = new Dictionary<string, Dispatcher>();

        internal static void AddSubscriber(Type type, RecommenderEventHandler handler)
        {
            lock (monitor)
            {
                List<string> additionalHierachy = new List<string>();

                TraceLog.EnterMethod();

                TraceLog.WriteIf(handler == null, LogEventType.Verbose, "No handler. Building hierarchy only.");

                Type parent = type;
                Dispatcher dispatcher;


                //Add dispatcher if necessary and insert handler to subscriber list
                if (!typeSubscriberMapping.ContainsKey(type.FullName))
                {
                    TraceLog.Write(LogEventType.Verbose, "Dispatcher not in List.");

                    if (handler != null)
                    {
                        dispatcher = new Dispatcher(handler);
                    }
                    //If no handler is provided just build empty dispatcher for event type
                    else
                    {
                        dispatcher = new Dispatcher();
                    }

                    TraceLog.Write(LogEventType.Verbose, "Adding new dispatcher instance.");
                    typeSubscriberMapping.Add(type.FullName, dispatcher);
                    additionalHierachy.Add(type.FullName);

                }
                //If dispatcher already exists just add handler if provided
                else
                {
                    TraceLog.Write(LogEventType.Verbose, "Existing dispatcher found.");
                    dispatcher = null;

                    if (handler != null)
                    {
                        TraceLog.Write(LogEventType.Verbose, "Adding Handler.");

                        typeSubscriberMapping[type.FullName].AddHandler(handler);
                    }
                }


                //Build hierachy if type did not already exist up until a base type already contains a dispatcher
                if (dispatcher != null && type.FullName != typeof(RecommenderEvent).FullName)
                {
                    TraceLog.Write(LogEventType.Verbose, "Building hierachy for parents.");

                    Dispatcher tmpDispatcher;

                    do
                    {
                        parent = parent.BaseType;

                        TraceLog.Write(LogEventType.Verbose, "Building hierachy for Type: {0}", parent);

                        if (!typeSubscriberMapping.ContainsKey(parent.FullName))
                        {
                            TraceLog.Write(LogEventType.Verbose, "Type {0} not in List", parent);

                            tmpDispatcher = new Dispatcher();
                            dispatcher.Parent = tmpDispatcher;
                            dispatcher.AddHandler(tmpDispatcher.Dispatch);

                            typeSubscriberMapping.Add(parent.FullName, tmpDispatcher);
                            additionalHierachy.Add(parent.FullName);
                            dispatcher = tmpDispatcher;
                        }
                        else
                        {
                            TraceLog.Write(LogEventType.Verbose, "Type {0} already in List", parent);
                            dispatcher.AddHandler(typeSubscriberMapping[parent.FullName].Dispatch);
                            dispatcher.Parent = typeSubscriberMapping[parent.FullName];
                            additionalHierachy.Add(parent.FullName);
                        }
                    }
                    while (parent.FullName != typeof(RecommenderEvent).FullName);
                }

                if (additionalHierachy.Count > 0)
                {
                    new RemoteRecommenderEvent().Fire(new RemoteRecommenderEventArgs(additionalHierachy.ToArray(), null));
                }



                TraceLog.ExitMethod();
            }
        }

        internal static void RemoveSubscriber(RecommenderEventHandler handler)
        {
            lock (monitor)
            {
                TraceLog.EnterMethod();
                //Walk all lists and remove handler from them.
                foreach (string type in typeSubscriberMapping.Keys)
                {
                    TraceLog.Write(LogEventType.Verbose, "Removing handler from type: {0}", type);
                    RemoveSubscriber(type, handler);
                }
                TraceLog.ExitMethod();
            }
        }

        internal static void RemoveSubscriber(Type type, RecommenderEventHandler handler)
        {
            lock (monitor)
            {
                TraceLog.EnterMethod();

                RemoveSubscriber(type.FullName, handler);

                TraceLog.ExitMethod();
            }
        }

        internal static void RemoveSubscriber(string typeName, RecommenderEventHandler handler)
        {
            lock (monitor)
            {
                TraceLog.EnterMethod();

                TraceLog.WriteIf(!typeSubscriberMapping.ContainsKey(typeName), LogEventType.Warning, "No dispatcher for specified type!");
                //Remove the handler but keep hierachy
                if (typeSubscriberMapping.ContainsKey(typeName))
                {
                    TraceLog.Write(LogEventType.Verbose, "Removing handler from dispatcher");
                    Dispatcher tmpDispatcher = typeSubscriberMapping[typeName];
                    tmpDispatcher.RemoveHandler((h) => h.Method.MethodHandle.Equals(handler.Method.MethodHandle) && h.Target.Equals(handler.Target));
                }
                TraceLog.ExitMethod();
            }
        }

        // Calls all handlers that have subscribed to an event of a certain type or a base class of this type.
        internal static void Fire(Type type, RecommenderEventArgs args)
        {
            TraceLog.EnterMethod();
            //Build up hierachy if the type is not yet in use
            if (!typeSubscriberMapping.ContainsKey(type.FullName))
            {
                TraceLog.Write(LogEventType.Verbose, "No subscribers for this type. Building hierarchy.");
                AddSubscriber(type, null);
            }

            if (args.Type != typeof(RemoteRecommenderEvent))
            {
                new RemoteRecommenderEvent().Fire(new RemoteRecommenderEventArgs(null, args));
            }

            FireLocally(type.FullName, args);

            TraceLog.ExitMethod();
        }

        // Calls all handlers that have subscribed to an event of a certain type or a base class of this type.
        internal static void Fire(string typeName, RecommenderEventArgs args)
        {
            TraceLog.EnterMethod();

            if (args.Type != typeof(RemoteRecommenderEvent))
            {
                new RemoteRecommenderEvent().Fire(new RemoteRecommenderEventArgs(null, args));
            }

            FireLocally(typeName, args);

            TraceLog.ExitMethod();
        }


        // Synchronizes the current instance to an remote instace.
        public static void Synchronize(RemoteRecommenderEventArgs args)
        {
            if (args.AdditionalHierarchy != null)
            {
                AddHierachy(args.AdditionalHierarchy);
            }
            if (args.Args != null)
            {
                FireLocally(args.Args.Type.FullName, args.Args);
            }
        }

        // Calls all handlers that have subscribed to an event of a certain type or a base class of this type without fireing an synchronization event.
        private static void FireLocally(string typeName, RecommenderEventArgs args)
        {
            TraceLog.EnterMethod();

            if (!typeSubscriberMapping.ContainsKey(typeName))
            {
                TraceLog.Write(LogEventType.Error, "No subscribers for this type name. Event is lost!");
            }
            else
            {
                TraceLog.Write(LogEventType.Verbose, "Calling dispatcher");

                //Call all subscibers
                typeSubscriberMapping[typeName].Dispatch(args);
            }

            TraceLog.ExitMethod();
        }

        // Adds additional event types to the hierachy.
        private static void AddHierachy(string[] additionalHierachy)
        {
            lock (monitor)
            {
                TraceLog.EnterMethod();

                Dispatcher dispatcher = new Dispatcher();
                Dispatcher tmpdispatcher = null;

                for (int i = 0; i < additionalHierachy.Length; i++)
                {
                    if (!typeSubscriberMapping.ContainsKey(additionalHierachy[i]))
                    {
                        TraceLog.Write(LogEventType.Verbose, "Adding new type: " + additionalHierachy[i]);
                        dispatcher.Parent = new Dispatcher();
                        if (tmpdispatcher != null)
                        {
                            tmpdispatcher.AddHandler(dispatcher.Dispatch);
                        }

                        typeSubscriberMapping.Add(additionalHierachy[i], dispatcher);
                        tmpdispatcher = dispatcher;
                        dispatcher = dispatcher.Parent;

                    }
                    else
                    {
                        if (tmpdispatcher != null)
                        {
                            tmpdispatcher.Parent = typeSubscriberMapping[additionalHierachy[i]];
                            tmpdispatcher.AddHandler(typeSubscriberMapping[additionalHierachy[i]].Dispatch);
                        }
                        break;
                    }
                }

                TraceLog.ExitMethod();
            }
        }
    }
}
