﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Represents the prediction logic used by the <see cref="Recommender&lt;TUserActionType&gt;"/> to generate <see cref="Prediction&lt;TCatalogItemType&gt;"/>s.
    /// </summary>
    /// <remarks>
    /// A flow object represents the prediction logic that is used to generate Predictions (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User, int, int)"/>) 
    /// (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User)"/>) from a <see cref="Recommender&lt;TUserActionType&gt;"/> object.
    /// Catalog items are being send through a flow in following order:
    /// 1) Through all <see cref="IPreFilter&lt;TCatalogItemType&gt;"/> 
    /// 2) Through all <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>
    /// 3) Through a hybridizer <see cref="IHybridizer&lt;TCatalogItemType&gt;"/> (Default is <see cref="SingleEngineHybridizer&lt;TCatalogItemType&gt;"/>.
    /// 4) Through all <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>
    /// <see cref="IPreFilter&lt;TCatalogItemType&gt;"/> can exclude catalog items from the flow.
    /// <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/> are algorithms that generate a prediction for a catalog item.
    /// <see cref="IHybridizer&lt;TCatalogItemType&gt;"/> combine the results from multiple engines to a single prediction.
    /// <see cref="IPostFilter&lt;TCatalogItemType&gt;"/> can exclude prediction results from the flow
    /// The flow object generates a <see cref="PredictionProgressEvent"/> to indicate progress.
    /// </remarks>
    /// <example>
    /// <code>
    /// FlowId flowId = new FlowId();
    /// Flow flow = new Flow(flowId);
    /// 
    /// MyPeopleEngine peopleEngine = new MyPeopleEngine();
    /// MySerieEngine serieEngine = new MySerieEngine();
    /// flow.AddEngine(peopleEngine);
    /// flow.AddEngine(serieEngine);
    /// 
    /// IHybridizer hybridizer = new AveragingHybridizer();
    /// flow.SetHybridizer(hybridizer);
    /// 
    /// Recommender.RegisterFlow(flow);
    /// </code>
    /// </example>
    public class Flow : Flow<CatalogItem, UserAction>
    {
        private Flow() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="Flow&lt;TCatalogItemType, TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="flowId">The flow id.</param>
        public Flow(FlowId flowId)
            : base(flowId)
        {
        }
    }

    /// <summary>
    /// Represents the prediction logic used by the <see cref="Recommender&lt;TUserActionType&gt;"/> to generate <see cref="Prediction&lt;TCatalogItemType&gt;"/>s.
    /// </summary>
    /// <remarks>
    /// A flow object represents the prediction logic that is used to generate Predictions (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User, int, int)"/>) 
    /// (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User)"/>) from a 
    /// <see cref="Recommender&lt;TUserActionType&gt;"/> object.
    /// Catalog items are being send through a flow in following order:
    /// 1) Through all <see cref="IPreFilter&lt;TCatalogItemType&gt;"/> 
    /// 2) Through all <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>
    /// 3) Through a hybridizer <see cref="IHybridizer&lt;TCatalogItemType&gt;"/> (Default is <see cref="SingleEngineHybridizer&lt;TCatalogItemType&gt;"/>.
    /// 4) Through all <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>
    /// <see cref="IPreFilter&lt;TCatalogItemType&gt;"/> can exclude catalog items from the flow.
    /// <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/> are algorithms that generate a prediction for a catalog item.
    /// <see cref="IHybridizer&lt;TCatalogItemType&gt;"/> combine the results from multiple engines to a single prediction.
    /// <see cref="IPostFilter&lt;TCatalogItemType&gt;"/> can exclude prediction results from the flow
    /// The flow object generates a <see cref="PredictionProgressEvent"/> to indicate progress.
    /// </remarks>
    /// <example>
    /// <code>
    /// FlowId flowId = new FlowId();
    /// Flow flow = new Flow(flowId);
    /// 
    /// MyPeopleEngine peopleEngine = new MyPeopleEngine();
    /// MySerieEngine serieEngine = new MySerieEngine();
    /// flow.AddEngine(peopleEngine);
    /// flow.AddEngine(serieEngine);
    /// 
    /// IHybridizer hybridizer = new AveragingHybridizer();
    /// flow.SetHybridizer(hybridizer);
    /// 
    /// Recommender.RegisterFlow(flow);
    /// </code>
    /// </example>
    /// <typeparam name="TCatalogItemType">The type of the catalog item type.</typeparam>
    /// <typeparam name="TUserActionType">The type of the feedback item type.</typeparam>
    public class Flow<TCatalogItemType, TUserActionType>
        where TCatalogItemType : CatalogItem
        where TUserActionType : UserAction
    {
        private List<Engine<TCatalogItemType, TUserActionType>> engines;
        private List<IPreFilter<TCatalogItemType>> preFilters;
        private List<IPostFilter<TCatalogItemType>> postFilters;
        private IHybridizer<TCatalogItemType> hybridizer;
        private FlowId id;
        internal bool Registered { get; set; }
        private PredictionProgressEvent predictionProgressEvent;

        /// <summary>
        /// Do not use; for inheritance only.
        /// </summary>
        protected Flow() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="Flow&lt;TCatalogItemType, TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="flowId">The flow id.</param>
        internal Flow(FlowId flowId)
        {
            this.id = flowId;
            this.engines = new List<Engine<TCatalogItemType, TUserActionType>>();
            this.preFilters = new List<IPreFilter<TCatalogItemType>>();
            this.postFilters = new List<IPostFilter<TCatalogItemType>>();
            this.predictionProgressEvent = new PredictionProgressEvent();
            this.hybridizer = new SingleEngineHybridizer<TCatalogItemType>();
        }

        private void SendProgressNotification(int percentageCompleted)
        {
            predictionProgressEvent.Fire(new PredictionProgressEventArgs(this.id, percentageCompleted, percentageCompleted == 100));
        }

        /// <summary>
        /// Adds a <see cref="IPreFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="preFilter">The PreFilter to add.</param>
        public void AddPreFilter(IPreFilter<TCatalogItemType> preFilter)
        {
            this.preFilters.Add(preFilter);
        }

        /// <summary>
        /// Inserts a <see cref="IPreFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="index">The position to insert the PreFilter.</param>
        /// <param name="preFilter">The PreFilter to insert.</param>
        public void InsertPreFilter(int index, IPreFilter<TCatalogItemType> preFilter)
        {
            this.preFilters.Insert(index, preFilter);
        }

        /// <summary>
        /// Removes a <see cref="IPreFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="preFilter">The PreFilter to remove.</param>
        public void RemovePreFilter(IPreFilter<TCatalogItemType> preFilter)
        {
            this.preFilters.Remove(preFilter);
        }

        /// <summary>
        /// Removes a <see cref="IPreFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="index">The index.</param>
        public void RemovePreFilter(int index)
        {
            this.preFilters.RemoveAt(index);
        }

        /// <summary>
        /// Gets a list of all <see cref="IPreFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <returns>a list of all PreFilters</returns>
        public ReadOnlyCollection<IPreFilter<TCatalogItemType>> PreFilters
        {
            get
            {
                return new ReadOnlyCollection<IPreFilter<TCatalogItemType>>(this.preFilters);
            }
        }

        /// <summary>
        /// Adds an <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>.
        /// </summary>
        /// <param name="engine">The Engine.</param>
        public void AddEngine(Engine<TCatalogItemType, TUserActionType> engine)
        {
            Contract.Precondition(this.Registered == false, "A Flow is immutable when it is registered. Please unregister first.");
            this.engines.Add(engine);
        }

        /// <summary>
        /// Removes an <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>.
        /// </summary>
        /// <param name="engine">The Engine.</param>
        public void RemoveEngine(Engine<TCatalogItemType, TUserActionType> engine)
        {
            Contract.Precondition(this.Registered == false, "A Flow is immutable when it is registered. Please unregister first.");
            this.engines.Remove(engine);
        }

        /// <summary>
        /// Gets a list of all <see cref="Engine&lt;TCatalogItemType, TUserActionType&gt;"/>.
        /// </summary>
        /// <returns>a list of engines."/></returns>
        public ReadOnlyCollection<Engine<TCatalogItemType, TUserActionType>> Engines
        {
            get
            {
                return new ReadOnlyCollection<Engine<TCatalogItemType, TUserActionType>>(this.engines);
            }
        }

        /// <summary>
        /// Gets or sets the hybridizer.
        /// </summary>
        /// <value>The hybridizer.</value>
        public IHybridizer<TCatalogItemType> Hybridizer
        {
            get
            {
                return this.hybridizer;
            }
            set
            {
                this.hybridizer = value;
            }
        }

        /// <summary>
        /// Adds the <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="postFilter">The PostFilter.</param>
        public void AddPostFilter(IPostFilter<TCatalogItemType> postFilter)
        {
            this.postFilters.Add(postFilter);
        }

        /// <summary>
        /// Inserts the <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="index">The index.</param>
        /// <param name="postFilter">The PostFilter.</param>
        public void InsertPostFilter(int index, IPostFilter<TCatalogItemType> postFilter)
        {
            this.postFilters.Insert(index, postFilter);
        }

        /// <summary>
        /// Removes the <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="postFilter">The PostFilter.</param>
        public void RemovePostFilter(IPostFilter<TCatalogItemType> postFilter)
        {
            this.postFilters.Remove(postFilter);
        }

        /// <summary>
        /// Removes the <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <param name="index">The index.</param>
        public void RemovePostFilter(int index)
        {
            this.postFilters.RemoveAt(index);
        }

        /// <summary>
        /// Gets the list of <see cref="IPostFilter&lt;TCatalogItemType&gt;"/>.
        /// </summary>
        /// <returns>a list of all PostFilters</returns>
        public ReadOnlyCollection<IPostFilter<TCatalogItemType>> PostFilters
        {
            get
            {
                return new ReadOnlyCollection<IPostFilter<TCatalogItemType>>(this.postFilters);
            }
        }

        /// <summary>
        /// Processes the specified query, returning the hybridized predicted ratings for the given items.
        /// </summary>
        /// <param name="query">The query.</param>
        /// <param name="user">The user.</param>
        /// <returns></returns>
        internal IList<Prediction<TCatalogItemType>> Process(IQueryable<TCatalogItemType> query, User user)
        {
            return this.Process(query, user, FlowRecommendation.PredictRating, -1, -1);
        }

        /// <summary>
        /// Processes the specified query, returning the top items.
        /// </summary>
        /// <param name="query">The query.</param>
        /// <param name="user">The user.</param>
        /// <param name="skip">The skip.</param>
        /// <param name="take">The take.</param>
        /// <returns></returns>
        internal IList<Prediction<TCatalogItemType>> Process(IQueryable<TCatalogItemType> query, User user, int skip, int take)
        {
            return this.Process(query, user, FlowRecommendation.TopN, skip, take);
        }

        /// <summary>
        /// Given an identifier for the relation type (e.g. "rating") and the proper key, the predicted value
        /// for this relation is returned.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="keys">The keys.</param>
        /// <returns></returns>
        internal double PredictRelation(RelationType relationType, int[] keys)
        {
            Contract.Precondition(this.hybridizer != null, "Hybridizer is not set.");
            Contract.Precondition(this.engines.Count > 0, "Flow must contain engines.");            
            if(this.engines.Count != 1)
            {
                throw new InvalidOperationException("PredictRelation is not defined for flows with more than one engine.");
            }
            if (this.engines[0] is IEntityRelationPredictor)
            {
                IEntityRelationPredictor entityRelationPredictor = (IEntityRelationPredictor)this.engines[0];
                return entityRelationPredictor.PredictRelation(relationType, keys);
            }
            else
            {
                throw new InvalidOperationException("Engine is not of type IEntityRelationPredictor.");
            }
        }

        /// <summary>
        /// This is the TopN method.
        /// </summary>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="keys">Combination of other relation's keys. E.g. for predicting top-n items, keys[0] contains the user_id for which top-n ratings should be predicted.</param>
        /// <param name="count">The count.</param>
        /// <returns>An ordered array of keys to that entity.</returns>
        internal int[] PredictEntity(EntityType entityType, int[] keys, int count)
        {
            Contract.Precondition(this.hybridizer != null, "Hybridizer is not set.");
            Contract.Precondition(this.engines.Count > 0, "Flow must contain engines.");
            if (this.engines.Count != 1)
            {
                throw new InvalidOperationException("PredictEntity is not defined for flows with more than one engine.");
            }
            if (!(this.engines[0] is IEntityRelationPredictor))
            {
                IEntityRelationPredictor entityRelationPredictor = (IEntityRelationPredictor)this.engines[0];
                return entityRelationPredictor.PredictEntity(entityType, keys, count);
            }
            else
            {
                throw new InvalidOperationException("Engine is not of type IEntityRelationPredictor.");
            }
        }

        private IList<Prediction<TCatalogItemType>> Process(IQueryable<TCatalogItemType> query, User user, FlowRecommendation flowRecommendation, int skip, int take)
        {
            int steps = this.preFilters.Count + this.engines.Count + 1 + this.postFilters.Count;
            int step = -1;

            this.SendProgressNotification((++step * 100) / steps);

            Contract.Precondition(query != null, "Query can not be null.");

            Contract.Precondition(this.hybridizer != null, "Hybridizer is not set.");

            Contract.Precondition(this.engines.Count > 0,"Flow must contain engines.");

            Contract.Precondition(FlowRecommendation.None != flowRecommendation, "flowRecommendation must be initialized.");

            // let the prefilters redefine the query
            foreach (IPreFilter<TCatalogItemType> pf in this.preFilters)
            {
                IQueryable<TCatalogItemType> tmpPrefilter = pf.PreFilter(query);

                Contract.Precondition(tmpPrefilter != null, "A pre-filter can not return null.");

                // note: we want to check that the new query is smaller than the original but this will
                //       probably cause the query to be executed so this test is only done for debug builds
#if DEBUG
                Contract.Precondition(tmpPrefilter.Count() <= query.Count(), "A pre-filter is not allowed to add items.");
#endif 

                query = tmpPrefilter;

                this.SendProgressNotification((++step * 100) / steps);
            }

            // hybridize the results from the engines

            Dictionary<TCatalogItemType, IList<EngineResult<TCatalogItemType>>> tmpEngine = new Dictionary<TCatalogItemType,IList<EngineResult<TCatalogItemType>>>(query.Count());
            foreach (Engine<TCatalogItemType, TUserActionType> e in this.engines)
            {
                IList<Prediction<TCatalogItemType>> tmp = e.Predict(query, user);

                this.SendProgressNotification((++step * 100) / steps);
                if (tmp != null)
                {
                    foreach (Prediction<TCatalogItemType> p in tmp)
                    {
                        if (p != null)
                        {
                            if (!tmpEngine.ContainsKey(p.CatalogItem))
                                tmpEngine[p.CatalogItem] = new List<EngineResult<TCatalogItemType>>(this.engines.Count);
                            tmpEngine[p.CatalogItem].Add(new EngineResult<TCatalogItemType>(p, e.Id));
                        }
                    }
                }
            }

            IList<Prediction<TCatalogItemType>> result = this.hybridizer.Hybridize(new ReadOnlyDictionary<TCatalogItemType, IList<EngineResult<TCatalogItemType>>>(tmpEngine));

            Contract.Precondition(result.Count == tmpEngine.Count, "A hybridizer must hybridize ALL engine results.");

            this.SendProgressNotification((++step * 100) / steps);

            // postfilters may exclude items from being returned
            foreach (IPostFilter<TCatalogItemType> pf in this.postFilters)
            {
                IList<Prediction<TCatalogItemType>> tmpPostFilter = pf.PostFilter(new ReadOnlyCollection<Prediction<TCatalogItemType>>(result));

                Contract.Precondition(tmpPostFilter != null, "A post-filter can not return null.");
                Contract.Precondition(tmpPostFilter.Count() <= result.Count(), "A post-filter is not allowed to add items.");

                result = tmpPostFilter;

                this.SendProgressNotification((++step * 100) / steps);
            }

            if (FlowRecommendation.TopN == flowRecommendation)
            {
                result = result.OrderByDescending(prediction => prediction.Value).
                    ThenByDescending(prediction => prediction.CatalogItem.ItemId).Skip(skip).Take(take).ToList();
            }

            return result;
        }

        /// <summary>
        /// Gets the Id of this flow.
        /// </summary>
        /// <remarks>
        /// Flows are assigned an Id implicitly when they are created. This Id has to be used, after the flow is registered with
        /// the core (<see cref="Recommender&lt;TUserActionType&gt;.RegisterFlow"/>), when triggering a prediction 
        /// (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User, int, int)"/>) 
        /// (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User)"/>). 
        /// </remarks>
        /// <value>The Id.</value>
        public FlowId Id
        {
            get
            {
                return this.id;
            }
        }

    }
}
