﻿using System;
using System.Globalization;

using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Represents a rating.
    /// </summary>
    /// <remarks>
    /// The value of a rating must be between (including) -1.0 and 1.0 (including).
    /// </remarks>
    [Obsolete("Use similar functionality in the RatingAction class", true)]
    [Serializable]
    public class Rating
    {
        private double ratingValue;

        /// <summary>
        /// Represents the smallest possible value of a Rating. This field is
        /// constant.
        /// </summary>
        public const double MinValue = -1.0;

        /// <summary>
        /// Represents the largest possible value of a Rating. This field is
        /// constant.
        /// </summary>
        public const double MaxValue = 1.0;

        /// <summary>
        /// Verifies the rating value is within range.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public static bool VerifyRatingValueIsWithinRange(double value)
        {
            return value >= Rating.MinValue && value <= Rating.MaxValue;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Rating"/> class.
        /// </summary>
        /// <param name="ratingValue">The rating value.</param>
        public Rating(double ratingValue)
        {
            this.Value = ratingValue;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Rating"/> class,
        /// normalizing the value to [-1,1].
        /// </summary>
        /// <param name="ratingValue">The rating value.</param>
        /// <param name="lowerRangeBound">The lower range bound to normalize to.</param>
        /// <param name="upperRangeBound">The upper range bound to normalize to.</param>
        public Rating(double ratingValue, double lowerRangeBound, double upperRangeBound)
        {
            Contract.Precondition(upperRangeBound - lowerRangeBound > 0, "Invalid range.  Range must be > 0.");
            Contract.Precondition(upperRangeBound > lowerRangeBound, "Invalid range.  upperRangeBound must be > lowerRangeBound.");
            Contract.Precondition(ratingValue <= upperRangeBound && ratingValue >= lowerRangeBound, "Invalid value.  Value must be <= upperRangeBound AND >= lowerRangeBound");

            //First normalize to [0,1].  There may be a better way to do this in one shot, but nothing comes to mind.
            double value0to1 = (ratingValue - lowerRangeBound) / (upperRangeBound - lowerRangeBound);

            //Now normalize to [-1,1].
            this.Value = (value0to1 * 2) - 1;
        }

        /// <summary>
        /// Gets or sets the value.
        /// </summary>
        /// <value>The value.</value>
        public Double Value
        {
            get 
            { 
                return this.ratingValue; 
            }

            set
            {
                Contract.Precondition(
                    Rating.VerifyRatingValueIsWithinRange(value),
                        "{0} out of range. Value must be <= {1} and >= {2}.", 
                        this.GetType().ToString(), Rating.MaxValue, Rating.MinValue);

                this.ratingValue = value;
            }
        }

        /// <summary>
        /// Returns a <see cref="T:System.String"/> that represents the current <see cref="T:System.Object"/>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String"/> that represents the current <see cref="T:System.Object"/>.
        /// </returns>
        public override string ToString()
        {
            return this.Value.ToString(CultureInfo.CurrentCulture);
        }

        /// <summary>
        /// Adjusteds the rating value to a new range.
        /// </summary>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        [Obsolete("Use RatingAction.ScaleRatingToApplication instead", true)]
        public double ScaledValue(double lowerRangeBound, double upperRangeBound)
        {
            return Rating.ScaledValue(this, lowerRangeBound, upperRangeBound);
        }

        /// <summary>
        /// Adjusteds the rating value to a new range.
        /// </summary>
        /// <param name="rating">The rating.</param>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        [Obsolete("Use RatingAction.ScaleRatingToApplication instead", true)]
        public static double ScaledValue(Rating rating, double lowerRangeBound, double upperRangeBound)
        {
            return Rating.ScaledValue(rating.Value, lowerRangeBound, upperRangeBound);
        }

        /// <summary>
        /// Adjusteds the rating value to a new range.
        /// </summary>
        /// <param name="ratingValue">The rating value.</param>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        [Obsolete("Use RatingAction.ScaleRatingToApplication instead", true)]
        public static double ScaledValue(double ratingValue, double lowerRangeBound, double upperRangeBound)
        {
            Contract.Precondition(Rating.VerifyRatingValueIsWithinRange(ratingValue), "ArgumentOutOfRange: ratingValue");

            double value0to1 = (ratingValue + 1) / 2;
            double scaledValue = (value0to1 * (upperRangeBound - lowerRangeBound)) + lowerRangeBound;
            return scaledValue;
        }

        /// <summary>
        /// Performs an implicit conversion from <see cref="MyMediaProject.RecommenderSystem.Core.Rating"/> to <see cref="System.Double"/>.
        /// </summary>
        /// <param name="ratingValue">The rating value.</param>
        /// <returns>The result of the conversion.</returns>
        public static implicit operator double(Rating ratingValue)
        {
            return ratingValue.Value;
        }

        /// <summary>
        /// Performs an implicit conversion from <see cref="System.Double"/> to <see cref="MyMediaProject.RecommenderSystem.Core.Rating"/>.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>The result of the conversion.</returns>
        public static implicit operator Rating(double value)
        {
            return new Rating(value);
        }

    }
}
