using System.Collections.Generic;
using System.Collections.ObjectModel;
using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Data.Common;
using System.Transactions;
using System.Data.Linq.Mapping;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization.Formatters;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core.Databases;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// An abstraction of the LINQ based class to database mapping.  Provides access to all of the entities logically associated
    /// with a <see cref="CatalogItem"/>.
    /// </summary>
    public sealed class CatalogItemStore :
        IEnumerable<CatalogItem>,
        IQueryable<CatalogItem>,
        ITrackableDataStore,
        IDisposable
    {
        private object locker = new object();
        private bool isInitialized;
        private string dbConnectionString;
        private CatalogItemStoreDataContext dataContext;
        private bool ownsDataContext;

        internal CatalogItemStoreDataContext DataContext
        {
            get { return this.GetDatabaseContext(); }
        }

        #region constructor and initialization

        /// <summary>
        /// Initializes a new instance of the <see cref="CatalogItemStore"/> class.
        /// </summary>
        /// <param name="programId">The program id.</param>
        /// <remarks>
        /// A store is identified by a program and a store id. The store id is used to distinguish multiple stores within an application.
        /// The program id is used to distinguish different program (instances).
        /// </remarks>
        public CatalogItemStore(ProgramId programId)
        {
            this.dbConnectionString = Recommender.GetDatabaseConnectionString(programId);
            this.dataContext = new CatalogItemStoreDataContext(this.dbConnectionString);
            this.ownsDataContext = true;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CatalogItemStore"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        public CatalogItemStore(CatalogItemStoreDataContext context)
        {
            this.dbConnectionString = context.Connection.ConnectionString;
            this.dataContext = context;
            this.ownsDataContext = false;
        }

        /// <summary>
        /// Gets the database context.
        /// </summary>
        /// <returns></returns>
        /// <remarks>
        /// The context is meant for logical operations on the datebase, and should be
        /// explicitly disposed of when no longer needed.  (i.e. use a "using" statement).
        /// Further, there is a bug in Linq-Sql, such that DataContext.Connection.Dispose should be called.
        /// </remarks>
        internal CatalogItemStoreDataContext GetDatabaseContext()
        {
            //CatalogItemStoreDataContext storeDB = new CatalogItemStoreDataContext(this.dbConnection);
            return this.dataContext;
        }

        /// <summary>
        /// Initializes this instance.
        /// </summary>
        public void Initialize()
        {
            CatalogItemStoreDataContext context = this.GetDatabaseContext();
            {
                lock (locker)
                {
                    if (!context.DatabaseExists())
                    {
                        TraceLog.Write(LogEventType.Information, "Recommender store database does not exist, creating it.");
                        context.CreateDatabase();
                    }

                    isInitialized = true;
                }
                //context.Connection.Dispose();  //Workaround for Linq-Sql bug
            }
        }

        /// <summary>
        /// Initializes and clears the store.
        /// </summary>
        public void InitializeAndClear()
        {
            lock (locker)
            {
                CatalogItemStoreDataContext context = this.GetDatabaseContext();
                {
                    if (context.DatabaseExists())
                    {
                        this.ClearDatabaseContent();
                    }
                    //context.Connection.Dispose();  //Workaround for Linq-Sql bug
                }
            }
            this.Initialize();
        }

        /// <summary>
        /// Returns whether or not the store exists.
        /// </summary>
        /// <returns></returns>
        public bool StoreExists()
        {
            bool storeExists = false;
            CatalogItemStoreDataContext context = this.GetDatabaseContext();
            {
                storeExists = context.DatabaseExists();
                //context.Connection.Dispose();  //Workaround for Linq-Sql bug
            }
            return storeExists;
        }

        /// <summary>
        /// Deletes the database.
        /// </summary>
        public void DeleteDatabase()
        {
            CatalogItemStoreDataContext context = this.GetDatabaseContext();
            {
                if (context.DatabaseExists())
                {
                    context.DeleteDatabase();
                }
                //context.Connection.Dispose();  //Workaround for Linq-Sql bug
            }
        }

        #endregion

        #region Modification Functions
        /// <summary>
        /// Removes all items.
        /// </summary>
        internal void RemoveAll()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            this.ClearDatabaseContent();
        }
        #endregion

        #region query functions

        /// <summary>
        /// Gets the change number.
        /// </summary>
        /// <remarks>
        /// For an unchanged store, 0 will be returned.
        /// </remarks>
        /// <value>The change number.</value>
        public int ChangeNumber
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

                int changeNumber = 0;
                lock (locker)
                {
                    CatalogItemStoreDataContext context = this.GetDatabaseContext();
                    {
                        changeNumber = StoreTracker<ItemStoreChange>.ChangeNumber(context.ItemStoreChanges);
                        //context.Connection.Dispose();  //Workaround for Linq-Sql bug
                    }
                }
                return changeNumber;
            }
        }

        /// <summary>
        /// Gets the UserAction with the specified item id.
        /// </summary>
        /// <value></value>
        public CatalogItem this[int itemId]
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
                CatalogItem item;

                CatalogItemStoreDataContext context = this.GetDatabaseContext();
                {
                    item = context.CatalogItems.Single<CatalogItem>(c => c.ItemId == itemId);
                    //context.Connection.Dispose();  //Workaround for Linq-Sql bug
                }
                return item;
            }
        }

        /// <summary>
        /// Gets the differences.
        /// </summary>
        /// <param name="startChangeNumber">The start change number.</param>
        /// <param name="endChangeNumber">The end change number.</param>
        /// <returns>a list with all changes.</returns>
        /// <remark>
        /// Gets a list of all changes that make up the differences between the two given revisions starting with and including the
        /// change number specified in startChangeNumberand ending with and including the change number specified in endChangeNumber.
        /// All change numbers are ascending integer numbers, starting with 1 for the first change of the store.
        /// The list of changes does not contain every single change, but a list of all changes that have to be taken into
        /// account to get a picture of the diff between the startChangeNumber and the endChangeNumber.
        /// I.e. if in change number 50 an item with the id "a" gets added, and then gets removed in change number 60, the list of changes between 40 and 70 will not show
        /// this change because it does not make a difference for the state of the store.
        /// </remark>
        public IList<IDataStoreChange> GetDifferences(int startChangeNumber, int endChangeNumber)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            IList<IDataStoreChange> differences;
            lock (locker)
            {
                CatalogItemStoreDataContext context = this.GetDatabaseContext();
                {
                    differences = StoreTracker<ItemStoreChange>.GetDifferences(context.ItemStoreChanges, startChangeNumber, endChangeNumber);
                    //context.Connection.Dispose();  //Workaround for Linq-Sql bug
                }
            }

            return differences;
        }
        #endregion

        #region private functions

        private void ClearDatabaseContent()
        {
            CatalogItemStoreDataContext context = this.GetDatabaseContext();
            if (context.DatabaseExists())
            {
                context.EmptyDatabase();
            }
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        public void Dispose(bool disposing)
        {
            if (this.ownsDataContext && disposing)
            {
                this.dataContext.Connection.Dispose();
                this.dataContext.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion


        #region IEnumerable<UserAction> Members

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1"/> that can be used to iterate through the collection.
        /// </returns>
        IEnumerator<CatalogItem> IEnumerable<CatalogItem>.GetEnumerator()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
            IEnumerator<CatalogItem> enumerator;
            CatalogItemStoreDataContext context = this.GetDatabaseContext();
            {
                enumerator = context.CatalogItems.GetEnumerator();

                //context.Connection.Dispose();  //Workaround for Linq-Sql bug
            }
            return enumerator;
        }

        /// <summary>
        /// The number of UserAction elements contained in the store.
        /// </summary>
        /// <remarks>
        /// This instance should be used in leiu of the IEnumerable extension count methods,
        /// which will not necessarily execute on the store server.
        /// </remarks>
        /// <returns>
        /// The number of UserAction elements contained in the store.
        /// </returns>
        public int Count()
        {
            return this.GetDatabaseContext().CatalogItems.Count<CatalogItem>();
        }
        #endregion

        #region IEnumerable Members

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator"/> object that can be used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
            IEnumerator enumerator;
            CatalogItemStoreDataContext context = this.GetDatabaseContext();
            {
                enumerator = context.CatalogItems.GetEnumerator();
                //context.Connection.Dispose();  //Workaround for Linq-Sql bug
            }
            return enumerator;
        }

        #endregion

        #region IQueryable Members

        Type IQueryable.ElementType
        {
            get { return (this.GetDatabaseContext().CatalogItems as IQueryable).ElementType; }
        }

        /// <summary>
        /// Gets the expression tree that is associated with the instance of <see cref="T:System.Linq.IQueryable"/>.
        /// </summary>
        /// <value></value>
        /// <returns>The <see cref="T:System.Linq.Expressions.Expression"/> that is associated with this instance of <see cref="T:System.Linq.IQueryable"/>.</returns>
        System.Linq.Expressions.Expression IQueryable.Expression
        {
            get { return (this.GetDatabaseContext().CatalogItems as IQueryable).Expression; }
        }

        IQueryProvider IQueryable.Provider
        {
            get { return (this.GetDatabaseContext().CatalogItems as IQueryable).Provider; }
        }

        #endregion
    }
}