using System.Collections.Generic;
using System.Collections.ObjectModel;
using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Data.Common;
using System.Data.Linq.Mapping;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization.Formatters;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Implements a versioned generic storage solution with notification support.
    /// </summary>
    /// <remarks>
    /// This class can be used to store any kind of information of type UserAction. It keeps all data in memory and uses SQL Express
    /// for back-end storage.
    /// It behaves like a dictionary: keys are of type <see cref="RecommenderId"/> and values are of type UserAction.
    /// Every change to the store is marked with a change number and is fired as an <see cref="RecommenderEvent"/>. Receiving components
    /// can then pull changes from this store based on the change numbers.
    /// Specialized versions of this class might fire specialized events.
    /// Note: TContType must be marked as serializable.
    /// </remarks>
    /// <example>
    /// <code>
    /// public MyChangeNotification : RecommenderEvent
    /// {
    ///   ....
    /// }
    /// 
    /// 
    /// [Serializable]
    /// public class MyClass
    /// {
    ///   ....
    /// }
    /// 
    /// public MyClassStore : UserActionStore&lt;MyClass&gt;
    /// {
    ///     protected void SendStoreChangedNotification(StoreChangedEventArgs storeChangedEventArgs)
    ///     {
    ///         new MyChangeNotification().Fire(storeChangedEventArgs);
    ///     }
    /// }
    /// </code>
    /// </example>
    public class UserActionStore :
        IEnumerable<UserAction>,
        IQueryable<UserAction>,
        ITrackableDataStore,
        IDisposable
    {
        private PersistentStoreId storeId;
        private object locker = new object();
        private bool isInitialized;
        private string dbConnectionString;
        private MyMediaDataContext dataContext;
        private bool ownsDataContext;

        internal MyMediaDataContext DataContext
        {
            get { return dataContext; }
        }

        #region constructor and initialization

        /// <summary>
        /// Initializes a new instance of the <see cref="UserActionStore"/> class.
        /// </summary>
        /// <param name="programId">The program id.</param>
        /// <remarks>
        /// A store is identified by a program and a store id. The store id is used to distinguish multiple stores within an application.
        /// The program id is used to distinguish different program (instances).
        /// </remarks>
        public UserActionStore(ProgramId programId)
        {
            this.storeId = Constants.FeedbackStoreId;

            this.dbConnectionString = Recommender.GetDatabaseConnectionString(programId);

            this.dataContext = new MyMediaDataContext(this.dbConnectionString);
            this.ownsDataContext = true;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="UserActionStore"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        public UserActionStore(MyMediaDataContext context)
        {
            this.storeId = Constants.FeedbackStoreId;
            this.dbConnectionString = context.Connection.ConnectionString;
            this.dataContext = context;
            this.ownsDataContext = false;
        }

        /// <summary>
        /// Initializes this instance.
        /// </summary>
        public void Initialize()
        {
            lock (locker)
            {
                if (!this.dataContext.DatabaseExists())
                {
                    TraceLog.Write(LogEventType.Information, "Recommender store database does not exist, creating it.");
                    this.dataContext.CreateDatabase();
                }

                isInitialized = true;
            }
        }

        /// <summary>
        /// Initializes and clears the store.
        /// </summary>
        public void InitializeAndClear()
        {
            lock (locker)
            {
                if (this.dataContext.DatabaseExists())
                {
                    this.ClearDatabaseContent();
                }
            }
            this.Initialize();
        }

        /// <summary>
        /// Returns whether or not the store exists.
        /// </summary>
        /// <returns></returns>
        public bool StoreExists()
        {
            bool storeExists = false;
            storeExists = this.dataContext.DatabaseExists();
            return storeExists;
        }

        /// <summary>
        /// Deletes the database.
        /// </summary>
        public void DeleteDatabase()
        {
            if (this.dataContext.DatabaseExists())
            {
                this.dataContext.DeleteDatabase();
            }
        }

        #endregion

        #region modification functions

        /// <summary>
        /// Processes feedback.
        /// </summary>
        /// <param name="feedbackItems">The feedback items.</param>
        /// <remarks>
        /// Feedback is (permanently) stored by the framework. A copy of the related catalog item is automatically
        /// (and permanently) stored. After storing it, the feedback is forwarded to all currently active engines that can
        /// then process it.
        /// It is possible to have multiple feedback items for a catalog item.
        /// </remarks>
        public void ProcessFeedback(IEnumerable<UserAction> feedbackItems)
        {
            List<UserAction> feedbackItemUpdateList = new List<UserAction>();
            List<UserAction> feedbackItemOriginalList = new List<UserAction>();
            List<UserAction> feedbackItemNewList = new List<UserAction>();

            lock (this)
            {
                //Validate all of the user feedback, to ensure it is 
                //consistent with the rest of the system.
                foreach (UserAction feedbackItem in feedbackItems)
                {
                    Contract.Precondition(feedbackItem.ItemId != default(int), "ItemId inside a feedback item must be set.");
                    Contract.Precondition(feedbackItem.UserId != default(int), "UserId inside a feedback item must be set.");

                    IQueryable<UserAction> existingFeedbackItems = this.dataContext.UserActions.Where(fb => fb.UserId == feedbackItem.UserId &&
                        fb.ItemId == feedbackItem.ItemId && fb.GetType() == feedbackItem.GetType()).Where(feedbackItem.EquivalenceQuery());

                    Contract.Postcondition(existingFeedbackItems.Count() <= 1, "EquivalenceQuery for type {0} returned more than one equivalent item." +
                        "  This is an invalid state.  Look at the implementation of EquvalenceQuery for this type." +
                        "  This query must return zero or one objects.",
                        feedbackItem.GetType());

                    UserAction existingFeedbackItem = existingFeedbackItems.SingleOrDefault();

                    if (null == existingFeedbackItem)
                    {
                        feedbackItemNewList.Add(feedbackItem);
                    }
                    else
                    {
                        feedbackItemUpdateList.Add(feedbackItem);
                        feedbackItemOriginalList.Add(existingFeedbackItem);
                    }
                }

                if (feedbackItemUpdateList.Count() > 0)
                    this.UpdateItems(feedbackItemUpdateList.ToArray(), feedbackItemOriginalList.ToArray());
                if (feedbackItemNewList.Count() > 0)
                    this.AddItems(feedbackItemNewList.ToArray());
            }
        }

        /// <summary>
        /// Adds the item.
        /// </summary>
        /// <param name="item">The item.</param>
        internal void AddItem(UserAction item)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            InternalAddItems(new UserAction[] { item });
        }

        /// <summary>
        /// Adds the items.
        /// </summary>
        /// <param name="newItems">The new items.</param>
        internal void AddItems(UserAction[] newItems)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(newItems != null, "There has to be at least one item to add.");

            Contract.Precondition(newItems.Length != 0, "There has to be at least one item to add.");

            InternalAddItems(newItems);
        }

        ///// <summary>
        ///// Updates the item.
        ///// </summary>
        ///// <param name="item">The item.</param>
        //internal void UpdateItem(UserAction item)
        //{
        //    Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

        //    Contract.Precondition(item != null, "CatalogItem can not be null.");

        //    InternalUpdateItems(new UserAction[] { item });
        //}

        /// <summary>
        /// Updates the items.
        /// </summary>
        /// <param name="updatedItems">The updated items.</param>
        /// <param name="originalItems">The original items.</param>
        public void UpdateItems(UserAction[] updatedItems, UserAction[] originalItems)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(updatedItems != null, "There has to be at least one item to update.");

            Contract.Precondition(updatedItems.Length != 0, "There has to be at least one item to update.");

            InternalUpdateItems(updatedItems, originalItems);
        }

        /// <summary>
        /// Removes the item.
        /// </summary>
        /// <param name="itemId">The item id.</param>
        public void RemoveItem(int itemId)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(itemId != 0, "ItemKey can not be zero.");

            InternalRemoveItems(new int[] { itemId });
        }

        /// <summary>
        /// Removes the items.
        /// </summary>
        /// <param name="itemIds">The item ids.</param>
        public void RemoveItems(int[] itemIds)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(itemIds != null, "There has to be at least one item to remove.");

            Contract.Precondition(itemIds.Length != 0, "There has to be at least one item to remove.");

            InternalRemoveItems(itemIds);
        }

        internal void SeparateForTestSet(int[] itemIds)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(itemIds != null, "There has to be at least one item to remove.");

            Contract.Precondition(itemIds.Length != 0, "There has to be at least one item to remove.");

            SetTestStatus(itemIds);
        }

        internal void RemoveAllTestItems()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            ResetTestStatus();
        }

        /// <summary>
        /// Removes all items.
        /// </summary>
        public void RemoveAll()
        {
            this.ClearDatabaseContent();
        }
        #endregion

        #region query functions

        /// <summary>
        /// Gets the id.
        /// </summary>
        /// <value>The id.</value>
        public PersistentStoreId Id
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
                return storeId;
            }
        }

        /// <summary>
        /// Determines whether the UserActionStore contains an item with the specified item id.
        /// </summary>
        /// <param name="itemId">The item id.</param>
        /// <returns>
        /// 	<c>true</c> if it contains the specified item id; otherwise, <c>false</c>.
        /// </returns>
        public bool Contains(int itemId)
        {
            bool containsItem = false;
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            containsItem = null != this.dataContext.UserActions.FirstOrDefault<UserAction>(c => c.Key == itemId);
            return containsItem;
        }


        /// <summary>
        /// Gets the change number.
        /// </summary>
        /// <remarks>
        /// For an unchanged store, 0 will be returned.
        /// </remarks>
        /// <value>The change number.</value>
        public int ChangeNumber
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

                int changeNumber = 0;
                lock (locker)
                {
                    changeNumber = StoreTracker<UserActionStoreChange>.ChangeNumber(this.dataContext.UserActionStoreChanges);
                }
                return changeNumber;
            }
        }

        /// <summary>
        /// Gets the UserAction with the specified item id.
        /// </summary>
        /// <value></value>
        public UserAction this[int itemKey]
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
                UserAction userAction;

                userAction = this.dataContext.UserActions.Single<UserAction>(c => c.Key == itemKey);
                return userAction;
            }
        }

        /// <summary>
        /// Gets the differences.
        /// </summary>
        /// <param name="startChangeNumber">The start change number.</param>
        /// <param name="endChangeNumber">The end change number.</param>
        /// <returns>a list with all changes.</returns>
        /// <remark>
        /// Gets a list of all changes that make up the differences between the two given revisions starting with and including the
        /// change number specified in startChangeNumberand ending with and including the change number specified in endChangeNumber.
        /// All change numbers are ascending integer numbers, starting with 1 for the first change of the store.
        /// The list of changes does not contain every single change, but a list of all changes that have to be taken into
        /// account to get a picture of the diff between the startChangeNumber and the endChangeNumber.
        /// I.e. if in change number 50 an item with the id "a" gets added, and then gets removed in change number 60, the list of changes between 40 and 70 will not show
        /// this change because it does not make a difference for the state of the store.
        /// </remark>
        public IList<IDataStoreChange> GetDifferences(int startChangeNumber, int endChangeNumber)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            IList<IDataStoreChange> differences;
            lock (locker)
            {
                differences = StoreTracker<UserActionStoreChange>.GetDifferences(this.dataContext.UserActionStoreChanges, startChangeNumber, endChangeNumber);
            }

            return differences;
        }
        #endregion

        #region IEnumerable Members

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator"/> object that can be used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
            IEnumerator enumerator;
            enumerator = this.dataContext.UserActions.GetEnumerator();
            return enumerator;
        }

        #endregion

        #region IEnumerable<UserAction> Members

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1"/> that can be used to iterate through the collection.
        /// </returns>
        IEnumerator<UserAction> IEnumerable<UserAction>.GetEnumerator()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
            IEnumerator<UserAction> enumerator;
            enumerator = (from dbUserAction in this.dataContext.UserActions
                          where dbUserAction is UserAction
                          select dbUserAction as UserAction).GetEnumerator();

            return enumerator;
        }

        /// <summary>
        /// The number of UserAction elements contained in the store.
        /// </summary>
        /// <remarks>
        /// This instance should be used in leiu of the IEnumerable extension count methods,
        /// which will not necessarily execute on the store server.
        /// </remarks>
        /// <returns>
        /// The number of UserAction elements contained in the store.
        /// </returns>
        public int Count()
        {
            return this.dataContext.UserActions.Count<UserAction>();
        }
        #endregion

        #region private functions

        /// <summary>
        /// Sends the store changed notification.
        /// </summary>
        /// <param name="storeChangedEventArgs">The <see cref="MyMediaProject.RecommenderSystem.Core.StoreChangedEventArgs"/> instance containing the event data.</param>
        protected virtual void SendStoreChangedNotification(StoreChangedEventArgs storeChangedEventArgs)
        {
            new FeedbackStoreChangedEvent().Fire(storeChangedEventArgs);
        }

        private void InternalAddItems(UserAction[] newItems)
        {
            int newChangeNumber;
            lock (locker)
            {
                newChangeNumber = ChangeNumber + 1;

                foreach (UserAction item in newItems)
                {
                    // Add item to the db
                    item.PrepareForInsert(this.dataContext);
                    this.dataContext.UserActions.InsertOnSubmit(item);
                }
                this.dataContext.SubmitChanges();
                foreach (UserAction item in newItems)
                {
                    Contract.Postcondition(0 != item.Key);
                    // Add change to the db
                    StoreTracker<UserActionStoreChange>.AddChange(
                        this.dataContext.UserActionStoreChanges,
                        newChangeNumber,
                        StoreChangeType.Added,
                        item.Key);
                }
                this.dataContext.SubmitChanges();
            }

            SendStoreChangedNotification(new StoreChangedEventArgs(storeId, newChangeNumber));
        }

        private void InternalUpdateItems(UserAction[] changedItems, UserAction[] originalItems)
        {
            int newChangeNumber;
            lock (locker)
            {
                newChangeNumber = ChangeNumber + 1;

                for(int i = 0;i < changedItems.Count();i++)
                {
                    UserAction newItem = changedItems[i];
                    UserAction originalItem = originalItems[i];

                    newItem.ResolveEquivalentAction(originalItem);

                    // Add change to the db
                    StoreTracker<UserActionStoreChange>.AddChange(
                        this.dataContext.UserActionStoreChanges,
                        newChangeNumber,
                        StoreChangeType.Changed,
                        originalItem.Key);
                }

                this.dataContext.SubmitChanges();
            }
            SendStoreChangedNotification(new StoreChangedEventArgs(storeId, newChangeNumber));
        }

        private void SetTestStatus(int[] itemIds)
        {
            lock (locker)
            {
                foreach (int itemId in itemIds)
                {
                    // Remove item from db
                    UserAction action = this.dataContext.UserActions.Single<UserAction>(c => c.Key == itemId);
                    action.TestAction = true;
                }

                this.dataContext.SubmitChanges();
            }
        }

        private void ResetTestStatus()
        {
            lock (locker)
            {
                this.dataContext.ExecuteCommand("UPDATE [UserAction] SET [TestAction] = 0");
            }
        }

        private void InternalRemoveItems(int[] itemIds)
        {
            int newChangeNumber;
            lock (locker)
            {
                newChangeNumber = ChangeNumber + 1;

                foreach (int itemId in itemIds)
                {
                    // Remove item from db
                    this.dataContext.UserActions.DeleteOnSubmit(this.dataContext.UserActions.Single<UserAction>(c => c.Key == itemId));

                    // Add change to the db
                    StoreTracker<UserActionStoreChange>.AddChange(
                        this.dataContext.UserActionStoreChanges,
                        newChangeNumber,
                        StoreChangeType.Removed,
                        itemId);
                }

                this.dataContext.SubmitChanges();
            }
            SendStoreChangedNotification(new StoreChangedEventArgs(storeId, newChangeNumber));
        }

        private void ClearDatabaseContent()
        {
            if (this.dataContext.DatabaseExists())
            {
                this.dataContext.EmptyDatabase();
            }
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.ownsDataContext)
            {
                this.dataContext.Connection.Dispose();
                this.dataContext.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        #region IQueryable Members

        Type IQueryable.ElementType
        {
            get { return (this.dataContext.UserActions as IQueryable).ElementType; }
        }

        /// <summary>
        /// Gets the expression tree that is associated with the instance of <see cref="T:System.Linq.IQueryable"/>.
        /// </summary>
        /// <value></value>
        /// <returns>The <see cref="T:System.Linq.Expressions.Expression"/> that is associated with this instance of <see cref="T:System.Linq.IQueryable"/>.</returns>
        System.Linq.Expressions.Expression IQueryable.Expression
        {
            get { return (this.dataContext.UserActions as IQueryable).Expression; }
        }

        IQueryProvider IQueryable.Provider
        {
            get { return (this.dataContext.UserActions as IQueryable).Provider; }
        }

        #endregion
    }
}