using System;
using System.Text;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using System.Data.Linq;
using MyMediaProject.RecommenderSystem.Core;
using System.Collections.ObjectModel;
using System.IO;

namespace RecommenderCoreTest
{
    [TestClass]
    public class UserActionStoreTest
    {
        private UserActionStore  userActionStore;
        private UserStore userStore;
        private CatalogItemStore itemStore;
        private List<int> eventedChangeNumbers;
        private StoreChangeEvent storeChangeEvent;
        private RecommenderEventHandler eventHandler;
        public const string TestStoreId = "DF6206D6-2A62-4671-900F-78AEF4342814";
        public const string CopyStoreId = "599178B4-9DA3-49f1-BD5F-6FA74F5324ED";
        public const string TestApplicationId = "00D74598-1516-416b-8A88-BFD58CADB71B";
        public const string TestApplicationFriendlyName = "UserActionTest";

        [ClassInitialize]
        static public void ClassInitialize(TestContext testContext)
        {
            UserActionStoreTest.ClassCleanUp();
        }
        
        [ClassCleanup]
        static public void ClassCleanUp()
        {
            using (UserActionStore userActionStore = new UserActionStore(
                new ProgramId(new Guid(TestApplicationId), TestApplicationFriendlyName)))

            {
                userActionStore.RemoveAll();
                userActionStore.DeleteDatabase();
            }
        }

        public UserActionStoreTest()
        {
        }

        [TestInitialize]
        public void Initialize()
        {
            eventedChangeNumbers = new List<int>();
            this.userActionStore = new UserActionStore(
                new ProgramId(new Guid(TestApplicationId), TestApplicationFriendlyName));
            this.itemStore = new CatalogItemStore(new ProgramId(new Guid(TestApplicationId), TestApplicationFriendlyName));
            this.userStore = new UserStore(new ProgramId(new Guid(TestApplicationId), TestApplicationFriendlyName));

            this.userActionStore.Initialize();
            this.itemStore.Initialize();
            this.userStore.Initialize();

            storeChangeEvent = new StoreChangeEvent();
            this.eventHandler = new RecommenderEventHandler(StoreChanged);
            storeChangeEvent.Subscribe(this.eventHandler);
        }

        [TestCleanup]
        public void TestCleanup()
        {
            this.Uninitialize();
            UserActionStoreTest.ClassCleanUp();
        }

        public void Uninitialize()
        {
            this.userActionStore.Dispose();
            this.itemStore.Dispose();
            this.userStore.Dispose();
            this.storeChangeEvent.Unsubscribe(this.eventHandler);
        }

        void StoreChanged(EventArgs e)
        {
            if (e is StoreChangedEventArgs)
            {
                StoreChangedEventArgs eventArgs = e as StoreChangedEventArgs;
                eventedChangeNumbers.Add(eventArgs.NewChangeNumber);
            }
        }

        private void AssertChange(ICollection<IDataStoreChange> changes,  StoreChangeType changeType, int itemId)
        {
            bool seenOnce = false;

            foreach (IDataStoreChange change in changes)
            {
                if (change.ChangedItemId == itemId)
                {
                    if (seenOnce)
                    {
                        Assert.Fail(string.Format("More than one change for changedItemId {0}: change type: {1}, change number: {2}", new object[] { itemId, change.ChangeType, change.ChangeNumber }));
                    }
                    else
                    {
                        Assert.IsTrue(change.ChangeType == changeType);
                        seenOnce = true;
                    }
                }
            }
            Assert.IsTrue(seenOnce, string.Format("There is no change for the changedItemId {0}", itemId));
        }

        [TestMethod]
        public void UserActionStoreTestHistory()
        {
            IList<IDataStoreChange> changes;
            User[] users = {
                new User(),
                new User(),
                new User(),
                new User()
            };
            this.userStore.AddItems(users);

            CatalogItem[] catalogItems = {
                new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } },
                new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } },
                new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } },
                new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } }
            };
            this.itemStore.GetDatabaseContext().CatalogItems.InsertAllOnSubmit(catalogItems);
            this.itemStore.GetDatabaseContext().SubmitChanges();

            RatingAction
                firstRating = new RatingAction(users[0], catalogItems[0], 1.0),
                secondRating = new RatingAction(users[1], catalogItems[1], 1.0),
                thirdRating = new RatingAction(users[2], catalogItems[2], 1.0),
                fourthRating = new RatingAction(users[3], catalogItems[3], 1.0);
            
            Assert.IsFalse(userActionStore.Contains(firstRating.ItemId), "The item was not yet added, so it should not be in the store.");

            userActionStore.AddItem(firstRating);

            Assert.IsTrue(userActionStore.Contains(firstRating.Key), "The item was just added, so it should be in the store.");

            userActionStore.AddItem(secondRating);

            userActionStore.AddItem(thirdRating);

            changes = userActionStore.GetDifferences(0, 0);
            Assert.AreEqual<int>(0, changes.Count, "There should be no changes for change number 0");

            changes = userActionStore.GetDifferences(0, 3);
            Assert.AreEqual<int>(3, changes.Count, "There should be 3 changes");
            AssertChange(changes, StoreChangeType.Added, firstRating.Key);
            AssertChange(changes, StoreChangeType.Added, secondRating.Key);
            AssertChange(changes, StoreChangeType.Added, thirdRating.Key);

            changes = userActionStore.GetDifferences(3, 3);
            Assert.AreEqual<int>(1, changes.Count);
            AssertChange(changes, StoreChangeType.Added, thirdRating.Key);

            userActionStore.RemoveItem(thirdRating.Key);
            changes = userActionStore.GetDifferences(userActionStore.ChangeNumber, userActionStore.ChangeNumber);
            AssertChange(changes, StoreChangeType.Removed, thirdRating.Key);

            userActionStore.AddItem(fourthRating);

            changes = userActionStore.GetDifferences(0, 5);
            Assert.AreEqual<int>(3, changes.Count);
            AssertChange(changes, StoreChangeType.Added, firstRating.Key);
            AssertChange(changes, StoreChangeType.Added, secondRating.Key);
            AssertChange(changes, StoreChangeType.Added, fourthRating.Key);

            Assert.AreEqual<int>(6, eventedChangeNumbers.Count, "Did not receive the expected number of changes");
            
            //Verify updates to catalog
            RatingAction
                firstDuplicateRating = new RatingAction(users[0], catalogItems[0], -1.0),
                secondDuplicateRating = new RatingAction(users[1], catalogItems[1], -1.0),
                thirdDuplicateRating = new RatingAction(users[2], catalogItems[2], -1.0),
                fourthDuplicateRating = new RatingAction(users[3], catalogItems[3], -1.0);
            RatingAction[] duplicateRatings = new RatingAction[] 
            {
                firstDuplicateRating, 
                secondDuplicateRating, 
                fourthDuplicateRating,
            };

            userActionStore.ProcessFeedback(duplicateRatings);

            Assert.AreEqual<int>(7, eventedChangeNumbers.Count, "Did not receive the expected number of changes");
            changes = userActionStore.GetDifferences(6, 6);
            Assert.AreEqual<int>(3, changes.Count);
            AssertChange(changes, StoreChangeType.Changed, firstRating.Key);
            AssertChange(changes, StoreChangeType.Changed, secondRating.Key);
            AssertChange(changes, StoreChangeType.Changed, fourthRating.Key);

            thirdRating = new RatingAction(users[2], catalogItems[2], 1.0);//Create a new feedback item.
            userActionStore.ProcessFeedback(new UserAction[]{thirdRating});
            Assert.AreEqual<int>(8, eventedChangeNumbers.Count, "Did not receive the expected number of changes");
            changes = userActionStore.GetDifferences(7, 7);
            Assert.AreEqual<int>(1, changes.Count);
            AssertChange(changes, StoreChangeType.Added, thirdRating.Key);

            userActionStore.ProcessFeedback(new UserAction[] { thirdDuplicateRating });
            Assert.AreEqual<int>(9, eventedChangeNumbers.Count, "Did not receive the expected number of changes");
            changes = userActionStore.GetDifferences(8, 8);
            Assert.AreEqual<int>(1, changes.Count);
            AssertChange(changes, StoreChangeType.Changed, thirdRating.Key);

            //Assert.AreEqual<int>(0, RecommenderStatistics.ActiveDataContexts, "There must be zero active data contexts in the system on close.");
        }

        /// <summary>
        /// Verifies that appropriate operations on the user action set occur on the server.
        /// Specicially this means checking the RecommenderStatistics to check the count
        /// of UserActions created.
        /// </summary>
        [TestMethod]
        public void UserActionStoreExecutionOnServer()
        {
            int userActionCount = 10;
            int expectedCount = RecommenderStatistics.UserActionsCreated;

            this.InsertRatings(userActionCount);
            expectedCount += userActionCount;
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Inserted a set of items.");

            this.InsertRatings(userActionCount);
            expectedCount += userActionCount;
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Inserted an additional set of items.");

            List<UserAction> userActions = this.userActionStore.Take<UserAction>(userActionCount).ToList<UserAction>();
            expectedCount += userActions.Count<UserAction>();
            Assert.AreNotEqual<int>(0, userActions.Count<UserAction>(), "Result set should not be empty.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Performed a subset operation expected to execute on the server, only created objects for the returned items.");

            userActionCount = this.userActionStore.Count();
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verify that count operation does not result in UserAction count increase; this should occur on the server.");

            int recommenderId = userActions.First<UserAction>().Key;

            IEnumerable<UserAction> userActionsEnumerable = this.userActionStore.Where<UserAction>(c => c.Key == recommenderId);
            Assert.AreEqual<int>(1, userActionsEnumerable.Count<UserAction>(), "Result set should contain one item.");
            expectedCount += 1;
            Assert.IsTrue(expectedCount >= RecommenderStatistics.UserActionsCreated,
                "Performed a subset operation selecting a specific item, which can be cached.");
            expectedCount = RecommenderStatistics.UserActionsCreated;

            userActionsEnumerable = from us in this.userActionStore
                                    where us.Key == recommenderId
                                    select us;
            Assert.AreEqual<int>(1, userActionsEnumerable.Count<UserAction>(), "Result set should contain one item.");
            expectedCount += 1;
            Assert.IsTrue(expectedCount >= RecommenderStatistics.UserActionsCreated,
                "Performed a subset operation selecting a specific item using alternate query style, which can be cached.");
            expectedCount = RecommenderStatistics.UserActionsCreated;

            this.Uninitialize();
            this.Initialize();

            userActions = this.userActionStore.Take<UserAction>(userActionCount).ToList<UserAction>();
            expectedCount += userActions.Count<UserAction>();
            Assert.AreNotEqual<int>(0, userActions.Count<UserAction>(), "Result set should not be empty.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Performed a subset operation after the store has been reinitialized.");

            int skipCount = userActionCount/2;
            int takeCount = userActionCount/2;
            userActions = this.userActionStore.Skip(skipCount).Take(takeCount).ToList<UserAction>();
            expectedCount += takeCount;
            Assert.AreNotEqual<int>(0, userActions.Count<UserAction>(), "Result set should not be empty.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verified skip/take semantics work as expected.");

            userActionCount = this.userActionStore.Count();
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verify the count operation after a the store has been reinitialized.");

            userActionStore.RemoveAll();
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verify clearing the store does not create new items.");
            //Assert.AreEqual<int>(0, RecommenderStatistics.ActiveDataContexts, "There must be zero active data contexts in the system on close.");
        }

        [TestMethod]
        public void UserActionStoreRatingClass()
        {
            int expectedCount = RecommenderStatistics.UserActionsCreated;

            User user = new User();
            this.userStore.AddItem(user);
            CatalogItem item = new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } };
            this.itemStore.GetDatabaseContext().CatalogItems.InsertOnSubmit(item);
            this.itemStore.GetDatabaseContext().SubmitChanges();

            RatingAction rating = new RatingAction(user, item, RatingAction.MaxValue);
            expectedCount += 1; //Created new class to insert.
            this.userActionStore.AddItem(rating);

            rating = null;
            rating = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is RatingAction) as RatingAction;
            expectedCount += 1;  //Retrieve exactly one instance from the server.

            Assert.AreNotEqual<RatingAction>(default(RatingAction), rating,
                "Verify retrieval of item.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Quering for a rating should execute on the server.");

            //Update rating
            RatingAction newRating = new RatingAction(rating.User, rating.CatalogItem, RatingAction.MinValue);
            this.userActionStore.ProcessFeedback(new UserAction[] { newRating }.AsEnumerable());

            this.Uninitialize();
            this.Initialize();

            expectedCount = RecommenderStatistics.UserActionsCreated;
            rating = null;
            rating = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is RatingAction) as RatingAction;
            expectedCount += 1;  //Retrieve exactly one instance from the server.
            Assert.AreNotEqual<RatingAction>(default(RatingAction), rating,
                "Verify item exists after initialization cycle.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Quering for a rating should execute on the server after store intialization cycle.");
            //Assert.AreEqual<int>(0, RecommenderStatistics.ActiveDataContexts, "There must be zero active data contexts in the system on close.");
        }

        [TestMethod]
        public void UserActionStoreTagClass()
        {
            int userActionCount = 10;
            int expectedCount = RecommenderStatistics.UserActionsCreated;

            User user = new User();
            this.userStore.AddItem(user);
            CatalogItem item = new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } };
            this.itemStore.GetDatabaseContext().CatalogItems.InsertOnSubmit(item);
            this.itemStore.GetDatabaseContext().SubmitChanges();

            this.InsertRatings(userActionCount);
            expectedCount += userActionCount;

            //Create a new tag and insert
            string tagValue = "UserActionStoreTagClassTest";
            TaggingAction tag = new TaggingAction(user, item, tagValue);
            expectedCount += 1; //Created new class to insert.
            this.userActionStore.AddItem(tag);

            Nullable<int> tagId = tag.TagId;

            tag = null;
            tag = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is TaggingAction
                && c.TagValue.Tag == tagValue) as TaggingAction;
            expectedCount += 1;  //Retrieve exactly one instance from the server.
            
            Assert.AreNotEqual<TaggingAction>(default(TaggingAction), tag,
                "Verify retrieval of item.");
            Assert.AreEqual<string>(tagValue, tag.TagValue.Tag,
                "Verify tag value after insert.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Quering for a tag should execute on the server.");

            tag = null;
            tag = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is TaggingAction
                && (c as TaggingAction).TagId == tagId) as TaggingAction;
            expectedCount += 1;  //Retrieve exactly one instance from the server.

            Assert.AreNotEqual<TaggingAction>(default(TaggingAction), tag,
                "Verify retrieval of item by TagId.");
            Assert.AreEqual<string>(tagValue, tag.TagValue.Tag,
                "Verify tag value after retrieval by TagId.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Quering for a tag, by TagId, should execute on the server.");

            this.Uninitialize();
            this.Initialize();

            tag = null;
            tag = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is TaggingAction
                && c.TagValue.Tag == tagValue) as TaggingAction;
            expectedCount += 1;  //Retrieve exactly one instance from the server.
            Assert.AreNotEqual<TaggingAction>(default(TaggingAction), tag,
                "Verify item exists after initialization cycle.");
            Assert.AreEqual<string>(tagValue, tag.TagValue.Tag, 
                "Verify tag value is the same after store intialization cycle.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Quering for a tag should execute on the server after store intialization cycle.");

            //Verify duplicate tag entries are handled apprpraitely.
            TaggingAction duplicateTag = new TaggingAction(user, item, tagValue);
            this.userActionStore.ProcessFeedback(new UserAction[] { duplicateTag });
            IQueryable<TaggingAction> tags = this.userActionStore.Where<UserAction>(
                c => c is TaggingAction
                && c.TagValue.Tag == tagValue).Cast<TaggingAction>();
            Assert.AreEqual<int>(1, tags.Count(), "Duplicate tag entries.");
            tag = this.userActionStore.FirstOrDefault<UserAction>(
                            c => c is TaggingAction
                            && c.TagValue.Tag == tagValue) as TaggingAction;
            Assert.AreEqual<string>(tagValue, tag.TagValue.Tag, "Tags are not the same.");

            //Assert.AreEqual<int>(0, RecommenderStatistics.ActiveDataContexts, "There must be zero active data contexts in the system on close.");
        }

        [TestMethod]
        public void UserActionStoreOtherPropertyClass()
        {
            int userActionCount = 10;
            int expectedCount = RecommenderStatistics.UserActionsCreated;

            this.InsertRatings(userActionCount);
            expectedCount += userActionCount;

            User user = new User();
            this.userStore.AddItem(user);
            CatalogItem item = new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } };
            this.itemStore.GetDatabaseContext().CatalogItems.InsertOnSubmit(item);
            this.itemStore.GetDatabaseContext().SubmitChanges();

            //Create a new tag and insert
            string name1 = "ActionName1";
            string property1 = "ActionProperty1";
            OtherAction otherAction = new OtherAction(user, item, name1, property1);
            expectedCount += 1;
            this.userActionStore.AddItem(otherAction);

            otherAction = null;
            otherAction = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is OtherAction 
                && c.OtherActionProperty.Name == name1 
                && c.OtherActionProperty.Property == property1) as OtherAction;
            expectedCount += 1;
            Assert.AreNotEqual<OtherAction>(default(OtherAction), otherAction,
                "Verify item exists.");
            Assert.AreEqual<string>(name1, otherAction.OtherActionProperty.Name,
                "Verify otheraction name is the same.");
            Assert.AreEqual<string>(property1, otherAction.OtherActionProperty.Property,
                "Verify otheraction property is the same.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verify query executed on the server.");

            string name2 = "ActionName2";
            string property2 = "ActionProperty2";
            otherAction = new OtherAction(user, item, name2, property2);
            expectedCount += 1;
            //Intentionally not inserting into DB.

            otherAction = null;
            otherAction = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is OtherAction
                && c.OtherActionProperty.Name == name2
                && c.OtherActionProperty.Property == property2) as OtherAction;

            Assert.AreEqual<OtherAction>(default(OtherAction), otherAction,
                "Verify item doesn't exist.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verify query executed on the server.");

            this.Uninitialize();
            this.Initialize();

            otherAction = null;
            otherAction = this.userActionStore.FirstOrDefault<UserAction>(
                c => c is OtherAction
                && c.OtherActionProperty.Name == name1
                && c.OtherActionProperty.Property == property1) as OtherAction;
            expectedCount += 1;
            Assert.AreNotEqual<OtherAction>(default(OtherAction), otherAction,
                "Verify item exists after initialization cycle.");
            Assert.AreEqual<string>(name1, otherAction.OtherActionProperty.Name,
                "Verify otheraction name is the same after store intialization cycle.");
            Assert.AreEqual<string>(property1, otherAction.OtherActionProperty.Property,
                "Verify otheraction property is the same after store intialization cycle.");
            Assert.AreEqual<int>(expectedCount, RecommenderStatistics.UserActionsCreated,
                "Verify query executed on the server after initialization cycle.");

            string property3 = "PropertyChange";
            OtherAction duplicateOtherAction1 = new OtherAction(user, item, name1, property3);
            this.userActionStore.ProcessFeedback(new UserAction[] { duplicateOtherAction1 });
            IQueryable<OtherAction> otherActions = this.userActionStore.Where<UserAction>(
                c => c is OtherAction
                && c.OtherActionProperty.Name == name1).Cast<OtherAction>();
            Assert.AreEqual<int>(1, otherActions.Count(), "Expected one item to be returned.");
            otherAction = otherActions.First();
            Assert.AreEqual<string>(property3, otherAction.OtherActionProperty.Property, "New property value not set.");

            //Assert.AreEqual<int>(0, RecommenderStatistics.ActiveDataContexts, "There must be zero active data contexts in the system on close.");
        }

        [TestMethod]
        public void UserActionStoreImplicitCastInQueryOnServer()
        {
            User user = new User();
            this.userStore.AddItem(user);
            CatalogItem item = new CatalogItem() { Module = new Module() { Name = string.Empty, Organization = string.Empty } };
            this.itemStore.GetDatabaseContext().CatalogItems.InsertOnSubmit(item);
            this.itemStore.GetDatabaseContext().SubmitChanges();

            userActionStore.AddItem(new RatingAction(user, item, RatingAction.MaxValue));
            IEnumerable<RatingAction> ratings = from rating in this.userActionStore
                                                where rating is RatingAction && (rating as RatingAction).Rating == RatingAction.MaxValue
                                    select rating as RatingAction;
            RatingAction ratingAction = ratings.FirstOrDefault() as RatingAction;
            Assert.AreEqual<double>(RatingAction.MaxValue, ratingAction.Rating.Value, 
                "Verify query against database using implicit casts executes on server.");

            this.Uninitialize();
            this.Initialize();

            ratings = from rating in this.userActionStore
                      where rating is RatingAction && (rating as RatingAction).Rating == RatingAction.MaxValue
                                                select rating as RatingAction;
            ratingAction = ratings.FirstOrDefault() as RatingAction;
            Assert.AreEqual<double>(RatingAction.MaxValue, ratingAction.Rating.Value,
                "Verify query against database using implicit casts after initialization cycle.");
            //Assert.AreEqual<int>(0, RecommenderStatistics.ActiveDataContexts, "There must be zero active data contexts in the system on close.");
        }

        private void InsertRatings(int numberOfItems)
        {
            Collection<RatingAction> ratings = new Collection<RatingAction>();

            for (int i = 0; i < numberOfItems; i++)
            {
                CatalogItem item = new CatalogItem() { Module = new Module() {Name = string.Empty, Organization = string.Empty } };
                User user = new User();
                
                this.userStore.AddItem(user);
                this.itemStore.GetDatabaseContext().CatalogItems.InsertOnSubmit(item);
                this.itemStore.GetDatabaseContext().SubmitChanges();

                RatingAction rating = new RatingAction(user, item, 0.0);
                ratings.Add(rating);
            }
            this.userActionStore.AddItems(ratings.ToArray());
        }
    }
}
