﻿using System;
using EMIC.Design;
using System.Collections.Generic;
using System.IO;
using System.Globalization;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using Microsoft.SqlServer.Management.Smo;
using Microsoft.SqlServer.Management.Common;
using System.Diagnostics;
using System.Data.SqlClient;

namespace MyMediaProject.RecommenderSystem.Framework.Databases
{
    /// <summary>
    /// Utility class for manipulating database objects.
    /// </summary>
    public static class DatabaseHelper
    {
        /// <summary>
        /// Backups the database in use by the given <see cref="Recommender"/>.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <remarks>
        /// The default backup location is a folder named "Databases" in the current working directory.
        /// The database backup name is of the format, "[DatabaseName].bak".
        /// </remarks>
        /// <example>
        /// For a database name of, "MyMediaProject.FieldTrial", a backup file name of "MyMediaProject.FieldTrial.bak" is used,
        /// by default.
        /// </example>
        public static void BackupRecommenderDatabase(Recommender recommender)
        {
            string defaultBackupFileFullPath = GetDefaultBackupFileName(recommender);
            DatabaseHelper.BackupRecommenderDatabase(recommender, defaultBackupFileFullPath);
        }

        /// <summary>
        /// Backups the database in use by the given <see cref="Recommender"/>.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <param name="backupFile">The backup file.</param>
        public static void BackupRecommenderDatabase(Recommender recommender, string backupFile)
        {
            DirectoryInfo parentFolderInfo = Directory.GetParent(backupFile);
            if (!parentFolderInfo.Exists)
            {
                Directory.CreateDirectory(parentFolderInfo.FullName);
            }
            DatabaseHelper.BackupDatabase(recommender.DatabaseName, backupFile, recommender.DatabaseSqlInstance);
        }

        /// <summary>
        /// Restores the recommender database.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        public static void RestoreRecommenderDatabase(Recommender recommender)
        {
            string defaultBackupFileFullPath = GetDefaultBackupFileName(recommender);
            DatabaseHelper.RestoreRecommenderDatabase(recommender, defaultBackupFileFullPath);
        }

        /// <summary>
        /// Restores the recommender database.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <param name="restoreFileFullPath">The restore file full path.</param>
        public static void RestoreRecommenderDatabase(Recommender recommender, string restoreFileFullPath)
        {
            DatabaseHelper.RestoreDatabase(recommender.DatabaseName, restoreFileFullPath, recommender.DatabaseSqlInstance);
        }

        private static string GetDefaultBackupFileName(Recommender recommender)
        {
            string defaultBackupFileShortName = string.Format(CultureInfo.InvariantCulture, "{0}.bak",
                recommender.DatabaseName);
            string defaultBackupFileFullPath = Path.Combine(Directory.GetCurrentDirectory(), "Databases");
            defaultBackupFileFullPath = Path.Combine(defaultBackupFileFullPath, defaultBackupFileShortName);
            return defaultBackupFileFullPath;
        }

        /// <summary>
        /// Will fail if targetDatabase already exists.
        /// </summary>
        /// <param name="targetDatabase">The database name.</param>
        /// <param name="backupFileFullPath">The backup file full path.</param>
        /// <param name="sqlServerInstance">The SQL server instance.</param>
        /// <returns></returns>
        public static bool RestoreDatabase(string targetDatabase, string backupFileFullPath, string sqlServerInstance)
        {
            bool result = false;
            if (File.Exists(backupFileFullPath))
            {
                // declare server side connection
                ServerConnection connection = new ServerConnection();
                connection.ApplicationName = "MyMedia.DatabaseHelper";
                connection.ServerInstance = sqlServerInstance;
                connection.DatabaseName = "master";

                // connect to server
                Server server = new Server(connection);

                Database database = new Database(server, targetDatabase);

                // prepare the restore
                Restore restore = new Restore();
                restore.Database = targetDatabase;
                restore.Devices.AddDevice(backupFileFullPath, DeviceType.File);
                restore.ReplaceDatabase = true;

                try
                {
                    database.Create();

                    // do the restore
                    restore.SqlRestore(server);

                    // set restored database online
                    server.Databases[targetDatabase].SetOnline();
                    result = true;
                }
                catch (SqlException e)
                {
                    TraceLog.OnException(e);
                }
                catch (Exception e)
                {
                    TraceLog.OnException(e);
                    throw;
                }
            }
            else
            {
                TraceLog.Write(LogEventType.Information, 
                    "Unable to restore database; file \"{0}\" does not exist.", 
                    backupFileFullPath);
            }
            return result;
        }


        /// <summary>
        /// Backs up the database.
        /// </summary>
        /// <param name="targetDatabaseName">Name of the database to back up.</param>
        /// <param name="backupFileFullPath">The backup file full path.</param>
        /// <param name="sqlServerInstance">The SQL server instance.</param>
        /// <returns></returns>
        /// <exception cref="FailedOperationException"/>
        public static void BackupDatabase(string targetDatabaseName, string backupFileFullPath, string sqlServerInstance)
        {
            // connect to server
            Server server = new Server(sqlServerInstance);

            // DEBUG
            //Console.WriteLine("Login: " + server.ConnectionContext.TrueLogin);

            // prepare the backup
            Backup backup = new Backup();
            backup.Database = targetDatabaseName;
            backup.Devices.AddDevice(backupFileFullPath, DeviceType.File);
            backup.Initialize = true;
            backup.Checksum = true;
            backup.Incremental = false;
            backup.LogTruncation = BackupTruncateLogType.Truncate;

            // debugging
            //backup.Information += new ServerMessageEventHandler(backup_Information);
            //backup.PercentComplete += new PercentCompleteEventHandler(backup_PercentComplete);

            // do the backup
            try
            {
                backup.SqlBackup(server);
            }
            catch (FailedOperationException ex)
            {
                Debug.WriteLine(ex.Message + ex.InnerException);
                throw;
            }
        }

        /// <summary>
        /// Drops the database.
        /// </summary>
        /// <param name="targetDatabase">The target database.</param>
        /// <param name="sqlServerInstance">The SQL server instance.</param>
        /// <returns></returns>
        public static bool DropDatabase(string targetDatabase, string sqlServerInstance)
        {
            // declare server side connection
            ServerConnection connection = new ServerConnection();
            connection.ApplicationName = "MyMedia.DatabaseHelper";
            connection.ServerInstance = sqlServerInstance;
            connection.DatabaseName = "master";

            // connect to server
            Server server = new Server(connection);

            // get & prepare the database
            Database dbToDrop = server.Databases[targetDatabase];
            server.KillAllProcesses(targetDatabase);

            // drop the database
            try
            {
                //If database is null, than it does not exist, and does not need to be dropped.
                if (null != dbToDrop)
                {
                    dbToDrop.Drop();
                }
            }
            catch (SqlException ex)
            {
                Debug.WriteLine(ex.Message + ex.InnerException);
                return false;
            }

            return true;
        }

        /// <summary>
        /// Gets the backup location.
        /// </summary>
        /// <param name="sqlServerInstance">The SQL server instance.</param>
        /// <param name="failed">if set to <c>true</c> [failed].</param>
        /// <returns></returns>
        public static string GetBackupLocation(string sqlServerInstance, ref bool failed)
        {
            failed = false;

            string location= string.Empty;
            // declare server side connection
            ServerConnection connection = new ServerConnection();
            connection.ApplicationName = "MyMedia.DatabaseHelper";
            connection.ServerInstance = sqlServerInstance;
            connection.DatabaseName = "master";

            // connect to server
            Server server = new Server(connection);
            try
            {
                location = server.Settings.BackupDirectory;
            }
            catch (SqlException ex)
            {
                Debug.WriteLine(ex.Message + ex.InnerException);
                failed = true;
                location = ex.Message;
            }

            return location;
        }

        /// <summary>
        /// Gets the temp backup file full path.
        /// </summary>
        /// <param name="sqlServerInstance">The SQL server instance.</param>
        /// <param name="targetDatabaseName">Name of the target database.</param>
        /// <returns></returns>
        public static string GetTempBackupFileFullPath(string sqlServerInstance, string targetDatabaseName)
        {
            bool failed = true;
            string tempBackupFileFullPath = null;
            string backupLocation = DatabaseHelper.GetBackupLocation(sqlServerInstance, ref failed);
            if (!failed)
            {
                tempBackupFileFullPath = Path.Combine(backupLocation, targetDatabaseName);
            }
            return tempBackupFileFullPath;
        }
    }
}
