﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Framework;

namespace MyMediaProject.RecommenderSystem.Framework.Metrics
{
    /// <summary>
    /// Utility metric for a list of recommendations, specifically applicable (but not limited to), top-n recommendations.
    /// This measures utility as if the user is scanning a list of recommendations 
    /// and stoping once they find something they like (i.e. in the test set).
    /// </summary>
    /// <typeparam name="TUserActionType">The type of the user action type.</typeparam>
    public class ExpectedUtilityMetric<TUserActionType> : MetricBase<TUserActionType>
        where TUserActionType : RatingAction
    {
        static private RecommenderId expectedUtilityMetricId = new RecommenderId(new Guid("{F9BBD089-0E6C-4024-A645-378AA180C8BE}"));
        static private string expectedUtilityeMetricDescription = 
            @"Expected utility metric.  This measures utility as if the user is scanning a list of recommendations" +
             " and stoping once they find something they like (i.e. in the test set).";

        private double halfLifePosition = 5;
        private double ratingConsideredLike;
        private double userUtilityAccumulator = 0;
        private double userCount = 0;

        /// <summary>
        /// Initializes a new instance of the <see cref="ExpectedUtilityMetric&lt;TUserActionType&gt;"/> class.
        /// With a default half life position of 5.
        /// </summary>
        /// <param name="ratingConsideredLike">The rating considered like.</param>
        public ExpectedUtilityMetric(double ratingConsideredLike)
            : base(expectedUtilityMetricId, expectedUtilityeMetricDescription)
        {
            this.ratingConsideredLike = ratingConsideredLike;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ExpectedUtilityMetric&lt;TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="ratingConsideredLike">The rating considered like.</param>
        /// <param name="halfLifePosition">The half life position, at which the utility function will be 0.5.</param>
        public ExpectedUtilityMetric(double ratingConsideredLike, double halfLifePosition)
            : base(expectedUtilityMetricId, expectedUtilityeMetricDescription)
        {
            this.ratingConsideredLike = ratingConsideredLike;
            this.halfLifePosition = halfLifePosition;
        }

        /// <summary>
        /// Compute per users metrics, to be saved in overriding class
        /// internal state.
        /// </summary>
        /// <param name="user">The user.</param>
        /// <param name="testFeedback">The test feedback.</param>
        /// <param name="trainingFeedback">The training feedback.</param>
        /// <param name="predictions">The predictions.</param>
        /// <remarks>
        /// This is an implementation of the utility function defined in
        /// "Dependency Networks for Inference, Collaborative Filtering, and Data Visualization"
        /// by Heckerman et all
        /// Located here:
        /// ftp://ftp.research.microsoft.com/pub/TR/TR-2000-16.pdf
        /// See section "Evaluation Criteria and Experimental Procedure," page 19, for the relevant equations.
        /// </remarks>
        public override void ComputeMetricsForUser(
            User user,
            IQueryable<TUserActionType> testFeedback,
            IQueryable<TUserActionType> trainingFeedback,
            IList<Prediction<CatalogItem>> predictions)
        {
            double index;
            double summationAccumulatorNumerator = 0;
            double summationAccumulatorDenomonator = 0;

            IEnumerable<Prediction<CatalogItem>> orderedPrediction = from prediction in predictions
                                                                       orderby prediction.Rating descending
                                                                       select prediction;

            index = 0;
            foreach (Prediction<CatalogItem> predictionItem in orderedPrediction)
            {
                TUserActionType feedback = testFeedback.SingleOrDefault(fb => fb.ItemId == predictionItem.CatalogItem.ItemId);
                if (null != feedback && feedback.Rating.Value >= this.ratingConsideredLike)
                {
                    summationAccumulatorNumerator += this.UtilityFunction(index);
                }
                index++;
            }

            //This is used to normalize across varying user feedback sizes.
            summationAccumulatorDenomonator = this.IdealUtility(testFeedback.Count());

            this.userUtilityAccumulator += summationAccumulatorNumerator / summationAccumulatorDenomonator;
            this.userCount++;
        }

        /// <summary>
        /// Returns the final result, after the intermediate metrics for each
        /// user have been computed.
        /// </summary>
        /// <returns>The final metric result as a double.</returns>
        public override double FinalMetric()
        {
            return 100 * this.userUtilityAccumulator / this.userCount;
        }

        /// <summary>
        /// An exponentially decaying function to represent the utility of a prediction
        /// at location "itemPosition" also occuring in the user test set.
        /// </summary>
        /// <param name="itemPosition">The item position.</param>
        /// <returns></returns>
        private double UtilityFunction(double itemPosition)
        {
            double exponent = -itemPosition / this.halfLifePosition;
            double probability = Math.Pow(2.0, exponent);
            return probability;
        }

        /// <summary>
        /// Computes the ideal utility of a list where all of the test set is at the top
        /// of the recommendation list.
        /// </summary>
        /// <param name="testSetSize">Size of the test set.</param>
        /// <returns></returns>
        private double IdealUtility(int testSetSize)
        {
            double accumulator = 0;
            for (int i = 0; i < testSetSize; i++)
            {
                accumulator += this.UtilityFunction(Convert.ToDouble(i));
            }
            return accumulator;
        }
    }
}
