﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Framework;

namespace MyMediaProject.RecommenderSystem.Framework.Metrics
{
    /// <summary>
    /// Base class for computing various types of recommendation metrics.
    /// </summary>
    /// <typeparam name="TUserActionType">The type of the user action type.</typeparam>
    public abstract class MetricBase<TUserActionType>
        where TUserActionType : UserAction
    {
        private RecommenderId metricId;
        private string metricDescription;

        /// <summary>
        /// Initializes a new instance of the <see cref="MetricBase&lt;TUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="metricId">The metric id.</param>
        /// <param name="metricDescription">The metric description.</param>
        protected MetricBase(RecommenderId metricId, string metricDescription)
        {
            this.metricId = metricId;
            this.metricDescription = metricDescription;
        }

        /// <summary>
        /// Compute per users metrics, to be saved in overriding class
        /// internal state.
        /// </summary>
        /// <param name="user">The user.</param>
        /// <param name="testFeedback">The test feedback.</param>
        /// <param name="trainingFeedback">The training feedback.</param>
        /// <param name="predictions">The predictions.</param>
        /// <remarks>
        /// This method must be overriden to make intermediate computations
        /// for the given user.  This method will be called for each user in
        /// the system.  The overriding class must maintain the relevant internal
        /// state for each metric computation, to produce the final result.
        /// </remarks>
        public abstract void ComputeMetricsForUser(
            User user,
            IQueryable<TUserActionType> testFeedback,
            IQueryable<TUserActionType> trainingFeedback,
            IList<Prediction<CatalogItem>> predictions);

        /// <summary>
        /// Returns the final result, after the intermediate metrics for each 
        /// user have been computed.
        /// </summary>
        /// <returns>The final metric result as a double.</returns>
        public abstract double FinalMetric();

        /// <summary>
        /// Gets the description of the metric.
        /// </summary>
        /// <value>The description.</value>
        public string Description 
        { 
            get
            {
                return this.metricDescription;
            }
        }

        /// <summary>
        /// Gets the id of the metric.
        /// </summary>
        /// <value>The id.</value>
        public RecommenderId Id 
        { 
            get
            {
                return this.metricId;
            }
        }
    }
}
