﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Framework;

namespace MyMediaProject.RecommenderSystem.Framework.Metrics
{
    /// <summary>
    /// This is a base implementation of a Root Means Squared Error calculator.  This implementation
    /// does not handle special situations such as items predicted but not in the test set, and vice-versa.
    /// Classes can derive from this, and handle these special situations.
    /// </summary>
    /// <typeparam name="TUserActionType">The type of the feedback item type.</typeparam>
    public class RootMeanSquaredErrorMetric<TUserActionType> : MetricBase<TUserActionType>
        where TUserActionType : RatingAction
    {
        static private RecommenderId rmseMetricId = new RecommenderId(new Guid("{B3D5C2E0-A3CA-4615-AEB1-29A5FC2CA5B0}"));
        static private string rmseMetricDescription = "Root mean squared error metric, the default rating criteria for recommender systems.";

        private double absoluteSquaredErrorAccumulator = 0;
        private double sampleCount = 0;

        /// <summary>
        /// Initializes a new instance of the <see cref="RootMeanSquaredErrorMetric&lt;TUserActionType&gt;"/> class.
        /// </summary>
        public RootMeanSquaredErrorMetric()
            : base(rmseMetricId, rmseMetricDescription)
        {
        }

        /// <summary>
        /// Compute per users metrics, to be saved in overriding class
        /// internal state.
        /// </summary>
        /// <param name="user">The user.</param>
        /// <param name="testFeedback">The test feedback.</param>
        /// <param name="trainingFeedback">The training feedback.</param>
        /// <param name="predictions">The predictions.</param>
        /// <remarks>
        /// This method must be overriden to make intermediate computations
        /// for the given user.  This method will be called for each user in
        /// the system.  The overriding class must maintain the relevant internal
        /// state for each metric computation, to produce the final result.
        /// </remarks>
        public override void ComputeMetricsForUser(
            User user,
            IQueryable<TUserActionType> testFeedback,
            IQueryable<TUserActionType> trainingFeedback,
            IList<Prediction<CatalogItem>> predictions)
        {
            this.ComputeMetricsForUser(testFeedback, predictions);
        }

        /// <summary>
        /// Returns the final result, after the intermediate metrics for each
        /// user have been computed.
        /// </summary>
        /// <returns>The final metric result as a double.</returns>
        public override double FinalMetric()
        {
            if (0 == this.sampleCount)
            {
                throw new InvalidOperationException(
                    "The method this.ComputeMetricsForUser must be called at least once before the final metrics can be computed");
            }
            double rootMean = Math.Sqrt(this.absoluteSquaredErrorAccumulator / sampleCount);
            return rootMean;
        }

        /// <summary>
        /// Gets or sets the absolute squared error accumulator.
        /// </summary>
        /// <value>The absolute squared error accumulator.</value>
        public double AbsoluteSquaredErrorAccumulator
        {
            get { return absoluteSquaredErrorAccumulator; }
            set { absoluteSquaredErrorAccumulator = value; }
        }

        /// <summary>
        /// Gets or sets the sample count.
        /// </summary>
        /// <value>The sample count.</value>
        public double SampleCount
        {
            get { return sampleCount; }
            set { sampleCount = value; }
        }

        /// <summary>
        /// Computes the metrics for user.
        /// </summary>
        /// <param name="testFeedbacks">The test feedbacks.</param>
        /// <param name="predictions">The predictions.</param>
        protected virtual void ComputeMetricsForUser(IQueryable<TUserActionType> testFeedbacks, IList<Prediction<CatalogItem>> predictions)
        {
            foreach (TUserActionType testFeedback in testFeedbacks)
            {
                Prediction<CatalogItem> prediction = predictions.FirstOrDefault<Prediction<CatalogItem>>(
                    p => p.CatalogItem.ItemId == testFeedback.ItemId);
                if(null != prediction)
                {
                    double error = prediction.Rating - testFeedback.Rating.Value;
                    this.AbsoluteSquaredErrorAccumulator += Math.Pow(error, 2.0);
                    this.SampleCount++;
                }
            }
        }
    }
}
