﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
   
   
    /// <summary>
    /// Memorybased implementation of the relational data reader.
    /// The implementation is for experimental purposes only.
    /// It works with CSV-files; no database like SQL is required.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public class DataSetMemoryBased : IEntityRelationDataReader
    {
        public enum DataType
        {
            dt_undef = 0,
            dt_int = 1,
            dt_double = 2,
        }

        public class DataArray
        {
            public virtual void setSize(int row_count) {}
            public virtual void read(int row_index, string s) {}
        }
        public class DataArrayInteger: DataArray
        {
            public int[] data;
            public override void setSize(int row_count) {
                data = new int[row_count];
            }
            public override void read(int row_index, string s) {
                data[row_index] = Convert.ToInt32(s);
            }
        }
        public class DataArrayDouble: DataArray
        {
            public double[] data;
            public override void setSize(int row_count) {
                data = new double[row_count];
            }
            public override void read(int row_index, string s) {
                data[row_index] = Convert.ToDouble(s);
            }
        }


        public class DataStorage
        {
            public List<DataArray> columns = new List<DataArray>();
            public DataStorage(DataType[] data_types)
            {
                foreach (DataType dt in data_types)
                {
                    switch (dt)
                    {
                        case DataType.dt_int:
                            columns.Add(new DataArrayInteger());
                            break;
                        case DataType.dt_double:
                            columns.Add(new DataArrayDouble());
                            break;
                        default:
                            throw new Exception("unknown data type");
                    }
                }
            }
            public void setSize(int num_rows) {
                for (int i = 0; i < columns.Count; i++) {
                    columns[i].setSize(num_rows);
                }
            }
        }

        // Dispose() calls Dispose(true)
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        // NOTE: Leave out the finalizer altogether if this class doesn't 
        // own unmanaged resources itself, but leave the other methods
        // exactly as they are. 
        ~DataSetMemoryBased() 
        {
            // Finalizer calls Dispose(false)
            Dispose(false);
        }
        // The bulk of the clean-up code is implemented in Dispose(bool)
        protected virtual void Dispose(bool disposing)
        {
            if (disposing) 
            {
            // free managed resources
            
            }
        }


        public DataStorage data;

        protected int current_row;
        protected int num_rows;

        string fileToRead;
        bool isOpen = false;

        public DataSetMemoryBased(DataType[] data_types, String filename)
        {
            data = new DataStorage(data_types);
            fileToRead = filename;
        }

        private void init(int num_rows) {
            this.num_rows = num_rows;
            data.setSize(num_rows);
            current_row = -1;
            isOpen = true;
        }

        public DataSetMemoryBased(DataType[] data_types, int num_rows)
        {
            data = new DataStorage(data_types);
            init(num_rows);
        }

        public void Close()
        {
            //isOpen = false;
        }

        public int ColumnCount
        {
            get { return data.columns.Count; }
        }
        public int RowCount
        {
            get { return num_rows; }
        }

        public double GetDouble(int colIndex)
        {
            if (current_row < 0)
            {
                throw new Exception("bof");
            }
            if (current_row >= num_rows)
            {
                throw new Exception("end of file");
            }
            if (colIndex >= ColumnCount)
            {
                throw new Exception("column index exceeds column count");
            }

            if (data.columns[colIndex] is DataArrayDouble)
            {
                return (data.columns[colIndex] as DataArrayDouble).data[current_row];
            }
            else if (data.columns[colIndex] is DataArrayInteger)
            {
                return (data.columns[colIndex] as DataArrayInteger).data[current_row];
            }
            else
            {
                throw new NotImplementedException();
            }
        }

        public short GetInt16(int colIndex)
        {
            throw new NotImplementedException();
        }

        public int GetInt32(int colIndex)
        {
            if (current_row < 0)
            {
                throw new Exception("bof");
            }
            if (current_row >= num_rows)
            {
                throw new Exception("end of file");
            }
            if (colIndex >= ColumnCount)
            {
                throw new Exception("column index exceeds column count");
            }

            if (data.columns[colIndex] is DataArrayDouble)
            {
                throw new Exception("column is double; cannot read an int");
            }
            else if (data.columns[colIndex] is DataArrayInteger)
            {
                return (data.columns[colIndex] as DataArrayInteger).data[current_row];
            }
            else
            {
                throw new NotImplementedException();
            }
        }

        public long GetInt64(int colIndex)
        {
            throw new NotImplementedException();
        }

        public String GetString(int colIndex)
        {
            throw new NotImplementedException();
        }

        public object GetObject(int colIndex)
        {
            throw new NotImplementedException();
        }


        public bool IsOpen
        {
            get { return isOpen; }
        }

        public void Open()
        {
            if (!isOpen)
            {
                int cnt = 0;
                {
                    StreamReader sr = new StreamReader(fileToRead);
                    String line;
                    while ((line = sr.ReadLine()) != null)
                    {
                        cnt++;
                    }
                }
                this.num_rows = cnt;
                init(num_rows);
                {
                    StreamReader sr = new StreamReader(fileToRead);
                    String line;
                    cnt = 0;
                    while ((line = sr.ReadLine()) != null)
                    {
                        String[] tokens = line.Split('\t');
                        for (int i = 0; i < ColumnCount; i++)
                        {
                            data.columns[i].read(cnt, tokens[i]);
                        }
                        cnt++;
                    }
                    sr.Close();
                }
                isOpen = true;
            }

            current_row = -1;
        }

        public bool Read()
        {
            if (current_row >= (num_rows - 1))
            {
                return false;// throw new Exception("end of file");
            }
            current_row++;

            return true;
        }
    }
}
