package nl.telin.mymedia.analysis.infgain;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Map.Entry;

/**
 * The SparseVector is a indexed vector where 0 values have been removed.
 * @author Christian.Wartena
 */
public class SparseVector<K extends Number> implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = -1767222368029239345L;
	private int size;
	private HashMap<Integer,K> data;
	
	
	/**
	 * Constructor, initializes size to 0
	 */
	public SparseVector() {
		size = 0;
		data = new HashMap<Integer,K>(8,0.75f);
	}
	
	/**
	 * @return the size of this vector
	 */
	public int getSize() {
		return size;
	}
	
	/**
	 * @return the number of non-zero values
	 */
	public int getNrOfNonZeroValues() {
		return data.size();
	}

	/**
	 * set value at indexed position
	 * @param position the position
	 * @param value the value
	 * @return true
	 */
	public boolean setAt(int position, K value) {
		boolean added = true;
		
		data.put(position,value);
		if(position > size) {
			size = position;
		}
		
		return added;
	}

	/**
	 * get Flaot value from position
	 * @param position the position
	 * @return Float value
	 */
	public Float getFloatAt(int position) {
		Float rc;
		rc = (Float) data.get(position);
		if(rc == null) {
			rc = new Float(0); //TODO dit is niet fraai
		}
		return rc;
	}

	/**
	 * get Integer value from position
	 * @param position the position
	 * @return Integer value
	 */
	public Integer getIntegerAt(int position) {
		Integer rc;
		K number;
		number =  data.get(position);
		if(number == null) {
			rc = new Integer(0); //TODO dit is niet fraai
		} else {
			rc = (Integer) number;
		}
		return rc;
	}
	
	
	/**
	 * add value to existing value at indexed position
	 * @param position the position
	 * @param a the value to be added
	 * @return the added value
	 */
	public Float addTo(int position, Float a) {
	
		Float value = (Float) data.get(position);
		if(value == null) {
			value = a;
		} else {
			int temp = value.intValue() + a.intValue();
			value = new Float(temp);
		}
		data.put(position,(K) value);
		if(position > size) {
			size = position;
		}
		
		return value;
	}
	
	/**
	 * add value to existing value at indexed position
	 * @param position the position
	 * @param a the value to be added
	 * @return the added value
	 */
	public Integer addTo(int position, Integer a) {
		
		Integer value = (Integer) data.get(position);
		if(value == null) {
			value = a;
		} else {
			int temp = value.intValue() + a.intValue();
			value = new Integer(temp);
		}
		data.put(position,(K) value);
		if(position > size) {
			size = position;
		}
		
		return value;
	}
	
	/**
	 * @ return iterator over index values
	 */
	public Iterator<Integer> getKeyIterator(){
		/*
		TreeSet<Integer> asSortedSet = new TreeSet<Integer>();
		asSortedSet.addAll(data.keySet());
		Iterator<Integer> i = asSortedSet.iterator();
		*/
		Iterator<Integer> i = data.keySet().iterator();
		return i;
	}
	
	/**
	 * divide existing values by a certain value
	 * @param fraction the divider
	 */
	public void divideElementsBy(int fraction) {
		//Iterator<Entry<Integer,K>> it = data.entrySet().iterator();
		for(Entry<Integer,K> e:data.entrySet()) {
			Float val = new Float(e.getValue().doubleValue()/fraction);
			e.setValue((K) val);
		}
	}
	
	/**
	 * divide existing values by a certain value
	 * @param fraction the divider
	 */
	public void divideElementsBy(float fraction) {
		//Iterator<Entry<Integer,K>> it = data.entrySet().iterator();
		for(Entry<Integer,K> e:data.entrySet()) {
			Float val = new Float(e.getValue().doubleValue()/fraction);
			e.setValue((K) val);
		}
	}

	
	/* 
	 public void positiveDiff(SparseVector<Float> distr) {
		Iterator<Entry<Integer,K>> it = data.entrySet().iterator();
		float sum = 0;
		while(it.hasNext()) {
			Entry<Integer,K> e = it.next();
			Float val = new Float(e.getValue().doubleValue() - distr.getFloatAt(e.getKey()));
			if(val > 0) {
				e.setValue((K) val);
				sum += (float) val;
			} else {
				it.remove();
			}
		}
		if(sum > 0) {
			divideElementsBy(sum);
		}
	}
	*/
	
	/**
	 * remove values that have a negative difference with provided distr.
	 * remaining items are divided by sum difference.  
	 * @param distr the distribution to compare to.
	 */
	public void positiveDiff(SparseVector<Float> distr) {
		Iterator<Entry<Integer,K>> it = data.entrySet().iterator();
		float sum = 0;
		while(it.hasNext()) {
			Entry<Integer,K> e = it.next();
			Float val = new Float(e.getValue().doubleValue() - distr.getFloatAt(e.getKey()));
			if(val > 0) {
				sum += (float) val;
			} else {
				it.remove();
			}
		}
		if(sum > 0) {
			divideElementsBy(sum);
		}
	}
	
	
}
