package nl.telin.mymedia.uima.annotator;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import nl.telin.mymedia.processor.CarrotProcessor;
import nl.telin.mymedia.uima.annotation.CarrotAnnotation;
import nl.telin.mymedia.uima.annotation.OntologyAnnotation;

import org.apache.uima.analysis_engine.ResultSpecification;
import org.apache.uima.analysis_engine.annotator.AnnotatorConfigurationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorContext;
import org.apache.uima.analysis_engine.annotator.AnnotatorContextException;
import org.apache.uima.analysis_engine.annotator.AnnotatorInitializationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorProcessException;
import org.apache.uima.analysis_engine.annotator.JTextAnnotator_ImplBase;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.jcas.JCas;
import org.apache.uima.util.Logger;


/**
 * @author Wout.Slakhorst
 * Uses OntologyAnnotations to determine the most important ones, according to Carrot
 */
public class CarrotAnnotator  extends JTextAnnotator_ImplBase {

	private Logger logger;
	private CarrotProcessor carrot;
	
	private static final String TT_ANNOTATION_NAME = "nl.telin.mymedia.uima.annotation.OntologyAnnotation";
	
	private Type ttAnnotation;
	
	@Override
	/**
	 * @inheritDoc
	 * config parameters:
	 * ONTOLOGY - path to ontology
	 */
	public void initialize(AnnotatorContext context)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {
		super.initialize(context);

		try {
			logger = getContext().getLogger();

			String path = (String) getContext().getConfigParameterValue("ONTOLOGY");
			
			carrot = new CarrotProcessor();
			carrot.loadModel(path, "");

		} catch (AnnotatorContextException e) {
			throw new AnnotatorConfigurationException(e);
		} catch (FileNotFoundException e) {
			throw new AnnotatorConfigurationException(e);
		}

	}
	
	@Override
	/**
	 * @inheritDoc
	 */
	public void typeSystemInit(TypeSystem aTypeSystem)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {

		// initialize cas token type
		this.ttAnnotation = aTypeSystem.getType(TT_ANNOTATION_NAME);		

	}
	
	//@Override
	/**
	 * @inheritDoc
	 */
	public void process(JCas jCas, ResultSpecification arg1)
			throws AnnotatorProcessException {
		
		String input = jCas.getDocumentText();
		int start = 0;
		int end = input.length() - 1;
		
		// iterate over all token annotations and put them in a Set
		Set<String> terms = new HashSet<String>();
		Map<String, Double> weights = new HashMap<String, Double>();
		//List<OntologyAnnotation> annotations = new ArrayList<OntologyAnnotation>();
		
		FSIterator ttIterator = jCas.getCas().getAnnotationIndex(
				this.ttAnnotation).iterator();

		while (ttIterator.hasNext()) {
			OntologyAnnotation annot = (OntologyAnnotation) ttIterator.next();
			String URI = annot.getOntologyURI();
			terms.add(URI);
			
			// calc weight
			Double i = weights.get(URI);
			if (i == null) {
				weights.put(URI, 1d);
			} else {
				weights.put(URI, i + 1d);
			}
		}
		
		carrot.setTerms(terms,weights);
		weights = carrot.recomputeWeights();
		
		// lets create an ordered list of the set of terms, sort
		// and pick the top sqrt(n)
		
		List<String> sortedList = new ArrayList<String>(terms.size());
		sortedList.addAll(terms);
		Collections.sort(sortedList, new CarrotComparator(weights));
		
		
		int itemsToDisplay = Math.min(6, (int) Math.floor(Math.sqrt(terms.size()))); 
		
		/*
		for (int i = 0; i < itemsToDisplay; i++) {
			String subject = sortedList.get(i);
			CarrotAnnotation newAnnot = new CarrotAnnotation(jCas, start, end);
			newAnnot.setSubject(subject);
			newAnnot.setRank(i + 1);
			
			// get some extra info from ontology
			newAnnot.setPrefLabel(carrot.getPrefLabel(subject));
			
			newAnnot.addToIndexes();
		}
		*/
		
		//Add them in a strage order to get them sorted in the displayed list...
		for (int i = 0; (double) i < (double) itemsToDisplay/2d; i++) {
			int last = itemsToDisplay - 1 - i ;
			
			String subject = sortedList.get(last);
			CarrotAnnotation newAnnot = new CarrotAnnotation(jCas, start, end);
			newAnnot.setSubject(subject);
			newAnnot.setRank(last+1);
			// get some extra info from ontology
			newAnnot.setPrefLabel(carrot.getPrefLabel(subject));
			newAnnot.addToIndexes();
			
			if(last != i) {
				subject = sortedList.get(i);
				newAnnot = new CarrotAnnotation(jCas, start, end);
				newAnnot.setSubject(subject);
				newAnnot.setRank(i + 1);
				// get some extra info from ontology
				newAnnot.setPrefLabel(carrot.getPrefLabel(subject));
				newAnnot.addToIndexes();
			}
		}
	}
	
	private class CarrotComparator implements Comparator<String> {

		private Map<String, Double> weights;
		
		public CarrotComparator(Map<String, Double> weights) {
			this.weights = weights;
		}
		
		//@Override
		public int compare(String o1, String o2) {
			if  (o1 == null || o2 == null) {
				return 0;
			}

			double w1 = weights.get(o1);
			double w2 = weights.get(o2);
			

			if (w1 > w2) {
				return -1;
			} else if (w1 < w2) {
				return +1;
			} else {
				return o1.compareTo(o2);
			}
		}
		
	}
}
