package nl.telin.mymedia.uima.trainer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import nl.telin.mymedia.analysis.infgain.SparseVector;
import nl.telin.mymedia.db.InfGainVector;
import nl.telin.mymedia.db.Lemma;
import nl.telin.mymedia.db.util.HibernateUtil;

import org.apache.uima.TokenAnnotation;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Type;
import org.apache.uima.collection.CasConsumer_ImplBase;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.resource.ResourceProcessException;
import org.apache.uima.util.ProcessTrace;

/**
 * @author Wout.Slakhorst
 * InfGainTrainer depends on TfIdfTrainer because the latter creates Lemma's
 * in the database
 */
public class InfGainTrainer extends CasConsumer_ImplBase {

	private int docs = 0;
	private List<String> lemmaList;
	private Map<String,Integer> invLemmaList;
	private ArrayList<SparseVector<Integer>> index;
	private Map<Integer,Integer> totalOccurences; //total nr of lemmata for each doc
	private float wordCounts[];

	int nrOfOccurrences;
	
	private Type tokenAnnotation;
	private static final String TOKEN_ANNOTATION_NAME = "org.apache.uima.TokenAnnotation";
	
	private static Set<String> semPos;
	
	static {
		semPos = new HashSet<String>();
		semPos.add("NN");
		semPos.add("NNP");
		semPos.add("JJ");
		semPos.add("VBG");
		semPos.add("VBD");
		semPos.add("VBN");
		semPos.add("VBP");
		semPos.add("VB");
		semPos.add("VBZ");
		semPos.add("FW");
		semPos.add("JJR");
		semPos.add("JJS");
		semPos.add("NNS");
		semPos.add("NNPS");
		semPos.add("NP");
		semPos.add("NPS");		
	}
	
	private static Set<String> verbPos;
	
	static {
		verbPos = new HashSet<String>();
		verbPos.add("VBG");
		verbPos.add("VBD");
		verbPos.add("VBN");
		verbPos.add("VBP");
		verbPos.add("VB");
		verbPos.add("VBZ");	
	}
	
	private static Set<String> auxList;
	
	static {
		auxList = new HashSet<String>();
		auxList.add("do");
		auxList.add("be");
		auxList.add("have");
		auxList.add("will");
		auxList.add("shall");
		auxList.add("can");
		auxList.add("may");
		auxList.add("go");
		auxList.add("make");
		auxList.add("let");
		auxList.add("try");
		auxList.add("get");
	}
	
	/**
	 * @inheritDoc
	 * removes all InfGainVectors from the database
	 */
	public void initialize() throws ResourceInitializationException {
		super.initialize();
		
		lemmaList = new ArrayList<String>();
		invLemmaList = new HashMap<String,Integer>();
		totalOccurences = new HashMap<Integer,Integer>();
		index = new ArrayList<SparseVector<Integer>>(100);
		nrOfOccurrences = 0;
		
		HibernateUtil.removeAllInfGain();
	}

	/**
	 * @inheritDoc
	 * only processes token with type WORD
	 */
	public void processCas(CAS acas) throws ResourceProcessException {
		this.tokenAnnotation = acas.getTypeSystem().getType(TOKEN_ANNOTATION_NAME);
		FSIterator tokenIterator = acas.getAnnotationIndex(this.tokenAnnotation).iterator();
		
		while (tokenIterator.hasNext()) {
			TokenAnnotation currentToken = (TokenAnnotation)tokenIterator.next();
					
			if ("WORD".equals(currentToken.getTokentype()) 
					&& currentToken.getCoveredText().length() > 1
					&& semanticLemma(currentToken)
					) {
				String lemma = currentToken.getLemma().toLowerCase();
				if (lemma == null) {
					lemma = currentToken.getCoveredText().toLowerCase();
				}
				addCount(lemma);	
			}
		}		
		docs++;
		
	}

	private boolean semanticLemma(TokenAnnotation currentToken) {
		String POS = currentToken.getPos();
		return (semPos.contains(POS) && !auxVerb(currentToken));
	}

	private boolean auxVerb(TokenAnnotation token) {
		boolean rc = false;
		String lemma = token.getLemma();
		if(auxList.contains(lemma)) {
			rc = true;
		}
		return rc;
	}

	//@Override
	/**
	 * @inheritDoc
	 * do post processing like compute probabilities
	 */
	public void destroy() {
		System.out.println("Computing InfGain");
		
		computeWordProbabilities();
		//reComputeProbabilities();
		computeAndSaveCoocDistr();
		computeGeneralDistribution();
		HibernateUtil.shutdown();
		
	}
	
	//@Override
	/**
	 * @inheritDoc
	 * do post processing like compute probabilities
	 */
	public void collectionProcessComplete(ProcessTrace arg0)
			throws ResourceProcessException, IOException {
		
		computeWordProbabilities();
		//reComputeProbabilities();
		computeAndSaveCoocDistr();			
		computeGeneralDistribution();
		HibernateUtil.shutdown();
	}
	
	private void computeGeneralDistribution() {
		List<Lemma> lemmata = HibernateUtil.getAllLemma();
		SparseVector<Float> generalDistr = new SparseVector<Float>();
		for(Lemma lemma:lemmata) {
			InfGainVector iv  = HibernateUtil.findInfGainVector(lemma);
			if(iv != null) {
				int index = iv.getIndex();
				generalDistr.setAt(index,(float) lemma.getFreq()/ (float) nrOfOccurrences);
			}
		}
		
		// store in db
		InfGainVector vector = new InfGainVector(new Integer(-1), generalDistr, null);
		HibernateUtil.saveCODistr(vector);		
	}

	private void addCount(String lemma) {

		Integer lemIndex = invLemmaList.get(lemma);
		if (lemIndex == null) {
			lemIndex = lemmaList.size();
			lemmaList.add(lemma);
			invLemmaList.put(lemma, lemIndex);
		}

		if (lemIndex < index.size()) {
			SparseVector<Integer> w = index.get(lemIndex);
			w.addTo(docs, 1);
		} else {
			SparseVector<Integer> w = new SparseVector<Integer>();
			w.setAt(docs, 1);
			index.add(w);
		}

		Integer docCnt = totalOccurences.get(docs);
		if (docCnt == null) {
			docCnt = 0;
		}
		docCnt++;
		totalOccurences.put(docs, docCnt);
		nrOfOccurrences++;

	}
	
	private void computeAndSaveCoocDistr() {
		
		int i = 0;
		
			
		for(Integer w:invLemmaList.values()) {
			String lemma = lemmaList.get(w);
			Lemma l = HibernateUtil.findLemma(lemma);
			
			if (l != null && l.getDocs() > docs/1000) {			
				SparseVector<Float> cODistr = wordProbsForKeyword(w);
				
				InfGainVector vector = new InfGainVector(w, cODistr, l);
				//l.setCODistr(vector);
				HibernateUtil.saveCODistr(vector);
				l.setVectorId(w);
				HibernateUtil.saveLemma(l);
				
//					writer.writeObject(w);
//					writer.writeObject(cODistr);
//					writer.writeObject(l.getId());
				System.out.println(++i + " saved " + lemma);
			}
			
		}
	}

	
	private void computeWordProbabilities() {
		int nrOfWords = lemmaList.size();
		wordCounts = new float[nrOfWords];
		for (int i = 0; i < nrOfWords; i++) {
			int total = 0;
			SparseVector<Integer> w = index.get(i);
			int j;
			Iterator<Integer> it = w.getKeyIterator();
			while (it.hasNext()) {
				j = it.next();
				total += w.getIntegerAt(j);
			}
			wordCounts[i] = total;
		}
	}
	
	private SparseVector<Float> wordProbsForKeyword(int z) {
		SparseVector<Float> p = new SparseVector<Float>();

		int Nzd, Ntd;
		SparseVector<Integer> distrZ = index.get(z);
		SparseVector<Integer> distrT;

		int nrOfWords = lemmaList.size();

		for (int t = 0; t < nrOfWords; t++) {
			distrT = index.get(t);
			float p_t = 0;
			int d;
	
			Iterator<Integer> it;
			if(distrT.getNrOfNonZeroValues() < distrZ.getNrOfNonZeroValues()) {
				it = distrT.getKeyIterator();
			} else {
				it = distrZ.getKeyIterator();
			}
			
			while(it.hasNext()){
				d = it.next();
				Nzd = distrZ.getIntegerAt(d);
				Ntd = distrT.getIntegerAt(d);
				if(Nzd > 0 && Ntd > 0) {
					p_t += ((float) Nzd / (float) wordCounts[z])
					* ((float) Ntd / (float) totalOccurences.get(d));
				}
			}

			if (p_t > 0) {
				p.setAt(t, p_t);
			}
		}
		return p;
	}
}
