package bbc.rd.subsyst;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

//==================================================================================================
/**
 * Class that reads a Timed Text stream and removes all the tags and any repeated text (commonly seen 
 * as live subtitles are entered on the fly) to produce a normalised text stream.
 * 
 * Note that a side effect of the processing is that repeated lines will be removed, even if they are 
 * legitimately repeated. For statistical purposes this may not be significant but it is not helpful 
 * if creating a transcript.
 */
public class NormalisingSubtitleReader {

  BufferedReader reader;
  String last;
  String next;
  
//--------------------------------------------------------------------------------------------------
  /**
   * Create a new instance of a NormalisingSubtitleReader.
   *
   * @param in an InputStream providing a subtitle stream.
   */
  NormalisingSubtitleReader(InputStream in) {
    reader = new BufferedReader(new InputStreamReader(in));
  }
//--------------------------------------------------------------------------------------------------
  /**
   * Read a line of text. This may lead to multiple lines being read from the underlying InputStream.
   * @return A String containing the contents of the line, not including any line-termination characters, 
   * or null if the end of the stream has been reached 
   * @throws IOException - If an I/O error occurs
   */
  public String readLine() throws IOException {
    String line;
    do {
      line = readRemovingOverlappingPrefix();
    } while(line != null && line.length() == 0);
    return line;
  }
//--------------------------------------------------------------------------------------------------
  /**
   * Read lines, removing prefixes that match the suffix of previous lines.
   *
   * e.g. the sequence:
   *
   *   Top Gear this week features
   *   features racing from Silverstone  
   *
   * would be reduced to:
   *
   *   Top Gear this week features
   *   racing from Silverstone  
   * 
   * @return
   * @throws IOException
   */
  protected String readRemovingOverlappingPrefix() throws IOException {
    String line = readWithAggregatedPrefix();
    int overlapIndex = getOverlapIndex(line, last);
    //System.out.println("last = " + last);
    //System.out.println("line = " + line);
    if(overlapIndex != -1) {
      line = line.substring(overlapIndex).trim();
    }
    last = line;
    //System.out.println("w/o prefix = " +  line);
    return line;
  }
//--------------------------------------------------------------------------------------------------
  /**
   * Read lines, ignoring any lines that are followed by a line which starts with the same prefix.
   *
   * e.g. the sequence:
   *
   *   Top 
   *   Top Gear
   *   Top Gear this
   *   Top Gear this week features
   *
   * would be reduced to:
   *
   *   Top Gear this week features
   * 
   * @return the first line that is not followed by a line which starts with the same prefix.
   * @throws IOException
   */
  protected String readWithAggregatedPrefix() throws IOException {
    //System.out.println("next =       " + next);
    String line;
    if(next == null) {
      line = readWithoutTags();
    } else {
      line = next;
    }
    if(line != null) {
      next = readWithoutTags();
      while(next != null && next.startsWith(line)) {
        //System.out.println("      line = " + line);
        //System.out.println("      next = " + next);
        line = next;
        next = readWithoutTags();
      }
      //System.out.println("Aggregated = " +  line);
      return line;
    } else {
      return null;
    }
  }
//--------------------------------------------------------------------------------------------------
  /**
   * Removes all tags, reduces multiple spaces to single spaces and removes leading and trailing spaces.
   * Skips lines that do not start with "<p"
   * @param in the input string
   * @return the processed string
   */
  protected String readWithoutTags() throws IOException {
    StringBuffer out = new StringBuffer();
    String in;
    do {
      in = reader.readLine();
      if(in == null) return null;
    } while(!in.trim().startsWith("<p"));

    boolean isTag = false;
    boolean lastWasSpace = false;
    for(int i=0; i<in.length(); i++) {
      char c = in.charAt(i);
      if(c == '<') {
        isTag = true;
      } else if(c == '>') {
        isTag = false;
      } else {
        if(!isTag) {
          if(c != ' ') {
            out.append(c);
            lastWasSpace = false;
          } else {
            if(!lastWasSpace) out.append(c);
            lastWasSpace = true;          
          }
        }
      }
    }
    String line = out.toString().trim();
    //System.out.println("w/o tags =   " +  line);
    return line;
  }
//--------------------------------------------------------------------------------------------------
  /**
   * If the beginning of line is identical to the end of last then this method returns the index of the substring
   * of line that needs to be removed to avoid duplication. Otherwise this method returns -1.
   * @param line
   * @param last
   * @return
   */
  protected static int getOverlapIndex(String line, String last) {
    if(line == null || last == null) return -1;
    for(int i=line.length(); i>0; i--) {
      String sub = line.substring(0, i);
      if(last.endsWith(sub)) return i;  
    }
    return -1;
  }
//--------------------------------------------------------------------------------------------------
}
//==================================================================================================
