package nl.telin.mymedia.ontology;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.hp.hpl.jena.rdf.model.InfModel;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Selector;
import com.hp.hpl.jena.rdf.model.SimpleSelector;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;

/**
 * @author Wout.Slakhorst
 * adds weights to ontology concepts based on distance to other subjects
 */
public class WeightCalculator {

	private InfModel ontologyInf;
	private Set<String> terms;
	Map<String, Double> weights;
	private Map<String, Set<String>> distance1;
	private Map<String, Set<String>> distance2;

	private Model model;

	public WeightCalculator() {
		terms = new HashSet<String>();
	}

	public static void main(String[] args) throws Exception {
		WeightCalculator wc = new WeightCalculator();
		if (args.length != 2) {
			printUsage();
			System.exit(1);
		}
		wc.loadModel(args[0]);
		wc.computeWeights();

		// write model
		wc.writeModel(args[1]);
	}

	public static void printUsage() {
		System.out.println("Usage: ");
		System.out.println("java WeightCalculator PATH_TO_ONTOLOGY OUTPUT");
	}

	public void writeModel(String outputFile) throws IOException {
		model.write(new FileWriter(outputFile));
	}

	public void loadModel(String ontology) throws FileNotFoundException {
		try {
			InputStream stream = new FileInputStream(ontology);

			// load jena
			model = ModelFactory.createDefaultModel();
			model.read(stream, "");
			ontologyInf = ModelFactory.createRDFSModel(model);

			ResIterator iter = model.listSubjects();

			while (iter.hasNext()) {
				Resource resource = (Resource) iter.next();
				terms.add(resource.getURI());
			}

		} catch (FileNotFoundException e) {
			throw e;
		}
	}

	public void computeWeights() {
		findAllDistance1Subjects();
		findAllDistance2Subjects();

		// total = 0;
		//double count = 0;
		
		for (String subject : terms) {
			double weight = distance1Relations(subject)
					+ distance2Relations(subject);
			model.add(model.createStatement(model.createResource(subject),
					model.createProperty(
							"http://www.w3.org/2004/02/skos/core#", "relations"),
					new Double(weight).toString()));
			
			//count++;
			//total += weight;
		}
		
		//System.out.println(total/count);

	}

	private int distance1Relations(String subject) {
		int rc = 0;
		Set<String> relatedTerms = distance1.get(subject);

		for (String term : relatedTerms) {
			if (terms.contains(term)) {
				rc++;
			}
		}

		return rc;
	}

	private int distance2Relations(String subject) {
		int rc = 0;
		Set<String> relatedTerms = distance2.get(subject);

		for (String term : relatedTerms) {
			if (terms.contains(term)) {
				rc++;
			}
		}

		return rc;
	}

	private void findAllDistance1Subjects() {
		distance1 = new HashMap<String, Set<String>>();
		for (String subject : terms) {
			distance1.put(subject, getDistance1Subjects(subject));
		}
	}

	private void findAllDistance2Subjects() {
		distance2 = new HashMap<String, Set<String>>();
		for (String subject : terms) {
			distance2.put(subject, getDistance2Subjects(subject));
		}
	}

	/**
	 * returns a set of subjects, each with the namespace
	 * 
	 * @param subject
	 *            with namespace
	 * @return
	 */
	private Set<String> getDistance1Subjects(String subject) {

		Set<String> result = new HashSet<String>();

		// related
		Selector query = new SimpleSelector(
				ontologyInf.createResource(subject), ontologyInf
						.createProperty("http://www.w3.org/2004/02/skos/core#",
								"related"), (RDFNode) null);
		StmtIterator iter = ontologyInf.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			result.add(stmt.getObject().toString());
		}

		// narrower
		query = new SimpleSelector(ontologyInf.createResource(subject),
				ontologyInf.createProperty(
						"http://www.w3.org/2004/02/skos/core#", "narrower"),
				(RDFNode) null);
		iter = ontologyInf.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			result.add(stmt.getObject().toString());
		}

		// broader
		query = new SimpleSelector(ontologyInf.createResource(subject),
				ontologyInf.createProperty(
						"http://www.w3.org/2004/02/skos/core#", "broader"),
				(RDFNode) null);
		iter = ontologyInf.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			result.add(stmt.getObject().toString());
		}

		result.remove(subject);

		return result;
	}

	/**
	 * get all level 2 subjects. For each subject the level 1 subjects are
	 * retrieved after which for each of those the level 1 subjects are
	 * retrieved.
	 * 
	 * @param subject
	 * @return
	 */
	private Set<String> getDistance2Subjects(String subject) {

		Set<String> result = new HashSet<String>();

		Set<String> l1 = distance1.get(subject);
		for (String s : l1) {
			result.addAll(getDistance1Subjects(s));
		}
		result.remove(subject);
		result.removeAll(l1); // CW 2008/12/23
		return result;
	}
}
