﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Data.Linq;
using System.Globalization;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Databases;
using MyMediaProject.RecommenderSystem.Framework;
using EMIC.Design;
using System.Data.SqlClient;
using System.Data;

namespace MovieLens
{
    partial class MovieLensDataFactory : CatalogItemFactory
    {
        const string description = "MovieLens catalog importer";
        const string factoryGuidString = "{24AC6470-C061-4d10-8D23-2673497F0D93}";
        const string imdbQueryString = @"http://us.imdb.com/find?s=all&q=";
        const string mlRatingsFile = @"MovieLensData\ratings.dat";
        const string mlUsersFile = @"MovieLensData\users.dat";
        const string mlMoviesFile = @"MovieLensData\movies.dat";

        Dictionary<int, User> users = new Dictionary<int, User>();

        public MovieLensDataFactory()
            : base(new CatalogItemFactoryId(new Guid(factoryGuidString)), description)
        {
        }

        Module module;
        Language language;
        Service service;

        public void ImportData()
        {
            if (this.DataContext.CatalogItems.Count() == 0)
            {
                this.EmptyStore();
                this.module = this.DataContext.CreateModule("MovieLensImporter", "MyMediaProject");
                this.language = this.DataContext.CreateLanguage(CultureInfo.CurrentCulture);
                this.language.Module = this.module;
                this.service = this.DataContext.CreateService("MovieLens");
                this.service.Owner = "MyMediaProject";

                this.ImportItems();
                this.ImportUsers();
                this.ImportRatings();
            }
        }

        private void ImportUsers()
        {
            using (TextReader file = new StreamReader(mlUsersFile))
            {
                int itemCount = 0;
                string line = file.ReadLine();
                while (null != line)
                {
                    string[] attributes = line.Split(new string[]{"::"}, StringSplitOptions.None);
                    int userId = int.Parse(attributes[0]);
                    User user = new User();
                    user.UserId = userId;
                    this.DataContext.Users.InsertOnSubmit(user);
                    TraceLog.Write(LogEventType.Verbose, "Imported user {0}", itemCount);
                    line = file.ReadLine();
                }
                this.SubmitChanges();
            }
        }

        private void ImportRatings()
        {
            using (TextReader file = new StreamReader(mlRatingsFile))
            {
                int itemCount = 0;
                string line = file.ReadLine();
                while (null != line)
                {
                    string[] attributes = line.Split(new string[] { "::" }, StringSplitOptions.None);
                    int userId = int.Parse(attributes[0]);
                    int item = int.Parse(attributes[1]);
                    double rating = double.Parse(attributes[2]);

                    RatingAction userAction = new RatingAction();
                    userAction.UserId = userId;
                    userAction.ItemId = item;
                    userAction.Rating = RatingAction.ScaleRatingToRecommender(rating, 1,5);
                    this.DataContext.UserActions.InsertOnSubmit(userAction);

                    if (itemCount++ > 10000)
                    {
                        this.SubmitChanges();
                        itemCount = 0;
                    }

                    line = file.ReadLine();
                }
                this.SubmitChanges();
            }
        }

        private void ImportItems()
        {
            this.ImportFromMovieLensFiles();
        }

        private void ImportFromMovieLensFiles()
        {
            string placeholderPropertyName = @"TEMP ITEM, TO BE DELETED";
            Contract.Precondition(this.DataContext.CatalogItems.Count() == 0);

            using (TextReader file = new StreamReader(mlMoviesFile))
            {
                System.Console.WriteLine("Importing MovieLens data...");

                int itemCount = 1;  //starting database identity value.
                string line = file.ReadLine();
                while (null != line)
                {
                    string[] attributes = line.Split(new string[] { "|", "::" }, StringSplitOptions.None);
                    int movieLensId = Convert.ToInt32(attributes[0]);

                    Contract.Precondition(itemCount <= movieLensId);
                    //Workaround for missing movielens movie id's.  In this dataset, they are not 
                    //continuous, so we insert "dummy" items which we will delete at the end of the 
                    //import.  This ensures that the id's for the catalog items, also referenced
                    //in the rating data, are consisten.
                    while (itemCount < movieLensId)
                    {
                        CatalogItem placeholderCatalogItem = this.DataContext.CreateItem();
                        placeholderCatalogItem.ItemId = itemCount++;
                        placeholderCatalogItem.AddProperty(this.DataContext, placeholderPropertyName, string.Empty);
                    }

                    Contract.Postcondition(itemCount == movieLensId);
                    CatalogItem catalogItem = this.DataContext.CreateItem();
                    catalogItem.ItemId = movieLensId;

                    ItemTitle title = new ItemTitle()
                    {
                        Title = attributes[1],
                        Language = this.language,
                        Module = this.module
                    };
                    catalogItem.AddTitle(this.DataContext, title);

                    AddImdbLink(catalogItem, attributes[1]);

                    //Get next line
                    line = file.ReadLine();
                    itemCount++;
                }
                TraceLog.Write(LogEventType.Verbose, "Submitting items.");
                this.SubmitChanges();
                TraceLog.Write(LogEventType.Verbose, "Submitting items complete.");

                //Delete placeholder items
                this.DataContext.CatalogItems.DeleteAllOnSubmit(this.DataContext.CatalogItems.Where(item =>
                    item.ItemProperties.Any(property => property.Property.Name == placeholderPropertyName)));
                this.DataContext.ItemProperties.DeleteAllOnSubmit(this.DataContext.ItemProperties.Where(
                    property => property.Property.Name == placeholderPropertyName));
                this.DataContext.Properties.DeleteAllOnSubmit(this.DataContext.Properties.Where(
                    property => property.Name == placeholderPropertyName));
                this.SubmitChanges();
            }
        }

        private void AddImdbLink(CatalogItem catalogItem, string title)
        {
            //Create an IMDB search string based on the title of the movie.
            //Effectively, this is just the title words separated by a "+"            
            string[] titleWords = title.Split(' ');
            string separatedTitle = string.Empty;
            foreach (string titleWord in titleWords)
            {
                if (separatedTitle != string.Empty)
                {
                    //separate the words of the title by a "+"
                    separatedTitle += "+";
                }
                separatedTitle += titleWord;
            }
            string uriString = imdbQueryString + separatedTitle;
            Uri url = new Uri(uriString);
            Property imdbLink = new Property()
            {
                Name = "ImdbLink",
                Value = url.AbsoluteUri,
                Module = this.module
            };
            ItemProperty imdbProperty = new ItemProperty()
            {
                Module = this.module,
                Property = imdbLink
            };
            catalogItem.ItemProperties.Add(imdbProperty);
        }
    }
}
