﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;

using System.Text;
using System.Threading;
using System.Windows.Forms;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Framework;
using MyMediaProject.RecommenderSystem.Core.Config;
using EMIC.Design;
using MovieLens;
using MyMediaProject.RecommenderSystem.Framework.Databases; // Database Backup/Restore

namespace RecommenderDisplay
{
    public partial class RecommenderDisplayForm : Form
    {
        MovieLensRecommender recommender;
        private const string imdbDatabase = "imdb";
        private const string imdbDatabaseBackupFile = "imdb.bak";
        private const string localDatabaseFolder = "Databases";

        public RecommenderDisplayForm()
        {
            InitializeComponent();
            // reset dummy values
            textBoxMovieName.Clear();
            textBoxMovieURL.Clear();

            System.AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(CurrentDomain_UnhandledException);
        }

        private void RecommenderDisplayForm_Load(object sender, EventArgs e)
        {
            this.recommender = RecommenderSingleton<MovieLensRecommender>.Instance;
            CheckBackups();
            this.recommender.Train();
            GetUsers();
            this.proxyTextBox.Text = ConfigManager.ProxyString;
        }

        private void GetUsers()
        {
            TraceLog.Write(LogEventType.Verbose, "Loading users list");
            using (IReleasableQueryable<User> userIds = this.recommender.Users)
            {
                IEnumerable<UserDisplay> userDisplayIds =
                    from ui in userIds
                    select new UserDisplay(this.recommender, ui);

                SortableBindingList<UserDisplay> userDisplayBindingList =
                    new SortableBindingList<UserDisplay>(userDisplayIds.ToList());
                userDisplayBindingList.AddingNew += new AddingNewEventHandler(AddingNewUserDisplayIds);
                this.usersDataGrid.DataSource = userDisplayBindingList;
                TraceLog.Write(LogEventType.Verbose, "Loading users list complete.");
            }
        }

        void AddingNewUserDisplayIds(object sender, AddingNewEventArgs e)
        {
            Contract.Precondition(this.usersDataGrid.DataSource is SortableBindingList<UserDisplay>);
            User user = new User();
            UserDisplay newUser = new UserDisplay(this.recommender, user);
            this.recommender.AddUsers(new User[]{user});
            e.NewObject = newUser;
            this.itemsDataGrid.Rows.Clear();
        }

        private void dataGridView1_SelectionChanged(object sender, EventArgs e) // Selection of user has changed (LHS column)
        {
            if (this.usersDataGrid.SelectedRows.Count != 0)
            {
                if (null != this.usersDataGrid.SelectedRows[0] && null != this.usersDataGrid.SelectedRows[0].Cells[0])
                {
                    User userId = this.usersDataGrid.SelectedRows[0].Cells[0].Value as User;
                    if (null != userId)
                    {
                        TraceLog.Write(LogEventType.Verbose, "Updating users ratings list for user: {0}.", userId);

                        IEnumerable<FeedbackDisplay> itemList;
                        IList<UserAction> userActions = recommender.GetUserFeedback(userId).ToList();
                        if (0 != userActions.Count())
                        {
                            IList<RatingAction> feedbackForUser = userActions.Select(item => item as RatingAction).ToList();
                            itemList = from prediction in this.recommender.Predict(this.recommender.CatalogItemStore, this.recommender.DefaultFlowId, userId)
                                       join action in feedbackForUser on prediction.CatalogItem.ItemId equals action.ItemId into oc
                                       from x in oc.DefaultIfEmpty()
                                       select new FeedbackDisplay(this.recommender, prediction.CatalogItem, x, prediction, this.usersDataGrid.SelectedRows[0].Cells[0].Value as User);
                        }
                        else
                        {
                            itemList = from ci in this.recommender.CatalogItemStore
                                       select new FeedbackDisplay(this.recommender, ci, null, new Prediction<CatalogItem>(ci, 0.0), userId);
                        }

                        SortableBindingList<FeedbackDisplay> catalogList = new SortableBindingList<FeedbackDisplay>(itemList.OrderByDescending(item => item.ActualRating).ThenByDescending(item => item.Title).ToList());
                        //SortableBindingList<FeedbackDisplay> catalogList = new SortableBindingList<FeedbackDisplay>(itemList.OrderByDescending(item => item.ActualRating.GetValueOrDefault(Rating.MinValue - 1)).ThenByDescending(item => item.Title).ToList());

                        this.itemsDataGrid.DataSource = catalogList;

                        TraceLog.Write(LogEventType.Verbose, "Updating users ratings list complete.");
                    }
                    else
                    {
                        MessageBox.Show("Invalid User.", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                        this.itemsDataGrid.SelectedCells[0].Value = null;
                    }
                }
                else
                {
                    ClearRecommendations();
                }
            }
        }

        private void resetToolStripMenuItem_Click(object sender, EventArgs e)
        {
            recommender.ResetAllSettings();

            ClearRecommendations();

            this.GetUsers();
        }

        private void ClearRecommendations()
        {
            this.dataGridView1_SelectionChanged(null, null);
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void RecommenderDisplayForm_KeyPress(object sender, KeyPressEventArgs e)
        {

        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            Contract.Precondition(this.itemsDataGrid.DataSource is SortableBindingList<FeedbackDisplay>);
            SortableBindingList<FeedbackDisplay> binding = this.itemsDataGrid.DataSource as SortableBindingList<FeedbackDisplay>;
            FeedbackDisplay targetElement = binding.FirstOrDefault(
                c => c.Title.StartsWith(this.ratedItemsSearchTextBox.Text, StringComparison.InvariantCultureIgnoreCase));
            if (null != targetElement)
            {
                int index = binding.IndexOf(targetElement);
                this.itemsDataGrid.FirstDisplayedScrollingRowIndex = index;
            }
        }

        private void button1_Click(object sender, EventArgs e)
        {
            this.recommender.Train();
            this.dataGridView1_SelectionChanged(sender, e);
        }

        //void FeedbackListChanged(object sender, ListChangedEventArgs e)
        //{
        //    Contract.Precondition(sender is SortableBindingList<FeedbackItem>);
        //    SortableBindingList<FeedbackItem> list = sender as SortableBindingList<FeedbackItem>;
        //    if (e.ListChangedType == ListChangedType.ItemChanged)
        //    {
        //        Contract.Assert(false, "Feedback item change is not implemented");
        //    }
        //    else if (e.ListChangedType == ListChangedType.ItemDeleted)
        //    {
        //        Contract.Assert(false, "Feedback item delete is not implemented");
        //    }
        //    else
        //    {
        //        Contract.Assert(e.ListChangedType != ListChangedType.ItemMoved, "Event type not handled.");
        //        Contract.Assert(e.ListChangedType != ListChangedType.PropertyDescriptorAdded, "Event type not handled.");
        //        Contract.Assert(e.ListChangedType != ListChangedType.PropertyDescriptorChanged, "Event type not handled.");
        //        Contract.Assert(e.ListChangedType != ListChangedType.PropertyDescriptorDeleted, "Event type not handled.");
        //    }
        //}


        private void CurrentDomain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
        {
            MessageBox.Show(e.ExceptionObject.ToString(),
                "An Unhandled Error Has Occurred: The Application Will Now Exit.",
                MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show("MovieLensRecommender Demonstrator V0.0.0" + Environment.NewLine + Environment.NewLine +
                "Developed as part of the EU FP7 project \"MyMedia\".",
                "About MovieLensRecommender", MessageBoxButtons.OK);
        }

        private void restoreDatabaseToolStripMenuItem_Click(object sender, EventArgs e)
        {
        }

        private void backupDatabaseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                DatabaseHelper.BackupRecommenderDatabase(this.recommender);
                MessageBox.Show("Database backup succeeded", "Database Backup", MessageBoxButtons.OK);
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Database Backup Failed",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void RestoreImdbDatabaseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            throw new NotImplementedException();
        }

        private void RestoreMyMediaDatabaseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                DatabaseHelper.RestoreRecommenderDatabase(this.recommender);
                MessageBox.Show("Sucessfully restored the database.",
                    "Database Restore", MessageBoxButtons.OK);                
            }
            catch (Exception exception)
            {
                if (null != sender && null != e) //Do this in "quiet" mode if the sender/eventargs are null
                {
                    MessageBox.Show("Could not restore the database" + imdbDatabase + "\r\n" + exception.ToString(),
                        "Database Restore Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void CheckBackups()
        {
            restoreDatabaseToolStripMenuItem.Enabled = false;
        }

        private static string LocalDatabaseBackupPath
        {
            get
            {
                string localDbPath = Path.Combine(Directory.GetCurrentDirectory(), localDatabaseFolder);
                if(!Directory.Exists(localDbPath))
                {
                    Directory.CreateDirectory(localDbPath);
                }
                return localDbPath;
            }
        }

        private void proxyTextBox_TextChanged(object sender, EventArgs e)
        {
            ConfigManager.ProxyString = proxyTextBox.Text;
        }

        private void dataGridView1_SelectionChanged(object sender, DataGridViewCellEventArgs e)
        {
            this.dataGridView1_SelectionChanged(sender, e as EventArgs);
        }
    }
}
