﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Linq;
using System.Text;
using System.Reflection;
using MyMediaProject.RecommenderSystem.Core.Config;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;

namespace ExternalSystemIntegration
{
    /// <summary>
    /// Integration components providing integration to external systems (typically, social networks)
    /// can derive from this class, inheriting many handy methods, including initialization and automatic 
    /// deployment via the constructor
    /// All ExternalSystemIntegrator members are decorated for the simplier usage
    /// </summary>
    public abstract partial class AbstractIntegrationDecorator : IExternalSystemIntegrator, IDisposable
    {
        /// <summary>
        /// Decorated ExternalSystemIntegrator's instance
        /// </summary>
        private ExternalSystemIntegrator Integrator;

        // UsersManager instance
        private UsersManager UsersMngr;

        // db context
        private MyMediaDataContext Context;

        private bool ownsDataContext;

        // text to be written to a configuration file as a description of the setting that is updated
        private String DescriptionForProgrammaticallyGeneratedConfigProperties = "Programmatically generated";

        /// <summary>
        /// Constructor. Initializes integration component's members (deploys and/or enforces
        /// constituents from configuration to the DB if needed) by the name of the component
        /// </summary>
        /// <param name="integratorName">Name of the integrator.</param>
        /// <example>
        /// If <c>BlaSocialNetwork</c> is an integration component extending <c>AbstractIntegrationDecorator</c>,
        /// the constructor would be:
        /// <code>
        /// // Constructor of the BlaSocialNetwork integration component
        /// public BlaSocialNetworkIntegrator()
        /// : base(integrationComponentName_as_registered_in_the_DB)
        /// </code>
        /// </example>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if no integration component with such name is registered in the DB.</exception>
        protected AbstractIntegrationDecorator(String integratorName)
            : this(integratorName, true)
        {
        }

        /// <summary>
        /// Constructor without specified DataContext
        /// </summary>
        /// <param name="integratorName"></param>
        /// <param name="isDeploymentToBeEnsured"></param>
        protected AbstractIntegrationDecorator(String integratorName, bool isDeploymentToBeEnsured) : this(integratorName, isDeploymentToBeEnsured, null)
        {
        }

        /// <summary>
        /// Constructor. Initializes integration component's members (deploys and/or enforces
        /// constituents from configuration to the DB if required) by the name of the component
        /// </summary>
        /// <param name="integratorName">Name of the integrator.</param>
        /// <param name="isDeploymentToBeEnsured">if true, ensures that the DB entries are in line with the configuration;</param>
        /// <param name="context">Initialized DataContext to be used by this class.</param>
        /// <example>
        /// If <c>BlaSocialNetwork</c> is an integration component extending <c>AbstractIntegrationDecorator</c>,
        /// the constructor would be:
        /// <code>
        /// // Constructor of the BlaSocialNetwork integration component
        /// public BlaSocialNetworkIntegrator()
        /// : base(integrationComponentName_as_registered_in_the_DB)
        /// </code>
        /// </example>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if no integration component with such name is registered in the DB.</exception>
        protected AbstractIntegrationDecorator(String integratorName, bool isDeploymentToBeEnsured, MyMediaDataContext context)
        {
            TraceLog.Write(LogEventType.Verbose, "AbstractIntegrationDecorator constructor with name = '" + integratorName + "' called.");

            String dbConnectionString = "";
            if (context != null)
            {
                this.Context = context;
                
                dbConnectionString = context.Connection.DataSource;
            }else{
                dbConnectionString = ConfigManager.DBConnectionString;
                this.Context = new MyMediaDataContext(dbConnectionString);
                ownsDataContext = true;
            }

            this.UsersMngr = new UsersManager(this.Context);

            if (isDeploymentToBeEnsured)
            {
                ConfigBasedDeployer deployer = new ConfigBasedDeployer(dbConnectionString);
                this.Integrator = deployer.EnsureIntegrationComponentIsDeployed(integratorName);
            }
            else
            {
                ExternalSystemIntegrator theIntegrator = Context.ExternalSystemIntegrators.FirstOrDefault<ExternalSystemIntegrator>(i => i.Name == integratorName);
                this.Integrator = theIntegrator;
            }
            TraceLog.Write(LogEventType.Information, "Number of IntegratorSupportedFeatures initialized: " + Integrator.IntegratorSupportedFeatures.Count);
        }


        /// <summary>
        /// Performs necessary steps to authenticate user at an external social network. 
        /// Different derived classes might use distinct approaches.
        /// </summary>
        /// <param name="dictionaryParameters">
        /// Name value pairs of the parameters needed by implementations(e.g. fallback URL, username and password pair).
        /// The names of the parameters are defined by implementations.
        /// </param>
        /// <returns>ExternalSystemIntegratorResult which could be WebBrowserRedirectRequest</returns>
        public abstract ExternalSystemIntegratorResult PerformExternalAuthentication(Dictionary<String, Object> dictionaryParameters);


        /// <summary>
        /// Invokes integrator's method(operation provided by external system's integration component) by given method name.
        /// Useful for performing callback of WebBrowserRedirectRequest.
        /// </summary>
        /// <param name="operationName">Name of the operation.</param>
        /// <param name="operationParameters">Parameters to be passed to the specified operation</param>
        /// <returns>ExternalSystemIntegratorResult object</returns>
        public virtual ExternalSystemIntegratorResult PerformOperationCall(string operationName, Object operationParameters)
        {
            ExternalSystemIntegratorResult result;
            MethodInfo mInfo;
            Object[] parameters;
            if (operationParameters != null)
            {
                Type[] paramTypes = new Type[] { operationParameters.GetType() };
                mInfo = this.GetType().GetMethod(operationName, paramTypes);
                parameters = new Object[] { operationParameters };
            }else{
                parameters = new Object[]{};
                mInfo = this.GetType().GetMethod(operationName, new Type[] { });
            }

            // if no method found
            if (mInfo == null)
            {
                throw new ExternalSystemIntegratorException("No method '" + operationName + "' found.");
            }

            try
            {
                result = (ExternalSystemIntegratorResult)mInfo.Invoke(this, parameters);
            }
            catch (System.Reflection.TargetInvocationException ex) // catching an exception that executed method might throw
            {
                TraceLog.Write(LogEventType.Error, "Exception occured while invoking a method '{0}': '{1}'.", new object[] { operationName, ex.ToString() });
                
                throw new ExternalSystemIntegratorException(ex.InnerException);
            }

            return result;
        }

        /// <summary>
        /// Initializes external user (creates it, if not registered in the DB yet)
        /// </summary>
        /// <param name="externallyAuthenticatedUser"></param>
        /// <returns></returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(ExternalSystemUser externallyAuthenticatedUser)
        {
            return UsersMngr.AssociateExternalAndLocalUser(externallyAuthenticatedUser, this.Integrator.ExternalSystemId, this.Integrator.ModuleId);
        }

        /// <summary>
        /// Initializes external user referencing the local one
        /// Creates both an external user, and a local one, if not registered in the DB yet.
        /// </summary>
        /// <param name="externallyAuthenticatedUser"></param>
        /// <param name="localUser">Local user to associate the external user to(provide null if does not exist)</param>
        /// <returns>Initialized external system user, referencing initialized local user</returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(ExternalSystemUser externallyAuthenticatedUser, User localUser)
        {
            return UsersMngr.AssociateExternalAndLocalUser(externallyAuthenticatedUser, localUser, this.Integrator.ExternalSystemId, this.Integrator.ModuleId);
        }


        /// <summary>
        /// Name of the configuration module that specifies component's features
        /// </summary>
        /// <param name="integratorName">Name of the integration component</param>
        /// <returns>The name of the configuration module that specifies data about component's provided features</returns>
        public static String GetIntegratorFeaturesModuleName(String integratorName)
        {
            return integratorName + "Features";
        }

        /// <summary>
        /// Ensures that the DB data about user's external group memberships is 
        /// up to date (holds all external groups, and notices cancelled memberships)
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="externalGroupIdNamePair"></param>
        public void EnsureUserExternalGroupMembershipsAreUpToDate(ExternalSystemUser externalUser, ICollection<ExternalGroupsManagement.ExternalGroupIdNamePair> externalGroupIdNamePair)
        {
            ExternalGroupsManagement externalGroupsManagment = new ExternalGroupsManagement(Context);
            externalGroupsManagment.EnsureUserExternalGroupMembershipsAreUpToDate(externalUser, externalGroupIdNamePair);
        }

        /**********************/
        /* Decorated  members */
        /**********************/

        /// <summary>
        /// Name of the decorator (used as ID in the DB and configuration)
        /// </summary>
        public String Name
        {
            get
            {
                return this.Integrator.Name;
            }
            set
            {
                this.Integrator.Name = value;
            }
        }

        /// <summary>
        /// Tracing ModuleId (table Module)
        /// </summary>
        public int ModuleId
        {
            get
            {
                return this.Integrator.ModuleId;
            }
        }

        /// <summary>
        /// The Id from the DB of the ExternalSystem object 
        /// </summary>
        public int ExternalSystemId {
            get
            {
                return this.Integrator.ExternalSystemId;
            }
            set
            {
                this.Integrator.ExternalSystemId = value;
            }
        }

        /// <summary>
        /// Retrieves the reference to the ExternalSystem object that is associated with the Integrator
        /// </summary>
        /// <returns>the ExternalSystem object that this Integrator interacts with</returns>
        public ExternalSystem ExternalSystem
        {
            get
            {
                return Integrator.ExternalSystem;
            }
            set
            {
                Integrator.ExternalSystem = value;
                Integrator.ExternalSystemId = value.ExternalSystemId;
            }
        }

        /// <summary>
        /// Retrieves the value of the requested property
        /// </summary>
        /// <param name="propertyName">the name of the property, as registered in the configuration</param>
        /// <returns>property value or null if property is not set</returns>
        public String GetProperty(String propertyName)
        {
            String ModuleName = this.Integrator.Name;
            if (ConfigManager.PropertyValueExists(ModuleName, propertyName))
            {
                return ConfigManager.Modules[ModuleName].Properties[propertyName].Value;
            }

            return null;
        }


        /// <summary>
        /// Sets the config value for the specified property
        /// </summary>
        /// <param name="propertyName"></param>
        /// <param name="propertyValue"></param>
        /// <returns></returns>
        public String SetProperty(String propertyName, String propertyValue)
        {
            String ModuleName = this.Integrator.Name;
            if (ConfigManager.PropertyValueExists(ModuleName, propertyName))
            {
                ConfigManager.Modules[ModuleName].Properties[propertyName].Value = propertyValue;
            }
            else
            {
                ConfigProperty property = new ConfigProperty(propertyName, propertyValue, DescriptionForProgrammaticallyGeneratedConfigProperties);
                ConfigManager.Modules[ModuleName].Properties.Add(propertyName, property);
            }

            ConfigManager.SaveModuleConfiguration(ModuleName);
            return null;
        }

        /// <summary>
        /// Gets the integrator supported features.
        /// </summary>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
        public EntitySet<IntegratorSupportedFeature> GetIntegratorSupportedFeatures()
        {
            return Integrator.IntegratorSupportedFeatures;
        }

        /// <summary>
        /// Checks if an external integration component supports the specified feature
        /// </summary>
        /// <param name="featureName">Name of the feature.</param>
        /// <returns>
        /// 	<c>true</c> if [is feature supported] [the specified feature name]; otherwise, <c>false</c>.
        /// </returns>
        public bool IsFeatureSupported(string featureName)
        {
            foreach (IntegratorSupportedFeature feature in GetIntegratorSupportedFeatures()/* integrator.IntegratorSupportedFeatures */)
            {
                if (featureName.Equals(feature.ExternalIntegrationFeature.Name))
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Checks if an external integration component supports the specified feature
        /// </summary>
        /// <param name="feature">the object holding ExternalIntegrationFeature</param>
        /// <returns>true if feature is supported by the integration component, false otherwise.</returns>
        public virtual bool IsFeatureSupported(ExternalIntegrationFeature feature)
        {
            foreach (IntegratorSupportedFeature f in GetIntegratorSupportedFeatures())
            {
                if (f.ExternalIntegrationFeatureId.Equals(feature.ExternalIntegrationFeatureId))
                {
                    return true;
                }
            }
            return false;
        }

        private ExternalIntegrationFeature GetIntegrationFeatureByItsName(String featureName)
        {
            ExternalIntegrationFeature feature = Context.ExternalIntegrationFeatures.FirstOrDefault<ExternalIntegrationFeature>(f => f.Name == featureName);
            if (feature != null)
            {
                return feature;
            }

            // the feature is not supported by the component, so it can't be used by the user
            throw new ArgumentException("The feature could not been found.");
        }

        /// <summary>
        /// Verifies if an integration feature is enabled by user
        /// </summary>
        /// <param name="featureName"></param>
        /// <param name="user"></param>
        /// <returns>true if feature enabled, false otherwise</returns>
        public virtual bool IsFeatureAllowedByUser(String featureName, User user)
        {
            int integrationFeatureId = default(int);
            foreach (IntegratorSupportedFeature feature in GetIntegratorSupportedFeatures()/* integrator.IntegratorSupportedFeatures */)
            {
                if (featureName.Equals(feature.ExternalIntegrationFeature.Name))
                {
                    integrationFeatureId = feature.ExternalIntegrationFeatureId;
                    break;
                }
            }

            if (integrationFeatureId == default(int))
            {
                // the feature is not supported by the component, so it can't be used by the user
                return false;
            }

            // reverse logic
            return !(Context.DisabledByUserFeatures.Any<DisabledByUserFeature>(f => f.ExternalIntegrationFeatureId == integrationFeatureId && f.UserId == user.UserId && f.ExternalSystemId == Integrator.ExternalSystemId));
        }


        /// <summary>
        /// Disables specified integration feature for given user
        /// </summary>
        /// <param name="featureName"></param>
        /// <param name="user"></param>
        /// <returns>true if the feature has been disabled, false if nothing needed to be done(if the feature was disabled already)</returns>
        public bool DisableIntegrationFeatureForUser(String featureName, User user)
        {
            // if the feature disabled already, do nothing
            if (!IsFeatureAllowedByUser(featureName, user))
            {
                return false;
            }

            ExternalIntegrationFeature feature;
            try
            {
                feature = GetIntegrationFeatureByItsName(featureName);
            }catch(ArgumentException ex){// the feature does not exist
                throw new ExternalSystemIntegratorException(ex);
            }

            // does this integration component support the feature?
            if (!IsFeatureSupported(feature)){
                throw new ArgumentException("This feature is not supported by the integration component.");
            }

            // the feature exists, is supported by the integration component and loaded, do disable it now
            return DisableIntegrationFeatureForUser(feature, user);
        }

        /// <summary>
        /// Disables specified integration feature for given user
        /// </summary>
        /// <param name="feature"></param>
        /// <param name="user"></param>
        /// <returns>true if the feature has been disabled, false if nothing needed to be done(if the feature was disabled already)</returns>
        public bool DisableIntegrationFeatureForUser(ExternalIntegrationFeature feature, User user)
        {
            // if the feature disabled already, do nothing
            if (!IsFeatureAllowedByUser(feature.Name, user))
            {
                return false;
            }
            DisabledByUserFeature disabledByUserFeature = new DisabledByUserFeature();
            disabledByUserFeature.UserId = user.UserId;
            disabledByUserFeature.ExternalSystemId = Integrator.ExternalSystem.ExternalSystemId;
            disabledByUserFeature.ExternalIntegrationFeatureId = feature.ExternalIntegrationFeatureId;
            disabledByUserFeature.Created = DateTime.Now;
            disabledByUserFeature.Modified = DateTime.Now;

            Context.DisabledByUserFeatures.InsertOnSubmit(disabledByUserFeature);
            Context.SubmitChanges();

            return true;
        }

        /// <summary>
        /// Removes user block(if present) from certain integration feature (specified by its name)
        /// </summary>
        /// <param name="featureName"></param>
        /// <param name="user"></param>
        /// <returns></returns>
        public bool EnableIntegrationFeatureForUser(String featureName, User user)
        {
            return EnableIntegrationFeatureForUser(GetIntegrationFeatureByItsName(featureName), user);
        }

        /// <summary>
        /// Removes user block(if present) from certain integration feature
        /// </summary>
        /// <param name="feature"></param>
        /// <param name="user"></param>
        /// <returns>true if the blocking had to be removed, false if blocking did not exist(and thus nothing has been changed)</returns>
        public bool EnableIntegrationFeatureForUser(ExternalIntegrationFeature feature, User user)
        {
            DisabledByUserFeature disabledFeature = Context.DisabledByUserFeatures.FirstOrDefault<DisabledByUserFeature>(f => f.ExternalIntegrationFeatureId == feature.ExternalIntegrationFeatureId && f.UserId == user.UserId && f.ExternalSystemId == this.Integrator.ExternalSystemId);
            // if the feature has not been blocked, there is nothing to do
            if (disabledFeature == null)
            {
                return false;
            }

            Context.DisabledByUserFeatures.DeleteOnSubmit(disabledFeature);
            Context.SubmitChanges();
            return true;
        }


        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.ownsDataContext && disposing)
            {
                this.Context.Connection.Dispose();
                this.Context.Dispose();
                this.UsersMngr.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion
    }
}
