﻿using System;
using MyMediaProject.RecommenderSystem.Core;
using System.Collections.Generic;
using System.Linq;
using System.Data.Linq;
using MyMediaProject.RecommenderSystem.Core.Config;
using EMIC.Design;

namespace ExternalSystemIntegration
{
    /// <summary>
    /// Ensures that ExternalIntegrationComponent is deployed correctly together with all its properties,
    /// and that the DB holds the updated values (in line with the configuration settings)
    /// </summary>
    public class ConfigBasedDeployer : IExternalIntegrationDeployer, IDisposable
    {
        private String DbConnectionString;

        private ExternalSystemIntegrator Integrator;

        private MyMediaDataContext Context;


        /// <summary>
        /// Constructor. Requires database connection string to be passed (needed to create DB if missing)
        /// </summary>
        /// <param name="dbConnectionString">The db connection string.</param>
        public ConfigBasedDeployer(String dbConnectionString)
        {
            this.DbConnectionString = dbConnectionString;
            this.Context = new MyMediaDataContext(DbConnectionString);
        }

        /// <summary>
        /// Constructor. Uses default connection string
        /// </summary>
        public ConfigBasedDeployer()
        {
            this.DbConnectionString = ConfigManager.DBConnectionString;
            this.Context = new MyMediaDataContext(DbConnectionString);
        }

        /// <summary>
        /// Deploys external system integration component, if not deployed yet (by the specified component's name)
        /// </summary>
        /// <param name="integratorName">The name(unique identifier) of the integration component</param>
        /// <returns>Initialized and deployed integration component</returns>
        public ExternalSystemIntegrator EnsureIntegrationComponentIsDeployed(String integratorName)
        {
            // make sure that the ExternalSystemIntegration tables exist in the DB
            CreateDatabaseUtility.CreateExternalSystemIntegrationDatabaseIfMissing(DbConnectionString);

            // Deploy and/or initialize the external system object
            ExternalSystem TheExternalSystem = DeployOrInitializeExternalSystem(integratorName);

            // initialize the decorated ExternalSystemIntegrator instance
            this.Integrator = DeployOrInitializeIntegrator(integratorName, TheExternalSystem);

            // Deploy and/or initialize integrator's supported features
            EntitySet<IntegratorSupportedFeature> configuredFeatures = DeployOrInitializeIntegratorFeatures();

            // removes obsolete features from DB
            RemoveObsoleteFeatures(configuredFeatures);

            Integrator.IntegratorSupportedFeatures = configuredFeatures;

            return Integrator;
        }


        private int InitializeIntegratorsTracingModuleId(String integratorName, String description)
        {
            Module module = Context.Modules.FirstOrDefault<Module>(m => m.Name == integratorName);
            if (module != null && module.ModuleId != default(int))
            {
                return module.ModuleId;
            }

            // Tracing Module is not deployed for this integration component yet
            module = new Module();
            module.Name = integratorName;
            module.Description = description;
            module.Organization = integratorName;
            module.Created = DateTime.Now;
            module.Modified = DateTime.Now;

            Context.Modules.InsertOnSubmit(module);
            Context.SubmitChanges();

            TraceLog.Write(LogEventType.Information, "New module('" + module.Name + "') has been created (ModuleId=" + module.ModuleId + ").");
            return module.ModuleId;
        }


        /// <summary>
        /// Reads the data about Integrator supported features from the configuration,
        /// and ensures the DB corresponds to the configuration(existing and up-to-date)
        /// </summary>
        /// <returns></returns>
        private EntitySet<IntegratorSupportedFeature> DeployOrInitializeIntegratorFeatures()
        {
            EntitySet<IntegratorSupportedFeature> result = new EntitySet<IntegratorSupportedFeature>();
            String IntegratorFeaturesModuleName = AbstractIntegrationDecorator.GetIntegratorFeaturesModuleName(this.Integrator.Name);

            // if there is no such ConfigModule, or it has no properties=features, return null
            if (!ConfigManager.Modules.Keys.Contains(IntegratorFeaturesModuleName)
                || ConfigManager.Modules[IntegratorFeaturesModuleName].Properties.Count == 0)
            {
                return null;
            }


            foreach (ConfigProperty property in
                        ConfigManager.Modules[IntegratorFeaturesModuleName].Properties.Values)
            {
                TraceLog.Write(LogEventType.Verbose, "Loaded " + property.ToString());
                IntegratorSupportedFeature integratorFeature = null;

                integratorFeature = DeployOrInitializeIntegratorSupportedFeature(property);
                result.Add(integratorFeature);
            }

            return result;
        }

        /// <summary>
        /// Ensures that an individual feature is registered in the DB, and is up-to-date
        /// (if missing, creates it from the configuration data)
        /// </summary>
        /// <param name="property">Configuration property encapsulating data of the integration feature</param>
        /// <returns>Initialized integration feature object</returns>
        private IntegratorSupportedFeature DeployOrInitializeIntegratorSupportedFeature(ConfigProperty property)
        {
            IntegratorSupportedFeature theFeature = null;

            // does the integrator's feature with that name exist?
            theFeature = Context.IntegratorSupportedFeatures.FirstOrDefault<IntegratorSupportedFeature>
                                (f => f.ExternalSystemIntegratorId == this.Integrator.ExternalSystemIntegratorId
                                    && f.ExternalIntegrationFeatureId == f.ExternalIntegrationFeature.ExternalIntegrationFeatureId
                                    && f.ExternalIntegrationFeature.Name == property.Name);
            // the feature exists
            if (theFeature != null)
            {
                bool IsVisibleAccordingToConfig = false;
                // make sure data is up to date
                if ("Visible".Equals(property.Value))
                {
                    IsVisibleAccordingToConfig = true;
                }

                // if object needs to be changed in the DB
                if (theFeature.ExternalIntegrationFeature.UpdateObjectIfNeeded(property.Description, IsVisibleAccordingToConfig))
                {
                    Context.SubmitChanges();
                }
            }


            // if the IntegratorSupportedFeature is not in the DB yet
            if (theFeature == null)
            {
                TraceLog.Write(LogEventType.Information, "The Feature '{0}' is not registered in the DB yet.", new object[] { property.Name });

                // make sure the feature is registered as global first
                ExternalIntegrationFeature globalFeature = CreateExternalIntegrationFeatureIfNeeded(property);

                // register the feature in the DB
                theFeature = AssociateFeatureWithTheIntegrator(globalFeature.ExternalIntegrationFeatureId);
            }

            return theFeature;
        }


        /// <summary>
        /// Registers ExternalIntegrationFeature in the DB if it does not exist yet
        /// </summary>
        /// <param name="property"></param>
        /// <returns>Initialized Linq ExternalIntegrationFeature object</returns>
        private ExternalIntegrationFeature CreateExternalIntegrationFeatureIfNeeded(ConfigProperty property)
        {
            ExternalIntegrationFeature theFeature = null;
			theFeature = Context.ExternalIntegrationFeatures.FirstOrDefault<ExternalIntegrationFeature>(f => f.Name == property.Name);

            if (theFeature != null)
            {
                TraceLog.Write(LogEventType.Information, "The Feature '{0}' is not registered in the DB yet.", new object[] { property.Name});
                return theFeature;
            }
            // feature is missing, create it
            theFeature = new ExternalIntegrationFeature();
            theFeature.Name = property.Name;
            theFeature.Description = property.Description;
            theFeature.IsVisible = false;
            if ("Visible".Equals(property.Value))
            {
                theFeature.IsVisible = true;
            }

            theFeature.Created = DateTime.Now;
            theFeature.Modified = DateTime.Now;

            // do the insert to the DB
            Context.ExternalIntegrationFeatures.InsertOnSubmit(theFeature);
            Context.SubmitChanges();

            return theFeature;
        }


        /// <summary>
        /// Removes integrator's features from the DB that are not enabled by the configuration 
        /// </summary>
        /// <param name="configuredFeatures"></param>
        private void RemoveObsoleteFeatures(EntitySet<IntegratorSupportedFeature> configuredFeatures)
        {
            bool IsSubmitRequired = false;

            // if the list of features that are specified in the DB differ from the ones in the configuration, 
            // remove the obsolete links to features in the DB
            foreach (IntegratorSupportedFeature dbFeature in Integrator.IntegratorSupportedFeatures)
            {
                bool IsDbFeatureEnabled = false;
                foreach (IntegratorSupportedFeature confFeature in configuredFeatures)
                {
                    if (dbFeature.ExternalIntegrationFeature.Name.Equals(confFeature.ExternalIntegrationFeature.Name))
                    {
                        IsDbFeatureEnabled = true;
                        break;
                    }
                }

                if (!IsDbFeatureEnabled)
                {
                    // this is an obsolete feature that is not enabled by the configuration, delete it from the DB
                    Context.IntegratorSupportedFeatures.DeleteOnSubmit(dbFeature);
                    TraceLog.Write(LogEventType.Information, "Integrator's feature '{0}' (IntegratorSupportedFeatureId={2}) of the integation component '{1}' has been removed from the DB, because it is not enabled in the component's configuration.", new object[] { dbFeature.ExternalIntegrationFeature.Name, Integrator.Name, dbFeature.IntegratorSupportedFeatureId });
                    IsSubmitRequired = true;
                }
            }

            // the deletion of the disabled features leads to the Context change, and should be done outside the cycles above
            if (IsSubmitRequired)
            {
                Context.SubmitChanges();
            }
        }

        /// <summary>
        /// Adds entry to the IntegratorSupportedFeatures table, specifying that the integration component 
        /// has the given ExternalIntegrationFeature enabled.
        /// </summary>
        /// <param name="globalFeatureId">Id of the ExternalIntegrationFeature to be anabled for the integrator</param>
        /// <returns>Initialized integrator's feature Linq object</returns>
        private IntegratorSupportedFeature AssociateFeatureWithTheIntegrator(int globalFeatureId)
        {
            IntegratorSupportedFeature theFeature = new IntegratorSupportedFeature();

            theFeature.ExternalSystemIntegratorId = this.Integrator.ExternalSystemIntegratorId;
            theFeature.ExternalIntegrationFeatureId = globalFeatureId;
            theFeature.Created = DateTime.Now;
            theFeature.Modified = DateTime.Now;

            // do the insert to the DB
            Context.IntegratorSupportedFeatures.InsertOnSubmit(theFeature);
            Context.SubmitChanges();

            return theFeature;
        }


        /// <summary>
        /// Returns intialized Linq ExternalSystemIntegrator object (creates it from Config, if missing in the DB)
        /// </summary>
        /// <param name="integratorName"></param>
        /// <param name="externalSystem"></param>
        /// <returns>intialized Linq ExternalSystemIntegrator object</returns>
        private ExternalSystemIntegrator DeployOrInitializeIntegrator(String integratorName, ExternalSystem externalSystem)
        {
            ExternalSystemIntegrator theIntegrator = null;
            theIntegrator = Context.ExternalSystemIntegrators.FirstOrDefault<ExternalSystemIntegrator>(i => i.Name == integratorName);

            if (theIntegrator != null)
            {
                return theIntegrator;
            }

            TraceLog.Write(LogEventType.Information, "The Integrator Component '{0}' needs to be registered.", new object[] { integratorName });
            theIntegrator = new ExternalSystemIntegrator();

            theIntegrator.ModuleId = InitializeIntegratorsTracingModuleId(integratorName, ConfigManager.Modules[integratorName].Description);

            theIntegrator.ExternalSystemId = externalSystem.ExternalSystemId;
            theIntegrator.ExternalSystem = externalSystem;
            theIntegrator.Name = integratorName;
            theIntegrator.Created = DateTime.Now;
            theIntegrator.Modified = DateTime.Now;

            // insert the object to the DB
            Context.ExternalSystemIntegrators.InsertOnSubmit(theIntegrator);
            Context.SubmitChanges();

            return theIntegrator;
        }

        /// <summary>
        /// Reads the data about the ExternalSystem from the configuration,
        /// and ensures the DB has the ExternalSystem registered properly (existing and up-to-date)
        /// </summary>
        /// <param name="integratorName">The Name(unique identifier) of the component, as specified by the configuration</param>
        /// <returns>Initialized ExternalSystem LINQ object</returns>
        public ExternalSystem DeployOrInitializeExternalSystem(String integratorName)
        {
            ExternalSystem TheExternalSystem = null;

            // if the Integrator does not set the name of the external system that is provides integration to
            if (!ConfigManager.PropertyValueExists(integratorName, ExternalSystemIntegrator.ConfigProperty_ExternalSystemName))
            {
                TraceLog.Write(LogEventType.Error, "The integrator '{0}' does not specify the property '{1}' in its configuration. This property is required for the proper deployment of the component.", new object[] { integratorName, ExternalSystemIntegrator.ConfigProperty_ExternalSystemName});
                throw new ArgumentException("The integrator '" + integratorName + "' does not specify the property '" + ExternalSystemIntegrator.ConfigProperty_ExternalSystemName + "' in its configuration. This property is required for the proper deployment of the component.");
            }
            String extSystemName = ConfigManager.Modules[integratorName].Properties[ExternalSystemIntegrator.ConfigProperty_ExternalSystemName].Value;

            // verify that the ExternalSystem (identified by its name) 
            // is registered in the DB. If not, insert it to the DB.
            TheExternalSystem = Context.ExternalSystems.FirstOrDefault<ExternalSystem>(externalSystem => externalSystem.Name == extSystemName);

            // if the ExternalSystem is not in the DB yet
            if (TheExternalSystem == null)
            {
	            TraceLog.Write(LogEventType.Information, "The ExternalSystem '{0}' is not registered in the DB yet.", new object[] { extSystemName });

                // register the ExternalSystem in the DB
                TheExternalSystem = CreateExternalSystemFromConfigData(extSystemName);
            }
            else
            {
                SortedDictionary<String, ConfigProperty> properties = ConfigManager.Modules[extSystemName].Properties;
                // if the data in the configuration differs from the DB, update the DB
                if (TheExternalSystem.UpdateObjectIfNeeded(properties[ExternalSystem.ConfigProperty_UrlToLogo].Value, properties[ExternalSystem.ConfigProperty_UrlToStartPage].Value))
                {
                    Context.SubmitChanges();
                }
            }
            return TheExternalSystem;
        }


        /// <summary>
        /// Registers the ExternalSystem in the DB(assuming that the ExternalSystem with that Name does not exist) by using data from the configuration 
        /// </summary>
        /// <param name="externalSystemName">The unique name(identifier) of the ExternalSystem</param>
        /// <returns>Initialized Linq ExternalSystem object</returns>
        private ExternalSystem CreateExternalSystemFromConfigData(String externalSystemName)
        {
            ExternalSystem TheExternalSystem = new ExternalSystem();
            TheExternalSystem.Name = ConfigManager.Modules[externalSystemName].Properties[ExternalSystem.ConfigProperty_Name].Value;
            TheExternalSystem.UrlToLogo = ConfigManager.Modules[externalSystemName].Properties[ExternalSystem.ConfigProperty_UrlToLogo].Value;
            TheExternalSystem.UrlToStartPage = ConfigManager.Modules[externalSystemName].Properties[ExternalSystem.ConfigProperty_UrlToStartPage].Value;
            TheExternalSystem.Created = DateTime.Now;
            TheExternalSystem.Modified = DateTime.Now;
            Context.ExternalSystems.InsertOnSubmit(TheExternalSystem);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information, "New ExternalSystem('{0}', with ID = '{1}') has been created(registered in the DB) out of the Config data.", new object[] { TheExternalSystem.Name, TheExternalSystem.ExternalSystemId });

            return TheExternalSystem;
        }

        #region IDisposable Members

            /// <summary>
            /// Releases unmanaged and - optionally - managed resources
            /// </summary>
            /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
            protected virtual void Dispose(bool disposing)
            {
                this.Context.Connection.Dispose();
                this.Context.Dispose();
            }

            /// <summary>
            /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
            /// </summary>
            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

        #endregion

    }
}
