﻿using System;
using System.Collections.Generic;
using System.Linq;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;

namespace ExternalSystemIntegration
{
    /// <summary>
    /// Provides methods to register and keep up to date the information about
    /// users' memberships in external groups.
    /// </summary>
    public class ExternalGroupsManagement
    {
        // db context
        private MyMediaDataContext Context;

        /// <summary>
        /// Constructor requiring DB context
        /// </summary>
        /// <param name="context"></param>
        public ExternalGroupsManagement(MyMediaDataContext context)
        {
            this.Context = context;
        }

        /// <summary>
        /// Structure to hold external group Identifier and Name pairs
        /// </summary>
        public struct ExternalGroupIdNamePair
        {
            /// <summary>
            /// Gets or sets the name.
            /// </summary>
            /// <value>The name.</value>
            public String Name { get; set; }
            /// <summary>
            /// Gets or sets the external identifier.
            /// </summary>
            /// <value>The external identifier.</value>
            public String ExternalIdentifier { get; set; }
        }


        /// <summary>
        /// Ensures that the DB data about user's external group memberships is 
        /// up to date (holds all external groups, and notices cancelled memberships)
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="externalGroupIdNamePair"></param>
        public void EnsureUserExternalGroupMembershipsAreUpToDate(ExternalSystemUser externalUser, ICollection<ExternalGroupIdNamePair> externalGroupIdNamePair)
        {
            HashSet<int> CurrentActiveExternalGroupIds = new HashSet<int>();
            // ensure external groups are registered in the DB
            foreach (ExternalGroupIdNamePair groupData in externalGroupIdNamePair)
            {
                ExternalGroup externalGroup = RegisterExternalGroup(externalUser.ExternalSystemId, groupData.ExternalIdentifier, groupData.Name);
                externalGroup = RegisterUserExternalGroupMembership(externalUser, externalGroup);
                CurrentActiveExternalGroupIds.Add(externalGroup.ExternalGroupId);
            }

            if (externalUser.ExternalGroupMemberships.Count == CurrentActiveExternalGroupIds.Count)
            {
                return;
            }

            bool isChangesSubmitRequired = false;
            // now eliminate cancelled group memberships: list DB memberships & make sure they active
            foreach (ExternalGroupMembership membership in externalUser.ExternalGroupMemberships)
            {
                if (!membership.IsCancelled && !CurrentActiveExternalGroupIds.Contains(membership.ExternalGroupId))
                {
                    membership.IsCancelled = true;
                    isChangesSubmitRequired = true;
                }
            }

            if (isChangesSubmitRequired)
            {
                Context.SubmitChanges();
            }
        }

        /// <summary>
        /// Makes sure DB has correct data about user's membership in some external group
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="externalGroup"></param>
        /// <returns></returns>
        public ExternalGroup RegisterUserExternalGroupMembership(ExternalSystemUser externalUser, ExternalGroup externalGroup)
        {
            ExternalGroupMembership currentGroupMembership = Context.ExternalGroupMemberships.FirstOrDefault<ExternalGroupMembership>(item => item.ExternalSystemUserId == externalUser.ExternalSystemUserId && item.ExternalGroupId == externalGroup.ExternalGroupId);
            if (currentGroupMembership != null)
            {
                if (currentGroupMembership.IsCancelled)
                {
                    currentGroupMembership.IsCancelled = false;
                    Context.SubmitChanges();
                }
                return externalGroup;
            }

            // register the membership
            ExternalGroupMembership newMembership = new ExternalGroupMembership();
            newMembership.ExternalGroupId = externalGroup.ExternalGroupId;
            newMembership.ExternalSystemUserId = externalUser.ExternalSystemUserId;
            newMembership.IsCancelled = false;
            newMembership.Created = DateTime.Now;
            newMembership.Modified = DateTime.Now;
            Context.ExternalGroupMemberships.InsertOnSubmit(newMembership);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information,
                "User's external group membership registered: groupId='" + externalGroup.ExternalGroupId
                + "', ExternalSystemUserId='" + externalUser.ExternalSystemUserId + "'.");

            return externalGroup;
        }

        /// <summary>
        /// Makes sure user's group membership is registered in the DB
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="groupExternalIdentifier"></param>
        /// <param name="groupName"></param>
        /// <returns></returns>
        public ExternalGroup RegisterUserExternalGroupMembership(ExternalSystemUser externalUser, String groupExternalIdentifier, String groupName)
        {
            ExternalGroup externalGroup = RegisterExternalGroup(externalUser.ExternalSystemId, groupExternalIdentifier, groupName);

            return RegisterUserExternalGroupMembership(externalUser, externalGroup);
        }

        /// <summary>
        /// Alias with empty name
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="groupExternalIdentifier"></param>
        /// <returns></returns>
        public ExternalGroup RegisterUserExternalGroupMembership(ExternalSystemUser externalUser, String groupExternalIdentifier)
        {
            return RegisterUserExternalGroupMembership(externalUser, groupExternalIdentifier, null);
        }


        /// <summary>
        /// Ensures that external group is registered in the DB
        /// </summary>
        /// <param name="externalSystemId"></param>
        /// <param name="groupExternalIdentifier"></param>
        /// <returns></returns>
        public ExternalGroup RegisterExternalGroup(int externalSystemId, String groupExternalIdentifier)
        {
            return RegisterExternalGroup(externalSystemId, groupExternalIdentifier, null);
        }

        /// <summary>
        /// Ensures that external group is registered in the DB
        /// </summary>
        /// <param name="externalSystemId"></param>
        /// <param name="groupExternalIdentifier"></param>
        /// <param name="groupName">optional, can be empty</param>
        /// <returns>Initialized (existing or just created ExternalGroup object)</returns>
        public ExternalGroup RegisterExternalGroup(int externalSystemId, String groupExternalIdentifier, String groupName)
        {
            ExternalGroup externalGroup = Context.ExternalGroups.FirstOrDefault<ExternalGroup>
                    (g => g.ExternalSystemId == externalSystemId && g.GroupExternalIdentifier == groupExternalIdentifier);
            if (externalGroup != null)
            {
                return externalGroup;
            }

            externalGroup = new ExternalGroup();
            externalGroup.ExternalSystemId = externalSystemId;
            externalGroup.GroupExternalIdentifier = groupExternalIdentifier;
            externalGroup.Name = groupName;

            externalGroup.Created = DateTime.Now;
            externalGroup.Modified = DateTime.Now;

            Context.ExternalGroups.InsertOnSubmit(externalGroup);
            Context.SubmitChanges();

            TraceLog.Write(LogEventType.Information, "New ExternalGroup registered in the DB: (" + externalGroup.ToString() + " )");
            return externalGroup;
        }
    }
}
