﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ExternalSystemIntegration;
using MyMediaProject.RecommenderSystem.Core;

namespace ExternalSystemIntegration
{
    /// <summary>
    /// This class encapsulates the result that the operations of external social network 
    /// integration components can return. The result might be void, a boolean value,
    /// more comlicated data structure, or even user's WEB browser's HTTP redirect rquest.
    /// </summary>
    public class ExternalSystemIntegratorResult
    {
        /* possible types of the result object */

        /// <summary>void result type</summary>
        public const int ReturnTypeCodeVoid = 1;
        
        /// <summary>boolean result type</summary>
        public const int ReturnTypeCodeBool = 2;
        
        /// <summary>WebBrowserRedirectRequest result type</summary>
        public const int ReturnTypeCodeRedirectRequest = 3;
        
        /// <summary>string result type</summary>
        public const int ReturnTypeCodeString = 4;

        /// <summary>ExternalSystemUser result type</summary>
        public const int ReturnTypeCodeExtUser = 5;

        /// <summary>List(ExternalSystemUser) result type</summary>
        public const int ReturnTypeCodeListOfExtUsers = 6;

        /// <summary>Dictionary(String, String) result type</summary>
        public const int ReturnTypeCodeStringDictionary = 7;
        
        /// <summary>any other result type that does not fall into one of other specified types</summary>
        public const int ReturnTypeCodeOtherObject = 10;

        // holds the value of RETURN_TYPE_CODE_x
        int ResultTypeCode { get; set; }

        Object Result { get; set; } // holds the result object of one of the types mentioned above

        /**********************/
        /*    Constructors    */
        /**********************/

        /// <summary>Constructor for the result type: void</summary>
        public ExternalSystemIntegratorResult()
        {
            ResultTypeCode = ReturnTypeCodeVoid;
        }

        /// <summary>Constructor for the result type: boolean</summary>
        public ExternalSystemIntegratorResult(bool result)
        {
            ResultTypeCode = ReturnTypeCodeBool;
            Result = result;
        }

        /// <summary>Constructor for the result type: WebBrowserRedirectRequest</summary>
        public ExternalSystemIntegratorResult(WebBrowserRedirectRequest result)
        {
            ResultTypeCode = ReturnTypeCodeRedirectRequest;
            Result = result;
        }

        /// <summary>Constructor for the result type: string</summary>
        public ExternalSystemIntegratorResult(string result)
        {
            ResultTypeCode = ReturnTypeCodeString;
            Result = result;
        }

        /// <summary>Constructor for the result type: ExternalSystemUser</summary>
        public ExternalSystemIntegratorResult(ExternalSystemUser result)
        {
            ResultTypeCode = ReturnTypeCodeExtUser;
            Result = result;
        }

        /// <summary>Constructor for the result type: List&lt;ExternalSystemUser&gt;</summary>
        public ExternalSystemIntegratorResult(List<ExternalSystemUser> result)
        {
            ResultTypeCode = ReturnTypeCodeListOfExtUsers;
            Result = result;
        }

        /// <summary>Constructor for the result type: Dictionary&lt;String, String&gt;</summary>
        public ExternalSystemIntegratorResult(Dictionary<String, String> result)
        {
            ResultTypeCode = ReturnTypeCodeStringDictionary;
            Result = result;
        }

        /// <summary>Constructor for the result type: Object (other than one of types defined above)</summary>
        public ExternalSystemIntegratorResult(Object result)
        {
            ResultTypeCode = ReturnTypeCodeOtherObject;
            Result = result;
        }


        /************************************/
        /*    Type determination methods    */
        /************************************/

        /// <summary>Returns true if the result type is void</summary>
        public bool IsVoidResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeVoid)
                return true;
            else
                return false;
        }
        /// <summary>Returns true if the result type is bool</summary>
        public bool IsBoolResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeBool)
                return true;
            else
                return false;
        }
        /// <summary>Returns true if the result type is string</summary>
        public bool IsStringResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeString)
                return true;
            else
                return false;
        }
        /// <summary>Returns true if the result type is WebBrowserRedirectRequest</summary>
        public bool IsRedirectRequestResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeRedirectRequest)
                return true;
            else
                return false;
        }
        /// <summary>Returns true if the result type is an external user</summary>
        public bool IsExternalUserResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeExtUser)
                return true;
            else
                return false;
        }
        /// <summary>Returns true if the result type is List of external users </summary>
        public bool IsListOfUsersResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeListOfExtUsers)
                return true;
            else
                return false;
        }
        /// <summary>Returns true if the result type is of type Dictionary(String, String)</summary>
        public bool IsStringDictionaryResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeStringDictionary)
                return true;
            else
                return false;
        }

        /// <summary>Returns true if the result type is an object that is not of one of the basic types supported</summary>
        public bool IsOtherObjectResultType()
        {
            if (ResultTypeCode == ReturnTypeCodeOtherObject)
                return true;
            else
                return false;
        }

        /**********************************/
        /*    Result retrieval methods    */
        /**********************************/

        /// <summary>Dummy implementation of void type result retrieval</summary>
        public void GetVoidResult()
        {
            if (ResultTypeCode != ReturnTypeCodeVoid)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return;
        }

        /// <summary>This method should be used, if boolean value is the expected result only.</summary>
        public bool GetBoolResult()
        {
            if (ResultTypeCode != ReturnTypeCodeBool)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return (bool)Result;
        }

        /// <summary>This method should be used, if string value is the expected result only.</summary>
        public string GetStringResult()
        {
            if (ResultTypeCode != ReturnTypeCodeString)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return (string)Result;
        }

        /// <summary>This method should be used, if WebBrowserRedirectRequest is the expected result only.</summary>
        public WebBrowserRedirectRequest GetRedirectRequestResult()
        {
            if (ResultTypeCode != ReturnTypeCodeRedirectRequest)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return (WebBrowserRedirectRequest)Result;
        }

        /// <summary>This method should be used, if an external user is the expected result only.</summary>
        public ExternalSystemUser GetExternalUserResult()
        {
            if (ResultTypeCode != ReturnTypeCodeExtUser)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return (ExternalSystemUser)Result;
        }

        /// <summary>This method should be used, if List of external users is the expected result only.</summary>
        public List<ExternalSystemUser> GetListOfExternalUsersResult()
        {
            if (ResultTypeCode != ReturnTypeCodeListOfExtUsers)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return (List<ExternalSystemUser>)Result;
        }

        /// <summary>This method should be used, if Dictionary(String, String) is the expected result only.</summary>
        public Dictionary<String, String> GetStringDictionaryResult()
        {
            if (ResultTypeCode != ReturnTypeCodeStringDictionary)
            {
                throw new ExternalSystemIntegratorException("Wrong return type requested.");
            }
            return (Dictionary<String, String>)Result;
        }


        /// <summary>This method should be used, if the return type is not one of the defined ones</summary>
        public Object GetObjectResult()
        {
            return (Object)Result;
        }
    }
}
