﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;

namespace ExternalSystemIntegration
{
    /// <summary>
    /// Provider methods to register and handle recommendations 
    /// </summary>
    public class RecommendationsManager : IDisposable
    {
        private MyMediaDataContext Context;
        private bool ownsDataContext;

        /// <summary>
        /// Constructor requiring initialized data context
        /// </summary>
        /// <param name="context"></param>
        public RecommendationsManager(MyMediaDataContext context)
        {
            if (context != null)
            {
                this.Context = context;
                ownsDataContext = true;
            }
            else
            {
                String dbConnectionString = ConfigManager.DBConnectionString;
                TraceLog.Write(LogEventType.Information, "RecommendationsManager() : using dbConnectionString='{0}'.", new object[] { dbConnectionString });
                this.Context = new MyMediaDataContext(dbConnectionString);
            }
        }

        /// <summary>
        /// Register the recommendation made to a local user
        /// </summary>
        /// <param name="catalogItemId">either of catalogItemId and virtualProfileId parameters must be specified</param>
        /// <param name="virtualProfileId">either of catalogItemId and virtualProfileId parameters must be specified</param>
        /// <param name="fromUser"></param>
        /// <param name="toUser"></param>
        /// <returns></returns>
        public ContentRecommendation RegisterNewRecommendation(int catalogItemId, int virtualProfileId, User fromUser, User toUser)
        {
            return InsertNewRecommendation(catalogItemId, virtualProfileId, fromUser, toUser, null);
        }


        /// <summary>
        /// Register the recommendation made to an external user
        /// </summary>
        /// <param name="catalogItemId">either of catalogItemId and virtualProfileId parameters must be specified</param>
        /// <param name="virtualProfileId">either of catalogItemId and virtualProfileId parameters must be specified</param>
        /// <param name="fromUser"></param>
        /// <param name="toExternalUser"></param>
        /// <returns></returns>
        public ContentRecommendation RegisterNewRecommendation(int catalogItemId, int virtualProfileId, User fromUser, ExternalSystemUser toExternalUser)
        {
            return InsertNewRecommendation(catalogItemId, virtualProfileId, fromUser, null, toExternalUser);
        }

        /// <summary>
        /// Retrives all active virtual profiles' recommendations made to the specified local user
        /// </summary>
        /// <param name="localUser"></param>
        /// <returns></returns>
        public IEnumerable<VirtualProfileUsageInvitation> GetActiveVirtualProfileInvitations(User localUser)
        {
            IEnumerable<VirtualProfileUsageInvitation> invitations =
                from invitation in Context.ContentRecommendations
                where invitation is ContentRecommendation &&
                invitation is VirtualProfileUsageInvitation &&
                invitation.IsActive == true && invitation.ToUserId == localUser.UserId
                select invitation as VirtualProfileUsageInvitation;

            return invitations;
        }

        /// <summary>
        /// Retrives all active catalog items recommendations made to the specified local user
        /// </summary>
        /// <param name="localUser"></param>
        /// <returns></returns>
        public IEnumerable<CatalogItemRecommendation> GetActiveCatalogItemRecommendations(User localUser)
        {
            IEnumerable<CatalogItemRecommendation> recommendations =
                from recommendation in Context.ContentRecommendations
                where recommendation is ContentRecommendation &&
                recommendation.IsActive == true && recommendation.ToUserId == localUser.UserId
                select recommendation as CatalogItemRecommendation;

            return recommendations;
        }

        /// <summary>
        /// Retirves all virtual profiles' recommendations made to the specified local user
        /// </summary>
        /// <param name="externalUser"></param>
        /// <returns></returns>
        public IEnumerable<ContentRecommendation> GetActiveVirtualProfileInvitations(ExternalSystemUser externalUser)
        {
            IEnumerable<ContentRecommendation> invitations =
                from invitation in Context.ContentRecommendations
                where invitation is ContentRecommendation &&
                        invitation.IsActive == true && invitation.ToExternalSystemUserId == externalUser.ExternalSystemUserId
                select invitation as ContentRecommendation;

            return invitations;
        }

        /// <summary>
        /// Marks recommendation as read(not active).
        /// </summary>
        /// <param name="contentRecommendationId"></param>
        /// <returns>true if the recommendation has been marked read, false if the recommendation did not exist, or had been inactive already</returns>
        public bool MarkRecommendationAsRead(int contentRecommendationId)
        {
            ContentRecommendation recommendation = GetRecommendationById(contentRecommendationId);
            if (recommendation == null || !recommendation.IsActive)
            {
                return false;
            }
            recommendation.IsActive = false;
            Context.SubmitChanges();
            return true;
        }

        /// <summary>
        /// Marks recommendation as accepted(and inactive).
        /// </summary>
        /// <param name="contentRecommendationId"></param>
        /// <returns>true on success, false if the recommendation did not exist, or had been inactive already</returns>
        public bool MarkRecommendationAsAccepted(int contentRecommendationId)
        {
            ContentRecommendation recommendation = GetRecommendationById(contentRecommendationId);
            if (recommendation == null || !recommendation.IsActive)
            {
                return false;
            }
            recommendation.IsActive = false;
            recommendation.IsAccepted = true;
            Context.SubmitChanges();
            return true;
        }

        /// <summary>
        /// Marks recommendation as rejected(and inactive).
        /// </summary>
        /// <param name="contentRecommendationId"></param>
        /// <returns>true on success, false if the recommendation did not exist, or had been inactive already</returns>
        public bool MarkRecommendationAsRejected(int contentRecommendationId)
        {
            ContentRecommendation recommendation = GetRecommendationById(contentRecommendationId);
            if (recommendation == null || !recommendation.IsActive)
            {
                return false;
            }
            recommendation.IsActive = false;
            recommendation.IsAccepted = false;
            Context.SubmitChanges();
            return true;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="catalogItemId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromLocalUser"></param>
        /// <param name="toLocalUser"></param>
        /// <param name="toExternalUser"></param>
        /// <returns></returns>
        private ContentRecommendation InsertNewRecommendation(int catalogItemId, int virtualProfileId, User fromLocalUser, User toLocalUser, ExternalSystemUser toExternalUser)
        {
            ContentRecommendation recommendation;

            int fromLocalUserId = 0;
            int toLocalUserId = 0;
            int toExternalUserId = 0;
            if (fromLocalUser != null){
                fromLocalUserId = fromLocalUser.UserId;
            }
            if (toLocalUser != null){
                toLocalUserId = toLocalUser.UserId;
            }
            if (toExternalUser != null){
                toExternalUserId = toExternalUser.ExternalSystemUserId;
            }

            // if active invitation with these params exists, don't invite again
            if ((recommendation = (CatalogItemRecommendation)GetRecommendationByParameters(catalogItemId, virtualProfileId, fromLocalUserId, toLocalUserId, toExternalUserId, true)) != null)
            {
                return recommendation;
            }

            // if catalog item is being recommended
            if (catalogItemId != default(int))
            {
                recommendation = new CatalogItemRecommendation(catalogItemId, fromLocalUser, toLocalUser);
                Context.ContentRecommendations.InsertOnSubmit(recommendation);
                Context.SubmitChanges();
                return recommendation;
            }
            else
            {
                if (toLocalUser != null)
                {
                    recommendation = new VirtualProfileUsageInvitation(virtualProfileId, fromLocalUser, toLocalUser);
                }
                else if (toExternalUser != null)
                {
                    recommendation = new VirtualProfileUsageInvitation(virtualProfileId, fromLocalUser, toExternalUser);
                }
                Context.ContentRecommendations.InsertOnSubmit(recommendation);
                Context.SubmitChanges();
                return recommendation;
            }
        }


        /// <summary>
        /// Returns existing recommendation according to the parameters specified
        /// </summary>
        /// <param name="catalogItemId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromLocalUserId"></param>
        /// <param name="toLocalUserId"></param>
        /// <param name="toExternalUserId"></param>
        /// <param name="isActive"></param>
        /// <returns></returns>
        public ContentRecommendation GetRecommendationByParameters(int catalogItemId, int virtualProfileId, int fromLocalUserId, int toLocalUserId, int toExternalUserId, bool isActive)
        {
            ContentRecommendation recommendation = null;
            if (catalogItemId != default(int))
            {
                recommendation = Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                                    (i => i is ContentRecommendation
                                        && i is CatalogItemRecommendation
                                        && (i as CatalogItemRecommendation).CatalogItemId == catalogItemId
                                        && i.FromUserId == fromLocalUserId
                                        && i.IsActive == isActive
                                        && (i.ToUserId == toLocalUserId || i.ToExternalSystemUserId == toExternalUserId)
                                     )  as CatalogItemRecommendation;
                return (CatalogItemRecommendation)recommendation;
            }

            recommendation = Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                                (i => i is ContentRecommendation
                                    && i is VirtualProfileUsageInvitation
                                    && (i as VirtualProfileUsageInvitation).VirtualProfileId == virtualProfileId
                                    && i.FromUserId == fromLocalUserId
                                    && i.IsActive == isActive
                                    && (i.ToUserId == toLocalUserId || i.ToExternalSystemUserId == toExternalUserId)
                                 ) as VirtualProfileUsageInvitation;
            return (VirtualProfileUsageInvitation)recommendation;
        }

        /// <summary>
        /// Returns existing recommendation according to the parameters specified
        /// </summary>
        /// <param name="catalogItemId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromLocalUserId"></param>
        /// <param name="toLocalUserId"></param>
        /// <param name="toExternalUserId"></param>
        /// <returns></returns>
        public ContentRecommendation GetRecommendationByParameters(int catalogItemId, int virtualProfileId, int fromLocalUserId, int toLocalUserId, int toExternalUserId)
        {
            ContentRecommendation recommendation = null;
            if (catalogItemId == default(int))
            {
                recommendation = Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                                    (i => i is ContentRecommendation
                                        && i is CatalogItemRecommendation
                                        && (i as CatalogItemRecommendation).CatalogItemId == catalogItemId
                                        && i.FromUserId == fromLocalUserId
                                        && (i.ToUserId == toLocalUserId || i.ToExternalSystemUserId == toExternalUserId)
                                     ) as CatalogItemRecommendation;
                return (CatalogItemRecommendation)recommendation;
            }

            recommendation = Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                                (i => i is ContentRecommendation
                                    && i is VirtualProfileUsageInvitation
                                    && (i as VirtualProfileUsageInvitation).VirtualProfileId == virtualProfileId
                                    && i.FromUserId == fromLocalUserId
                                    && (i.ToUserId == toLocalUserId || i.ToExternalSystemUserId == toExternalUserId)
                                 ) as VirtualProfileUsageInvitation;
            return (VirtualProfileUsageInvitation)recommendation;
        }

        /// <summary>
        /// Retrieves recommendation by its Id
        /// </summary>
        /// <param name="contentRecommendationId"></param>
        /// <returns></returns>
        public ContentRecommendation GetRecommendationById(int contentRecommendationId)
        {
            ContentRecommendation recommendation = Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                  (i => i.ContentRecommendationId == contentRecommendationId && i is ContentRecommendation);
            return recommendation;
        }

        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.ownsDataContext && disposing)
            {
                this.Context.Connection.Dispose();
                this.Context.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion
    }
}
