﻿using System;
using System.Collections.Generic;
using System.Data.Linq;
using System.Linq;
using System.Reflection;
using System.Text;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;

namespace ExternalSystemIntegration
{
    /// <summary>
    /// Class providing methods to handle external and local users 
    /// </summary>
    public class UsersManager : IDisposable
    {
        private MyMediaDataContext Context;
        private bool ownsDataContext;

        /// <summary>
        /// The length of the 'ExternalSystemUserProperty.Name' DB field.
        /// </summary>
        public const int ExternalSystemUserPropertyNameMaximalLength = 60;

        /// <summary>
        /// Constructor initializing a data context
        /// </summary>
        public UsersManager()
        {
            String dbConnectionString = ConfigManager.DBConnectionString;
            TraceLog.Write(LogEventType.Information, "UsersManager() : using dbConnectionString='{0}'.", new object[] { dbConnectionString });
            this.Context = new MyMediaDataContext(dbConnectionString);
            ownsDataContext = true;
        }

        /// <summary>
        /// Constructor that uses given DataContext. This DataContext will not be disposed by this class.
        /// </summary>
        /// <param name="context">Initialized data context that this class will use</param>
        public UsersManager(MyMediaDataContext context)
        {
            this.Context = context;
        }

        /// <summary>
        /// Retrieves registered external system user by external system identity
        /// </summary>
        /// <param name="externalSystemId"></param>
        /// <param name="externalSystemIdentity"></param>
        /// <returns>ExternalSystemUser object or null</returns>
        public ExternalSystemUser GetExternalSystemUserByExternalIdentity(int externalSystemId, String externalSystemIdentity)
        {
            ExternalSystemUser extUser = Context.ExternalSystemUsers.FirstOrDefault<ExternalSystemUser>(u => u.ExternalSystemId == externalSystemId && u.UserExternalIdentifier == externalSystemIdentity);
            return extUser;
        }

        /// <summary>
        /// Retrieves registered external system user by MyMedia's id
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <returns>ExternalSystemUser object or null</returns>
        public ExternalSystemUser GetExternalSystemUserById(int externalSystemUserId)
        {
            ExternalSystemUser extUser = null;
            extUser = Context.ExternalSystemUsers.FirstOrDefault<ExternalSystemUser>(u => u.ExternalSystemUserId == externalSystemUserId);
            
            if (extUser != null)
            {
                TraceLog.Write(LogEventType.Verbose, "GetExternalSystemUserById('{0}') = '{1}'. User:'{2}'.{3}", new object[] { externalSystemUserId, extUser.ToString(), extUser.User, extUser.UserVPAccessRights });
            }
            else
            {
                TraceLog.Write(LogEventType.Verbose, "GetExternalSystemUserById('{0}'): user not found.", new object[]{externalSystemUserId});
            }

            return extUser;
        }


        /// <summary>
        /// Retrieves registered local user by MyMedia's id
        /// </summary>
        /// <param name="userId"></param>
        /// <returns>User object or null</returns>
        public User GetLocalUserById(int userId)
        {
            User user = null;
            user = Context.Users.FirstOrDefault<User>(u => u.UserId == userId);

            return user;
        }

        /// <summary>
        /// Retrieves external systems' accounts associated with the specified local user
        /// </summary>
        /// <param name="userId"></param>
        /// <returns></returns>
        public ICollection<ExternalSystemUser> GetLocalUsersExternalAccounts(int userId)
        {
            User user = GetLocalUserById(userId);
            return user.ExternalSystemUsers.ToList();
        }

        /// <summary>
        /// Associates existing external and local users
        /// </summary>
        /// <param name="externalSystemUserId">Id of an existing external system user</param>
        /// <param name="userId">Id of an existing local user</param>
        /// <returns></returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(int externalSystemUserId, int userId)
        {
            User localUser = GetLocalUserById(userId);
            ExternalSystemUser extUser = GetExternalSystemUserById(externalSystemUserId);

            return AssociateExternalAndLocalUser(extUser, localUser, extUser.ExternalSystemId, default(int));
        }

        /// <summary>
        /// Associates existing external and local users
        /// </summary>
        /// <param name="extUser"></param>
        /// <param name="localUser"></param>
        /// <returns></returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(ExternalSystemUser extUser, User localUser)
        {
            return AssociateExternalAndLocalUser(extUser, localUser, extUser.ExternalSystemId, default(int));
        }

        /// <summary>
        /// Associates existing external and local users
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="userId"></param>
        /// <param name="moduleId"></param>
        /// <returns></returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(int externalSystemUserId, int userId, int moduleId)
        {
            User localUser = GetLocalUserById(userId);
            ExternalSystemUser extUser = GetExternalSystemUserById(externalSystemUserId);

            return AssociateExternalAndLocalUser(extUser, localUser, extUser.ExternalSystemId, moduleId);
        }


        /// <summary>
        /// Initializes external user referencing the local one
        /// Creates both an external user, and a local one, if not registered in the DB yet.
        /// </summary>
        /// <param name="externallyAuthenticatedUser"></param>
        /// <param name="externalSystemId"></param>
        /// <param name="moduleId">Existing module Id</param>
        /// <returns>Initialized external system user, referencing initialized local user</returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(ExternalSystemUser externallyAuthenticatedUser, int externalSystemId, int moduleId)
        {
            return AssociateExternalAndLocalUser(externallyAuthenticatedUser, null, externalSystemId, moduleId);
        }


        /// <summary>
        /// Initializes external user referencing the local one
        /// Creates both an external user, and a local one, if not registered in the DB yet.
        /// </summary>
        /// <param name="externallyAuthenticatedUser"></param>
        /// <param name="localUser">Local user to associate the external user to(provide null if does not exist)</param>
        /// <param name="externalSystemId"></param>
        /// <param name="moduleId">Existing Module id. Will only be used if local user is to be created(inserted into the DB)</param>
        /// <returns>Initialized external system user, referencing initialized local user</returns>
        public ExternalSystemUser AssociateExternalAndLocalUser(ExternalSystemUser externallyAuthenticatedUser, User localUser, int externalSystemId, int moduleId)
        {
            // resulting object that will hold reference to associated local user
            ExternalSystemUser initializedExtUser = externallyAuthenticatedUser;

            if (externallyAuthenticatedUser.ExternalSystemUserId == default(int))
            {
                // get the user with this external ID from the DB (if present)
                initializedExtUser = Context.ExternalSystemUsers.FirstOrDefault<ExternalSystemUser>
                                    (u => u.ExternalSystemId == externalSystemId
                                        && u.UserExternalIdentifier == externallyAuthenticatedUser.UserExternalIdentifier);
                if (initializedExtUser == null)
                {
                    initializedExtUser = externallyAuthenticatedUser;
                    initializedExtUser.Created = DateTime.Now;
                }
            }

            // if local user is associated with the external one
            if (initializedExtUser.User != null)
            {
                if (localUser != null && localUser.UserId != initializedExtUser.User.UserId)
                {
                    TraceLog.Write(LogEventType.Warning, "The external user account (" + initializedExtUser.ToString() + ") is associated with another local user already (Can't associate with local user localUserId='" + localUser.UserId + "').");
                    throw new InvalidOperationException("The external user account is associated with a local user already.");
                }
                return initializedExtUser;
            }

            // if local user does not exist yet
            if (localUser == null)
            {
                localUser = CreateNewUser(null, null, moduleId);
            }

            TraceLog.Write(LogEventType.Information, "New Local user(UserId='{0}') has been created.", new object[] { localUser.UserId });
            initializedExtUser.User = localUser;
            initializedExtUser.Modified = DateTime.Now;
            Context.SubmitChanges();

            return CreateExternalUserIfMissing(initializedExtUser);
        }


        /// <summary>
        /// Creates new local user
        /// </summary>
        /// <param name="firstName"></param>
        /// <param name="lastName"></param>
        /// <param name="moduleId"></param>
        /// <returns>Newly created local user instance</returns>
        public User CreateNewUser(String firstName, String lastName, int moduleId)
        {
            User localUser = new User();

            localUser.FirstName = firstName;
            localUser.LastName = lastName;

            localUser.ModuleId = moduleId;
            localUser.IsDisabled = false;

            localUser.Created = DateTime.Now;
            localUser.Modified = DateTime.Now;
            localUser.LastLogOn = null;

            Context.Users.InsertOnSubmit(localUser);
            Context.SubmitChanges();

            return localUser;
        }

        /// <summary>
        /// Removes the association between the local and external system user.
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="userId"></param>
        /// <returns></returns>
        public bool DissociateExternalAndLocalUser(int externalSystemUserId, int userId)
        {
            User localUser = GetLocalUserById(userId);
            ExternalSystemUser extUser = GetExternalSystemUserById(externalSystemUserId);

            return DissociateExternalAndLocalUser(extUser, localUser);
        }

        /// <summary>
        /// Removes the association between the local and external system user.
        /// </summary>
        /// <param name="extUser">initalized ExternalSystemUser object</param>
        /// <param name="localUser">initialized MyMedia User object</param>
        /// <returns>true if association has been deleted, false if the association did not exist.</returns>
        public bool DissociateExternalAndLocalUser(ExternalSystemUser extUser, User localUser)
        {
            if (extUser.LocalUserId == localUser.UserId)
            {
                extUser.User = null;
                extUser.Modified = DateTime.Now;
                Context.SubmitChanges();
                return true;
            }
            return false;
        }

        /// <summary>
        /// Creates an external user out of a fully initialized object, if is not present in the DB yet
        /// </summary>
        /// <returns>Initialized external system user, referencing initialized local user</returns>
        public ExternalSystemUser CreateExternalUserIfMissing(ExternalSystemUser extUser)
        {
            // register an external user, if not registered in the DB yet
            if (extUser.ExternalSystemUserId == default(int))
            {
                extUser.Created = DateTime.Now;
                extUser.Modified = DateTime.Now;

                Context.ExternalSystemUsers.InsertOnSubmit(extUser);
                Context.SubmitChanges();
                if (extUser.User == null)
                {
                    TraceLog.Write(LogEventType.Information, "New ExternalSystemUser created(ExternalSystemUserId='{2}', ExternalSystemId='{0}', UserExternalIdentifier='{1}')", new object[] { extUser.ExternalSystemId, extUser.UserExternalIdentifier, extUser.ExternalSystemUserId});
                }
                else
                {
                    TraceLog.Write(LogEventType.Information, "New ExternalSystemUser created(ExternalSystemUserId='{2}', ExternalSystemId='{0}', UserExternalIdentifier='{1}', localUserId='{3}')", new object[] { extUser.ExternalSystemId, extUser.UserExternalIdentifier, extUser.ExternalSystemUserId, extUser.User.UserId });
                }
            }

            return extUser;
        }

        /// <summary>
        /// Stores(inserts or updates, if existed already) given property for the specified external system user
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="propertyName"></param>
        /// <param name="propertyValue"></param>
        /// <returns>resulting ExternalSystemUserProperty object</returns>
        public ExternalSystemUserProperty StoreExternalSystemUserProperty(int externalSystemUserId, String propertyName, String propertyValue)
        {
            if (propertyName == null || propertyName.Length == 0 || propertyName.Length > ExternalSystemUserPropertyNameMaximalLength)
            {
                throw new ExternalSystemIntegratorException("The length of the property name must be between 1 and " + ExternalSystemUserPropertyNameMaximalLength + " characters.");
            }

            ExternalSystemUserProperty property = GetExternalSystemUserProperty(externalSystemUserId, propertyName);

            if (property == null)
            {
                property = new ExternalSystemUserProperty();
                property.ExternalSystemUserId = externalSystemUserId;
                property.Name = propertyName;
                property.Created = DateTime.Now;
            }

            property.Value = propertyValue;
            property.Modified = DateTime.Now;

            // if it is a new property, it must be inserted into the DB, otherwise will be updated on submit automatically.
            if (property.ExternalSystemUserPropertyId == default(int))
            {
                Context.ExternalSystemUserProperties.InsertOnSubmit(property);
            }
            Context.SubmitChanges();

            return property;
        }

        /// <summary>
        /// Retrieves the external user's property object by external userId and property name
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="propertyName"></param>
        /// <returns>ExternalSystemUserProperty object or null</returns>
        public ExternalSystemUserProperty GetExternalSystemUserProperty(int externalSystemUserId, String propertyName)
        {
            if (propertyName == null || propertyName.Length == 0 || propertyName.Length > ExternalSystemUserPropertyNameMaximalLength)
            {
                throw new ExternalSystemIntegratorException("The length of the property name must be between 1 and " + ExternalSystemUserPropertyNameMaximalLength + " characters.");
            }

            ExternalSystemUserProperty property = Context.ExternalSystemUserProperties.FirstOrDefault<ExternalSystemUserProperty>(p => p.ExternalSystemUserId == externalSystemUserId && p.Name == propertyName);
            return property;
        }

        /// <summary>
        /// Retrieves the value of the external user's property or null (if the property does not exist)
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="propertyName"></param>
        public String GetExternalSystemUserPropertyValue(int externalSystemUserId, String propertyName)
        {
            ExternalSystemUserProperty property = GetExternalSystemUserProperty(externalSystemUserId, propertyName);
            if (property == null)
            {
                return null;
            }

            return property.Value;
        }

        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.ownsDataContext && disposing)
            {
                this.Context.Connection.Dispose();
                this.Context.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

    }
}
