﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;

namespace VirtualProfiles
{
    public partial class _VirtualProfilesManager
    {
        /// <summary>
        /// Register the recommendation to use a virtual profile made to a local user
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromUser"></param>
        /// <param name="toUser"></param>
        /// <returns></returns>
        public VirtualProfileUsageInvitation InviteUserToVirtualProfile(int virtualProfileId, User fromUser, User toUser)
        {
            ContentRecommendation recommendation = InsertNewInvitation(default(int), virtualProfileId, fromUser.UserId, toUser.UserId, default(int));
            return (VirtualProfileUsageInvitation)recommendation;
        }


        /// <summary>
        /// Register the recommendation to use a virtual profile made to an external user
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromUser"></param>
        /// <param name="toUser"></param>
        /// <returns></returns>
        public VirtualProfileUsageInvitation InviteUserToVirtualProfile(int virtualProfileId, User fromUser, ExternalSystemUser toExternalUser)
        {
            return (VirtualProfileUsageInvitation)InsertNewInvitation(default(int), virtualProfileId, fromUser.UserId, default(int), toExternalUser.ExternalSystemUserId);
        }

        /// <summary>
        /// Retirves all virtual profiles' invitations made to the specified local user
        /// </summary>
        /// <param name="localUser"></param>
        /// <returns></returns>
        public IEnumerable<VirtualProfileUsageInvitation> GetActiveVirtualProfileInvitations(User localUser)
        {
            IEnumerable<VirtualProfileUsageInvitation> invitations =
                from invitation in Context.ContentRecommendations
                where invitation is VirtualProfileUsageInvitation && 
                invitation.IsActive == true && invitation.ToUserId == localUser.UserId
                select invitation as VirtualProfileUsageInvitation;

            return invitations;
        }

        /// <summary>
        /// Retirves all virtual profiles' invitations made to the specified local user
        /// </summary>
        /// <param name="localUser"></param>
        /// <returns></returns>
        public IEnumerable<VirtualProfileUsageInvitation> GetActiveVirtualProfileInvitations(ExternalSystemUser externalUser)
        {
            IEnumerable<VirtualProfileUsageInvitation> invitations =
                from invitation in Context.ContentRecommendations
                where invitation is VirtualProfileUsageInvitation &&
                        invitation.IsActive == true && invitation.ToExternalSystemUserId == externalUser.ExternalSystemUserId
                select invitation as VirtualProfileUsageInvitation;

            return invitations;
        }

        /// <summary>
        /// Marks recommendation as read(not active).
        /// </summary>
        /// <param name="recommendationId"></param>
        /// <returns>true if the recommendation has been marked read, false if the recommendation did not exist, or had been inactive already</returns>
        public bool MarkRecommendationAsRead(int recommendationId)
        {
            ContentRecommendation recommendation = GetRecommendationById(recommendationId);
            if (recommendation == null || !recommendation.IsActive)
            {
                return false;
            }
            recommendation.IsActive = false;
            Context.SubmitChanges();
            return true;
        }


        /// <summary>
        /// Utility method to register an recommendation made to EITHER local OR external one.
        /// If recommendation has been made previously, new recommendation will NOT be registered and 
        /// the old instance will be returned instead.
        /// </summary>
        /// <param name="catalogItemId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromLocalUserId"></param>
        /// <param name="toLocalUserId"></param>
        /// <param name="toExternalUserId"></param>
        /// <returns>either newly inserted recommendation, or (if duplicate) the previous instance of it</returns>
        private ContentRecommendation InsertNewInvitation(int catalogItemId, int virtualProfileId, int fromLocalUserId, int toLocalUserId, int toExternalUserId)
        {
            if (fromLocalUserId == toLocalUserId || fromLocalUserId == toExternalUserId)
            {
                throw new ArgumentException("User can't invite himself to use a vp.");
            }

            VirtualProfileUsageInvitation invitation = null;
            // if invited already, don't invite again
            if ((invitation = GetRecommendationById(virtualProfileId, fromLocalUserId, toLocalUserId, default(int))) != null)
            {
                return invitation;
            }

            // create recommendation
            invitation = new VirtualProfileUsageInvitation();
            invitation.VirtualProfileId = virtualProfileId;
            invitation.FromUserId = fromLocalUserId;

            if (toLocalUserId != default(int))
            {
                invitation.ToUserId = toLocalUserId;
            }
            if (toExternalUserId != default(int))
            {
                invitation.ToExternalSystemUserId = toExternalUserId;
            }

            invitation.IsActive = true;
            invitation.Created = DateTime.Now;
            invitation.Modified = DateTime.Now;

            Context.ContentRecommendations.InsertOnSubmit(invitation);
            Context.SubmitChanges();

            return invitation;
        }


        /// <summary>
        /// Tries to retrieve registered recommendation made to EITHER local OR external user
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="fromLocalUserId"></param>
        /// <param name="toLocalUserId"></param>
        /// <param name="toExternalUserId"></param>
        /// <returns></returns>
        private VirtualProfileUsageInvitation GetRecommendationById(int virtualProfileId, int fromLocalUserId, int toLocalUserId, int toExternalUserId)
        {
            // hack fixme: how to include the 'virtualProfileId' into the query??!?!?!?!?!?!?
            return (VirtualProfileUsageInvitation)Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                    (i => i is VirtualProfileUsageInvitation
                        && i.FromUserId == fromLocalUserId
                        && (i.ToUserId == toLocalUserId || i.ToExternalSystemUserId == toExternalUserId)
                        && (i as VirtualProfileUsageInvitation).VirtualProfileId == virtualProfileId
                     );
        }

        /// <summary>
        /// Retrieves recommendation by its Id
        /// </summary>
        /// <param name="ContentRecommendationId"></param>
        /// <returns></returns>
        private ContentRecommendation GetRecommendationById(int contentRecommendationId)
        {
            ContentRecommendation recommendation = Context.ContentRecommendations.FirstOrDefault<ContentRecommendation>
                  (i => i.ContentRecommendationId == contentRecommendationId && i is VirtualProfileUsageInvitation);
            return recommendation;
        }
    }
}