﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;
using ExternalSystemIntegration;

namespace VirtualProfiles
{
    public partial class VirtualProfilesManager
    {
        /// <summary>
        /// Access right that allows to use the VP, receive recommendations given to the virtual profile
        /// </summary>
        public static VPAccessRight AccessRightGetRecommendations;

        /// <summary>
        /// Access right that allows to modify VP by submitting feeback(normally, ratings) to the virtual profile
        /// </summary>
        public static VPAccessRight AccessRightSubmitFeedback;

        /// <summary>
        /// Access right that allows to modify VP's properties (to change the access level, description, etc)
        /// </summary>
        public static VPAccessRight AccessRightAdministrate;

        /// <summary>
        /// All available access rights in one collection (convenient for listing them). 
        /// Are initialized by constructor.
        /// </summary>
        public static HashSet<VPAccessRight> AllAccessRights;

        /// <summary>
        /// Id of the 'GetRecommendations' access right 
        /// </summary>
        public const int AccessRightId_GetRecommendations = 1;

        /// <summary>
        /// Id of the 'SubmitFeedback' access right 
        /// </summary>
        public const int AccessRightId_SubmitFeedback = 2;

        /// <summary>
        /// Id of the 'AdministerProfile' access right 
        /// </summary>
        public const int AccessRightId_AdministerProfile = 3;

        /// <summary>
        /// Name of the corresponding access right
        /// </summary>
        public const String AccessRightName_GetRecommendations = "GetRecommendations";

        /// <summary>
        /// Name of the corresponding access right
        /// </summary>
        public const String AccessRightName_SubmitFeedback = "SubmitFeedback";

        /// <summary>
        /// Name of the corresponding access right
        /// </summary>
        public const String AccessRightName_AdministerProfile = "AdministerProfile";

        /// <summary>
        /// Makes sure the access rights are deployed in the DB, and initializes them (sets data members)
        /// </summary>
        private void InitializeAccessRights()
        {
            AccessRightGetRecommendations = DeployAccessRightToDB(1, "GetRecommendations", "Access right that allows to use the VP, receive recommendations given to the virtual profile");
            AccessRightSubmitFeedback = DeployAccessRightToDB(2, "SubmitFeedback", "Access right that allows to modify VP by submitting feeback(normally, ratings) to the virtual profile");
            AccessRightAdministrate = DeployAccessRightToDB(3, "AdministerProfile", "Access right that allows to modify VP's properties (changing the access level, description, etc)");

            AllAccessRights = new HashSet<VPAccessRight>();
            AllAccessRights.Add(AccessRightGetRecommendations);
            AllAccessRights.Add(AccessRightSubmitFeedback);
            AllAccessRights.Add(AccessRightAdministrate);
        }


        /// <summary>
        /// Registers access right in the DB if missing, and returns 
        /// corresponding initialized VPAccessRight object
        /// </summary>
        /// <param name="accessRightId"></param>
        /// <param name="name"></param>
        /// <param name="description"></param>
        /// <returns></returns>
        private VPAccessRight DeployAccessRightToDB(int accessRightId, String name, String description)
        {
            VPAccessRight dbEntry = Context.VPAccessRights.FirstOrDefault<VPAccessRight>(a => a.Name == name);

            // if is not in the DB yet, register it
            if (dbEntry == null)
            {
                // insert to the DB
                dbEntry = new VPAccessRight();
                dbEntry.Name = name;
                dbEntry.Description = description;
                dbEntry.VPAccessRightId = accessRightId;

                dbEntry.Created = DateTime.Now;
                dbEntry.Modified = DateTime.Now;

                Context.VPAccessRights.InsertOnSubmit(dbEntry);
                Context.SubmitChanges();
                TraceLog.Write(LogEventType.Information, "New AccessRight(name='{0}', id='{1}') has been deployed.", new object[] { dbEntry.Name, dbEntry.VPAccessRightId });
            }

            return dbEntry;
        }


        /// <summary>
        /// Grants the specified access right on certain virtual profile to the specified MyMedia user
        /// </summary>
        /// <param name="userId">MyMedia user id</param>
        /// <param name="virtualProfileId">Id of a virtual profile</param>
        /// <param name="accessRightId">Id of the access right that is to be granted</param>
        /// <returns>true if access right has been granted, false if the user had the access right already.</returns>
        public bool GrantAccessRightToLocalUser(int userId, int virtualProfileId, int accessRightId)
        {
            if (IsAccessRightGrantedToLocalUser(userId, virtualProfileId, accessRightId))
            {
                return false;
            }

            UserVPAccessRight Result = new UserVPAccessRight();
            Result.UserId = userId;
            Result.VirtualProfileId = virtualProfileId;
            Result.VPAccessRightId = accessRightId;
            Result.Created = DateTime.Now;
            Result.Modified = DateTime.Now;

            Context.UserVPAccessRights.InsertOnSubmit(Result);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information, "VP Access Right '{0}' granted to local user '{1}' on the VP_id='{2}'.", new object[] { Result.VPAccessRight.Name, Result.UserId, Result.VPAccessRightId });

            return true;
        }

        /// <summary>
        /// Grants specified access right to the specified external user
        /// </summary>
        /// <param name="externalUserId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="accessRightId"></param>
        /// <returns>true if access right has been granted, false if the user had the access right already.</returns>
        public bool GrantAccessRightToExternalUser(int externalUserId, int virtualProfileId, int accessRightId)
        {
            if (IsAccessRightGrantedToExternalUser(externalUserId, virtualProfileId, accessRightId))
            {
                return false;
            }
            UserVPAccessRight Result = new UserVPAccessRight();
            Result.ExternalSystemUserId = externalUserId;
            Result.VirtualProfileId = virtualProfileId;
            Result.VPAccessRightId = accessRightId;
            Result.Created = DateTime.Now;
            Result.Modified = DateTime.Now;

            Context.UserVPAccessRights.InsertOnSubmit(Result);
            Context.SubmitChanges();

            return true;
        }

        /// <summary>
        /// Grants the specified access right on certain virtual profile to 
        /// the specified external user (user that is registered in an external system, 
        /// and does not have a local account yet)
        /// </summary>
        /// <param name="externalUser">Object holding an external user</param>
        /// <param name="virtualProfile">Object holding initialized existing virtual profile</param>
        /// <param name="accessRight">Access right that is to be granted</param>
        /// <returns>true if access right has been granted, false if the user had the access right already.</returns>
        public bool GrantAccessRightToExternalUser(ExternalSystemUser externalUser, VirtualProfile virtualProfile, VPAccessRight accessRight)
        {
            if (IsAccessRightGranted(externalUser, virtualProfile, accessRight.VPAccessRightId))
            {
                return false;
            }
            UserVPAccessRight Result = new UserVPAccessRight();
            Result.ExternalSystemUserId = externalUser.ExternalSystemUserId;
            Result.VirtualProfile = virtualProfile;
            Result.VPAccessRight = accessRight;
            Result.Created = DateTime.Now;
            Result.Modified = DateTime.Now;
            if (externalUser.User != null)
            {
                Result.UserId = externalUser.User.UserId;
            }

            Context.UserVPAccessRights.InsertOnSubmit(Result);
            Context.SubmitChanges();

            return true;
        }

        /// <summary>
        /// Grants the specified access right on certain virtual profile to the specified MyMedia user
        /// </summary>
        /// <param name="user">MyMedia user object</param>
        /// <param name="virtualProfile">Object holding initialized existing virtual profile</param>
        /// <param name="accessRight">Access right that is to be granted</param>
        /// <returns>true if access right has been granted, false if the user had the access right already.</returns>
        public bool GrantAccessRightToLocalUser(User user, VirtualProfile virtualProfile, VPAccessRight accessRight)
        {
            if (IsAccessRightGranted(user, virtualProfile, accessRight.VPAccessRightId))
            {
                return false;
            }

            UserVPAccessRight Result = new UserVPAccessRight();
            Result.UserId = user.UserId;
            Result.VirtualProfile = virtualProfile;
            Result.VPAccessRight = accessRight;
            Result.Created = DateTime.Now;
            Result.Modified = DateTime.Now;

            Context.UserVPAccessRights.InsertOnSubmit(Result);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information, "New AccessRight(name='{0}', id='{1}') has been deployed.", new object[] { Result.VPAccessRight.Name, Result.VPAccessRightId });

            return true;
        }

        /// <summary>
        /// Returns access rights that the specified user has on the given virtual profile
        /// </summary>
        /// <param name="userId">Local user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <returns>Set of access right objects</returns>
        public static HashSet<VPAccessRight> GetVPAccessRightsByLocalUserId(int userId, VirtualProfile virtualProfile){
            // if user is VP's creator, all rights are granted to him
            if (IsVirtualProfileOwner(userId, virtualProfile))
            {
                return AllAccessRights;
            }

            IEnumerable<UserVPAccessRight> userRights = virtualProfile.UserVPAccessRights.Where<UserVPAccessRight>(r => r.UserId == userId);
            HashSet<VPAccessRight> result = new HashSet<VPAccessRight>();
            foreach (UserVPAccessRight right in userRights)
            {
                result.Add(right.VPAccessRight);
            }

            if (virtualProfile.IsPublicUsageAllowed)
            {
                result.Add(AccessRightGetRecommendations);
            }

            if (virtualProfile.IsModificationAllowed)
            {
                result.Add(AccessRightSubmitFeedback);
            }

            return result;
        }

        /// <summary>
        /// Returns access rights that the specified user has on the given virtual profile
        /// </summary>
        /// <param name="user">Local user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <returns>Set of access right objects</returns>
        public HashSet<VPAccessRight> GetVPAccessRights(User user, VirtualProfile virtualProfile)
        {
            return GetVPAccessRightsByLocalUserId(user.UserId, virtualProfile);
        }

        /// <summary>
        /// Returns access rights that the specified external user has on the given virtual profile
        /// </summary>
        /// <param name="externalUser">External user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <returns>Set of access right objects</returns>
        public HashSet<VPAccessRight> GetVPAccessRights(ExternalSystemUser externalUser, VirtualProfile virtualProfile)
        {
            // if user is VP's creator, all rights are granted to him
            if (IsVirtualProfileOwner(externalUser, virtualProfile))
            {
                return AllAccessRights;
            }

            IEnumerable<UserVPAccessRight> userRights = virtualProfile.UserVPAccessRights.Where<UserVPAccessRight>(r => r.UserId == externalUser.LocalUserId || r.ExternalSystemUserId == externalUser.ExternalSystemUserId);

            HashSet<VPAccessRight> result = new HashSet<VPAccessRight>();
            foreach (UserVPAccessRight right in userRights)
            {
                result.Add(right.VPAccessRight);
            }

            if (virtualProfile.IsPublicUsageAllowed)
            {
                result.Add(AccessRightGetRecommendations);
            }

            if (virtualProfile.IsModificationAllowed)
            {
                result.Add(AccessRightSubmitFeedback);
            }

            return result;
        }

        /// <summary>
        /// Checks if user is the owner of the specified virtual profile
        /// </summary>
        /// <param name="user"></param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <returns></returns>
        public static bool IsVirtualProfileOwner(User user, VirtualProfile virtualProfile)
        {
            if (user != null)
            {
                return IsVirtualProfileOwner(user.UserId, virtualProfile);
            }
            return false;
        }

        /// <summary>
        /// Checks if user is the owner of the specified virtual profile
        /// </summary>
        /// <param name="userId">Local userId</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <returns></returns>
        public static bool IsVirtualProfileOwner(int userId, VirtualProfile virtualProfile)
        {
            if (userId == virtualProfile.CreatorUserId)
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// Checks if user is the owner of the specified virtual profile
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <returns></returns>
        public static bool IsVirtualProfileOwner(ExternalSystemUser externalUser, VirtualProfile virtualProfile)
        {
            if (externalUser != null && externalUser.User != null && externalUser.User.UserId == virtualProfile.CreatorUserId)
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// Verifies if certain access right is granted to the local MyMedia user on the given virtual profile
        /// </summary>
        /// <param name="localUserId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="accessRightId"></param>
        /// <returns></returns>
        public bool IsAccessRightGrantedToLocalUser(int localUserId, int virtualProfileId, int accessRightId)
        {
            UsersManager UsersManager = new UsersManager();
            User user = UsersManager.GetLocalUserById(localUserId);
            VirtualProfile vp = this.GetActiveVirtualProfileById(virtualProfileId);

            return IsAccessRightGranted(user, vp, accessRightId);
        }

        /// <summary>
        /// Verifies if certain access right is granted to the external user on the given virtual profile
        /// </summary>
        /// <param name="externalUserId"></param>
        /// <param name="virtualProfileId"></param>
        /// <param name="accessRightId"></param>
        /// <returns></returns>
        public bool IsAccessRightGrantedToExternalUser(int externalUserId, int virtualProfileId, int accessRightId)
        {
            UsersManager UsersManager = new UsersManager();
            ExternalSystemUser extUser = UsersManager.GetExternalSystemUserById(externalUserId);
            VirtualProfile vp = this.GetActiveVirtualProfileById(virtualProfileId);
            
            return IsAccessRightGranted(extUser, vp, accessRightId);
        }

        /// <summary>
        /// Verifies if certain access right is granted to the user on the given virtual profile
        /// </summary>
        /// <param name="user">Local user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <param name="vpAccessRightId">Id of the access right to check</param>
        /// <returns>true if access right is granted to the user, false - otherwise</returns>
        public static bool IsAccessRightGranted(User user, VirtualProfile virtualProfile, int vpAccessRightId)
        {
            if (IsVirtualProfileOwner(user, virtualProfile)
                || virtualProfile.IsPublicUsageAllowed && AccessRightId_GetRecommendations == vpAccessRightId
                || virtualProfile.IsModificationAllowed && AccessRightId_SubmitFeedback == vpAccessRightId
                )
            {
                return true;
            }

            // if the access is granted to the local user
            if (virtualProfile.UserVPAccessRights.Any<UserVPAccessRight>(r => r.UserId == user.UserId && r.VPAccessRightId == vpAccessRightId))
            {
                return true;
            }

            // check if the access is granted to one of the user's possible external accounts
            if (user.ExternalSystemUsers.Count() > 0)
            {
                foreach (ExternalSystemUser externalUser in user.ExternalSystemUsers)
                {
                    if (virtualProfile.UserVPAccessRights.Any<UserVPAccessRight>(r => r.VPAccessRightId == vpAccessRightId && (r.ExternalSystemUserId == externalUser.ExternalSystemUserId)))
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        /// <summary>
        /// Verifies if certain access right is granted to the user on the given virtual profile
        /// </summary>
        /// <param name="user">Local user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <param name="vpAccessRight">The access right object to be checked</param>
        /// <returns>true if access right is granted to the user, false - otherwise</returns>
        public static bool IsAccessRightGranted(User user, VirtualProfile virtualProfile, VPAccessRight vpAccessRight)
        {
            return IsAccessRightGranted(user, virtualProfile, vpAccessRight.VPAccessRightId);
        }

        /// <summary>
        /// Verifies if certain access right is granted to the external user on the given virtual profile
        /// </summary>
        /// <param name="externalUser">External user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <param name="vpAccessRightId">Id of the access right to check</param>
        /// <returns>true if access right is granted to the user, false - otherwise</returns>
        public static bool IsAccessRightGranted(ExternalSystemUser externalUser, VirtualProfile virtualProfile, int vpAccessRightId)
        {
            if (IsVirtualProfileOwner(externalUser, virtualProfile)
                || virtualProfile.IsPublicUsageAllowed && AccessRightId_GetRecommendations == vpAccessRightId
                || virtualProfile.IsModificationAllowed && AccessRightId_SubmitFeedback == vpAccessRightId
                )
            {
                return true;
            }

            return virtualProfile.UserVPAccessRights.Any<UserVPAccessRight>(r => r.VPAccessRightId == vpAccessRightId && (r.UserId == externalUser.LocalUserId || r.ExternalSystemUserId == externalUser.ExternalSystemUserId));
        }

        /// <summary>
        /// Verifies if certain access right is granted to the external user on the given virtual profile
        /// </summary>
        /// <param name="externalUser">External user object</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <param name="vpAccessRight">The access right object to be checked</param>
        /// <returns>true if access right is granted to the user, false - otherwise</returns>
        public static bool IsAccessRightGranted(ExternalSystemUser externalUser, VirtualProfile virtualProfile, VPAccessRight vpAccessRight)
        {
            return IsAccessRightGranted(externalUser, virtualProfile, vpAccessRight.VPAccessRightId);
        }


        /// <summary>
        /// Removes specified user's access right on the virtual channel, if present and user is not the creator of the VP
        /// Throws ArgumentException if user is an owner of the virtual profile
        /// </summary>
        /// <param name="user">MyMedia user</param>
        /// <param name="virtualProfile">Virtual profile object</param>
        /// <param name="accessRightId">Id of the access right that is to be removed</param>
        /// <returns>true if access right has been removed, false if user did not have the access right granted</returns>
        public bool RemoveAccessRight(User user, VirtualProfile virtualProfile, int accessRightId)
        {
            // if user is VP's owner, access rigth cannot be removed.
            if (IsVirtualProfileOwner(user, virtualProfile))
            {
                throw new ArgumentException("AccessRights cannot be removed from the owner of the virtual profile.");
            }

            // find it first
            UserVPAccessRight userVPAccessRight = virtualProfile.UserVPAccessRights.FirstOrDefault<UserVPAccessRight>(r => r.UserId == user.UserId && r.VPAccessRightId == accessRightId);
            if (userVPAccessRight != null)
            {
                Context.UserVPAccessRights.DeleteOnSubmit(userVPAccessRight);
                Context.SubmitChanges();
                return true;
            }

            return false;
        }

        /// <summary>
        /// Removes specified user's access right on the virtual channel, if present and user is not the creator of the VP
        /// Throws ArgumentException if user is an owner of the virtual profile
        /// </summary>
        /// <param name="externalUser"></param>
        /// <param name="virtualProfile"></param>
        /// <param name="accessRightId"></param>
        /// <returns>true if access right has been removed, false if user did not have the access right granted</returns>
        public bool RemoveAccessRight(ExternalSystemUser externalUser, VirtualProfile virtualProfile, int accessRightId)
        {
            // if user is VP's owner, access rigth cannot be removed.
            if (IsVirtualProfileOwner(externalUser, virtualProfile))
            {
                throw new ArgumentException("AccessRights cannot be removed from the owner of the virtual profile.");
            }

            // find it first
            UserVPAccessRight userVPAccessRight = virtualProfile.UserVPAccessRights.FirstOrDefault<UserVPAccessRight>(r => r.VPAccessRightId == accessRightId && (r.UserId == externalUser.LocalUserId || r.ExternalSystemUserId == externalUser.ExternalSystemUserId));
            if (userVPAccessRight != null)
            {
                Context.UserVPAccessRights.DeleteOnSubmit(userVPAccessRight);
                Context.SubmitChanges();
                return true;
            }

            return false;
        }

        /// <summary>
        /// Returns access right Id by its Name
        /// </summary>
        /// <param name="accessRightId"></param>
        /// <returns></returns>
        public static int GetAccessRightIdByName(String accessRightName)
        {
            if (AccessRightName_GetRecommendations.Equals(accessRightName))
            {
                return AccessRightId_GetRecommendations;
            }
            if (AccessRightName_AdministerProfile.Equals(accessRightName))
            {
                return AccessRightId_AdministerProfile;
            }
            if (AccessRightName_SubmitFeedback.Equals(accessRightName))
            {
                return AccessRightId_SubmitFeedback;
            }

            throw new ArgumentException("No VP access right exists with the specified name ('" + accessRightName + "').");
        }
    }
}
